// $Id: EODetailDataSource.m,v 1.2 2003/01/07 11:30:31 helge Exp $

#include <EOControl/EODetailDataSource.h>
#include <EOControl/EOClassDescription.h>
#include <EOControl/EOKeyValueCoding.h>
#include "common.h"

@implementation EODetailDataSource

- (id)initWithMasterClassDescription:(EOClassDescription *)_cd
  detailKey:(NSString *)_relKey
{
  if ((self = [super init])) {
    self->masterClassDescription = [_cd retain];

    [self qualifyWithRelationshipKey:_relKey ofObject:nil];
  }
  return self;
}

- (id)initWithMasterDataSource:(EODataSource *)_ds
  detailKey:(NSString *)_relKey
{
  if ((self = [self initWithMasterClassDescription:nil detailKey:_relKey])) {
    self->masterDataSource = [_ds retain];
  }
  return self;
}

- (id)init {
  return [self initWithMasterClassDescription:nil detailKey:nil];
}

- (void)dealloc {
  [self->detailKey              release];
  [self->masterObject           release];
  [self->masterClassDescription release];
  [self->masterDataSource       release];
  [super dealloc];
}

/* reflection */

- (void)setMasterClassDescription:(EOClassDescription *)_cd {
  ASSIGN(self->masterClassDescription, _cd);
}
- (EOClassDescription *)masterClassDescription {
  return self->masterClassDescription;
}

- (EODataSource *)masterDataSource {
  return self->masterDataSource;
}

/* editing context */

- (id)editingContext {
  return [[self masterObject] editingContext];
}

/* master-detail */

- (id)masterObject {
  return self->masterObject;
}
- (NSString *)detailKey {
  return self->detailKey;
}

- (void)qualifyWithRelationshipKey:(NSString *)_relKey ofObject:(id)_object {
  id tmp;

  tmp = self->detailKey;
  self->detailKey = [_relKey copy];
  [tmp release];

  ASSIGN(self->masterObject, _object);
}

/* operations */

- (NSArray *)fetchObjects {
  id       eo;
  NSString *dk;
  
  if ((eo = [self masterObject]) == nil)
    return [NSArray array];

  if ((dk = [self detailKey]) == nil)
    return [NSArray arrayWithObject:eo];

  return [eo valueForKey:dk];
}

- (void)insertObject:(id)_object {
  id       eo;
  NSString *dk;
  
  if ((eo = [self masterObject]) == nil) {
    [NSException raise:@"NSInternalInconsistencyException"
                 format:
                   @"detail datasource %@ has no master object set "
                   @"for insertion of object %@",
                   self, _object];
  }
  if ((dk = [self detailKey]) == nil) {
    [NSException raise:@"NSInternalInconsistencyException"
                 format:
                   @"detail datasource %@ has no detail key set "
                   @"for insertion of object %@ into master %@",
                   self, _object, eo];
  }
  
  [eo addObject:_object toBothSidesOfRelationshipWithKey:dk];
}

- (void)deleteObject:(id)_object {
  id       eo;
  NSString *dk;
  
  if ((eo = [self masterObject]) == nil) {
    [NSException raise:@"NSInternalInconsistencyException"
                 format:
                   @"detail datasource %@ has no master object set "
                   @"for deletion of object %@",
                   self, _object];
  }
  if ((dk = [self detailKey]) == nil) {
    [NSException raise:@"NSInternalInconsistencyException"
                 format:
                   @"detail datasource %@ has no detail key set "
                   @"for deletion of object %@ from master %@",
                   self, _object, eo];
  }
  
  [eo removeObject:_object fromPropertyWithKey:dk];
}

@end /* EODetailDataSource */
