// $Id: EOFaultHandler.m,v 1.3 2003/04/01 13:32:30 gnustep Exp $

#include "EOFaultHandler.h"
#include "EOFault.h"
#include "common.h"

#if NeXT_RUNTIME
#  if !defined(METHOD_NULL)
#    define METHOD_NULL NULL
#  endif
#endif

@implementation EOFaultHandler

- (void)setTargetClass:(Class)_class extraData:(void *)_extraData {
  self->targetClass = _class;
  self->extraData   = _extraData;
}

- (Class)targetClass; {
  return self->targetClass;
}
- (void *)extraData {
  return self->extraData;
}

/* firing */

- (BOOL)shouldPerformInvocation:(NSInvocation *)_invocation {
  return YES;
}

- (void)faultWillFire:(EOFault *)_fault {
}

- (void)completeInitializationOfObject:(id)_object {
  [self doesNotRecognizeSelector:_cmd];
}

/* fault reflection */

- (Class)classForFault:(EOFault *)_fault {
  return (object_is_instance(_fault))
    ? [self targetClass]
    : (*(Class *)_fault);
}

- (BOOL)respondsToSelector:(SEL)_selector forFault:(EOFault *)_fault {
  Class class;

  /* first check whether fault itself responds to selector */
  if (class_get_instance_method(*(Class *)_fault, _selector) != METHOD_NULL)
    return YES;

  /* then check whether the target class does */
  class = [self targetClass];
  return (class_get_instance_method(class, _selector) != NULL) ? YES : NO;
}

- (BOOL)conformsToProtocol:(Protocol *)_protocol forFault:(EOFault *)_fault {
  struct objc_protocol_list* protos;
  Class class, sClass;
  int i;

  class = object_is_instance(_fault) ? [self targetClass] : (Class)_fault;

  for (protos = class->protocols; protos; protos = protos->next) {
    for (i = 0; i < protos->count; i++) {
      if ([protos->list[i] conformsTo:_protocol])
        return YES;
    }
  }

  return ((sClass = [class superclass]))
    ? [sClass conformsToProtocol:_protocol]
    : NO;
}

- (BOOL)isKindOfClass:(Class)_class forFault:(EOFault *)_fault {
  Class class;

  class = object_is_instance(_fault) ? [self targetClass] : (Class)_fault;
  
  for (; class != Nil; class = class_get_super_class(class)) {
    if (class == _class)
      return YES;
  }
  return NO;
}

- (BOOL)isMemberOfClass:(Class)_class forFault:(EOFault *)_fault {
  Class class;

  class = object_is_instance(_fault) ? [self targetClass] : (Class)_fault;
  return class == _class ? YES : NO;
}

- (NSMethodSignature *)methodSignatureForSelector:(SEL)_selector
  forFault:(EOFault *)_fault
{
#if NeXT_Foundation_LIBRARY
  // probably incorrect
  return [_fault methodSignatureForSelector:_selector];
#else
  register const char *types = NULL;

  if (_selector == NULL) // invalid selector
    return nil;

#if GNU_RUNTIME && 0
  // GNU runtime selectors may be typed, a lookup may not be necessary
  types = aSelector->sel_types;
#endif

  /* first check for EOFault's own methods */
  
  if (types == NULL) {
    // lookup method for selector
    struct objc_method *mth;
    mth = class_get_instance_method(*(Class *)_fault, _selector);
    if (mth) types = mth->method_types;
  }
  
  /* then check in target class methods */
  
  if (types == NULL) {
    // lookup method for selector
    struct objc_method *mth;
    mth = class_get_instance_method([self targetClass], _selector);
    if (mth) types = mth->method_types;
  }
  
#if GNU_RUNTIME
  // GNU runtime selectors may be typed, a lookup may not be necessary
  if (types == NULL)
    types = _selector->sel_types;
#endif
  if (types == NULL)
    return nil;
  
  return [NSMethodSignature signatureWithObjCTypes:types];
#endif
}

/* description */

- (NSString *)descriptionForObject:(id)_fault {
  return [NSString stringWithFormat:@"<%@[0x%08X]: on=%@>",
                     NSStringFromClass(*(Class *)_fault),
                     _fault,
                     NSStringFromClass([self targetClass])];
}

@end /* EOFaultHandler */
