// $Id: EOFetchSpecification.m,v 1.7 2003/01/07 11:30:31 helge Exp $

#include "EOFetchSpecification.h"
#include "EOQualifier.h"
#include "EOSortOrdering.h"
#include "common.h"

@implementation EOFetchSpecification

+ (EOFetchSpecification *)fetchSpecificationWithEntityName:(NSString *)_ename
  qualifier:(EOQualifier *)_qualifier
  sortOrderings:(NSArray *)_sortOrderings
{
  EOFetchSpecification *fs = nil;

  fs = [[self alloc] initWithEntityName:_ename
                     qualifier:_qualifier
                     sortOrderings:_sortOrderings
                     usesDistinct:NO];
  return [fs autorelease];
}

- (id)initWithEntityName:(NSString *)_name
  qualifier:(EOQualifier *)_qualifier
  sortOrderings:(NSArray *)_sortOrderings
  usesDistinct:(BOOL)_dflag
{
  if ((self = [super init])) {
    self->entityName    = [_name copyWithZone:[self zone]];
    self->qualifier     = [_qualifier     retain];
    self->sortOrderings = [_sortOrderings retain];
    self->usesDistinct  = _dflag;
    self->fetchLimit    = 0;
    self->hints         = nil;
  }
  return self;
}

- (id)init {
  if ((self = [super init])) {
    self->usesDistinct = NO;
    self->fetchLimit   = 0;
    self->entityName    = nil;
    self->qualifier     = nil;
    self->sortOrderings = nil;
    self->hints         = nil;
  }
  return self;
}

- (id)initWithDictionary:(NSDictionary *)_dictionary {
  if ((self = [self init])) {
    id tmp;
    
    if ((tmp = [_dictionary objectForKey:@"qualifier"])) {
      if ([tmp isKindOfClass:[NSDictionary class]])
        tmp = [EOQualifier qualifierToMatchAllValues:tmp];
      else
        tmp = [EOQualifier qualifierWithQualifierFormat:tmp];
      
      [self setQualifier:tmp];
    }
    
    if ((tmp = [_dictionary objectForKey:@"sortOrderings"])) {
      NSArray        *sos = nil;
      EOSortOrdering *so;
      
      if ([tmp respondsToSelector:@selector(objectEnumerator)]) {
        NSMutableArray *result  = nil;
        NSEnumerator   *objEnum;
        id             obj;
        
        objEnum = [tmp objectEnumerator];
        result = [NSMutableArray arrayWithCapacity:8];
        while ((obj = [objEnum nextObject])) {
          so = [[EOSortOrdering alloc] initWithPropertyList:obj];
          [so autorelease];
          if (so)
            [result addObject:so];
        }
        sos = [NSArray arrayWithArray:result];
      }
      else {
        so = [[[EOSortOrdering alloc] initWithPropertyList:tmp] autorelease];
          
        if (so)
          sos = [NSArray arrayWithObject:so];
      }
      [self setSortOrderings:sos];
    }
    
    if ((tmp = [_dictionary objectForKey:@"fetchLimit"])) {
      if ([tmp respondsToSelector:@selector(intValue)])
        [self setFetchLimit:[tmp intValue]];
      else
        NSLog(@"%s: invalid fetchLimit key !", __PRETTY_FUNCTION__);
    }
    
    if ((tmp = [_dictionary objectForKey:@"hints"])) {
      if ([tmp isKindOfClass:[NSDictionary class]])
        [self setHints:tmp];
      else
        NSLog(@"%s: invalid hints key !", __PRETTY_FUNCTION__);
    }
    
    if ([self->hints objectForKey:@"addDocumentsAsObserver"] == nil) {
      NSMutableDictionary *hnts;

      hnts = [[NSMutableDictionary alloc] initWithDictionary:self->hints];
      [hnts setObject:[NSNumber numberWithBool:NO]
            forKey:@"addDocumentsAsObserver"];
      [self setHints:hnts];
      [hnts release];
    }
  }
  return self;
}
- (id)initWithString:(NSString *)_string {
  EOQualifier *q;
  
  q = [EOQualifier qualifierWithQualifierFormat:_string];
  
  return [self initWithEntityName:nil
               qualifier:q
               sortOrderings:nil
               usesDistinct:NO];
}

- (id)initWithPropertyList:(id)_plist {
  if ([_plist isKindOfClass:[NSDictionary class]])
    return [self initWithDictionary:_plist];
  else if ([_plist isKindOfClass:[NSString class]])
    return [self initWithString:_plist];
  else if ([_plist isKindOfClass:[self class]]) {
    [self release];
    return [_plist copy];
  }
  else {
    [self release];
    return nil;
  }
}

- (void)dealloc {
  [self->hints         release];
  [self->entityName    release];
  [self->qualifier     release];
  [self->sortOrderings release];
  [super dealloc];
}

- (void)setEntityName:(NSString *)_name {
  if (_name != self->entityName) {
#if !LIB_FOUNDATION_BOEHM_GC
    id tmp = self->entityName;
#endif
    self->entityName = [_name copyWithZone:[self zone]];
    [tmp release];
  }
}
- (NSString *)entityName {
  return self->entityName;
}

- (void)setQualifier:(EOQualifier *)_qualifier {
  ASSIGN(self->qualifier, _qualifier);
}
- (EOQualifier *)qualifier {
  return self->qualifier;
}

- (void)setSortOrderings:(NSArray *)_orderings {
  ASSIGN(self->sortOrderings, _orderings);
}
- (NSArray *)sortOrderings {
  return self->sortOrderings;
}

- (void)setUsesDistinct:(BOOL)_flag {
  self->usesDistinct = _flag;
}
- (BOOL)usesDistinct {
  return self->usesDistinct;
}

- (void)setLocksObjects:(BOOL)_flag {
  self->locksObjects = _flag;
}
- (BOOL)locksObjects {
  return self->locksObjects;
}

- (void)setFetchLimit:(unsigned)_limit {
  self->fetchLimit = _limit;
}
- (unsigned)fetchLimit {
  return self->fetchLimit;
}

- (void)setHints:(NSDictionary *)_hints {
  ASSIGN(self->hints, _hints);
}
- (NSDictionary *)hints {
  return self->hints;
}

/* bindings */

- (EOFetchSpecification *)
  fetchSpecificationWithQualifierBindings:(NSDictionary *)_bindings
{
  EOQualifier          *q     = nil;
  EOFetchSpecification *newfs = nil;

  q     = [[self qualifier] qualifierWithBindings:_bindings
                            requiresAllVariables:NO];
  newfs = [[[self class] alloc]
                  initWithEntityName:[self entityName]
                  qualifier:q
                  sortOrderings:[self sortOrderings]
                  usesDistinct:[self usesDistinct]];
  
  [newfs setLocksObjects:[self locksObjects]];
  [newfs setFetchLimit:[self fetchLimit]];
  
  return [newfs autorelease];
}

/* NSCopying */

- (id)copyWithZone:(NSZone *)_zone {
  EOFetchSpecification *fspec;
  NSDictionary *hdict;
  
  hdict = [[self hints] copy];
  
  fspec = [[[self class] alloc] initWithEntityName:[self entityName]
                                qualifier:[self qualifier]
                                sortOrderings:[self sortOrderings]
                                usesDistinct:[self usesDistinct]];
  [fspec setLocksObjects:[self locksObjects]];
  [fspec setFetchLimit:[self fetchLimit]];
  [fspec setHints:hdict];
  [hdict release];
  
  return fspec;
}

/* Equality */

- (BOOL)isEqualToFetchSpecification:(EOFetchSpecification *)_fspec {
  id t1, t2;
  if (_fspec == self)
    return YES;

  t1 = [self entityName];
  t2 = [_fspec entityName];
  if (t1 != t2) {
    if (![t1 isEqualToString:t2])
      return NO;
  }
  
  t1 = [self sortOrderings];
  t2 = [_fspec sortOrderings];
  if (t1 != t2) {
    if (![t1 isEqual:t2])
      return NO;
  }

  t1 = [self qualifier];
  t2 = [_fspec qualifier];
  if (t1 != t2) {
    if (![t1 isEqual:t2])
      return NO;
  }
  
  if ([self usesDistinct] != [_fspec usesDistinct])
    return NO;
  if ([self locksObjects] != [_fspec locksObjects])
    return NO;
  if ([self fetchLimit] != [_fspec fetchLimit])
    return NO;

  t1 = [self hints];
  t2 = [_fspec hints];
  if (t1 != t2) {
    if (![t1 isEqual:t2])
      return NO;
  }
  
  return YES;
}
- (BOOL)isEqual:(id)_other {
  if ([_other isKindOfClass:[EOFetchSpecification class]])
    return [self isEqualToFetchSpecification:_other];
  
  return NO;
}

/* remapping keys */

- (EOFetchSpecification *)fetchSpecificationByApplyingKeyMap:(NSDictionary *)_m {
  NSAutoreleasePool    *pool;
  EOFetchSpecification *fs;
  NSMutableDictionary  *lHints;
  EOQualifier    *q = nil;
  NSMutableArray *o = nil;

  pool = [[NSAutoreleasePool alloc] init];
  
  /* process qualifier */
  
  q = [self->qualifier qualifierByApplyingKeyMap:_m];
  
  /* process attributes */
  
  if (self->hints) {
    NSArray  *a;
    unsigned len;
    
    a = [self->hints objectForKey:@"attributes"];
    if ((len = [a count]) > 0) {
      NSMutableArray *ma;
      unsigned i;
      
      ma = [[NSMutableArray alloc] initWithCapacity:(len + 1)];
      for (i = 0; i < len; i++) {
	NSString *key, *tkey;
	
	key  = [a objectAtIndex:i];
	tkey = [_m objectForKey:key];
	
	[ma addObject:(tkey ? tkey : key)];
      }
      
      lHints = [self->hints mutableCopy];
      [lHints setObject:ma forKey:@"attributes"];
      [ma release];
    }
    else
      lHints = [self->hints retain];
  }
  else 
    lHints = nil;
  
  /* process orderings */
  
  if (self->sortOrderings) {
    unsigned i, len;
    
    len = [self->sortOrderings count];
    o   = [[NSMutableArray alloc] initWithCapacity:len];
    for (i = 0; i < len; i++) {
      EOSortOrdering *so, *tso;
      
      so  = [self->sortOrderings objectAtIndex:i];
      tso = [so sortOrderingByApplyingKeyMap:_m];
      [o addObject:tso ? tso : so];
    }
  }
  else
    o = nil;
  
  /* construct result */
  
  fs = [[EOFetchSpecification alloc] initWithEntityName:self->entityName
				     qualifier:q
				     sortOrderings:o
				     usesDistinct:self->usesDistinct];
  [fs setLocksObjects:self->locksObjects];
  [fs setFetchLimit:self->fetchLimit];
  if (lHints) {
    [fs setHints:lHints];
    [lHints release];
  }
  [o release];
  [pool release];
  return [fs autorelease];
}

/* description */

- (NSString *)description {
  NSMutableString *ms;
  id tmp;
  
  ms = [NSMutableString stringWithCapacity:128];
  [ms appendFormat:@"<%@[0x%08X]:", NSStringFromClass([self class]), self];

  if ((tmp = [self entityName]))
    [ms appendFormat:@" entity=%@", tmp];
  if ((tmp = [self qualifier]))
    [ms appendFormat:@" qualifier=%@", tmp];
  
  if ((tmp = [self sortOrderings]))
    [ms appendFormat:@" orderings=%@", tmp];
  
  if ([self locksObjects]) [ms appendString:@" locks"];
  if ([self usesDistinct]) [ms appendString:@" distinct"];
  
  if ([self fetchLimit] > 0)
    [ms appendFormat:@" limit=%i", [self fetchLimit]];

  if ((tmp = [self hints])) {
    NSEnumerator *e;
    NSString *hint;
    BOOL isFirst = YES;
    
    [ms appendString:@" hints:"];
    e = [tmp keyEnumerator];
    while ((hint = [e nextObject])) {
      if (isFirst) isFirst = NO;
      else [ms appendString:@","];
      [ms appendString:hint];
      [ms appendString:@"="];
      [ms appendString:[[tmp objectForKey:hint] stringValue]];
    }
  }
  
  [ms appendString:@">"];
  return ms;
}

@end /* EOFetchSpecification */
