// $Id: EOKeyValueArchiver.h,v 1.1.1.1 2002/05/22 17:27:30 helge Exp $

#ifndef __EOControl_EOKeyValueArchiver_H__
#define __EOControl_EOKeyValueArchiver_H__

#import <Foundation/NSObject.h>

@class NSString, NSDictionary, NSMutableArray, NSMutableDictionary;
@class NSMutableSet;

@interface EOKeyValueArchiver : NSObject
{
  NSMutableDictionary *plist;
  id delegate; // non-retained
}

/* coding */

- (void)encodeObject:(id)_obj            forKey:(NSString *)_key;
- (void)encodeReferenceToObject:(id)_obj forKey:(NSString *)_key;
- (void)encodeBool:(BOOL)_flag           forKey:(NSString *)_key;
- (void)encodeInt:(int)_value            forKey:(NSString *)_key;

- (NSDictionary *)dictionary;

/* delegate */

- (void)setDelegate:(id)_delegate;
- (id)delegate;

@end

@interface EOKeyValueUnarchiver : NSObject
{
  NSDictionary   *plist;
  NSMutableArray *unarchivedObjects;
  NSMutableSet   *awakeObjects;
  id parent;
  
  id delegate; // non-retained
}

- (id)initWithDictionary:(NSDictionary *)_dict;

/* decoding */

- (id)decodeObjectForKey:(NSString *)_key;
- (id)decodeObjectReferenceForKey:(NSString *)_key;
- (BOOL)decodeBoolForKey:(NSString *)_key;
- (int)decodeIntForKey:(NSString *)_key;

/* operations */

- (void)ensureObjectAwake:(id)_object;
- (void)finishInitializationOfObjects;
- (void)awakeObjects;
- (id)parent;

/* delegate */

- (void)setDelegate:(id)_delegate;
- (id)delegate;

@end

@protocol EOKeyValueArchiving

- (id)initWithKeyValueUnarchiver:(EOKeyValueUnarchiver *)_unarchiver;
- (void)encodeWithKeyValueArchiver:(EOKeyValueArchiver *)_archiver;

@end

@interface NSObject(EOKeyValueArchivingAwakeMethods)

- (void)finishInitializationWithKeyValueUnarchiver:(EOKeyValueUnarchiver *)_una;
- (void)awakeFromKeyValueUnarchiver:(EOKeyValueUnarchiver *)_unarchiver;

@end

/* delegates */

@interface NSObject(KVCArchiverDelegates)

- (id)archiver:(EOKeyValueArchiver *)_archiver
  referenceToEncodeForObject:(id)_obj;

@end

@interface NSObject(KVCUnarchiverDelegates)

- (id)unarchiver:(EOKeyValueUnarchiver *)_unarchiver
  objectForReference:(id)_obj;

@end

#endif /* __EOControl_EOKeyValueArchiver_H__ */
