// $Id: EONotQualifier.m,v 1.4 2003/02/03 14:56:45 helge Exp $

#include <EOControl/EOQualifier.h>
#include "common.h"

@interface NSObject(QualifierDescription)
- (NSString *)qualifierDescription;
@end

@interface EOQualifier(EvalContext)
- (BOOL)evaluateWithObject:(id)_object inEvalContext:(id)_ctx;
@end

@implementation EONotQualifier

- (id)initWithQualifier:(EOQualifier *)_qualifier {
  self->qualifier = [_qualifier retain];
  return self;
}

- (void)dealloc {
  [self->qualifier release];
  [super dealloc];
}

- (EOQualifier *)qualifier {
  return self->qualifier;
}

/* bindings */

- (EOQualifier *)qualifierWithBindings:(NSDictionary *)_bindings
  requiresAllVariables:(BOOL)_reqAll
{
  EOQualifier *nq;

  nq = [self->qualifier qualifierWithBindings:_bindings
                        requiresAllVariables:_reqAll];
  if (nq == nil)
    return self;

  if (nq == self->qualifier)
    return self;
  
  return [[[[self class] alloc] initWithQualifier:nq] autorelease];
}

- (NSArray *)bindingKeys {
  return [self->qualifier bindingKeys];
}

/* keys */

- (void)addQualifierKeysToSet:(NSMutableSet *)_keys {
  /* new in WO 4.5 */
  [self->qualifier addQualifierKeysToSet:_keys];
}

/* evaluation */

- (BOOL)evaluateWithObject:(id)_object inEvalContext:(id)_ctx {
  return
    [self->qualifier evaluateWithObject:_object inEvalContext:_ctx]
    ? NO : YES;
}
- (BOOL)evaluateWithObject:(id)_object {
  return [self evaluateWithObject:_object inEvalContext:nil];
}

/* NSCoding */

- (void)encodeWithCoder:(NSCoder *)_coder {
  [_coder encodeObject:self->qualifier];
}

- (id)initWithCoder:(NSCoder *)_coder {
  self->qualifier = [[_coder decodeObject] retain];
  return self;
}

/* Comparing */

- (BOOL)isEqualToQualifier:(EOQualifier *)_qual {
  return [self->qualifier isEqual:[(EONotQualifier *)_qual qualifier]];
}

/* remapping keys */

- (EOQualifier *)qualifierByApplyingTransformer:(id)_transformer
  inContext:(id)_ctx
{
  if ([_transformer respondsToSelector:
                      @selector(transformNotQualifier:inContext:)]) {
    return [_transformer transformNotQualifier:self inContext:_ctx];
  }
  else {
    EONotQualifier *nq;
    EOQualifier *q;
    
    q = [self->qualifier
             qualifierByApplyingTransformer:_transformer inContext:_ctx];
    nq = [[EONotQualifier alloc] initWithQualifier:(q ? q : self->qualifier)];
    return [nq autorelease];
  }
}

- (EOQualifier *)qualifierByApplyingKeyMap:(NSDictionary *)_map {
  EONotQualifier *nq;
  EOQualifier *q;
  
  q = [self->qualifier qualifierByApplyingKeyMap:_map];
  nq = [[EONotQualifier alloc] initWithQualifier:(q ? q : self->qualifier)];
  return [nq autorelease];
}

/* description */

- (NSString *)description {
  NSString *qd;
  
  qd = [self->qualifier qualifierDescription];
  
  return [[@"NOT (" stringByAppendingString:qd] stringByAppendingString:@")"];
}

@end /* EONotQualifier */
