// $Id: EOSortOrdering.h,v 1.3 2002/12/23 14:34:16 helge Exp $

#ifndef __EOControl_EOSortOrdering_H__
#define __EOControl_EOSortOrdering_H__

#import <Foundation/NSObject.h>
#include <EOControl/EOControlDecls.h>

@class NSDictionary, NSString;

#if NeXT_RUNTIME || APPLE_RUNTIME
#define EOCompareAscending  @selector(compareAscending:)
#define EOCompareDescending @selector(compareDescending:)
#define EOCompareCaseInsensitiveAscending  @selector(compareCaseInsensitiveAscending:)
#define EOCompareCaseInsensitiveDescending @selector(compareCaseInsensitiveDescending:)
#else
#  ifdef __APPLE__
#    warning did you define the proper runtime ? (eg APPLE_RUNTIME)
#  endif
EOControl_EXPORT SEL EOCompareAscending;
EOControl_EXPORT SEL EOCompareDescending;
EOControl_EXPORT SEL EOCompareCaseInsensitiveAscending;
EOControl_EXPORT SEL EOCompareCaseInsensitiveDescending;
#endif

@interface EOSortOrdering : NSObject
{
  NSString *key;
  SEL      selector;
}

+ (EOSortOrdering *)sortOrderingWithKey:(NSString *)_key selector:(SEL)_selector;
- (id)initWithKey:(NSString *)_key selector:(SEL)_selector;

/* propertylist initializers */

- (id)initWithDictionary:(NSDictionary *)_dict;
- (id)initWithString:(NSString *)_string;
- (id)initWithPropertyList:(id)_plist;

/* accessors */

- (NSString *)key;
- (SEL)selector;

/* remapping keys */

- (EOSortOrdering *)sortOrderingByApplyingKeyMap:(NSDictionary *)_map;

@end

#import <Foundation/NSArray.h>

@interface NSArray(EOSortOrdering)

- (NSArray *)sortedArrayUsingKeyOrderArray:(NSArray *)_orderings;

@end

@interface NSMutableArray(EOSortOrdering)

- (void)sortUsingKeyOrderArray:(NSArray *)_orderings;

@end

#import <Foundation/NSString.h>

@interface NSString(EOSortOrdering)
- (int)compareAscending:(id)_object;
- (int)compareDescending:(id)_object;
- (int)compareCaseInsensitiveAscending:(id)_object;
- (int)compareCaseInsensitiveDescending:(id)_object;
@end

#endif /* __EOControl_EOSortOrdering_H__ */
