// $Id: common.h,v 1.3 2003/04/01 13:32:30 gnustep Exp $

#ifndef __EOControl_COMMON_H__
#define __EOControl_COMMON_H__

#include <stdlib.h>
#include <objc/Protocol.h>

#import <Foundation/Foundation.h>
#import <Foundation/NSObjCRuntime.h>

#if LIB_FOUNDATION_LIBRARY
#  import <extensions/objc-runtime.h>
#  import <Foundation/exceptions/GeneralExceptions.h>
#elif GNUSTEP_BASE_LIBRARY
#  import <extensions/objc-runtime.h>
#  import <extensions/exceptions/GeneralExceptions.h>
#else
#  import <FoundationExt/objc-runtime.h>
#  import <FoundationExt/GeneralExceptions.h>
#  import <FoundationExt/MissingMethods.h>
#  import <FoundationExt/NSString+Ext.h>
#endif

#if NeXT_RUNTIME || APPLE_RUNTIME
#  define objc_free(__mem__)    free(__mem__)
#  define objc_malloc(__size__) malloc(__size__)
#  define objc_calloc(__cnt__, __size__) calloc(__cnt__, __size__)
#  define objc_realloc(__ptr__, __size__) realloc(__ptr__, __size__)
#  ifndef sel_eq
#    define sel_eq(sela,selb) (sela==selb?YES:NO)
#  endif
#endif

#ifndef ASSIGN
#  define ASSIGN(object, value) \
       ({id __object = (id)object;    \
         id __value = (id)value;      \
         if (__value != __object) { if (__value) [__value retain]; \
          if (__object) [__object release]; \
          object = __value;}})
#endif
#ifndef ASSIGNCOPY
#  define ASSIGNCOPY(object, value) \
       ({id __object = (id)object;    \
         id __value = (id)value;      \
         if (__value != __object) { if (__value) __value = [__value copy];   \
          if (__object) [__object release]; \
          object = __value;}})
#endif

// ******************** common functions ********************

static inline void *Malloc(int) __attribute__((unused));
static inline void *Calloc(int, int) __attribute__((unused));
static inline void *Realloc(void*, int) __attribute__((unused));

#if !LIB_FOUNDATION_BOEHM_GC
static inline void Free(void*) __attribute__((unused));
#endif

static inline int Strlen(const char*) __attribute__((unused));
static inline char* Strdup(const char*) __attribute__((unused));
static inline char* Strcpy (char*, const char*) __attribute__((unused));
static inline char* Strncpy (char*, const char*, unsigned)
    __attribute__((unused));
static inline char* Strcat (char*, const char*) __attribute__((unused));
static inline char* Strncat (char*, const char*, unsigned)
    __attribute__((unused));
static inline int Strcmp(const char*, const char*) __attribute__((unused));
static inline int Strncmp(const char*, const char*, unsigned)
    __attribute__((unused));
static inline int Atoi(const char*) __attribute__((unused));
static inline long Atol(const char*) __attribute__((unused));

#if LIB_FOUNDATION_LIBRARY

#if LIB_FOUNDATION_BOEHM_GC

# include <gc.h>

extern void __raiseMemoryException (void* pointer, int size);


static inline void *Malloc(int size)
{
    void* p = (void*)GC_MALLOC (size);
    if(!p)
        __raiseMemoryException (&p, size);
    return p;
}

static inline void *MallocAtomic(int size)
{
    void* p = (void*)GC_MALLOC_ATOMIC (size);
    if(!p)
        __raiseMemoryException (&p, size);
    return p;
}

static inline void Free(void* p)
{}

static inline void *Calloc(int elem, int size)
{
    int howMuch = elem * size;
    void* p = (void*)GC_MALLOC (howMuch);

    if(!p)
        __raiseMemoryException (&p, howMuch);
    memset (p, 0, howMuch);
    return p;
}

static inline void *CallocAtomic(int elem, int size)
{
    int howMuch = elem * size;
    void* p = (void*)GC_MALLOC_ATOMIC (howMuch);

    if(!p)
        __raiseMemoryException (&p, howMuch);
    memset (p, 0, howMuch);
    return p;
}

static inline void *Realloc(void* p, int size)
{
    void* new_p = GC_REALLOC (p, size);

    if(!new_p)
        __raiseMemoryException (&new_p, size);
    return new_p;
}

#else /* !LIB_FOUNDATION_BOEHM_GC */

#import <Foundation/NSZone.h>

static inline void *Malloc(int size)
{
    return [[NSZone defaultZone] malloc:size];
}

static inline void *MallocAtomic(int size)
{
    return [[NSZone defaultZone] mallocAtomic:size];
}

static inline void Free(void* p)
{
    return [[NSZone zoneFromPointer:p] freePointer:p];
}

static inline void *Calloc(int elem, int size)
{
    return [[NSZone defaultZone] calloc:elem byteSize:size];
}

static inline void *CallocAtomic(int elem, int size)
{
    return [[NSZone defaultZone] callocAtomic:elem byteSize:size];
}

static inline void *Realloc(void* p, int size)
{
    return [[NSZone zoneFromPointer:p] realloc:p size:size];
}

#endif /* !LIB_FOUNDATION_BOEHM_GC */


#else /* ! LIB_FOUNDATION_LIBRARY */

static inline void *Malloc(int size)
{
    return objc_malloc(size);
}

static inline void Free(void* p)
{
  if(p) { objc_free(p); p = NULL; }
}

static inline void *Calloc(int elem, int size)
{
    return objc_calloc(elem, size);
}

static inline void *Realloc(void* p, int size)
{
    return objc_realloc(p, size);
}

#endif /* ! LIB_FOUNDATION_LIBRARY */

static inline int Strlen(const char* s)
{
    return s ? strlen(s) : 0;
}

static inline char* Strdup(const char* s)
{
    return s ? strcpy(Malloc(strlen(s) + 1), s) : NULL;
}

static inline char* Strcpy (char* d, const char* s)
{
    return s && d ? strcpy(d, s) : d;
}

static inline char* Strncpy (char* d, const char* s, unsigned size)
{
    return s && d ? strncpy(d, s, size) : d;
}

static inline char* Strcat (char* d, const char* s)
{
    return s && d ? strcat(d, s) : d;
}

static inline char* Strncat (char* d, const char* s , unsigned size)
{
    return s && d ? strncat(d, s , size) : d;
}

static inline int Strcmp(const char* p, const char* q)
{
    if(!p) {
        if(!q)
            return 0;
        else return -1;
    }
    else {
        if(!q)
            return 1;
        else return strcmp(p, q);
    }
}

static inline int Strncmp(const char* p, const char* q, unsigned size)
{
    if(!p) {
        if(!q)
            return 0;
        else return -1;
    }
    else {
        if(!q)
            return 1;
        else return strncmp(p, q, size);
    }
}

static inline int Atoi(const char* str)
{
    return str ? atoi(str) : 0;
}

static inline long Atol(const char *str)
{
  return str ? atol(str) : 0;
}

#ifndef MAX
#define MAX(a, b) \
    ({typedef _ta = (a), _tb = (b);   \
	_ta _a = (a); _tb _b = (b);     \
	_a > _b ? _a : _b; })
#endif

#ifndef MIN
#define MIN(a, b) \
    ({typedef _ta = (a), _tb = (b);   \
	_ta _a = (a); _tb _b = (b);     \
	_a < _b ? _a : _b; })
#endif

#if !LIB_FOUNDATION_LIBRARY

static inline char *Ltoa(long nr, char *str, int base)
{
    char buff[34], rest, is_negative;
    int ptr;

    ptr = 32;
    buff[33] = '\0';
    if(nr < 0) {
	is_negative = 1;
	nr = -nr;
    }
    else
	is_negative = 0;

    while(nr != 0) {
	rest = nr % base;
	if(rest > 9)
	    rest += 'A' - 10;
	else
	    rest += '0';
	buff[ptr--] = rest;
	nr /= base;
    }
    if(ptr == 32)
	buff[ptr--] = '0';
    if(is_negative)
	buff[ptr--] = '-';

    Strcpy(str, &buff[ptr+1]);

    return(str);
}
#endif

#endif /* __EOControl_COMMON_H__ */
