// $Id: EOFilterDataSource.m,v 1.2 2003/01/07 11:33:35 helge Exp $

#include "EOFilterDataSource.h"
#import <NGExtensions/EODataSource+NGExtensions.h>
#import <EOControl/EOControl.h>
#import <NGExtensions/EOGrouping.h>
#import "common.h"

@interface NSDictionary(EOFilterDataSource)

- (NSArray *)flattenedArrayWithHint:(unsigned int)_hint
  andKeys:(NSArray *)_keys;

@end

@implementation EOFilterDataSource

- (id)initWithDataSource:(EODataSource *)_ds {
  if ((self = [super init])) {
    [self setSource:_ds];
  }
  return self;
}
- (id)init {
  return [self initWithDataSource:nil];
}

- (void)dealloc {
  [[NSNotificationCenter defaultCenter] removeObserver:self];
  [self->sortOrderings      release];
  [self->groupings          release];
  [self->auxiliaryQualifier release];
  [self->source             release];
  [super dealloc];
}

/* accessors */

- (void)setSource:(EODataSource *)_source {
  if (self->source != _source) {
    NSNotificationCenter *nc;
    
    nc = [NSNotificationCenter defaultCenter];

    if (self->source) {
      [nc removeObserver:self
          name:EODataSourceDidChangeNotification object:self->source];
    }
    
    ASSIGN(self->source, _source);
    
    if (self->source) {
      [nc addObserver:self selector:@selector(_sourceDidChange:)
          name:EODataSourceDidChangeNotification object:self->source];
    }
    
    [self postDataSourceChangedNotification];
  }
}
- (EODataSource *)source {
  return self->source;
}

- (void)setAuxiliaryQualifier:(EOQualifier *)_q {
  if (![_q isEqual:self->auxiliaryQualifier]) {
    ASSIGN(self->auxiliaryQualifier, _q);
    [self postDataSourceChangedNotification];
  }
}
- (EOQualifier *)auxiliaryQualifier {
  return self->auxiliaryQualifier;
}

- (void)setSortOrderings:(NSArray *)_so {
  if (self->sortOrderings != _so) {
    _so = [_so shallowCopy];
    [self->sortOrderings release];
    self->sortOrderings = _so;
    [self postDataSourceChangedNotification];
  }
}
- (NSArray *)sortOrderings {
  return self->sortOrderings;
}

- (void)setGroupings:(NSArray *)_groupings {
  if (self->groupings != _groupings) {
    _groupings = [_groupings shallowCopy];
    [self->groupings release];
    self->groupings = _groupings;
    [self postDataSourceChangedNotification];  
  }
}
- (NSArray *)groupings {
  return self->groupings;
}

- (void)setFetchSpecification:(EOFetchSpecification *)_fspec {
  [[self source] setFetchSpecification:_fspec];
}
- (EOFetchSpecification *)fetchSpecification {
  return [[self source] fetchSpecification];
}

/* notifications */

- (void)_sourceDidChange:(NSNotification *)_notification {
  [self postDataSourceChangedNotification];
}

/* operations */

- (NSArray *)fetchObjects {
  NSAutoreleasePool *pool;
  NSArray *objs;
  NSArray *groups;

  pool = [[NSAutoreleasePool alloc] init];
  
  objs = [[self source] fetchObjects];
  
  if ([self auxiliaryQualifier])
    objs = [objs filteredArrayUsingQualifier:[self auxiliaryQualifier]];

  if ((groups = [self groupings]) != nil) {
    unsigned int cnt;
    EOGrouping   *grouping;
    NSArray      *allKeys;
    NSArray      *sos;
    NSDictionary *groupDict;

    cnt = [objs count];

    grouping = [groups lastObject];
    
    if ((sos = [self sortOrderings]) != nil) {
      [grouping setSortOrderings:sos];
    }
    groupDict = [objs arrayGroupedBy:grouping];

    allKeys = [groupDict allKeys];
    allKeys = [allKeys sortedArrayUsingSelector:@selector(compare:)];
    objs    = [groupDict flattenedArrayWithHint:cnt andKeys:allKeys];
  }
  else if ([self sortOrderings])
    objs = [objs sortedArrayUsingKeyOrderArray:[self sortOrderings]];
  
  objs = [objs copy];
  [pool release];
  
  return [objs autorelease];
}

- (void)insertObject:(id)_obj {
  [[self source] insertObject:_obj];
}

- (void)deleteObject:(id)_obj {
  [[self source] deleteObject:_obj];
}

- (void)updateObject:(id)_obj {
  [self->source updateObject:_obj];
}

- (id)createObject {
  return [[self source] createObject];
}

- (EOClassDescription *)classDescriptionForObjects {
  return [[self source] classDescriptionForObjects];
}

@end /* EOFilterDataSource */

@implementation NSDictionary(EOFilterDataSource)

- (NSArray *)flattenedArrayWithHint:(unsigned int)_hint
  andKeys:(NSArray *)_keys
{
  NSMutableArray *result = nil;
  unsigned int   i, cnt;
  
  result =
    [[NSMutableArray alloc] initWithCapacity:_hint]; // should be improved
  
  for (i = 0, cnt = [_keys count]; i < cnt; i++) {
    NSString *key;
    NSArray  *tmp;
    
    key = [_keys objectAtIndex:i];
    tmp = [self objectForKey:key];
    [result addObjectsFromArray:tmp];
  }
  
  return [result autorelease];
}

@end /* NSDictionary(EOFilterDataSource) */
