// $Id: EOGrouping.m,v 1.2 2003/01/07 11:33:35 helge Exp $

#include <NGExtensions/EOGrouping.h>
#include "common.h"

@implementation EOGrouping

static BOOL ProfileComponents = NO;

- (id)initWithDefaultName:(NSString *)_defaultName {
  if ((self = [super init])) {
    ASSIGN(self->defaultName, _defaultName);
  }
  return self;
}

- (id)init {
  return [self initWithDefaultName:nil];
}

- (void)dealloc {
  [self->defaultName   release];
  [self->sortOrderings release];
  
  [super dealloc];
}

// accessors
- (void)setDefaultName:(NSString *)_defaultName {
  ASSIGN(self->defaultName, _defaultName);
}
- (NSString *)defaultName {
  return self->defaultName;
}

- (void)setSortOrderings:(NSArray *)_sortOrderings {
  ASSIGN(self->sortOrderings, _sortOrderings);
}

- (NSArray *)sortOrderings {
  return self->sortOrderings;
}

// ------------------

- (NSString *)groupNameForObject:(id)_object {
  [self doesNotRecognizeSelector:_cmd]; // subclass
  return nil;
}

- (NSArray *)orderedGroupNames {
  [self doesNotRecognizeSelector:_cmd]; // subclass
  return nil;
}

- (NSString *)description {
  return @"EOGrouping";
}

@end /* EOGrouping */

#include <EOControl/EOSortOrdering.h>

@implementation NSArray(EOGrouping)

- (NSDictionary *)arrayGroupedBy:(EOGrouping *)_grouping {
  NSMutableDictionary *result;
  NSZone              *zone;
  NSEnumerator        *keyEnum;
  NSString            *key;
  NSArray             *sortings;
  int                 i, cnt;
  IMP                 objAtIndex;
  IMP                 groupForObj;
  NSTimeInterval      st = 0.0;
  
  if (ProfileComponents)
    st = [[NSDate date] timeIntervalSince1970];

  zone   = [self zone];
  cnt    = [self count];
  result = [[NSMutableDictionary allocWithZone:zone] initWithCapacity:cnt];

  objAtIndex  = [self methodForSelector:@selector(objectAtIndex:)];
  groupForObj = [_grouping methodForSelector:@selector(groupNameForObject:)];
  
  for (i = 0; i < cnt; i++) {
    NSString       *gName = nil; // groupName
    NSMutableArray *tmp   = nil;
    id   obj              = nil;

    obj   = objAtIndex(self, @selector(objectAtIndex:), i);
    gName = groupForObj(_grouping, @selector(groupNameForObject:), obj);

    if (gName == nil) continue;

    if (!(tmp = [result objectForKey:gName])) {
      tmp = [[[NSMutableArray alloc] initWithCapacity:4] autorelease];
      [result setObject:tmp forKey:gName];
    }
    [tmp addObject:obj];
  }
  
  sortings = [_grouping sortOrderings];

  if ([sortings count] > 0) {
    // sort each group
    keyEnum = [result keyEnumerator];
    while ((key = [keyEnum nextObject])) {
      NSArray *tmp;
    
      tmp = [result objectForKey:key];
      tmp = [tmp  sortedArrayUsingKeyOrderArray:sortings];
      [result setObject:tmp forKey:key];
    }
  }
  
  if (ProfileComponents) {
    NSTimeInterval diff;
    diff = [[NSDate date] timeIntervalSince1970] - st;
    
    printf("NSArray+Grouping: %0.4fs\n", diff);
  }
  
  return [result autorelease];
}

@end /* NSArray(EOGrouping) */

NSString *EOGroupingHint = @"EOGroupingHint";

@implementation EOFetchSpecification(Groupings)

- (void)setGroupings:(NSArray *)_groupings {
  NSDictionary        *lhints;
  NSMutableDictionary *md;

  lhints = [self hints];
  md = lhints ? [lhints mutableCopy] : [[NSMutableDictionary alloc] init];
  if (_groupings)
    [md setObject:_groupings forKey:EOGroupingHint];
  else
    [md removeObjectForKey:EOGroupingHint];
  lhints = [md copy];
  [md release];
  [self setHints:lhints];
  [lhints release];
}
- (NSArray *)groupings {
  return [[self hints] objectForKey:EOGroupingHint];
}

@end /* EOFetchSpecification(Groupings) */
