/* 
   NGBitSet.h

   Copyright (C) 2002 SKYRIX Software AG, Helge Hess
   All rights reserved.

   Author: Helge Hess (helge.hess@skyrix.com)

   This file is part of the SKYRiX Core Framework

   Permission to use, copy, modify, and distribute this software and its
   documentation for any purpose and without fee is hereby granted, provided
   that the above copyright notice appear in all copies and that both that
   copyright notice and this permission notice appear in supporting
   documentation.

   We disclaim all warranties with regard to this software, including all
   implied warranties of merchantability and fitness, in no event shall
   we be liable for any special, indirect or consequential damages or any
   damages whatsoever resulting from loss of use, data or profits, whether in
   an action of contract, negligence or other tortious action, arising out of
   or in connection with the use or performance of this software.
*/
// $Id: NGBitSet.h,v 1.2 2002/12/27 09:42:32 helge Exp $

#ifndef __NGExtensions_NGBitSet_H__
#define __NGExtensions_NGBitSet_H__

#import <Foundation/NSObject.h>
#import <Foundation/NSRange.h>

@class NSArray, NSEnumerator;

typedef unsigned int NGBitSetStorage;

@protocol NGBitSet < NSObject >

// state

- (unsigned int)count;

// membership

- (BOOL)isMember:(unsigned int)_element;
- (void)addMember:(unsigned int)_element;
- (void)addMembersInRange:(NSRange)_range;
- (void)removeMember:(unsigned int)_element;
- (void)removeMembersInRange:(NSRange)_range;
- (void)removeAllMembers;

@end

@interface NGBitSet : NSObject < NGBitSet, NSCopying, NSCoding >
{
@protected
  unsigned int    universe;
  unsigned int    count;
  NGBitSetStorage *storage;
}

+ (id)bitSet;
+ (id)bitSetWithCapacity:(unsigned)_capacity;
+ (id)bitSetWithBitSet:(NGBitSet *)_set;
- (id)init;
- (id)initWithCapacity:(unsigned)_capacity; // designated initializer
- (id)initWithBitSet:(NGBitSet *)_set;
- (id)initWithNullTerminatedArray:(unsigned int *)_array;

// state

- (unsigned int)capacity;

// membership

- (unsigned int)firstMember;
- (unsigned int)lastMember;
- (void)addMembersFromBitSet:(NGBitSet *)_set;

// equality

- (BOOL)isEqual:(id)_object;
- (BOOL)isEqualToSet:(NGBitSet *)_set;

// enumerator

- (NSEnumerator *)objectEnumerator;

// NSCopying

- (id)copy;
- (id)copyWithZone:(NSZone *)_zone;

// NSCoding

- (void)encodeWithCoder:(NSCoder *)_coder;
- (id)initWithCoder:(NSCoder *)_coder;

// description

- (NSString *)description;
- (NSArray *)toArray;

@end

NSString *stringValueForBitset(unsigned int _set, char _setC, char _unsetC,
                               short _wide);

#endif /* __NGExtensions_NGBitSet_H__ */
