/* 
   NGCString.h

   Copyright (C) 2002 SKYRIX Software AG, Helge Hess
   All rights reserved.

   Author: Helge Hess (helge.hess@skyrix.com)

   This file is part of the SKYRiX Core Framework

   Permission to use, copy, modify, and distribute this software and its
   documentation for any purpose and without fee is hereby granted, provided
   that the above copyright notice appear in all copies and that both that
   copyright notice and this permission notice appear in supporting
   documentation.

   We disclaim all warranties with regard to this software, including all
   implied warranties of merchantability and fitness, in no event shall
   we be liable for any special, indirect or consequential damages or any
   damages whatsoever resulting from loss of use, data or profits, whether in
   an action of contract, negligence or other tortious action, arising out of
   or in connection with the use or performance of this software.
*/
// $Id: NGCString.h,v 1.2 2002/12/27 09:42:32 helge Exp $

#ifndef __NGExtensions_NGCString_H__
#define __NGExtensions_NGCString_H__

#import <Foundation/NSObject.h>

@class NSString;

@interface NGCString : NSObject < NSCopying, NSMutableCopying, NSCoding >
{
  char         *value;
  unsigned int len;
}

+ (id)stringWithCString:(const char *)_value length:(unsigned)_len;
+ (id)stringWithCString:(const char *)_value;
- (id)initWithCString:(const char *)_value length:(unsigned)_len;
- (id)initWithCString:(const char *)_value;

// unicode

- (unsigned int)length;
- (unichar)characterAtIndex:(unsigned int)_idx;

// comparing

- (BOOL)isEqual:(id)_object;
- (BOOL)isEqualToCString:(NGCString *)_cstr;
- (BOOL)isEqualToString:(NSString *)_str;
- (unsigned)hash;

// copying

- (id)copyWithZone:(NSZone *)_zone;
- (id)mutableCopyWithZone:(NSZone *)_zone;

// archiving

- (void)encodeWithCoder:(NSCoder *)_coder;
- (id)initWithCoder:(NSCoder *)_decoder;

// getting C strings

- (const char *)cString;
- (unsigned int)cStringLength;
- (void)getCString:(char *)_buffer;
- (void)getCString:(char *)_buffer maxLength:(unsigned int)_maxLength;
- (char)charAtIndex:(unsigned int)_idx;

// getting numeric values

- (double)doubleValue;
- (float)floatValue;
- (int)intValue;
- (NSString *)stringValue;

// description

- (NSString *)description;

@end

@interface NGMutableCString : NGCString
{
  unsigned int capacity;
}

// appending

- (void)appendString:(id)_str; // either NSString or NGCString
- (void)appendCString:(const char *)_cstr;
- (void)appendCString:(const char *)_cstr length:(unsigned)_len;

// removing

- (void)removeAllContents;

@end

#endif /* __NGExtensions_NGCString_H__ */
