// $Id: NGCustomFileManager.h,v 1.1.1.1 2002/05/22 17:27:16 helge Exp $

#ifndef __NGCustomFileManager_H__
#define __NGCustomFileManager_H__

#include <NGExtensions/NGFileManager.h>

/*
  An abstract baseclass for developing custom filemanagers which are ideally
  based on other filemanager classes.
*/

@class NGCustomFileManagerInfo;

@interface NGCustomFileManager : NGFileManager
{
}

/* customization */

- (NSString *)makeAbsolutePath:(NSString *)_path;
- (NGCustomFileManagerInfo *)fileManagerInfoForPath:(NSString *)_path;

@end

@interface NGCustomFileManager(NGFileManagerVersioning)

/* versioning */

- (BOOL)checkoutFileAtPath:(NSString *)_path handler:(id)_handler;
- (BOOL)releaseFileAtPath:(NSString *)_path  handler:(id)_handler;
- (BOOL)rejectFileAtPath:(NSString *)_path   handler:(id)_handler;
- (BOOL)checkoutFileAtPath:(NSString *)_path version:(NSString *)_version
  handler:(id)_handler;

/* versioning data */

- (NSString *)lastVersionAtPath:(NSString *)_path;
- (NSArray *)versionsAtPath:(NSString *)_path;
- (NSData *)contentsAtPath:(NSString *)_path version:(NSString *)_version;
- (NSDictionary *)fileAttributesAtPath:(NSString *)_path
  traverseLink:(BOOL)_followLink
  version:(NSString *)_version;

@end

@interface NGCustomFileManager(NGFileManagerDataSources)

/* datasources (work on folders) */

- (EODataSource *)dataSourceAtPath:(NSString *)_path;
- (EODataSource *)dataSource; // works on current-directory-path

@end

@interface NGCustomFileManager(NGFileManagerLocking)

- (BOOL)lockFileAtPath:(NSString *)_path handler:(id)_handler;
- (BOOL)unlockFileAtPath:(NSString *)_path handler:(id)_handler;
- (BOOL)isFileLockedAtPath:(NSString *)_path;

/* access rights */
- (BOOL)isLockableFileAtPath:(NSString *)_path;
- (BOOL)isUnlockableFileAtPath:(NSString *)_path;

@end

@interface NGCustomFileManagerInfo : NSObject
{
@private
  NGCustomFileManager        *master;      /* non retained */
  id<NGFileManager,NSObject> fileManager;
}

- (id)initWithCustomFileManager:(NGCustomFileManager *)_master
  fileManager:(id<NGFileManager,NSObject>)_fm;

- (void)resetMaster;

/* accessors */

- (NGCustomFileManager *)master;
- (id<NGFileManager,NSObject>)fileManager;

/* operations */

- (NSString *)rewriteAbsolutePath:(NSString *)_path;

/* capabilities */

- (BOOL)supportsGlobalIDs;

@end

#endif /* __NGCustomFileManager_H__ */
