/* 
   NGHashMap.h

   Copyright (C) 2002 SKYRIX Software AG, Helge Hess
   All rights reserved.

   Author: Jan Reichmann (jan.reichmann@skyrix.com)
           Helge Hess (helge.hess@skyrix.com)

   This file is part of the SKYRiX Core Framework

   Permission to use, copy, modify, and distribute this software and its
   documentation for any purpose and without fee is hereby granted, provided
   that the above copyright notice appear in all copies and that both that
   copyright notice and this permission notice appear in supporting
   documentation.

   We disclaim all warranties with regard to this software, including all
   implied warranties of merchantability and fitness, in no event shall
   we be liable for any special, indirect or consequential damages or any
   damages whatsoever resulting from loss of use, data or profits, whether in
   an action of contract, negligence or other tortious action, arising out of
   or in connection with the use or performance of this software.
*/
// $Id: NGHashMap.h,v 1.2 2002/12/27 09:42:33 helge Exp $

#ifndef __NGExtensions_NGHashMap_H__
#define __NGExtensions_NGHashMap_H__

#import <Foundation/NSObject.h>
#import <Foundation/NSMapTable.h>

@class NSArray;

@interface NGHashMap : NSObject < NSCopying, NSMutableCopying, NSCoding >
{
@protected
  NSMapTable *table;  
}

+ (id)hashMap;
+ (id)hashMapWithHashMap:(NGHashMap *)_hashMap;
+ (id)hashMapWithObjects:(NSArray *)_objects forKey:(id)_key;
+ (id)hashMapWithDictionary:(NSDictionary *)_dict;

- (id)init;
- (id)initWithCapacity:(unsigned int)_size;
- (id)initWithObjects:(NSArray *)_objects forKey:(id)_key;
- (id)initWithHashMap:(NGHashMap *)_hashMap;
- (id)initWithDictionary:(NSDictionary *)_dictionary;

- (BOOL)isEqual:(id)anObject;
- (BOOL)isEqualToHashMap:(NGHashMap *)_other;

- (id)objectForKey:(id)_key;
- (NSArray *)objectsForKey:(id)_key;
- (id)objectAtIndex:(unsigned int)_index forKey:(id)_key;

- (NSArray *)allKeys;
- (NSArray *)allObjects;

- (NSEnumerator *)keyEnumerator;
- (NSEnumerator *)objectEnumerator;
- (NSEnumerator *)objectEnumeratorForKey:(id)_key;

- (id)propertyList;
- (NSString *)description;
- (NSDictionary *)asDictionary;

- (unsigned int)hash;
- (unsigned int)count; // returns the number of keys
- (unsigned int)countObjectsForKey:(id)_key;

@end

@interface NGMutableHashMap : NGHashMap
{
}

+ (id)hashMapWithCapacity:(unsigned int)_numItems;

- (id)init;
 
- (void)insertObject:(id)_object atIndex:(unsigned int)_index forKey:(id)_key;
- (void)insertObjects:(NSArray *)_object
  atIndex:(unsigned int)_index forKey:(id)_key;
- (void)insertObjects:(id*)_objects count:(unsigned int)_count
  atIndex:(unsigned int)_index forKey:(id)_key;

- (void)addObject:(id)_object forKey:(id)_key;
- (void)addObjects:(NSArray *)_objects forKey:(id)_key;
- (void)addObjects:(id*)_objects count:(unsigned int)_count
  forKey:(id)_key;

- (void)setObject:(id)_object forKey:(id)_key;
- (void)setObjects:(NSArray *)_objects forKey:(id)_key;

- (void)removeAllObjects;
- (void)removeAllObjects:(id)_object forKey:(id)_key;
- (void)removeAllObjectsForKey:(id)_key;
- (void)removeAllObjectsForKeys:(NSArray *)_keyArray;

@end

#endif /* __NGExtensions_NGHashMap_H__ */
