/* 
   NGQuotedPrintableCoding.h

   Copyright (C) 2002 SKYRIX Software AG, Helge Hess
   All rights reserved.

   Author: Jan Reichman (jr@mdlink.de)
           Helge Hess (helge.hess@mdlink.de)

   This file is part of the SKYRiX Core Framework

   Permission to use, copy, modify, and distribute this software and its
   documentation for any purpose and without fee is hereby granted, provided
   that the above copyright notice appear in all copies and that both that
   copyright notice and this permission notice appear in supporting
   documentation.

   We disclaim all warranties with regard to this software, including all
   implied warranties of merchantability and fitness, in no event shall
   we be liable for any special, indirect or consequential damages or any
   damages whatsoever resulting from loss of use, data or profits, whether in
   an action of contract, negligence or other tortious action, arising out of
   or in connection with the use or performance of this software.
*/
// $Id: NGQuotedPrintableCoding.h,v 1.2 2002/12/27 09:43:13 helge Exp $

#ifndef __NGExtensions_NGQuotedPrintableCoding_H__
#define __NGExtensions_NGQuotedPrintableCoding_H__

#import <Foundation/NSString.h>
#import <Foundation/NSData.h>
#include <NGExtensions/NGExtensionsDecls.h>

/*
  Quoted Printable encoder/decoder

  As specified in RFC 822
*/

@interface NSString(QuotedPrintableCoding)

- (NSString *)stringByDecodingQuotedPrintable;
- (NSString *)stringByEncodingQuotedPrintable;

@end

@interface NSData(QuotedPrintableCoding)

/*
  Decode a quoted printable encoded data. Returns nil if decoding failed.
*/
- (NSData *)dataByDecodingQuotedPrintable;

/*
  Decode data in quoted printable encoding. Returns nil if encoding failed.
*/
- (NSData *)dataByEncodingQuotedPrintable;

@end

NGExtensions_EXPORT int
NGEncodeQuotedPrintable(const char *_src, unsigned _srcLen,
                        char *_dest, unsigned _destLen);
NGExtensions_EXPORT int
NGDecodeQuotedPrintable(const char *_src, unsigned _srcLen,
                        char *_dest, unsigned _destLen);

#endif /* __NGExtensions_NGQuotedPrintableCoding_H__ */
