/* 
   NGStack.h

   Copyright (C) 2002 SKYRIX Software AG, Helge Hess
   All rights reserved.

   Author: Helge Hess (helge.hess@skyrix.com)

   This file is part of the SKYRiX Core Framework

   Permission to use, copy, modify, and distribute this software and its
   documentation for any purpose and without fee is hereby granted, provided
   that the above copyright notice appear in all copies and that both that
   copyright notice and this permission notice appear in supporting
   documentation.

   We disclaim all warranties with regard to this software, including all
   implied warranties of merchantability and fitness, in no event shall
   we be liable for any special, indirect or consequential damages or any
   damages whatsoever resulting from loss of use, data or profits, whether in
   an action of contract, negligence or other tortious action, arising out of
   or in connection with the use or performance of this software.
*/
// $Id: NGStack.h,v 1.2 2002/12/27 09:43:14 helge Exp $

#ifndef __NGExtensions_NGStack_H__
#define __NGExtensions_NGStack_H__

#import <Foundation/NSObject.h>
#import <Foundation/NSException.h>
#import <Foundation/NSArray.h>

@class NSArray;

@protocol NGStack < NSObject >

// state

- (unsigned int)stackPointer;
- (unsigned int)count;
- (BOOL)isEmpty;

// operations

- (void)push:(id)_obj;
- (id)pop;
- (void)clear;

// elements

- (id)elementAtTop;
- (NSEnumerator *)topDownEnumerator;
- (NSEnumerator *)bottomUpEnumerator;

@end

@interface NGStack : NSObject < NGStack, NSCoding, NSCopying >
{
@protected
  unsigned int stackPointer;
  unsigned int capacity;
  id           *stack;
}

+ (id)stackWithCapacity:(unsigned int)_capacity;
+ (id)stack;
+ (id)stackWithArray:(NSArray *)_array;
- (id)init;
- (id)initWithCapacity:(unsigned int)_capacity; // designated initializer
- (id)initWithArray:(NSArray *)_array;

// state

- (unsigned int)capacity;

- (unsigned int)stackPointer;
- (unsigned int)count;
- (BOOL)isEmpty;

// elements

- (id)elementAtTop;
- (id)elementAtBottom;
- (NSEnumerator *)topDownEnumerator;
- (NSEnumerator *)bottomUpEnumerator;

// operations

- (void)push:(id)_obj;
- (id)pop;
- (void)clear;

// description

- (NSArray *)toArray; // array representation, bottom element first

@end

@interface NGStackException : NSException
@end

@interface NSMutableArray(Stack) < NGStack >
@end

#endif /* __NGExtensions_NGStack_H__ */
