/* 
   NGStringScanEnumerator.h

   Copyright (C) 2002 SKYRIX Software AG, Helge Hess
   All rights reserved.

   Author: Bjoern Stierand (bs@skyrix.com)
   
   This file is part of the skyrix-core-42 package.
   
   Permission to use, copy, modify, and distribute this software and its
   documentation for any purpose and without fee is hereby granted, provided
   that the above copyright notice appear in all copies and that both that
   copyright notice and this permission notice appear in supporting
   documentation.

   We disclaim all warranties with regard to this software, including all
   implied warranties of merchantability and fitness, in no event shall
   we be liable for any special, indirect or consequential damages or any
   damages whatsoever resulting from loss of use, data or profits, whether in
   an action of contract, negligence or other tortious action, arising out of
   or in connection with the use or performance of this software.
*/
// $Id: NGStringScanEnumerator.h,v 1.1 2002/10/17 14:19:58 helge Exp $

#ifndef __NGExtensions_NGStringScanEnumerator_H__
#define __NGExtensions_NGStringScanEnumerator_H__

#import <Foundation/NSEnumerator.h>
#import <Foundation/NSData.h>

/*
  NGStringScanEnumerator

  This class is used to scan (binary) NSData objects for printable strings,
  pretty much like the "strings" Unix program.
  
  Example:
  
    NSData *data = [NSData dataWithContentsOfMappedFile:@"/bin/ls"];
    NSEnumerator *e;
    NSString *s;
    
    e = [data stringScanEnumerator];
    while ((s = [e nextObject])) {
      if ([s hasPrefix:@"4"] && [s length] > 5) {
        NSLog(@"ls version: %@", s);
        break;
      }
    }
*/

@interface NGStringScanEnumerator : NSEnumerator
{
  unsigned int curPos;
  NSData       *data;
  unsigned int maxLength;
}

+ (id)enumeratorWithData:(NSData *)_data maxLength:(unsigned int)_maxLength;

@end /* StringEnumerator */

@interface NSData(NGStringScanEnumerator)

/* scan strings with up to 256 chars length */
- (NSEnumerator *)stringScanEnumerator;

- (NSEnumerator *)stringScanEnumeratorWithMaxStringLength:(unsigned int)_max;

@end

#endif /* __NGExtensions_NGStringScanEnumerator_H__ */
