/* 
   NGStringScanEnumerator.m

   Copyright (C) 2002 SKYRIX Software AG, Helge Hess
   All rights reserved.

   Author: Bjoern Stierand (bs@skyrix.com)
   
   This file is part of the skyrix-core-42 package.
   
   Permission to use, copy, modify, and distribute this software and its
   documentation for any purpose and without fee is hereby granted, provided
   that the above copyright notice appear in all copies and that both that
   copyright notice and this permission notice appear in supporting
   documentation.

   We disclaim all warranties with regard to this software, including all
   implied warranties of merchantability and fitness, in no event shall
   we be liable for any special, indirect or consequential damages or any
   damages whatsoever resulting from loss of use, data or profits, whether in
   an action of contract, negligence or other tortious action, arising out of
   or in connection with the use or performance of this software.
*/
// $Id: NGStringScanEnumerator.m,v 1.2 2002/10/21 14:27:04 helge Exp $

#include "NGStringScanEnumerator.h"
#include "common.h"
#include <ctype.h>

@implementation NGStringScanEnumerator

- (id)initWithData:(NSData *)_data maxLength:(unsigned int)_length {
  if (_data == nil) {
    [self release];
    return nil;
  }
  
  if ((self = [super init])) {
    self->data      = [_data retain];
    self->curPos    = 0;
    self->maxLength = _length;
  }
  return self;
}

+ (id)enumeratorWithData:(NSData *)_data maxLength:(unsigned int)_length {
  return [[[self alloc] initWithData:_data maxLength:_length] autorelease];
}

- (void)dealloc {
  [self->data release];
  [super dealloc];
}

- (NSString *)nextObject {
  register int i;
  const unsigned char *bytes;
  unsigned int length;
  register int startPos = -1;
  
  bytes  = [self->data bytes];
  length = [self->data length];

  if (length == 0) {
    [self->data release]; self->data = nil;
    return nil;
  }
  
  for (i = self->curPos; i < length; i++) {

    if (isprint(bytes[i])) {
      if (startPos == -1)
        startPos = i;
    }
    else {
      if (startPos != -1) {
        if ((i - startPos) >= self->maxLength) {
          self->curPos = i;
          
          return [NSString stringWithCString:(bytes + startPos)
                           length:(i - startPos)];
        }
        startPos = -1;
      }
    }
  }
  /* end reached (can release data) */
  [self->data release]; self->data = nil;
  return nil;
}

@end /* NGStringScanEnumerator */

@implementation NSData(NGStringScanEnumerator)

- (NSEnumerator *)stringScanEnumeratorWithMaxStringLength:(unsigned int)_max {
  return [NGStringScanEnumerator enumeratorWithData:self maxLength:_length];
}

- (NSEnumerator *)stringScanEnumerator {
  return [self stringScanEnumeratorWithMaxStringLength:256];
}

@end /* NSData(Strings) */
