/* 
   NSArray+enumerator.h

   Copyright (C) 2002 SKYRIX Software AG, Helge Hess
   All rights reserved.

   Author: Jan Reichmann (jan.reichmann@skyrix.com)

   This file is part of the SKYRiX Core Framework

   Permission to use, copy, modify, and distribute this software and its
   documentation for any purpose and without fee is hereby granted, provided
   that the above copyright notice appear in all copies and that both that
   copyright notice and this permission notice appear in supporting
   documentation.

   We disclaim all warranties with regard to this software, including all
   implied warranties of merchantability and fitness, in no event shall
   we be liable for any special, indirect or consequential damages or any
   damages whatsoever resulting from loss of use, data or profits, whether in
   an action of contract, negligence or other tortious action, arising out of
   or in connection with the use or performance of this software.
*/
// $Id: NSArray+enumerator.h,v 1.2 2002/12/27 09:43:14 helge Exp $

#ifndef __NGExtensions_NSArray_enumerator_H__
#define __NGExtensions_NSArray_enumerator_H__

#import <Foundation/NSArray.h>

@class NSSet;

@interface NSArray(enumerator)

- (id)initWithObjectsFromEnumerator:(NSEnumerator *)_enumerator;

/*
  Returns an array contructed using this algorithm:

    for (i = 0; i < [self count]; i++)
      newArray[i] = [array[i] performSelector:_selector];

  If the selector returns nil, NSNull is placed in the resulting array.
*/
- (NSArray *)mappedArrayUsingSelector:(SEL)_selector;
- (NSArray *)mappedArrayUsingSelector:(SEL)_selector withObject:(id)_object;
- (NSSet *)mappedSetUsingSelector:(SEL)_selector;
- (NSSet *)mappedSetUsingSelector:(SEL)_selector withObject:(id)_object;

@end

@interface NSMutableArray(enumerator);

- (id)initWithObjectsFromEnumerator:(NSEnumerator *)_enumerator;

@end

#endif /* __NGExtensions_NSArray_enumerator_H__ */
