/* 
   NSArray+enumerator.m

   Copyright (C) 2002 SKYRIX Software AG, Helge Hess
   All rights reserved.

   Author: Jan Reichmann (jan.reichmann@skyrix.com), Helge Hess (hh@mdlink.de)

   This file is part of the SKYRiX Core Framework

   Permission to use, copy, modify, and distribute this software and its
   documentation for any purpose and without fee is hereby granted, provided
   that the above copyright notice appear in all copies and that both that
   copyright notice and this permission notice appear in supporting
   documentation.

   We disclaim all warranties with regard to this software, including all
   implied warranties of merchantability and fitness, in no event shall
   we be liable for any special, indirect or consequential damages or any
   damages whatsoever resulting from loss of use, data or profits, whether in
   an action of contract, negligence or other tortious action, arising out of
   or in connection with the use or performance of this software.
*/
// $Id: NSArray+enumerator.m,v 1.3 2003/01/07 11:33:35 helge Exp $

#include "common.h"
#include "NSArray+enumerator.h"

@implementation NSArray(enumerator)

- (id)initWithObjectsFromEnumerator:(NSEnumerator *)_enumerator {
  NSMutableArray *array = nil;
  
  array = [[NSMutableArray alloc]
                           initWithObjectsFromEnumerator:_enumerator];
  self = [self initWithArray:array];
  [array release]; array = nil;
  return self;
}

// mapping

- (NSArray *)mappedArrayUsingSelector:(SEL)_selector {
  int i, count = [self count];
  id  objects[count];
  IMP objAtIndex;
  
  if (_selector == NULL) return self;

  objAtIndex = [self methodForSelector:@selector(objectAtIndex:)];

  for (i = 0; i < count; i++) {
    objects[i] = objAtIndex(self, @selector(objectAtIndex:), i);
    objects[i] = [objects[i] performSelector:_selector];
    if (objects[i] == nil)
      objects[i] = [NSNull null];
  }

  return [NSArray arrayWithObjects:objects count:count];
}
- (NSArray *)mappedArrayUsingSelector:(SEL)_selector withObject:(id)_object {
  int i, count = [self count];
  id  objects[count];
  IMP objAtIndex;
  
  if (_selector == NULL) return self;

  objAtIndex = [self methodForSelector:@selector(objectAtIndex:)];

  for (i = 0; i < count; i++) {
    objects[i] = [objAtIndex(self, @selector(objectAtIndex:), i)
                            performSelector:_selector withObject:_object];

    if (objects[i] == nil)
      objects[i] = [NSNull null];
  }

  return [NSArray arrayWithObjects:objects count:count];
}

- (NSSet *)mappedSetUsingSelector:(SEL)_selector {
  return [NSSet setWithArray:[self mappedArrayUsingSelector:_selector]];
}
- (NSSet *)mappedSetUsingSelector:(SEL)_selector withObject:(id)_object {
  return [NSSet setWithArray:[self mappedArrayUsingSelector:_selector
                                   withObject:_object]];
}

@end

@implementation NSMutableArray(enumerator) 

- (id)initWithObjectsFromEnumerator:(NSEnumerator *)_enumerator {
  if ((self = [self init])) {
    id obj = nil;
     
    while ((obj = [_enumerator nextObject]))
      [self addObject:obj];
  }
  return self;
}

@end

void __link_NGExtensions_NSArrayEnumerator() {
  __link_NGExtensions_NSArrayEnumerator();
}
