/* 
   NSAutoreleasePool+misc.m

   Copyright (C) 2002 SKYRIX Software AG, Helge Hess
   All rights reserved.

   Author: Helge Hess (helge.hess@skyrix.com)

   This file is part of the SKYRiX Core Framework

   Permission to use, copy, modify, and distribute this software and its
   documentation for any purpose and without fee is hereby granted, provided
   that the above copyright notice appear in all copies and that both that
   copyright notice and this permission notice appear in supporting
   documentation.

   We disclaim all warranties with regard to this software, including all
   implied warranties of merchantability and fitness, in no event shall
   we be liable for any special, indirect or consequential damages or any
   damages whatsoever resulting from loss of use, data or profits, whether in
   an action of contract, negligence or other tortious action, arising out of
   or in connection with the use or performance of this software.
*/
// $Id: NSAutoreleasePool+misc.m,v 1.2 2002/12/27 09:43:14 helge Exp $

#import "common.h"
#import "NSAutoreleasePool+misc.h"

#if defined(LIB_FOUNDATION_LIBRARY)

BOOL __autoreleaseEnableRetainRemove = NO;

#if !LIB_FOUNDATION_BOEHM_GC

@implementation NSAutoreleasePool(misc)

// retain/remove check

+ (void)enableRetainRemove:(BOOL)enable {
    __autoreleaseEnableRetainRemove = enable;
}

+ (NSAutoreleasePool *)findReleasingPoolForObject:(id)_obj {
    NSAutoreleasePool *pool = nil;
    
    for (pool = [self defaultPool]; pool; pool = pool->parentPool) {
        NSAutoreleasePoolChunk *ch;
        int i;

        for (ch = pool->firstChunk; ch; ch = ch->next) {
            for (i = 0; i < ch->used; i++) {
                if (ch->objects[i] == _obj)
                    return pool;
            }
        }
        //if ([pool doesReleaseObject:_obj])
        //    return pool;
    }
    return nil;
}

+ (BOOL)retainAutoreleasedObject:(id)_obj {
    register NSAutoreleasePool *pool = nil;
    
    for (pool = [self defaultPool]; pool; pool = pool->parentPool) {
        register NSAutoreleasePoolChunk *ch;
        register int i;

        for (ch = pool->firstChunk; ch; ch = ch->next) {
            for (i = 0; i < ch->used; i++) {
                if (ch->objects[i] == _obj) {
                    ch->objects[i] = nil;
                    return YES;
                }
            }
        }
        //if ([pool doesReleaseObject:_obj])
        //    return pool;
    }
    return NO;
}

- (BOOL)retainAutoreleasedObject:(id)_obj {
    NSAutoreleasePoolChunk *ch;
    int i;

    for (ch = firstChunk; ch; ch = ch->next) {
	for (i = 0; i < ch->used; i++) {
	    if (ch->objects[i] == _obj) {
                ch->objects[i] = nil;
                return YES;
            }
        }
    }
    return NO;
}

@end

@implementation NSObject(RC)

- (oneway void)release
{
#if BUILD_libFoundation_DLL && defined(__WIN32__)
    extern __declspec(dllimport) BOOL __autoreleaseEnableCheck;
#else
    extern BOOL __autoreleaseEnableCheck;
#endif

    // check if retainCount is Ok
    if (__autoreleaseEnableCheck) {
	unsigned int toCome = [NSAutoreleasePool autoreleaseCountForObject:self];
	if (toCome+1 > [self retainCount]) {
	    NSLog(@"Release[0x%08X<%@>] release check for object %@ "
                  @"has %d references "
	    	  @"and %d pending calls to release in autorelease pools\n", 
		  (unsigned)self, NSStringFromClass([self class]),
                  self,
                  [self retainCount], toCome);
            NSLog(@"  description='%@'", [self description]);
            abort(); // core dump for debugging
	    return;
	}
    }
    if (NSExtraRefCount(self) == 1)
	[self dealloc];
    else
	NSDecrementExtraRefCountWasZero(self);
}

- (id)retain
{
    extern BOOL __autoreleaseEnableRetainRemove;

    if (__autoreleaseEnableRetainRemove) {
        if ([NSAutoreleasePool retainAutoreleasedObject:self]) {
            NSLog(@"retained autoreleased object ..");
            return self;
        }
    }
    
    NSIncrementExtraRefCount(self);
    return self;
}

@end

#endif // !LIB_FOUNDATION_BOEHM_GC

#endif // defined(LIB_FOUNDATION_LIBRARY)

void __link_NSAutoreleasePool_misc(void) {
  __link_NSAutoreleasePool_misc();
}
