/* 
   NSCalendarDate+misc.h

   Copyright (C) 2002 SKYRIX Software AG, Helge Hess
   All rights reserved.

   Author: Helge Hess (helge.hess@skyrix.com)

   This file is part of the SKYRiX Core Framework

   Permission to use, copy, modify, and distribute this software and its
   documentation for any purpose and without fee is hereby granted, provided
   that the above copyright notice appear in all copies and that both that
   copyright notice and this permission notice appear in supporting
   documentation.

   We disclaim all warranties with regard to this software, including all
   implied warranties of merchantability and fitness, in no event shall
   we be liable for any special, indirect or consequential damages or any
   damages whatsoever resulting from loss of use, data or profits, whether in
   an action of contract, negligence or other tortious action, arising out of
   or in connection with the use or performance of this software.
*/
// $Id: NSCalendarDate+misc.h,v 1.2 2002/12/27 09:43:14 helge Exp $

#ifndef __NGExtensions_NSCalendarDate_misc_H__
#define __NGExtensions_NSCalendarDate_misc_H__

#import <Foundation/NSDate.h>

#if NeXT_Foundation_LIBRARY || GNUSTEP_BASE_LIBRARY
#  import <Foundation/NSCalendarDate.h>
#endif

@interface NSCalendarDate(misc)

- (int)weekOfMonth;
- (int)weekOfYear;
- (short)numberOfWeeksInYear;
+ (NSArray *)mondaysOfYear:(int)_year timeZone:(NSTimeZone *)_tz;
- (NSArray *)mondaysOfYear;
- (NSCalendarDate *)firstMondayAndLastWeekInYear:(short *)_lastWeek;
+ (NSCalendarDate *)mondayOfWeek:(int)_weekNumber inYear:(int)_year
  timeZone:(NSTimeZone *)_tz;
- (NSCalendarDate *)mondayOfWeek:(int)_weekNumber;

- (NSCalendarDate *)firstDayOfMonth;
- (NSCalendarDate *)lastDayOfMonth;
- (NSCalendarDate *)mondayOfWeek;
- (NSCalendarDate *)beginOfDay;
- (NSCalendarDate *)endOfDay;
- (int)numberOfDaysInMonth;

- (BOOL)isDateOnSameDay:(NSCalendarDate *)_date;
- (BOOL)isDateInSameWeek:(NSCalendarDate *)_date;

- (BOOL)isToday;
- (NSCalendarDate *)yesterday;
- (NSCalendarDate *)tomorrow;
- (BOOL)isForenoon;
- (BOOL)isAfternoon;

- (NSCalendarDate *)nextYear;
- (NSCalendarDate *)lastYear;

/* returns a date on the same day with the specified time */
- (NSCalendarDate *)hour:(int)_hour minute:(int)_minute second:(int)_second;
- (NSCalendarDate *)hour:(int)_hour minute:(int)_minute;

/*
  applies the following modifications:
    if year >= 70 && year < 135
      year = 1900 + year
    elif year >= 0 && year < 70
      year = 2000 + year
*/
      
- (NSCalendarDate *)y2kDate;

/*
  adding years, months and days while *keeping* the clock time, eg:

    d1 = [NSCalendarDate dateWithYear:2000 month:2 day:15
                         hour:12 minute:0 second:0
                         timeZone:@"MET"];
    d2 = [d1 dateByAddingYear:0 month:3 day:0];

    [d2 hourOfDay] will be '15' though the timezone changed from
    MET to MET-DST.

    -dateByAddingYears:months:days:hours:minutes:seconds: which can
    be found in NSCalendarDate will not keep the clock time (the time
    will be adjusted in the new DST timezone
*/

- (NSCalendarDate *)dateByAddingYears:(int)_years
  months:(int)_months
  days:(int)_days;

/* calculate easter ... */

- (NSCalendarDate *)easterOfYear;

@end

#endif /* __NGExtensions_NSCalendarDate_misc_H__ */
