/* 
   NSCalendarDate+misc.m

   Copyright (C) 2002 SKYRIX Software AG, Helge Hess
   All rights reserved.

   Author: Helge Hess (helge.hess@skyrix.com)

   This file is part of the SKYRiX Core Framework

   Permission to use, copy, modify, and distribute this software and its
   documentation for any purpose and without fee is hereby granted, provided
   that the above copyright notice appear in all copies and that both that
   copyright notice and this permission notice appear in supporting
   documentation.

   We disclaim all warranties with regard to this software, including all
   implied warranties of merchantability and fitness, in no event shall
   we be liable for any special, indirect or consequential damages or any
   damages whatsoever resulting from loss of use, data or profits, whether in
   an action of contract, negligence or other tortious action, arising out of
   or in connection with the use or performance of this software.
*/
// $Id: NSCalendarDate+misc.m,v 1.6 2003/01/20 10:16:23 helge Exp $

#import "common.h"
#import "NSCalendarDate+misc.h"

#define NUMBER_OF_SECONDS_IN_DAY (24 * 60 * 60)

@implementation NSCalendarDate(misc)

+ (NSCalendarDate *)mondayOfWeek:(int)_weekNumber inYear:(int)_year
  timeZone:(NSTimeZone *)_tz
{
  NSCalendarDate *janFirst;
  
  janFirst  = [NSCalendarDate dateWithYear:_year month:1 day:1 
                            hour:0 minute:0 second:0
                            timeZone:_tz];
  return [janFirst mondayOfWeek:_weekNumber];
}

- (NSCalendarDate *)mondayOfWeek:(int)_weekNumber {
  NSCalendarDate *mondayOfWeek;
  short          lastWeek;
  
  mondayOfWeek = [self firstMondayAndLastWeekInYear:&lastWeek];
  
  if (_weekNumber == 1)
    return mondayOfWeek;
  
  return [mondayOfWeek dateByAddingYears:0
                       months:0
                       days:(7 * (_weekNumber - 1))];
}

+ (NSArray *)mondaysOfYear:(int)_year timeZone:(NSTimeZone *)_tz {
  NSCalendarDate *janFirst;
  
  janFirst  = [NSCalendarDate dateWithYear:_year month:1 day:1 
                            hour:0 minute:0 second:0
                            timeZone:_tz];
  return [janFirst mondaysOfYear];
}

- (NSArray *)mondaysOfYear {
  NSArray        *array;
  NSMutableArray *mondays;
  NSCalendarDate *mondayOfWeek;
  short          lastWeek;
  int            i;
  
  mondayOfWeek = [self firstMondayAndLastWeekInYear:&lastWeek];
  mondays = [[NSMutableArray alloc] initWithCapacity:55];
  
  for (i = 0; i < lastWeek; i++) {
#if 0 // hh: can someone explain this ?!
    if (i > 0) {
      mondayOfWeek =
        [mondayOfWeek addYear:0 month:0 day:7 hour:0 minute:0 second:0];
    }
    mondayOfWeek = [[mondayOfWeek copy] autorelease];
    [mondays addObject:mondayOfWeek];
#else
    NSCalendarDate *tmp;
    tmp = [mondayOfWeek dateByAddingYears:0 months:0 days:(i * 7)];
    [mondays addObject:tmp];
#endif
  }
  
  array = [mondays copy];
  [mondays release];
  return [array autorelease];
}

- (NSCalendarDate *)firstMondayAndLastWeekInYear:(short *)_lastWeek {
  NSTimeZone     *tz;
  int            currentYear;
  short          lastWeek;
  NSCalendarDate *janFirst;
  NSCalendarDate *silvester;
  NSCalendarDate *mondayOfWeek;
  
  tz          = [self timeZone];
  currentYear = [self yearOfCommonEra];
  
  if ([self weekOfYear] == 53) {
    NSCalendarDate *nextJanFirst = nil;

    nextJanFirst = [NSCalendarDate dateWithYear:(currentYear + 1)
                                 month:1 day:1
                                 hour:0 minute:0 second:0
                                 timeZone:tz];

    if ([nextJanFirst weekOfYear] == 1)
      currentYear++;
  }
  
  janFirst  = [NSCalendarDate dateWithYear:currentYear
                            month:1 day:1
                            hour:0 minute:0 second:0
                            timeZone:tz];
  silvester = [NSCalendarDate dateWithYear:currentYear
                            month:12 day:31
                            hour:23 minute:59 second:59
                            timeZone:tz];

  lastWeek = [silvester weekOfYear];

  if (lastWeek == 53) {
    NSCalendarDate *nextJanFirst = nil;
    
    nextJanFirst = [NSCalendarDate dateWithYear:currentYear+1
                                 month:1 day:1
                                 hour:0 minute:0 second:0
                                 timeZone:tz];
    
    if ([nextJanFirst weekOfYear] == 1)
      lastWeek = 52;
  }
  
  if ([janFirst weekOfYear] != 1) {
    mondayOfWeek = [janFirst mondayOfWeek];
#if NeXT_Foundation_LIBRARY || COCOA_Foundation_LIBRARY
    mondayOfWeek =
      [mondayOfWeek dateByAddingYears:0 months:0 days:7 
		    hours:0 minutes:0 seconds:0];
#else
    mondayOfWeek =
      [mondayOfWeek addYear:0 month:0 day:7 hour:0 minute:0 second:0];
#endif
  }
  else {
    mondayOfWeek = [janFirst mondayOfWeek];
  }

  if (_lastWeek) *_lastWeek = lastWeek;
  return mondayOfWeek;
}

- (NSCalendarDate *)firstDayOfMonth {
  NSTimeInterval tv;
  NSCalendarDate *first;
  int            dayOfMonth;
  NSTimeZone     *tz;
  
  dayOfMonth = [self dayOfMonth];
  tz = [self timeZone];
  
  tv = (1 - dayOfMonth) * NUMBER_OF_SECONDS_IN_DAY;
  tv = [self timeIntervalSince1970] + tv;
  first = [NSCalendarDate dateWithTimeIntervalSince1970:tv];
  [first setTimeZone:tz];
  return first;
}

- (NSCalendarDate *)lastDayOfMonth {
  NSCalendarDate *day = [self firstDayOfMonth];
  day = [day dateByAddingYears:0 months:1 days:0];
  return [day dateByAddingYears:0 months:0 days:-1];
}

- (int)numberOfDaysInMonth {
  return [[self lastDayOfMonth] dayOfMonth];
}

- (int)weekOfMonth {
  /* returns 1-6 */
  int dayOfMonth;
  int weekOfYear;
  int firstWeekOfYear;
  
  dayOfMonth = [self dayOfMonth]; /* 1-31 */
  if (dayOfMonth == 1)
    return 1;

  /* could be done smarter (by calculating on dayOfWeek) */
  weekOfYear      = [self weekOfYear];
  firstWeekOfYear = [[self firstDayOfMonth] weekOfYear];
  
  return (weekOfYear - firstWeekOfYear + 1);
}

- (int)weekOfYear {
  static int whereToStart[] = { 6, 7, 8, 9, 10, 4, 5 };
  NSCalendarDate *janFirst;
  int            year, day, weekOfYear;
  NSTimeZone     *tz;
  
  year = [self yearOfCommonEra];
  day  = [self dayOfYear] - 1; 
  
  tz = [self timeZone];

  janFirst  = [NSCalendarDate dateWithYear:year month:1 day:1 
                            hour:0 minute:0 second:0
                            timeZone:tz];
  
  weekOfYear = (day + whereToStart[[janFirst dayOfWeek]]) / 7;

  if (weekOfYear == 0) {
    NSCalendarDate *silvesterLastYear;

    silvesterLastYear = [NSCalendarDate dateWithYear:(year - 1)
                                      month:12 day:31 
                                      hour:0 minute:0 second:0
                                      timeZone:tz];
    return [silvesterLastYear weekOfYear];
  }
  
#if 0  
  if (weekOfYear == 53) {
    NSCalendarDate *nextJanFirst = nil;
    int            janYear       = year + 1;
    int            janDay;
    int            week;

    nextJanFirst = [NSCalendarDate dateWithYear:janYear
                                 month:1 day:1
                                 hour:0 minute:0 second:0
                                 timeZone:[self timeZone]];

    janDay = [nextJanFirst dayOfYear];
    week   = (janDay + whereToStart[[nextJanFirst dayOfWeek]]) / 7;

    if (week == 1)
      return 52;
  }
#endif  
  return weekOfYear;
}

- (short)numberOfWeeksInYear {
  NSCalendarDate *silvester;
  NSCalendarDate *dayOfLastWeek;
  NSTimeZone     *tz;
  short currentYear;

  currentYear = [self yearOfCommonEra];
  tz          = [self timeZone];

  silvester = [NSCalendarDate dateWithYear:currentYear
                              month:12 day:31
                              hour:23 minute:59 second:59
                              timeZone:tz];
  dayOfLastWeek = [silvester dateByAddingYears:0 months:0 days:-3 hours:0
                             minutes:0 seconds:0];

  return [dayOfLastWeek weekOfYear];
}

- (NSCalendarDate *)mondayOfWeek {
  NSTimeInterval tv;
  NSCalendarDate *monday;
  int            dayOfWeek;
  NSTimeZone     *tz;
  
  dayOfWeek = [self dayOfWeek];
  tz = [self timeZone];
  
  if (dayOfWeek == 0) dayOfWeek = 7; // readjust Sunday
  
  tv = (1 - dayOfWeek) * NUMBER_OF_SECONDS_IN_DAY;
  tv = [self timeIntervalSince1970] + tv;
  monday = [NSCalendarDate dateWithTimeIntervalSince1970:tv];
  [monday setTimeZone:tz];
  return monday;
}

- (NSCalendarDate *)beginOfDay {
  NSTimeZone     *tz;

  tz = [self timeZone];
  
  return [NSCalendarDate dateWithYear:[self yearOfCommonEra]
                         month:       [self monthOfYear]
                         day:         [self dayOfMonth]
                         hour:0 minute:0 second:0
                         timeZone:    tz];
}

- (NSCalendarDate *)endOfDay {
  NSTimeZone     *tz;

  tz = [self timeZone];
  
  return [NSCalendarDate dateWithYear:[self yearOfCommonEra]
                         month:       [self monthOfYear]
                         day:         [self dayOfMonth]
                         hour:23 minute:59 second:59
                         timeZone:    tz];
}

- (BOOL)isDateOnSameDay:(NSCalendarDate *)_date {
  if ([self dayOfYear]       != [_date dayOfYear])       return NO;
  if ([self yearOfCommonEra] != [_date yearOfCommonEra]) return NO;
  return YES;
}
- (BOOL)isDateInSameWeek:(NSCalendarDate *)_date {
  if ([self weekOfYear]      != [_date weekOfYear])      return NO;
  if ([self yearOfCommonEra] != [_date yearOfCommonEra]) return NO;
  return YES;
}

- (BOOL)isToday {
  NSCalendarDate *d;
  BOOL result;
  
  d = [[NSCalendarDate alloc] init];
  [d setTimeZone:[self timeZone]];
  result = [self isDateOnSameDay:d];
  [d release];
  return result;
}

- (BOOL)isForenoon {
  return [self hourOfDay] >= 12 ? NO : YES;
}
- (BOOL)isAfternoon {
  return [self hourOfDay] >= 12 ? NO : YES;
}

- (NSCalendarDate *)yesterday {
  return [self dateByAddingYears:0 months:0 days:-1 
	       hours:0 minutes:0 seconds:0];
}
- (NSCalendarDate *)tomorrow {
  return [self dateByAddingYears:0 months:0 days:1 
	       hours:0 minutes:0 seconds:0];
}

- (NSCalendarDate *)lastYear {
  return [self dateByAddingYears:-1 months:0 days:0 
	       hours:0 minutes:0 seconds:0];
}
- (NSCalendarDate *)nextYear {
  return [self dateByAddingYears:1 months:0 days:0 
	       hours:0 minutes:0 seconds:0];
}

- (NSCalendarDate *)hour:(int)_hour minute:(int)_minute second:(int)_second {
  NSTimeZone *tz;

  tz = [self timeZone];
  
  return [NSCalendarDate dateWithYear:[self yearOfCommonEra]
                         month:       [self monthOfYear]
                         day:         [self dayOfMonth]
                         hour:_hour minute:_minute second:_second
                         timeZone:    tz];
}
- (NSCalendarDate *)hour:(int)_hour minute:(int)_minute {
  return [self hour:_hour minute:_minute second:0];
}

/* Y2K support */

- (NSCalendarDate *)y2kDate {
  NSCalendarDate *newDate;
  int year;
  
  year = [self yearOfCommonEra];
  if (year >= 70 && year < 135) {
    newDate = [[NSCalendarDate
                      alloc]
                      initWithYear:(year + 1900)
                      month:[self monthOfYear]
                      day:[self dayOfMonth]
                      hour:[self hourOfDay]
                      minute:[self minuteOfHour]
                      second:[self secondOfMinute]
                      timeZone:[self timeZone]];
  }
  else if (year >= 0 && year < 70) {
    newDate = [[NSCalendarDate
                      alloc]
                      initWithYear:(year + 2000)
                      month:[self monthOfYear]
                      day:[self dayOfMonth]
                      hour:[self hourOfDay]
                      minute:[self minuteOfHour]
                      second:[self secondOfMinute]
                      timeZone:[self timeZone]];
  }
  else
    newDate = [self retain];
    
  return [newDate autorelease];
}

- (NSCalendarDate *)dateByAddingYears:(int)_years
  months:(int)_months
  days:(int)_days
{
#if 0
  /* this expects that NSCalendarDate accepts invalid days, like
     2000-02-31 */
  int newYear, newMonth, newDay;
  
  newYear  = [self yearOfCommonEra] + _years;
  newMonth = [self monthOfYear]     + _months;
  newDay   = [self dayOfMonth]      + _days;

  // this doesn't check month overflow !!
  return [NSCalendarDate dateWithYear:newYear month:newMonth day:newDay
                         hour:[self hourOfDay]
                         minute:[self minuteOfHour]
                         second:[self secondOfMinute]
                         timeZone:[self timeZone]];
#else
  // but this does it 
  return [self dateByAddingYears:_years months:_months days:_days
               hours:0 minutes:0 seconds:0];
#endif
}

/* calculate easter ... */

- (NSCalendarDate *)easterOfYear {
  /*
    algorithm taken from:
      http://www.uni-bamberg.de/~ba1lw1/fkal.html#Algorithmus
  */
  int      y;
  unsigned m, n;
  int      a, b, c, d, e;
  unsigned easterMonth, easterDay;
  
  y = [self yearOfCommonEra];

  if ((y > 1699) && (y < 1800)) {
    m = 23;
    n = 3;
  }
  else if ((y > 1799) && (y < 1900)) {
    m = 23;
    n = 4;
  }
  else if ((y > 1899) && (y < 2100)) {
    m = 24;
    n = 5;
  }
  else if ((y > 2099) && (y < 2200)) {
    m = 24;
    n = 6;
  }
  else
    return nil;
    
  a = y % 19;
  b = y % 4;
  c = y % 7;
  d = (19 * a + m) % 30;
  e = (2 * b + 4 * c + 6 * d + n) % 7;
  
  easterMonth = 3;
  easterDay   = 22 + d + e;
  if (easterDay > 31) {
    easterDay  -= 31;
    easterMonth = 4;
    if (easterDay == 26)
      easterDay = 19;
    if ((easterDay == 25) && (d == 28) && (a > 10))
      easterDay = 18;
  }
  
  return [NSCalendarDate dateWithYear:y
                         month:easterMonth
                         day:easterDay
                         hour:0 minute:0 second:0
                         timeZone:[self timeZone]];
}

@end /* NSCalendarDate(misc) */

/* static linking */

void __link_NSCalendarDate_misc(void) {
  __link_NSCalendarDate_misc();
}
