/* 
   NSEnumerator+misc.m

   Copyright (C) 2000 MDlink online service center, Helge Hess
   All rights reserved.

   Author: Helge Hess (helge.hess@skyrix.com)

   This file is part of the MDlink Object Framework 3 (MOF 3)

   Permission to use, copy, modify, and distribute this software and its
   documentation for any purpose and without fee is hereby granted, provided
   that the above copyright notice appear in all copies and that both that
   copyright notice and this permission notice appear in supporting
   documentation.

   We disclaim all warranties with regard to this software, including all
   implied warranties of merchantability and fitness, in no event shall
   we be liable for any special, indirect or consequential damages or any
   damages whatsoever resulting from loss of use, data or profits, whether in
   an action of contract, negligence or other tortious action, arising out of
   or in connection with the use or performance of this software.
*/
// $Id: NSEnumerator+misc.m,v 1.4 2003/01/07 11:33:35 helge Exp $

#import "NSEnumerator+misc.h"
#import <Foundation/Foundation.h>
#import <EOControl/EOQualifier.h>
#include "common.h"

@interface _NGFilterEnumerator : NSEnumerator
{
  NSEnumerator *source;
}

+ (NSEnumerator *)filterEnumeratorWithSource:(NSEnumerator *)_e;

- (NSEnumerator *)source;

@end

@interface _NGQualifierFilterEnumerator : _NGFilterEnumerator
{
  EOQualifier *q;
}
- (void)setQualifier:(EOQualifier *)_q;
- (EOQualifier *)qualifier;
@end

@interface _NGSelFilterEnumerator : _NGFilterEnumerator
{
  SEL  sel;
  id   arg;
}
- (void)setSelector:(SEL)_s;
- (SEL)selector;
- (void)setArgument:(id)_arg;
- (id)argument;
@end

@implementation NSEnumerator(misc)

- (NSEnumerator *)filterWithQualifier:(EOQualifier *)_qualifier {
  id e;
  
  e = [_NGQualifierFilterEnumerator filterEnumeratorWithSource:self];
  [e setQualifier:_qualifier];
  return e;
}
- (NSEnumerator *)filterWithQualifierString:(NSString *)_s {
  EOQualifier *q;
  
  q = [EOQualifier qualifierWithQualifierFormat:_s];
  return [self filterWithQualifier:q];
}

- (NSEnumerator *)filterWithSelector:(SEL)_selector withObject:(id)_argument {
  id e;
  
  e = [_NGSelFilterEnumerator filterEnumeratorWithSource:self];
  [e setSelector:_selector];
  [e setArgument:_argument];
  return e;
}

@end /* NSEnumerator(misc) */

@implementation _NGFilterEnumerator

- (id)initWithSource:(NSEnumerator *)_e {
  self->source = [_e retain];
  return self;
}
- (void)dealloc {
  [self->source release];
  [super dealloc];
}

+ (NSEnumerator *)filterEnumeratorWithSource:(NSEnumerator *)_e {
  return [[(_NGFilterEnumerator *)[self alloc] initWithSource:_e] autorelease];
}

- (NSEnumerator *)source {
  return self->source;
}

- (id)nextObject {
  return [self->source nextObject];
}

@end /* _NGFilterEnumerator */

@implementation _NGQualifierFilterEnumerator

- (void)dealloc {
  [self->q release];
  [super dealloc];
}

- (void)setQualifier:(EOQualifier *)_q {
  ASSIGN(self->q, _q);
}
- (EOQualifier *)qualifier {
  return self->q;
}

- (id)nextObject {
  while (YES) {
    id obj;
    
    if ((obj = [self->source nextObject]) == nil)
      return nil;
    if (self->q == nil)
      return obj;

    if ([(id<EOQualifierEvaluation>)self->q evaluateWithObject:obj])
      return obj;
  }
}

@end /* _NGQualifierFilterEnumerator */

@implementation _NGSelFilterEnumerator

- (void)dealloc {
  [self->arg release];
  [super dealloc];
}

- (void)setSelector:(SEL)_s {
  self->sel = _s;
}
- (SEL)selector {
  return self->sel;
}

- (void)setArgument:(id)_arg {
  ASSIGN(self->arg, _arg);
}
- (id)argument {
  return self->arg;
}

- (id)nextObject {
  while (YES) {
    id obj;
    BOOL (*m)(id,SEL,id);
    
    if ((obj = [self->source nextObject]) == nil)
      return nil;
    
    if ((m = (void *)[obj methodForSelector:self->sel]) == NULL)
      continue;

    if (m(obj, self->sel, self->arg))
      return obj;
  }
}

@end /* _NGSelFilterEnumerator */
