// $Id: NSObject+Logs.m,v 1.3 2003/01/07 11:33:35 helge Exp $

#include "NSObject+Logs.h"
#include "common.h"

@implementation NSObject(NGLogs)

static Class StringClass = Nil;

static inline Class NSStringClass(void) {
  if (StringClass == Nil) StringClass = [NSString class];
  return StringClass;
}

- (BOOL)isDebuggingEnabled {
#if DEBUG
  return YES;
#else
  return NO;
#endif
}
- (NSString *)loggingPrefix {
  /* improve perf ... */
  return [NSStringClass() stringWithFormat:@"<0x%08X[%@]>",
                       self, NSStringFromClass([self class])];
}

- (void)logWithFormat:(NSString *)_format arguments:(va_list)ap {
  NSString *value = nil;
  
  value = [[NSStringClass() alloc] initWithFormat:_format arguments:ap];
  NSLog(@"%@ %@", [self loggingPrefix], value);
  [value release];
}

- (void)debugWithFormat:(NSString *)_format arguments:(va_list)ap {
#if DEBUG
  NSString *value = nil;
  
  if (![self isDebuggingEnabled]) return;
  
  value = [[NSStringClass() alloc] initWithFormat:_format arguments:ap];
  NSLog(@"<%@>D %@", [self loggingPrefix], value);
  [value release];
#else
#  warning debug is disabled, debugWithFormat wont print anything ..
#endif
}

- (void)logWithFormat:(NSString *)_format, ... {
  va_list ap;
  
  va_start(ap, _format);
  [self logWithFormat:_format arguments:ap];
  va_end(ap);
}
- (void)debugWithFormat:(NSString *)_format, ... {
  va_list ap;
  
  va_start(ap, _format);
  [self debugWithFormat:_format arguments:ap];
  va_end(ap);
}

@end /* NSObject(NGLogs) */
