/* 
   NSObject+Values.m

   Copyright (C) 2002 SKYRIX Software AG, Helge Hess
   All rights reserved.

   Author: Helge Hess (helge.hess@skyrix.com)

   This file is part of the SKYRiX Core Framework

   Permission to use, copy, modify, and distribute this software and its
   documentation for any purpose and without fee is hereby granted, provided
   that the above copyright notice appear in all copies and that both that
   copyright notice and this permission notice appear in supporting
   documentation.

   We disclaim all warranties with regard to this software, including all
   implied warranties of merchantability and fitness, in no event shall
   we be liable for any special, indirect or consequential damages or any
   damages whatsoever resulting from loss of use, data or profits, whether in
   an action of contract, negligence or other tortious action, arising out of
   or in connection with the use or performance of this software.
*/
// $Id: NSObject+Values.m,v 1.3 2002/12/27 09:43:59 helge Exp $

#import "common.h"
#import "NSObject+Values.h"

@implementation NSObject(NGValues)

- (BOOL)boolValue {
  // this returns always YES (the id != nil)
  return YES;
}

- (char)charValue {
  return (char)[self intValue];
}
- (unsigned char)unsignedCharValue {
  return (unsigned char)[self intValue];
}
- (short)shortValue {
  return (short)[self intValue];
}
- (unsigned short)unsignedShortValue {
  return (unsigned short)[self unsignedIntValue];
}

- (int)intValue {
  return [[self stringValue] intValue];
}
- (unsigned int)unsignedIntValue {
  return (unsigned int)[self intValue];
}

- (long)longValue {
  return (long)[self intValue];
}
- (unsigned long)unsignedLongValue {
  return (unsigned long)[self unsignedIntValue];
}

- (long long)longLongValue {
  return [[self stringValue] longLongValue];
}
- (unsigned long long)unsignedLongLongValue {
  return [[self stringValue] unsignedLongLongValue];
}

- (float)floatValue {
  return [[self stringValue] floatValue];
}
- (double)doubleValue {
  return [[self stringValue] doubleValue];
}

- (NSString *)stringValue {
  return [self description];
}

@end /* NSObject(Values) */

@implementation NSString(NGValues)

- (BOOL)boolValue {
  if ([self isEqualToString:@"YES"])   return YES;
  if ([self isEqualToString:@"NO"])    return NO;
  if ([self isEqualToString:@"yes"])   return YES;
  if ([self isEqualToString:@"no"])    return NO;
  if ([self isEqualToString:@"TRUE"])  return YES;
  if ([self isEqualToString:@"FALSE"]) return NO;
  if ([self isEqualToString:@"true"])  return YES;
  if ([self isEqualToString:@"false"]) return NO;
  if ([self isEqualToString:@"1"])     return YES;
  if ([self isEqualToString:@"0"])     return NO;
  return NO;
}

- (NSString *)stringValue {
  return self;
}

@end /* NSString(Values) */

@implementation NSMutableString(NGValues)

- (NSString *)stringValue {
  return [[self copy] autorelease];
}

@end /* NSMutableString(Values) */

void __link_NGExtensions_NSObjectValues(void) {
  /* required for static linking */
  __link_NGExtensions_NSObjectValues();
}
