/* 
   NSString+Formatting.h

   Copyright (C) 2002 SKYRIX Software AG, Helge Hess
   All rights reserved.

   Author: Helge Hess (helge.hess@skyrix.com)

   This file is part of the SKYRiX Core Framework

   Permission to use, copy, modify, and distribute this software and its
   documentation for any purpose and without fee is hereby granted, provided
   that the above copyright notice appear in all copies and that both that
   copyright notice and this permission notice appear in supporting
   documentation.

   We disclaim all warranties with regard to this software, including all
   implied warranties of merchantability and fitness, in no event shall
   we be liable for any special, indirect or consequential damages or any
   damages whatsoever resulting from loss of use, data or profits, whether in
   an action of contract, negligence or other tortious action, arising out of
   or in connection with the use or performance of this software.
*/
// $Id: NSString+Formatting.h,v 1.2 2002/12/27 09:43:59 helge Exp $

#ifndef __NGExtensions_NSString_Formatting_H__
#define __NGExtensions_NSString_Formatting_H__

#import <Foundation/NSString.h>

// category

@interface NSString(XSFormatting)

+ (id)stringWithCFormat:(const char *)_format arguments:(va_list)_ap;
+ (id)stringWithCFormat:(const char *)_format, ...;

@end

@interface NSMutableString(XSFormatting)

- (void)appendFormat:(NSString *)_format arguments:(va_list)_ap;
- (void)appendFormat:(NSString *)_format, ...;

@end

// C support functions

static inline int xs_vsnprintf(char *str, size_t max, const char *fmt, va_list _ap) {
  NSString *obj = [NSString stringWithCFormat:str arguments:_ap];
  [obj getCString:str maxLength:(max - 1)];
  return [obj cStringLength]; // return the length the string would have consumed
}

static inline int xs_vsprintf (char *_str, const char *_fmt, va_list _ap) {
  NSString *obj = [NSString stringWithCFormat:_str arguments:_ap];
  [obj getCString:_str];
  return [obj cStringLength]; // return the length of the string
}

/*
  Could use formats ..
     //     __attribute__ ((format (printf, 2, 3)));
     //     __attribute__ ((format (printf, 3, 4)));
*/
int xs_sprintf (char *str, const char *format, ...);
int xs_snprintf(char *str, size_t size, const char *format, ...);

#endif /* __NGExtensions_NSString_Formatting_H__ */
