/* 
   NSString+misc.h

   Copyright (C) 2002 SKYRIX Software AG, Helge Hess
   All rights reserved.

   Author: Helge Hess (helge.hess@skyrix.com)

   This file is part of the SKYRiX Core Framework

   Permission to use, copy, modify, and distribute this software and its
   documentation for any purpose and without fee is hereby granted, provided
   that the above copyright notice appear in all copies and that both that
   copyright notice and this permission notice appear in supporting
   documentation.

   We disclaim all warranties with regard to this software, including all
   implied warranties of merchantability and fitness, in no event shall
   we be liable for any special, indirect or consequential damages or any
   damages whatsoever resulting from loss of use, data or profits, whether in
   an action of contract, negligence or other tortious action, arising out of
   or in connection with the use or performance of this software.
*/
// $Id: NSString+misc.h,v 1.3 2002/12/27 09:43:59 helge Exp $

#ifndef __NGExtensions_NSString_misc_H__
#define __NGExtensions_NSString_misc_H__

#import <Foundation/NSString.h>

@class NSSet, NSDictionary, NSString;

@interface NSString(misc)

- (NSString *)stringByApplyingCEscaping;

/*
  Replaces keys, which enclosed in '$', with values from _bindings.
  For using of $ escape it with $$.

  Example:
    source: @"du da blah $var$ doof"

    dest = [source stringByReplacingVariablesWithBindings:
                     @{ var = @"dummy"; }];
    =>
    dest:   @"du da blah dummy doof"
*/
- (NSString *)stringByReplacingVariablesWithBindings:(NSDictionary *)_bindings;

/*
  If there are variables with no binding, _unkown is used instead.
  If _unkown is nil and there are unknown bindings, an exception will be thrown.
*/
- (NSString *)stringByReplacingVariablesWithBindings:(NSDictionary *)_bindings
  stringForUnknownBindings:(NSString *)_unknown;

/*
  Returns all binding variables. ('aaa $doof$ $bla$' --> (doof, bla) )
*/
- (NSSet *)bindingVariables;

@end

@interface NSString(FilePathVersioningMethods)

/*
  "/PATH/file.txt;1"
*/

- (NSString *)pathVersion;
- (NSString *)stringByDeletingPathVersion;
- (NSString *)stringByAppendingPathVersion:(NSString *)_version;

@end /* NSString(FilePathMethodsVersioning) */

@interface NSString(URLEscaping)

/*
  These functions encode/decode HTTP style URL paths, which can escape
  spaces and special chars.
  Chars are escaped using the '%' hex-notation:

  Encode:
      'Hello World' => 'Hello%20World'
      '& ?' => '%26%20%3F'
*/

- (BOOL)containsURLEscapeCharacters;
- (NSString *)stringByUnescapingURL;
- (NSString *)stringByEscapingURL;

@end

@interface NSString(HTMLEscaping)
- (NSString *)stringByEscapingHTMLString;
- (NSString *)stringByEscapingHTMLAttributeValue;
@end

@interface NSString(XMLEscaping)

- (NSString *)stringByEscapingXMLString;
- (NSString *)stringByEscapingXMLAttributeValue;

/*
  The following methods work in "fully-qualified XML names", in this
  format:
    '{namespace}name'
*/
- (BOOL)xmlIsFQN;
- (NSString *)xmlNamespaceURI;
- (NSString *)xmlLocalName;

@end

#endif /* __NGExtensions_NSString_misc_H__ */
