/* 
   NSString+misc.m

   Copyright (C) 2002 SKYRIX Software AG, Helge Hess
   All rights reserved.

   Author: Helge Hess (helge.hess@skyrix.com)

   This file is part of the SKYRiX Core Framework

   Permission to use, copy, modify, and distribute this software and its
   documentation for any purpose and without fee is hereby granted, provided
   that the above copyright notice appear in all copies and that both that
   copyright notice and this permission notice appear in supporting
   documentation.

   We disclaim all warranties with regard to this software, including all
   implied warranties of merchantability and fitness, in no event shall
   we be liable for any special, indirect or consequential damages or any
   damages whatsoever resulting from loss of use, data or profits, whether in
   an action of contract, negligence or other tortious action, arising out of
   or in connection with the use or performance of this software.
*/
// $Id: NSString+misc.m,v 1.6 2003/04/01 13:38:20 gnustep Exp $

#import "common.h"
#import "NGMemoryAllocation.h"
#import "NSString+misc.h"

#if !LIB_FOUNDATION_LIBRARY && !GNUSTEP_BASE_LIBRARY
#  import <FoundationExt/objc-api.h>
#  import <FoundationExt/objc-runtime.h>
#endif

@interface NSStringVariableBindingException : NSException
@end

@implementation NSStringVariableBindingException
@end

@implementation NSString(misc)

- (NSString *)stringByApplyingCEscaping {
  const char   *cstr = [self cString];
  char         *buffer = NGMallocAtomic([self cStringLength] * 2 + 1);
  register int pos = 0;
    
  while (*cstr) {
    switch (*cstr) {
      case '\n':
        buffer[pos] = '\\'; pos++;
        buffer[pos] = 'n';
        break;
      case '\r':
        buffer[pos] = '\\'; pos++;
        buffer[pos] = 'r';
        break;
      case '\t':
        buffer[pos] = '\\'; pos++;
        buffer[pos] = 't';
        break;

      default:
        buffer[pos] = *cstr;
        break;
    }
    cstr++;
    pos++;
  }
  buffer[pos] = '\0';

#if NeXT_Foundation_LIBRARY
  {
    NSString *s;

    s = [NSString stringWithCString:buffer];
    free(buffer);
    return s;
  }
#else
  return [NSString stringWithCStringNoCopy:buffer freeWhenDone:YES];
#endif
}
- (NSSet *)bindingVariables
{
  unsigned        len, pos = 0;
  const char      *buf     = NULL;
  NSMutableSet    *result  = nil;

  result = [NSMutableSet set];
  len    = [self cStringLength];  
  buf    = [self cString];


  while (pos < len) {
    if (pos + 1 == len) { /* last entry */
      if (buf[pos] == '$') { /* found $ without end-char */
        [[[NSStringVariableBindingException alloc]
                                            initWithFormat:@"didn`t find end of "
                                            @"variable for string %@", self]
                                            raise];
      }
      break;
    }
    if (buf[pos] == '$') {
      if (buf[pos + 1] == '$') { /* found $$ --> ignore*/
        pos++;
      }
      else {
        unsigned startPos = pos;

        pos += 2; /* buf[pos + 1] != '$' */
        while (pos < len) {
          if (buf[pos] != '$')
            pos++;
          else
            break;
        }
        if (pos == len) /* end of string was reached */
          [[[NSStringVariableBindingException alloc]
                                            initWithFormat:@"didn`t find end of "
                                            @"variable for string %@", self]
                                              raise];
        else {
          NSString *key   = nil;

          key = [[NSString alloc]
                           initWithCStringNoCopy:(char*)buf + startPos + 1
                           length:pos - startPos - 1
                           freeWhenDone:NO];
          [result addObject:key];
          [key release]; key = nil;
        }
      }
    }
    pos++;
  }
  return [[result copy] autorelease];
}

- (NSString *)stringByReplacingVariablesWithBindings:(NSDictionary *)_bindings
  stringForUnknownBindings:(NSString *)_unknown
{
  unsigned        len, pos = 0;
  const char      *buf     = NULL;
  NSMutableString *str     = nil;

  str = [self mutableCopyWithZone:[self zone]];
  len = [str cStringLength];  
  buf = [str cString];


  while (pos < len) {
    if (pos + 1 == len) { /* last entry */
      if (buf[pos] == '$') { /* found $ without end-char */
        [[[NSStringVariableBindingException alloc]
                                            initWithFormat:@"didn`t find end of "
                                            @"variable for string %@", self]
                                            raise];
      }
      break;
    }
    if (buf[pos] == '$') {
      if (buf[pos + 1] == '$') { /* found $$ --> $ */
        [str deleteCharactersInRange:NSMakeRange(pos, 1)];
        buf = [str cString];
        len = [str cStringLength];
      }
      else {
        unsigned startPos = pos;

        pos += 2; /* buf[pos + 1] != '$' */
        while (pos < len) {
          if (buf[pos] != '$')
            pos++;
          else
            break;
        }
        if (pos == len) /* end of string was reached */
          [[[NSStringVariableBindingException alloc]
                                            initWithFormat:@"didn`t find end of "
                                            @"variable for string %@", self]
                                              raise];
        else {
          NSString *key   = nil;
          NSString *value = nil;

          key = [[NSString alloc]
                           initWithCStringNoCopy:(char*)buf + startPos + 1
                           length:pos - startPos - 1
                           freeWhenDone:NO];

          if ((value = [[_bindings objectForKey:key] stringValue]) == nil) {
            if (_unknown == nil) {
              [[[NSStringVariableBindingException alloc]
                          initWithFormat:@"didn`t find binding for "
                                         @"name %@ in binding-dictionary %@",
                                         [key autorelease], _bindings] raise];
            }
            else
              value = _unknown;
          }
          [key release]; key = nil;
          [str replaceCharactersInRange:NSMakeRange(startPos, pos - startPos + 1)
               withString:value];
          buf = [str cString];
          len = [str cStringLength];
          pos = startPos - 1 + [value length];
        }
      }
    }
    pos++;
  }
  {
    id tmp = str;
    str = [str copy];
    [tmp release]; tmp = nil;
  }
  return [str autorelease];
}

- (NSString *)stringByReplacingVariablesWithBindings:(NSDictionary *)_bindings {
  return [self stringByReplacingVariablesWithBindings:_bindings
               stringForUnknownBindings:nil];
}

@end /* NSString(misc) */


@implementation NSString(FilePathVersioningMethods)

/*
  "/path/file.txt;1"
*/
- (NSString *)pathVersion {
  NSRange r;

  r = [self rangeOfString:@";"];
  if (r.length > 0) {
    return ([self length] > r.location)
      ? [self substringFromIndex:(r.location + r.length)]
      : @"";
  }
  return nil;
}

- (NSString *)stringByDeletingPathVersion {
  NSRange r;

  r = [self rangeOfString:@";"];
  return (r.length > 0)
    ? [self substringToIndex:r.location]
    : self;
}

- (NSString *)stringByAppendingPathVersion:(NSString *)_version {
  return [[self stringByAppendingString:@";"] stringByAppendingString:_version];
}

@end /* NSString(FilePathMethodsVersioning) */

@implementation NSString(URLEscaping)

static inline BOOL isUrlAlpha(unsigned char _c) {
  return
    (((_c >= 'a') && (_c <= 'z')) ||
     ((_c >= 'A') && (_c <= 'Z')))
    ? YES : NO;
}
static inline BOOL isUrlDigit(unsigned char _c) {
  return ((_c >= '0') && (_c <= '9')) ? YES : NO;
}
static inline BOOL isUrlSafeChar(unsigned char _c) {
  switch (_c) {
    case '$': case '-': case '_': case '@':
    case '.': case '&': case '+':
      return YES;

    default:
      return NO;
  }
}
static inline BOOL isUrlExtraChar(unsigned char _c) {
  switch (_c) {
    case '!': case '*': case '"': case '\'':
    case '|': case ',':
      return YES;
  }
  return NO;
}
static inline BOOL isUrlEscapeChar(unsigned char _c) {
  return (_c == '%') ? YES : NO;
}
static inline BOOL isUrlReservedChar(unsigned char _c) {
  switch (_c) {
    case '=': case ';': case '/':
    case '#': case '?': case ':':
    case ' ':
      return YES;
  }
  return NO;
}

static inline BOOL isUrlXalpha(unsigned char _c) {
  if (isUrlAlpha(_c))      return YES;
  if (isUrlDigit(_c))      return YES;
  if (isUrlSafeChar(_c))   return YES;
  if (isUrlExtraChar(_c))  return YES;
  if (isUrlEscapeChar(_c)) return YES;
  return NO;
}

static inline BOOL isUrlHexChar(unsigned char _c) {
  if (isUrlDigit(_c))
    return YES;
  if ((_c >= 'a') && (_c <= 'f'))
    return YES;
  if ((_c >= 'A') && (_c <= 'F'))
    return YES;
  return NO;
}

static inline BOOL isUrlAlphaNum(unsigned char _c) {
  return (isUrlAlpha(_c) || isUrlDigit(_c)) ? YES : NO;
}

static inline BOOL isToBeEscaped(unsigned char _c) {
  return (isUrlAlphaNum(_c) || (_c == '_')) ? NO : YES;
}

static void
NGEscapeUrlBuffer(const unsigned char *_source, unsigned char *_dest)
{
  register const unsigned char *src = (void*)_source;
  while (*src) {
    //if (*src == ' ') { // a ' ' becomes a '+'
    //  *_dest = '+'; _dest++;
    //}
    if (!isToBeEscaped(*src)) {
      *_dest = *src;
      _dest++;
    } 
    else { // any other char is escaped ..
      *_dest = '%'; _dest++;
      sprintf(_dest, "%02X", (unsigned)*src);
      _dest += 2;
    }
    src++;
  }
  *_dest = '\0';
}

static inline int _valueOfHexChar(register unichar _c) {
  switch (_c) {
    case '0': case '1': case '2': case '3': case '4':
    case '5': case '6': case '7': case '8': case '9':
      return (_c - 48); // 0-9 (ascii-char)'0' - 48 => (int)0
      
    case 'A': case 'B': case 'C':
    case 'D': case 'E': case 'F':
      return (_c - 55); // A-F, A=10..F=15, 'A'=65..'F'=70
      
    case 'a': case 'b': case 'c':
    case 'd': case 'e': case 'f':
      return (_c - 87); // a-f, a=10..F=15, 'a'=97..'f'=102

    default:
      return -1;
  }
}
static inline BOOL _isHexDigit(register unichar _c) {
  switch (_c) {
    case '0': case '1': case '2': case '3': case '4':
    case '5': case '6': case '7': case '8': case '9':
    case 'A': case 'B': case 'C':
    case 'D': case 'E': case 'F':
    case 'a': case 'b': case 'c':
    case 'd': case 'e': case 'f':
      return YES;

    default:
      return NO;
  }
}

static void
NGUnescapeUrlBuffer(const unsigned char *_source, unsigned char *_dest)
{
  BOOL done = NO;

  while (!done && (*_source != '\0')) {
    char c = *_source;

    //if (c == '+') // '+' stands for a space
    //  *_dest = ' ';
    if (c == '%') {
      _source++; c = *_source;
      
      if (c == '\0') {
        *_dest = '%';
        done = YES;
      }
      else if (_isHexDigit(c)) { // hex-escaped char, like '%F3'
        int decChar = _valueOfHexChar(c);
        _source++;
        c = *_source;
        decChar = decChar * 16 + _valueOfHexChar(c);
        *_dest = (unsigned char)decChar;
      }
      else // escaped char, like '%%' -> '%'
        *_dest = c;
    }
    else // char passed through
      *_dest = c;

    _dest++;
    _source++;
  }
  *_dest = '\0';
}

- (BOOL)containsURLEscapeCharacters {
  register unsigned i, len;
  register unichar (*charAtIdx)(id,SEL,unsigned);
  
  if ((len = [self length]) == 0) return NO;
  
  charAtIdx = (void*)[self methodForSelector:@selector(characterAtIndex:)];
  for (i = 0; i < len; i++) {
    if (charAtIdx(self, @selector(characterAtIndex:), i) == '%')
      return YES;
  }
  return NO;
}
- (BOOL)containsURLInvalidCharacters {
  register unsigned i, len;
  register unichar (*charAtIdx)(id,SEL,unsigned);
  
  if ((len = [self length]) == 0) return NO;
  
  charAtIdx = (void*)[self methodForSelector:@selector(characterAtIndex:)];
  for (i = 0; i < len; i++) {
    if (isToBeEscaped(charAtIdx(self, @selector(characterAtIndex:), i)))
      return YES;
  }
  return NO;
}

- (NSString *)stringByUnescapingURL {
  unsigned len;
  char     *cstr;
  char     *buffer = NULL;
  NSString *s;
  
  if (![self containsURLEscapeCharacters])
    return [[self copy] autorelease];
  
  if ((len = [self cStringLength]) == 0) return @"";
    
  cstr = malloc(len + 10);
  [self getCString:cstr];
  cstr[len] = '\0';
    
  buffer = objc_atomic_malloc(len + 2);
  NGUnescapeUrlBuffer(cstr, buffer);
  s = [[NSString alloc]
                 initWithCStringNoCopy:buffer
                 length:strlen(buffer)
                 freeWhenDone:YES];
  if (cstr) free(cstr);
  return [s autorelease];
}

- (NSString *)stringByEscapingURL {
  unsigned len;
  char     *cstr;
  NSString *s;
  char     *buffer = NULL;
  
  if ((len = [self cStringLength]) == 0) return @"";
  
  if (![self containsURLInvalidCharacters]) // needs to be escaped ?
    return [[self copy] autorelease];
  
  cstr = malloc(len + 1);
  [self getCString:cstr];
  cstr[len] = '\0';
    
  buffer = objc_atomic_malloc([self cStringLength] * 3 + 2);
  NGEscapeUrlBuffer(cstr, buffer);
    
  s = [[NSString alloc]
                 initWithCStringNoCopy:buffer
                 length:strlen(buffer)
                 freeWhenDone:YES];
  free(cstr);
  return [s autorelease];
}

@end /* NSString(URLEscaping) */

@implementation NSString(HTMLEscaping)

- (NSString *)stringByEscapingHTMLStringUsingCharacters {
  register unsigned i, len, j;
  register unichar  *chars, *buf;
  unsigned escapeCount;
  
  if ((len = [self length]) == 0) return @"";
  
  chars = objc_atomic_malloc((len + 3) * sizeof(unichar));
  [self getCharacters:chars];
  
  /* check for characters to escape ... */
  for (i = 0, escapeCount = 0; i < len; i++) {
    switch (chars[i]) {
      case '&': case '"': case '<': case '>':
        escapeCount++;
        break;
      default:
        if (chars[i] > 127)
          escapeCount++;
        break;
    }
  }
  if (escapeCount == 0 ) {
    /* nothing to escape ... */
    if (chars) objc_free(chars);
    return [[self copy] autorelease];
  }
  
  buf = objc_atomic_malloc(((len + 3) * sizeof(unichar)) +
                           (escapeCount * 8 * sizeof(unichar)));
  for (i = 0, j = 0; i < len; i++) {
    switch (chars[i]) {
      /* escape special chars */
      case '&':
        buf[j] = '&'; j++; buf[j] = 'a'; j++; buf[j] = 'm'; j++;
        buf[j] = 'p'; j++; buf[j] = ';'; j++;
        break;
      case '"':
        buf[j] = '&'; j++; buf[j] = 'q'; j++; buf[j] = 'u'; j++;
        buf[j] = 'o'; j++; buf[j] = 't'; j++; buf[j] = ';'; j++;
        break;
      case '<':
        buf[j] = '&'; j++; buf[j] = 'l'; j++; buf[j] = 't'; j++;
        buf[j] = ';'; j++;
        break;
      case '>':
        buf[j] = '&'; j++; buf[j] = 'g'; j++; buf[j] = 't'; j++;
        buf[j] = ';'; j++;
        break;
      case 223: /* &szlig; */
        buf[j] = '&'; j++; buf[j] = 's'; j++; buf[j] = 'z'; j++;
        buf[j] = 'l'; j++; buf[j] = 'i'; j++; buf[j] = 'g'; j++;
        buf[j] = ';'; j++;
        break;
      case 252: /* &uuml; */
        buf[j] = '&'; j++; buf[j] = 'u'; j++; buf[j] = 'u'; j++;
        buf[j] = 'm'; j++; buf[j] = 'l'; j++; buf[j] = ';'; j++;
        break;
      case 220: /* &Uuml; */
        buf[j] = '&'; j++; buf[j] = 'U'; j++; buf[j] = 'u'; j++;
        buf[j] = 'm'; j++; buf[j] = 'l'; j++; buf[j] = ';'; j++;
        break;
      case 228: /* &auml; */
        buf[j] = '&'; j++; buf[j] = 'a'; j++; buf[j] = 'u'; j++;
        buf[j] = 'm'; j++; buf[j] = 'l'; j++; buf[j] = ';'; j++;
        break;
        
      default:
        /* escape big chars */
        if (chars[i] > 127) {
          unsigned char nbuf[16];
          unsigned int k;
          
          sprintf(nbuf, "&#%i;", (int)chars[i]);
          for (k = 0; nbuf[k] != '\0'; k++) {
            buf[j] = nbuf[k];
            j++;
          }
        }
        else {
          /* nothing to escape */
          buf[j] = chars[i];
          j++;
        }
        break;
    }
  }
  
  self = [NSString stringWithCharacters:buf length:j];
  
  if (chars) objc_free(chars);
  if (buf)   objc_free(buf);
  return self;
}

- (NSString *)stringByEscapingHTMLAttributeValueUsingCharacters {
  register unsigned i, len, j;
  register unichar  *chars, *buf;
  unsigned escapeCount;
  
  if ((len = [self length]) == 0) return @"";
  
  chars = objc_atomic_malloc((len + 3) * sizeof(unichar));
  [self getCharacters:chars];
  
  /* check for characters to escape ... */
  for (i = 0, escapeCount = 0; i < len; i++) {
    switch (chars[i]) {
      case '&':
      case '"':
      case '<':
      case '>':
      case '\t':
      case '\n':
      case '\r':
        escapeCount++;
        break;
      default:
        if (chars[i] > 127)
          escapeCount++;
        break;
    }
  }
  if (escapeCount == 0 ) {
    /* nothing to escape ... */
    if (chars) objc_free(chars);
    return [[self copy] autorelease];
  }
  
  buf = objc_atomic_malloc(((len + 3) * sizeof(unichar)) +
                           (escapeCount * 8 * sizeof(unichar)));
  for (i = 0, j = 0; i < len; i++) {
    switch (chars[i]) {
      /* escape special chars */
      case '&':
        buf[j] = '&'; j++; buf[j] = 'a'; j++; buf[j] = 'm'; j++;
        buf[j] = 'p'; j++; buf[j] = ';'; j++;
        break;
      case '"':
        buf[j] = '&'; j++; buf[j] = 'q'; j++; buf[j] = 'u'; j++;
        buf[j] = 'o'; j++; buf[j] = 't'; j++; buf[j] = ';'; j++;
        break;
      case '<':
        buf[j] = '&'; j++; buf[j] = 'l'; j++; buf[j] = 't'; j++;
        buf[j] = ';'; j++;
        break;
      case '>':
        buf[j] = '&'; j++; buf[j] = 'g'; j++; buf[j] = 't'; j++;
        buf[j] = ';'; j++;
        break;
        
      case '\t':
        buf[j] = '&'; j++; buf[j] = '#'; j++; buf[j] = '9'; j++;
        buf[j] = ';'; j++;
        break;
      case '\n':
        buf[j] = '&'; j++; buf[j] = '#'; j++; buf[j] = '1'; j++;
        buf[j] = '0'; j++; buf[j] = ';'; j++;
        break;
      case '\r':
        buf[j] = '&'; j++; buf[j] = '#'; j++; buf[j] = '1'; j++;
        buf[j] = '3'; j++; buf[j] = ';'; j++;
        break;
        
      default:
        /* escape big chars */
        if (chars[i] > 127) {
          unsigned char nbuf[16];
          unsigned int k;
          
          sprintf(nbuf, "&#%i;", (int)chars[i]);
          for (k = 0; nbuf[k] != '\0'; k++) {
            buf[j] = nbuf[k];
            j++;
          }
        }
        else {
          /* nothing to escape */
          buf[j] = chars[i]; j++;
        }
        break;
    }
  }
  
  self = [NSString stringWithCharacters:buf length:j];
  
  if (chars) objc_free(chars);
  if (buf)   objc_free(buf);
  return self;
}

- (NSString *)stringByEscapingHTMLString {
  return [self stringByEscapingHTMLStringUsingCharacters];
}
- (NSString *)stringByEscapingHTMLAttributeValue {
  return [self stringByEscapingHTMLAttributeValueUsingCharacters];
}

@end /* NSString(HTMLEscaping) */

@implementation NSString(XMLEscaping)

- (NSString *)stringByEscapingXMLStringUsingCharacters {
  register unsigned i, len, j;
  register unichar  *chars, *buf;
  unsigned escapeCount;
  
  if ((len = [self length]) == 0) return @"";
  
  chars = objc_atomic_malloc((len + 3) * sizeof(unichar));
  [self getCharacters:chars];
  
  /* check for characters to escape ... */
  for (i = 0, escapeCount = 0; i < len; i++) {
    switch (chars[i]) {
      case '&': case '"': case '<': case '>':
        escapeCount++;
        break;
      default:
        if (chars[i] > 127)
          escapeCount++;
        break;
    }
  }
  if (escapeCount == 0 ) {
    /* nothing to escape ... */
    if (chars) objc_free(chars);
    return [[self copy] autorelease];
  }
  
  buf = objc_atomic_malloc(((len + 3) * sizeof(unichar)) +
                           (escapeCount * 8 * sizeof(unichar)));
  for (i = 0, j = 0; i < len; i++) {
    switch (chars[i]) {
      /* escape special chars */
      case '&':
        buf[j] = '&'; j++; buf[j] = 'a'; j++; buf[j] = 'm'; j++;
        buf[j] = 'p'; j++; buf[j] = ';'; j++;
        break;
      case '"':
        buf[j] = '&'; j++; buf[j] = 'q'; j++; buf[j] = 'u'; j++;
        buf[j] = 'o'; j++; buf[j] = 't'; j++; buf[j] = ';'; j++;
        break;
      case '<':
        buf[j] = '&'; j++; buf[j] = 'l'; j++; buf[j] = 't'; j++;
        buf[j] = ';'; j++;
        break;
      case '>':
        buf[j] = '&'; j++; buf[j] = 'g'; j++; buf[j] = 't'; j++;
        buf[j] = ';'; j++;
        break;
        
      default:
        /* escape big chars */
        if (chars[i] > 127) {
          unsigned char nbuf[16];
          unsigned int k;
          
          sprintf(nbuf, "&#%i;", (int)chars[i]);
          for (k = 0; nbuf[k] != '\0'; k++) {
            buf[j] = nbuf[k];
            j++;
          }
        }
        else {
          /* nothing to escape */
          buf[j] = chars[i];
          j++;
        }
        break;
    }
  }
  
  self = [NSString stringWithCharacters:buf length:j];
  
  if (chars) objc_free(chars);
  if (buf)   objc_free(buf);
  return self;
}

- (NSString *)stringByEscapingXMLString {
  return [self stringByEscapingXMLStringUsingCharacters];
}

- (NSString *)stringByEscapingXMLAttributeValue {
  return [self stringByEscapingHTMLAttributeValue];
}

/* XML FQNs */

- (BOOL)xmlIsFQN {
  if ([self length] == 0) return NO;
  return [self characterAtIndex:0] == '{' ? YES : NO;
}

- (NSString *)xmlNamespaceURI {
  NSRange r;
  
  r = [self rangeOfString:@"}" options:(NSLiteralSearch | NSBackwardsSearch)];
  if (r.length == 0) return nil;
  if ([self characterAtIndex:0] != '{') return nil;
  
  r.length   = (r.location - 1);
  r.location = 1;
  return [self substringWithRange:r];
}

- (NSString *)xmlLocalName {
  NSRange r;
  
  r = [self rangeOfString:@"}" options:(NSLiteralSearch | NSBackwardsSearch)];
  if (r.length == 0) return nil;
  if ([self characterAtIndex:0] != '{') return nil;
  
  return [self substringFromIndex:(r.location + r.length)];
}

@end /* NSString(XMLEscaping) */

// linking

void __link_NSString_misc(void) {
  __link_NSString_misc();
}
