/* 
   common.h

   Copyright (C) 2002 SKYRIX Software AG, Helge Hess
   All rights reserved.

   Author: Helge Hess (helge.hess@skyrix.com)

   This file is part of the SKYRiX Core Framework

   Permission to use, copy, modify, and distribute this software and its
   documentation for any purpose and without fee is hereby granted, provided
   that the above copyright notice appear in all copies and that both that
   copyright notice and this permission notice appear in supporting
   documentation.

   We disclaim all warranties with regard to this software, including all
   implied warranties of merchantability and fitness, in no event shall
   we be liable for any special, indirect or consequential damages or any
   damages whatsoever resulting from loss of use, data or profits, whether in
   an action of contract, negligence or other tortious action, arising out of
   or in connection with the use or performance of this software.
*/
// $Id: common.h,v 1.3 2003/01/07 11:33:35 helge Exp $

#ifndef __NGExtensions_common_h__
#define __NGExtensions_common_h__

#import <Foundation/Foundation.h>
#import <Foundation/NSMapTable.h>

#if defined(WIN32)
#  include <windows.h>
#elif defined(NeXT) || NeXT_Foundation_LIBRARY
#  include <netinet/in.h>
#else
#  include <arpa/inet.h>
#endif

#if !defined(WIN32)
#include <unistd.h>
#endif

#if defined(__MINGW32__)
#  define WITH_OPENSTEP 0
#  define GNUSTEP  1
#elif defined(NeXT) || defined(WIN32)
#  define WITH_OPENSTEP 1
#  define GNUSTEP  0
#  ifndef NeXT_RUNTIME
#    define NeXT_RUNTIME 1
#  endif
#else
#  define WITH_OPENSTEP 0
#  define GNUSTEP  1
#endif

#if LIB_FOUNDATION_LIBRARY
#  include <time.h>
#  import <objc/objc-api.h>
#  import <objc/objc.h>
#  import <objc/encoding.h>
#  import <extensions/objc-runtime.h>
#  import <extensions/exceptions/GeneralExceptions.h>
#endif

#ifndef ASSIGN
#  define ASSIGN(object, value) \
       ({id __object = (id)object;    \
         id __value = (id)value;      \
         if (__value != __object) { if (__value) [__value retain]; \
          if (__object) [__object release]; \
          object = __value;}})
#endif
#ifndef ASSIGNCOPY
#  define ASSIGNCOPY(object, value) \
       ({id __object = (id)object;    \
         id __value = (id)value;      \
         if (__value != __object) { if (__value) __value = [__value copy];   \
          if (__object) [__object release]; \
          object = __value;}})
#endif

#if defined(LIB_FOUNDATION_LIBRARY)
#  define NoZone nil
#else
#  define NoZone NULL
#endif

#include <ctype.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#ifndef __MINGW32__
#include <strings.h>
#endif

#if defined(WIN32)
static inline const char *index(const char *str, char c) __attribute__((unused));

static const char *index(const char *str, char c) {
  while ((*str != '\0') && (*str != c)) str++;
  if (*str == '\0') return NULL;
  else return str;
}
#endif

#if PROFILE
#  define BEGIN_PROFILE \
     { NSTimeInterval __ti = [[NSDate date] timeIntervalSince1970];

#  define END_PROFILE \
     __ti = [[NSDate date] timeIntervalSince1970] - __ti;\
     if (__ti > 0.05) \
       printf("***PROF[%s]: %0.3fs\n", __PRETTY_FUNCTION__, __ti);\
     else if (__ti > 0.005) \
       printf("PROF[%s]: %0.3fs\n", __PRETTY_FUNCTION__, __ti);\
     }

#  define PROFILE_CHECKPOINT(__key__) \
       printf("---PROF[%s] CP %s: %0.3fs\n", __PRETTY_FUNCTION__, __key__,\
              [[NSDate date] timeIntervalSince1970] - __ti)

#else
#  define BEGIN_PROFILE {
#  define END_PROFILE   }
#  define PROFILE_CHECKPOINT(__key__)
#endif

#endif
