// $Id: EOQualifier+LDAP.m,v 1.2 2003/01/07 11:34:54 helge Exp $

#include "EOQualifier+LDAP.h"
#include "common.h"

@interface EOQualifier(LDAPPrivates)

- (void)addToLDAPFilterString:(NSMutableString *)_s inContext:(id)_ctx;

@end

@implementation EOQualifier(LDAP)

- (id)initWithLDAPFilterString:(NSString *)_filter {
  return nil;
}

- (void)addToLDAPFilterString:(NSMutableString *)_s inContext:(id)_ctx {
  [self doesNotRecognizeSelector:_cmd]; // subclass
}

- (NSString *)ldapFilterString {
  NSMutableString *s;
  NSString *is;

  s = [[NSMutableString alloc] initWithCapacity:100];
  [self addToLDAPFilterString:s inContext:nil];
  is = [s copy];
  [s release];
  return [is autorelease];
}

@end /* EOQualifier(LDAP) */

@implementation EOAndQualifier(LDAP)

- (void)addToLDAPFilterString:(NSMutableString *)_s inContext:(id)_ctx {
  unsigned i, cnt;
  NSArray  *array;

  array = [self qualifiers];
  cnt   = [array count];
  
  [_s appendString:@"(&"];
  
  for (i = 0; i < cnt; i++) {
    EOQualifier *sq;

    sq = [array objectAtIndex:i];
    [sq addToLDAPFilterString:_s inContext:_ctx];
  }
  
  [_s appendString:@")"];
}

@end

@implementation EOOrQualifier(LDAP)

- (void)addToLDAPFilterString:(NSMutableString *)_s inContext:(id)_ctx {
  unsigned i, cnt;
  NSArray  *array;

  array = [self qualifiers];
  cnt   = [array count];
  
  [_s appendString:@"(|"];
  
  for (i = 0; i < cnt; i++) {
    EOQualifier *sq;

    sq = [array objectAtIndex:i];
    [sq addToLDAPFilterString:_s inContext:_ctx];
  }
  
  [_s appendString:@")"];
}

@end

@implementation EONotQualifier(LDAP)

- (void)addToLDAPFilterString:(NSMutableString *)_s inContext:(id)_ctx {
  [_s appendString:@"(!"];
  [[self qualifier] addToLDAPFilterString:_s inContext:_ctx];
  [_s appendString:@")"];
}

@end

@implementation EOKeyValueQualifier(LDAP)

- (void)addToLDAPFilterString:(NSMutableString *)_s inContext:(id)_ctx {
  SEL sel;

  sel = [self selector];

  if (sel_eq(sel,  EOQualifierOperatorNotEqual))
    [_s appendString:@"(!"];
  
  [_s appendString:@"("];
  [_s appendString:[self key]];

  if (sel_eq(sel,  EOQualifierOperatorEqual))
    [_s appendString:@"="];
  else if (sel_eq(sel,  EOQualifierOperatorNotEqual))
    [_s appendString:@"="];
  else if (sel_eq(sel,  EOQualifierOperatorLessThan))
    [_s appendString:@"<"];
  else if (sel_eq(sel,  EOQualifierOperatorGreaterThan))
    [_s appendString:@">"];
  else if (sel_eq(sel,  EOQualifierOperatorLessThanOrEqualTo))
    [_s appendString:@"<="];
  else if (sel_eq(sel,  EOQualifierOperatorGreaterThanOrEqualTo))
    [_s appendString:@">="];
  else if (sel_eq(sel,  EOQualifierOperatorContains))
    [_s appendString:@"=*"];
  else if (sel_eq(sel,  EOQualifierOperatorLike))
    [_s appendString:@"="];
  else if (sel_eq(sel,  EOQualifierOperatorCaseInsensitiveLike))
    [_s appendString:@"="];
  else {
    NSLog(@"UNKNOWN operator: %@", NSStringFromSelector([self selector]));
    [_s appendString:@"="];
  }

  [_s appendString:[[self value] description]];
  [_s appendString:@")"];
  
  if (sel_eq(sel,  EOQualifierOperatorNotEqual))
    [_s appendString:@")"];
}

@end

@implementation EOKeyComparisonQualifier(LDAP)

- (void)addToLDAPFilterString:(NSMutableString *)_s inContext:(id)_ctx {
  /* ldap supports no comparison operations on keys */
  SEL sel;

  sel = [self selector];

  if (sel_eq(sel,  EOQualifierOperatorNotEqual))
    [_s appendString:@"(!"];
  
  [_s appendString:@"("];
  [_s appendString:[self leftKey]];

  if (sel_eq(sel,  EOQualifierOperatorEqual))
    [_s appendString:@"="];
  else if (sel_eq(sel,  EOQualifierOperatorNotEqual))
    [_s appendString:@"="];
  else if (sel_eq(sel,  EOQualifierOperatorLessThan))
    [_s appendString:@"<"];
  else if (sel_eq(sel,  EOQualifierOperatorGreaterThan))
    [_s appendString:@">"];
  else if (sel_eq(sel,  EOQualifierOperatorLessThanOrEqualTo))
    [_s appendString:@"<="];
  else if (sel_eq(sel,  EOQualifierOperatorGreaterThanOrEqualTo))
    [_s appendString:@">="];
  else if (sel_eq(sel,  EOQualifierOperatorContains))
    [_s appendString:@"=*"];
  else if (sel_eq(sel,  EOQualifierOperatorLike))
    [_s appendString:@"="];
  else if (sel_eq(sel,  EOQualifierOperatorCaseInsensitiveLike))
    [_s appendString:@"="];
  else {
    NSLog(@"UNKNOWN operator: %@", NSStringFromSelector([self selector]));
    [_s appendString:@"="];
  }

  [_s appendString:[self rightKey]];
  [_s appendString:@")"];
  
  if (sel_eq(sel,  EOQualifierOperatorNotEqual))
    [_s appendString:@")"];
}

@end
