// $Id: NGLdapConnection.h,v 1.2 2002/09/30 09:02:10 helge Exp $

#ifndef __NGLdapConnection_H__
#define __NGLdapConnection_H__

#import <Foundation/NSObject.h>
#import <Foundation/NSDate.h>

@class NSString, NSArray, NSEnumerator;
@class EOQualifier;
@class NGLdapEntry;

@interface NGLdapConnection : NSObject
{
  void           *handle;
  NSString       *hostName;
  int            port;
  
  NSTimeInterval cacheTimeout;
  long           cacheMaxMemory; /* in bytes */
  BOOL           isCacheEnabled;

  struct {
    BOOL isBound:1;
  } flags;
}

- (id)initWithHostName:(NSString *)_hostName port:(int)_port; // designated init
- (id)initWithHostName:(NSString *)_hostName;

/* settings */

- (NSString *)hostName;
- (int)port;

/* binding */

- (BOOL)isBound;
- (void)unbind;

- (BOOL)bindWithMethod:(NSString *)_method
  binddn:(NSString *)_login credentials:(NSString *)_cred;

/* running queries */

- (NSEnumerator *)flatSearchAtBaseDN:(NSString *)_base
  qualifier:(EOQualifier *)_q
  attributes:(NSArray *)_attributes;
- (NSEnumerator *)deepSearchAtBaseDN:(NSString *)_base
  qualifier:(EOQualifier *)_q
  attributes:(NSArray *)_attributes;
- (NSEnumerator *)baseSearchAtBaseDN:(NSString *)_base
  qualifier:(EOQualifier *)_q
  attributes:(NSArray *)_attributes;

- (NGLdapEntry *)entryAtDN:(NSString *)_dn attributes:(NSArray *)_attrs;

/* cache */

- (void)setCacheTimeout:(NSTimeInterval)_to;
- (NSTimeInterval)cacheTimeout;

- (void)setCacheMaxMemoryUsage:(long)_maxMem;
- (long)cacheMaxMemoryUsage;

- (void)setUseCache:(BOOL)_flag;
- (BOOL)doesUseCache;

- (void)flushCache;
- (void)destroyCache;

- (void)cacheForgetEntryWithDN:(NSString *)_dn;

/* modifications */

- (BOOL)addEntry:(NGLdapEntry *)_entry;
- (BOOL)removeEntryWithDN:(NSString *)_dn;
- (BOOL)modifyEntryWithDN:(NSString *)_dn changes:(NSArray *)_mods;

/* root DSE */

- (NGLdapEntry *)schemaEntry;
- (NGLdapEntry *)rootDSE;
- (NGLdapEntry *)configEntry;
- (NSArray *)namingContexts;

@end

@interface NGLdapConnection(PlainPasswordCheck)

/* specialized password check routine */

+ (BOOL)checkPassword:(NSString *)_pwd ofLogin:(NSString *)_login
  atBaseDN:(NSString *)_baseDN
  onHost:(NSString *)_hostName port:(int)_port;

@end

#endif /* __NGLdapConnection_H__ */
