// $Id: NGImap4Client.m,v 1.38 2003/02/26 16:53:00 jan Exp $

#include <unistd.h>

#include "NGImap4Client.h"
#include "NGImap4Context.h"
#include "NGImap4Support.h"
#include "NGImap4Functions.h"
#include "NGImap4ResponseParser.h"
#include "NSString+Imap4.h"
#include "common.h"
#include <sys/time.h>
#include <../timeMacros.h>

@interface NGImap4Client(ConnectionRegistration)

- (void)removeFromConnectionRegister;
- (void)registerConnection;
- (NGCTextStream *)textStream;

@end /* NGImap4Client(ConnectionRegistration); */

@interface NGImap4Client(Private)

- (NGHashMap *)processCommand:(NSString *)_command;
- (NGHashMap *)processCommand:(NSString *)_command withTag:(BOOL)_tag;
- (NGHashMap *)processCommand:(NSString *)_command withTag:(BOOL)_tag
  withNoticication:(BOOL)_notification;
- (NGHashMap *)processCommand:(NSString *)_command logText:(NSString *)_txt;

- (void)sendCommand:(NSString *)_command;
- (void)sendCommand:(NSString *)_command withTag:(BOOL)_tag;
- (void)sendCommand:(NSString *)_command withTag:(BOOL)_tag
        logText:(NSString *)_txt;

- (void)sendResponseNotification:(NGHashMap *)map;

- (NSMutableDictionary *)normalizeResponse:(NGHashMap *)_map;
- (NSDictionary *)normalizeFetchResponse:(NGHashMap *)_map;
- (NSDictionary *)normalizeOpenConnectionResponse:(NGHashMap *)_map;
- (NSDictionary *)normalizeListResponse:(NGHashMap *)_map;
- (NSDictionary *)normalizeSelectResponse:(NGHashMap *)_map;
- (NSDictionary *)normalizeStatusResponse:(NGHashMap *)_map;
- (NSDictionary *)normalizeSearchResponse:(NGHashMap *)_map;
- (NSDictionary *)normalizeSortResponse:(NGHashMap *)_map;
- (NSDictionary *)normalizeThreadResponse:(NGHashMap *)_map;
- (NSDictionary *)normalizeCapabilityRespone:(NGHashMap *)_map;
- (NSDictionary *)normalizeQuotaResponse:(NGHashMap *)_map;
- (NSDictionary *)login;

@end

/*"
**  An implementation of an Imap4 client
**
** A folder name always looks like an absolute filename (/inbox/doof) 
**
"*/

@implementation NGImap4Client

static BOOL ProfileImapEnabled;

static inline NSArray *_flags2ImapFlags(NGImap4Client *self, NSArray *_flags);
static inline NSArray *_imapFlags2Flags(NGImap4Client *self, NSArray *_flags);
static inline NSString *_folder2ImapFolder(NGImap4Client *self,NSString *_folder);
static inline NSString *_imapFolder2Folder(NGImap4Client *self,NSString *_folder);

static NSNumber *YesNumber     = nil;
static NSNumber *NoNumber      = nil;
static NSArray  *FlagKeyWords  = nil;
static NSArray  *OtherKeyWords = nil;

static Class AndQualifier      = Nil;
static Class OrQualifier       = Nil;
static Class NotQualifier      = Nil;
static Class KeyValueQualifier = Nil;

static unsigned int CountClient    = 0;
static id           *ImapClients   = NULL;
static unsigned int MaxImapClients = 0;

static int PREVENT_EXCEPTIONS = -1;

- (BOOL)useSSL {
  return self->useSSL;
}

+ (void)initialize {

  if (PREVENT_EXCEPTIONS == -1) {
    PREVENT_EXCEPTIONS =
      [[NSUserDefaults standardUserDefaults]
                       boolForKey:@"ImapPreventConnectionExceptions"]?1:0;
  }
  if (YesNumber == nil)
    YesNumber = [[NSNumber numberWithBool:YES] retain];
  if (NoNumber == nil)
    NoNumber = [[NSNumber numberWithBool:NO] retain];
  
  if (AndQualifier == Nil)    AndQualifier = [EOAndQualifier class];
  if (OrQualifier == Nil)     OrQualifier = [EOOrQualifier class];
  if (NotQualifier == Nil)    NotQualifier = [EONotQualifier class];
  if (KeyValueQualifier == Nil)
    KeyValueQualifier = [EOKeyValueQualifier class];
  
  if (FlagKeyWords == nil) {
    FlagKeyWords = [[NSArray alloc] initWithObjects:
                            @"answered", @"deleted", @"draft", @"flagged", @"new",
                            @"old", @"recent", @"seen", @"unanswered",
                            @"undeleted", @"undraft", @"unflagged", @"unseen",
                            nil];
  }
  if (OtherKeyWords == nil) {
    OtherKeyWords = [[NSArray alloc] initWithObjects:
                             @"bcc", @"body", @"cc", @"from", @"subject",
                             @"text", @"to", @"keyword", @"unkeyword",
                             nil];
  }
  ProfileImapEnabled = [[NSUserDefaults standardUserDefaults]
                                         boolForKey:@"ProfileImapEnabled"];

  if (MaxImapClients < 1) {
    MaxImapClients =
      [[NSUserDefaults standardUserDefaults]
                       integerForKey:@"NGImapMaxConnectionCount"];
    if (MaxImapClients < 1)
      MaxImapClients = 50;
  }
  if (!ImapClients) {
    ImapClients = NSZoneCalloc(NULL, MaxImapClients, sizeof(id));
  }
}

+ (id)clientWithURL:(NSURL *)_url {
  return [[(NGImap4Client *)[self alloc] initWithURL:_url] autorelease];
}
+ (id)clientWithAddress:(id<NGSocketAddress>)_address {
  return
    [[(NGImap4Client *)[self alloc] initWithAddress:_address] autorelease];
}

+ (id)clientWithHost:(id)_host {
  return [[[self alloc] initWithHost:_host] autorelease];
}

- (id)initWithHost:(id)_host {
  NGInternetSocketAddress *a;
  a = [NGInternetSocketAddress addressWithPort:143 onHost:_host];
  return [self initWithAddress:a];
}
- (id)initWithURL:(NSURL *)_url {
  NGInternetSocketAddress *a;
  int port;
  id  tmp;
  
  if ((self->useSSL = [[_url scheme] isEqualToString:@"imaps"])) {
    if (NSClassFromString(@"NGActiveSSLSocket") == nil) {
      [self logWithFormat:
	      @"no SSL support available, cannot connect: %@", _url];
      [self release];
      return nil;
    }
  }
  if ((tmp = [_url port])) {
    port = [tmp intValue];
    if (port <= 0) port = self->useSSL ? 993 : 143;
  }
  else
    port = self->useSSL ? 993 : 143;
  
  self->login    = [[_url user]     copy];
  self->password = [[_url password] copy];
  a = [NGInternetSocketAddress addressWithPort:port onHost:[_url host]];
  return [self initWithAddress:a];
}

/**"
 ** designated initializer
"**/

- (id)initWithAddress:(id<NGSocketAddress>)_address {
  if ((self = [super init])) {
    self->tagId      = 0;
    self->address    = [_address retain];
    self->isLogin    = NO;
    
    self->debug      =
      [[NSUserDefaults standardUserDefaults] boolForKey:@"ImapDebugEnabled"];
    
    self->responseReceiver =
      [[NSMutableArray allocWithZone:[self zone]] initWithCapacity:128];
  }
  return self;
}

- (void)dealloc {
  [self removeFromConnectionRegister];
  [self->text    release];
  [self->address release];
  [self->socket  release];
  [self->parser  release];
  [self->responseReceiver release];
  [self->login          release];
  [self->password       release];
  [self->selectedFolder release];
  [self->delimiter      release];

  self->context = nil; /* not retained */
  [super dealloc];
}

/* equality (required for adding clients to Foundation sets) */

- (BOOL)isEqual:(id)_obj {
  if (_obj == self)
    return YES;
  if ([_obj isKindOfClass:[NGImap4Client class]])
    return [self isEqualToClient:_obj];
  return NO;
}

- (BOOL)isEqualToClient:(NGImap4Client *)_obj {
  if (_obj == self) return YES;
  return [(NGInternetSocketAddress *)[_obj address]
                                     isEqualToAddress:self->address];
}

// accessors

- (id<NGActiveSocket>)socket {
  return self->socket;
}

- (id<NGSocketAddress>)address {
  return self->address;
}

- (NSString *)delimiter {
  return self->delimiter;
}

/* connection */

/*"
** Opens a connection to given Host.
"*/

- (NSDictionary *)_openConnection {
  struct timeval tv;
  double         ti = 0.0;
  Class socketClass = Nil;

  NSDictionary  *res;
  
  
  if (ProfileImapEnabled) {
    gettimeofday(&tv, NULL);
    ti =  (double)tv.tv_sec + ((double)tv.tv_usec / 1000000.0);
  }

  [self->socket release]; self->socket = nil;
  [self->parser release]; self->parser = nil;
  [self->text   release]; self->text   = nil;
  
  socketClass = [self useSSL] 
    ? NSClassFromString(@"NGActiveSSLSocket")
    : [NGActiveSocket class];

  [self->context resetLastException];

  NS_DURING {
    self->socket =
      [[socketClass socketConnectedToAddress:self->address] retain];
  }
  NS_HANDLER {
    [self->context setLastException:localException];
  }
  NS_ENDHANDLER;
  if ([self->context lastException])
    return nil;
  
  if (self->socket == nil)
    return nil;

  {
    NGBufferedStream *buffer;

    buffer     = [[NGBufferedStream alloc] initWithSource:self->socket];
    self->text = [[NGCTextStream alloc] initWithSource:buffer];
    [buffer release]; buffer = nil;
  }
  self->parser = [[NGImap4ResponseParser alloc] initWithStream:self->socket];
  self->tagId  = 0;
  /* receive greeting from server without tag-id */
  
  if (ProfileImapEnabled) {
    gettimeofday(&tv, NULL);
    ti = (double)tv.tv_sec + ((double)tv.tv_usec / 1000000.0) - ti;
    fprintf(stderr, "[%s] <openConnection> : time needed: %4.4fs\n",
           __PRETTY_FUNCTION__, ti < 0.0 ? -1.0 : ti);    
  }
  [self registerConnection];
  [self->context resetLastException];
  res = nil;
  NS_DURING {
    res = [self normalizeOpenConnectionResponse:
                [self->parser parseResponseForTagId:-1]];
  }
  NS_HANDLER {
    [self->context setLastException:localException];
  }
  NS_ENDHANDLER;
  
  if (_checkResult(self->context, res, __PRETTY_FUNCTION__))
    return res;

  return nil;
}


- (NSDictionary *)openConnection {
  return [self _openConnection];
}

/*"
** Check whether stream is already open (could be closed because server-timeout)
"*/

- (NSNumber *)isConnected {
  return [NSNumber numberWithBool:
                     (self->socket == nil)
                     ? NO : [(NGActiveSocket *)self->socket isAlive]];
}

/*"
** Close a consisting connection.
"*/

- (void)closeConnection {
  NS_DURING {
    [self->text release]; 
  }
  NS_HANDLER {
    fprintf(stderr, "got exception during release stream %s\n",
            [[localException description] cString]);
  }
  NS_ENDHANDLER;
  self->text = nil;
  
  NS_DURING {
    [self->socket close];
  }
  NS_HANDLER {
    fprintf(stderr, "got exception during close socket %s\n",
            [[localException description] cString]);
  }
  NS_ENDHANDLER;
  
  NS_DURING {
    [self->socket release];
  }
  self->socket = nil;
  NS_HANDLER {
    fprintf(stderr, "got exception during release socket %s\n",
            [[localException description] cString]);
  }
  NS_ENDHANDLER;
    
  [self->parser    release]; self->parser = nil;
  [self->delimiter release]; self->delimiter = nil;
  [self removeFromConnectionRegister];
}

// ResponseNotifications

- (void)registerForResponseNotification:(id<NGImap4ResponseReceiver>)_obj {
  [self->responseReceiver addObject:[NSValue valueWithNonretainedObject:_obj]];
}

- (void)removeFromResponseNotification:(id<NGImap4ResponseReceiver>)_obj {
  [self->responseReceiver removeObject:
       [NSValue valueWithNonretainedObject:_obj]];
}

- (void)sendResponseNotification:(NGHashMap *)_map {
  NSValue                     *val;
  id<NGImap4ResponseReceiver> obj;
  NSEnumerator                *enumerator;
  NSDictionary                *resp;

  resp       = [self normalizeResponse:_map];
  enumerator = [self->responseReceiver objectEnumerator];

  while ((val = [enumerator nextObject])) {
    obj = [val nonretainedObjectValue];
    [obj responseNotificationFrom:self response:resp];
  }
}

// commands

/*"
** login with plaintext password authenticating
"*/

- (NSDictionary *)login:(NSString *)_login password:(NSString *)_passwd {

  if ((_login == nil) || (_passwd == nil))
    return nil;

  [self->login    release]; self->login    = nil;
  [self->password release]; self->password = nil;
  
  self->login = [_login copy];

  /* check pwd for specials */
  {
    unsigned char       *buffer;
    const unsigned char *chars;
    unsigned            len, i, j;

    len   = [_passwd cStringLength];
    chars = [_passwd cString];

    buffer = calloc(sizeof(char), len*2 + 1);
    buffer[len*2] = '\0';

    for (i=0, j=0; i < len; i++, j++) {
      BOOL conv = NO;
      
      if (chars[i] <= 0x1F || chars[i] > 0x7F) {
        conv = YES;
      }
      switch (chars[i]) {
        case '(':
        case ')':
        case '{':
        case ' ':
        case '%':
        case '*':
        case '"':
        case '\\':
          conv = YES;
      }
      if (conv) {
        buffer[j++] = '\\';
      }
      buffer[j] = chars[i];
    }
    self->password = [[NSString alloc] initWithCString:buffer length:j];
  }
  return [self login];
}

- (void)reconnect {
  if ([self->context lastException] != nil)
    return;
    
  [self closeConnection];  
  self->tagId = 0;
  [self openConnection];

  if ([self->context lastException] != nil)
    return;
  
  [self login];
}

- (NSDictionary *)login {
  NGHashMap *map;

  self->isLogin = YES;

  map = [self processCommand:[NSString stringWithFormat:
                                       @"login \"%@\" \"%@\"",
                                       self->login, self->password]
              logText:[NSString stringWithFormat:@"login %@ <%@>",
                                self->login,
                                (self->password != nil) ? @"PASSWORD"
                                                        : @"NO PASSWORD"]];

  if (self->selectedFolder != nil)
    [self select:self->selectedFolder];
  
  self->isLogin = NO;
  
  return [self normalizeResponse:map];
}

/*"
** logout from the connected host and close the connection
*/

- (NSDictionary *)logout {
  NGHashMap *map;

  map = [self processCommand:@"logout"];
  [self closeConnection];
  
  return [self normalizeResponse:map];
}


// Authenticated State

/*"
** Implementation of the LIST command. The method build statements
** like 'LIST "_folder" "_pattern"'.
** The Cyrus IMAP4 v1.5.14 server ignores the given folder.
** Instead of you should use the pattern to get the expected result.
** If folder is NIL it would be set to empty string ''.
** If pattern is NIL it would be set to ''.
"*/


- (NSDictionary *)list:(NSString *)_folder pattern:(NSString *)_pattern {
  NSAutoreleasePool *pool;
  NGHashMap         *map;
  NSDictionary      *result;
  
  pool = [[NSAutoreleasePool alloc] init];
  
  if (_folder == nil)
    _folder = @"";

  if ([_folder length] > 0)
    if (!(_folder = _folder2ImapFolder(self, _folder)))
      return nil;
  
  if (_pattern == nil)
    _pattern = @"";

  if ([_pattern length] > 0)
    if (!(_pattern = _folder2ImapFolder(self, _pattern)))
      return nil;

  
  map = [self processCommand:[NSString stringWithFormat:@"list \"%@\" \"%@\"",
                                       _folder, _pattern]];
  
  if (self->delimiter == nil) {
    self->delimiter = [[[[map objectEnumeratorForKey:@"list"] nextObject]
                              objectForKey:@"delimiter"]
                              copyWithZone:[self zone]];
  }

  result = [[self normalizeListResponse:map] copy];
  [pool release];
  return [result autorelease];
}

/*"
** capabitlity command
"*/

- (NSDictionary *)capability {
  return [self normalizeCapabilityRespone:
               [self processCommand:@"capability"]];
}

/*"
** Implementation of the LSUB command. The method build statements
** like 'LSUB "_folder" "_pattern"'.
** The returnvalue is the same like the list:pattern: method
"*/

- (NSDictionary *)lsub:(NSString *)_folder pattern:(NSString *)_pattern {
  NGHashMap *map;

  if (_folder == nil)
    _folder = @"";

  if ([_folder length] > 0)
    if (!(_folder = _folder2ImapFolder(self, _folder)))
      return nil;
  
  if (_pattern == nil)
    _pattern = @"";

  if ([_pattern length] > 0)
    if (!(_pattern = _folder2ImapFolder(self, _pattern)))
      return nil;


  map = [self processCommand:[NSString stringWithFormat:@"lsub \"%@\" \"%@\"",
                                       _folder, _pattern]];

  if (self->delimiter == nil) {
    self->delimiter = [[[[map objectEnumeratorForKey:@"LIST"] nextObject]
                              objectForKey:@"delimiter"] copy];
  }
  return [self normalizeListResponse:map];
}

/*"
** Implementation of the SELECT command
"*/

- (NSDictionary *)select:(NSString *)_folder {
  id tmp;

  tmp = self->selectedFolder;

  if ([_folder length] == 0)
    return nil;

  if (!(_folder = _folder2ImapFolder(self, _folder)))
    return nil;

  self->selectedFolder = [_folder copyWithZone:[self zone]];
  
  [tmp release]; tmp = nil;

  return [self normalizeSelectResponse:
               [self processCommand:
                     [NSString stringWithFormat:@"select \"%@\"",
                               self->selectedFolder]]];
}

/*"
** Implementation of the STATUS command
"*/

- (NSDictionary *)status:(NSString *)_folder flags:(NSArray *)_flags {
  NSString *cmd;
  
  if (_folder == nil)
    return nil;

  if ((_flags == nil) || ([_flags count] == 0))
    return nil;
  
  if (!(_folder = _folder2ImapFolder(self, _folder)))
    return nil;
  
  cmd     = [NSString stringWithFormat:@"status \"%@\" (%@)",
                      _folder, [_flags componentsJoinedByString:@" "]];
  return [self normalizeStatusResponse:[self processCommand:cmd]];
}

/*"
** Implementation of the NOOP command
"*/ 

- (NSDictionary *)noop {
  // at any state
  return [self normalizeResponse:[self processCommand:@"noop"]];
}

/*"
** Implementation of the RENAME command
*/

- (NSDictionary *)rename:(NSString *)_folder to:(NSString *)_newName {
  NSString *cmd;
  
  if (!(_folder  = _folder2ImapFolder(self, _folder)))
    return nil;
  
  if (!(_newName = _folder2ImapFolder(self, _newName)))
    return nil;

  cmd = [NSString stringWithFormat:@"rename \"%@\" \"%@\"", _folder, _newName];
  
  return [self normalizeResponse:[self processCommand:cmd]];
}

/*"
** Implementation of the DELETE command
"*/

- (NSDictionary *)delete:(NSString *)_folder {
  NSString *cmd;
  
  if (!(_folder = _folder2ImapFolder(self, _folder)))
    return nil;
  
  cmd     = [NSString stringWithFormat:@"delete \"%@\"", _folder];
  
  return [self normalizeResponse:[self processCommand:cmd]];
}

/*"
** Implementation of the CREATE command
"*/

- (NSDictionary *)create:(NSString *)_name {
  NSString *cmd;
  
  if (!(_name = _folder2ImapFolder(self, _name)))
    return nil;
  
  cmd   = [NSString stringWithFormat:@"create \"%@\"", _name];
  
  return [self normalizeResponse:[self processCommand:cmd]];
}

/*"
** Implementation of the SUBSCRIBE command
"*/

- (NSDictionary *)subscribe:(NSString *)_name {
  NSString *cmd;
  
  if (!(_name = _folder2ImapFolder(self, _name)))
    return nil;
  
  cmd   = [NSString stringWithFormat:@"subscribe \"%@\"", _name];
  
  return [self normalizeResponse:[self processCommand:cmd]];
}

/*"
** Implementation of the UNSUBSCRIBE command
"*/

- (NSDictionary *)unsubscribe:(NSString *)_name {
  NSString *cmd;
  
  if (!(_name = _folder2ImapFolder(self, _name)))
    return nil;
  
  cmd   = [NSString stringWithFormat:@"unsubscribe \"%@\"", _name];
  
  return [self normalizeResponse:[self processCommand:cmd]];
}


/*"
** Implementation of the EXPUNGE command
"*/

- (NSDictionary *)expunge {
  return [self normalizeResponse:[self processCommand:@"expunge"]];
}

/*"
** Implementation of the UID FETCH command
"*/ 

- (NSDictionary *)fetchUids:(NSArray *)_uids parts:(NSArray *)_parts {
  NSAutoreleasePool *pool;
  NSString          *cmd;
  NSDictionary      *result;
  
  pool = [[NSAutoreleasePool alloc] init];
  cmd  = [NSString stringWithFormat:@"uid fetch %@ (%@)",
                   [_uids componentsJoinedByString:@","],
                   [_parts componentsJoinedByString:@" "]];
  
  result = [[self normalizeFetchResponse:[self processCommand:cmd]] retain];
  [pool release];
  
  return [result autorelease];
}

- (NSDictionary *)fetchUid:(unsigned)_uid parts:(NSArray *)_parts {
  NSAutoreleasePool *pool;
  NSString          *cmd;
  NSDictionary      *result;
  
  pool   = [[NSAutoreleasePool alloc] init];
  cmd    = [NSString stringWithFormat:@"uid fetch %d (%@)", _uid,
                     [_parts componentsJoinedByString:@" "]];
  result = [[self normalizeFetchResponse:[self processCommand:cmd]] retain];
  
  [pool release];
  return [result autorelease];
}

/*"
** Implementation of the FETCH command
"*/ 

- (NSDictionary *)fetchFrom:(unsigned)_from to:(unsigned)_to
  parts:(NSArray *)_parts
{
  NSAutoreleasePool *pool;
  NSString          *cmd;
  NSDictionary      *result; 
  NGHashMap         *rawResult;
 
  if (_to == 0)
    return [self noop];

  if (_from == 0)
    _from = 1;
  
  pool      = [[NSAutoreleasePool alloc] init];
  cmd       = [NSString stringWithFormat:@"fetch %d:%d (%@)", _from, _to,
                        [_parts componentsJoinedByString:@" "]];
  rawResult = [self processCommand:cmd];
  result    = [[self normalizeFetchResponse:rawResult] retain];
  [pool release];
  return [result autorelease];
}


/*"
** Implementation of the UID STORE command
"*/

- (NSDictionary *)storeUid:(unsigned)_uid add:(NSNumber *)_add
  flags:(NSArray *)_flags
{
  NSString *icmd, *iflags;
  
  iflags = [_flags2ImapFlags(self, _flags) componentsJoinedByString:@" "];
  icmd   = [NSString stringWithFormat:@"uid store %d %cFLAGS (%@)",
                     _uid, ([_add boolValue] == YES) ? '+' : '-',
                     iflags];
  return [self normalizeResponse:[self processCommand:icmd]];
}

/*"
** Implementation of the STORE command
"*/

- (NSDictionary *)storeFrom:(unsigned)_from to:(unsigned)_to
  add:(NSNumber *)_add
  flags:(NSArray *)_flags
{
  NSString *cmd;

  if (_to == 0)
    return [self noop];
    
  if (_from == 0)
    _from = 1;

  cmd = [NSString stringWithFormat:@"store %d:%d %cFLAGS (%@)",
                    _from, _to, ([_add boolValue] == YES) ? '+' : '-',
                    [_flags2ImapFlags(self, _flags)
                                     componentsJoinedByString:@" "]];
  
  return [self normalizeResponse:[self processCommand:cmd]];
}

/*"
** Implementation of the COPY command
"*/

- (NSDictionary *)copyFrom:(unsigned)_from to:(unsigned)_to
  toFolder:(NSString *)_folder
{
  NSString *cmd;

  if (_to == 0)
    return [self noop];
  
  if (_from == 0)
    _from = 1;

  if (!(_folder = _folder2ImapFolder(self, _folder)))
    return nil;
  
  cmd     = [NSString stringWithFormat:@"copy %d:%d \"%@\"",
                      _from, _to, _folder];
  return [self normalizeResponse:[self processCommand:cmd]];
}

/*"
** Implementation of the UID COPY command
"*/

- (NSDictionary *)copyUid:(unsigned)_uid toFolder:(NSString *)_folder {
  NSString *cmd;
  
  if (!(_folder = _folder2ImapFolder(self, _folder)))
    return nil;
  

  cmd = [NSString stringWithFormat:@"uid copy %d \"%@\"", _uid, _folder];
  
  return [self normalizeResponse:[self processCommand:cmd]];
}

- (NSDictionary *)getQuotaRoot:(NSString *)_folder {
  NSString *cmd;

  if (!(_folder = _folder2ImapFolder(self, _folder)))
    return nil;
  
  
  cmd = [NSString stringWithFormat:@"getquotaroot \"%@\"", _folder];
  
  return [self normalizeQuotaResponse:[self processCommand:cmd]];
}

/*"
** Implementation of the append command
** Does only append messages with less than 15000 Byte
"*/

- (NSDictionary *)append:(NSData *)_message toFolder:(NSString *)_folder
  withFlags:(NSArray *)_flags
{
  NSArray   *flags;
  NGHashMap *result;
  NSString  *message, *icmd;

  flags   = _flags2ImapFlags(self, _flags);
  if (!(_folder = _folder2ImapFolder(self, _folder)))
    return nil;
  

  /* Remove bare newlines */
  {
    char       *new;
    const char *old;
    int         cntOld   = 0;
    int         cntNew   = 0;
    int         len      = 0;

      
    old = [_message bytes];
    len = [_message length];
      
#if NeXT_Foundation_LIBRARY
    new = NSZoneMalloc([self zone], sizeof(char) * len * 2);
#else
    new = NSZoneMallocAtomic([self zone], sizeof(char) * len * 2);
#endif

    while (cntOld < (len - 1)) {
      if (old[cntOld] == '\n') {
        new[cntNew++] = '\r';
        new[cntNew++] = '\n';
      }
      else if (old[cntOld] != '\r') {
        new[cntNew++] = old[cntOld];
      }
      cntOld++;
    }
    if (old[cntOld] == '\n') {
      new[cntNew++] = '\r';
      new[cntNew++] = '\n';
    }
    else if (old[cntOld] != '\r') {
      new[cntNew++] = old[cntOld];
    } 
    message = [[NSString alloc] initWithCString:new length:cntNew];
    NGFree(new); new = NULL;
  }
  
  icmd = [NSString stringWithFormat:@"append \"%@\" (%@) {%d}",
                     _folder,
                     [flags componentsJoinedByString:@" "],
                     [message cStringLength]];
  result = [self processCommand:icmd
                 withTag:YES withNoticication:NO];
  
  if ([[result objectForKey:@"ContinuationResponse"] boolValue] == YES)
    result = [self processCommand:message withTag:NO];

  [message release]; message = nil;

  return [self normalizeResponse:result];
}


- (NSString *)_searchExprForQual:(EOQualifier *)_qualifier {

#define THROW_SEARCH_EXCEPTION_RELEASE_SEARCH(reason)                   \
{                                                                       \
  if (!PREVENT_EXCEPTIONS) {                                            \
  NSString *descr    = @"Couldn`t evaluate qualifier for imap search "; \
  id       exception = nil;                                             \
                                                                        \
  if (_qualifier == nil) _qualifier = (id)[EONull null];                \
                                                                        \
  descr = [descr stringByAppendingString:(NSString *)reason];           \
                                                                        \
  exception = [[NGImap4SearchException alloc] initWithFormat:descr];    \
  [(NSException *)exception setUserInfo:                                \
                  [NSDictionary dictionaryWithObject:_qualifier         \
                                forKey:@"qualifier"]];                  \
  [search release];                                                     \
  [self->context setLastException:exception];                           \
  }                                                                     \
}                                                                       \

/* end of define */
  
  BOOL            disjunction = NO; /* OR */
  NSEnumerator    *quals;
  id              qualifier;
  NSMutableString *search;

  if (_qualifier == nil)
    return @" ALL";
  
  search = [[NSMutableString alloc] initWithCapacity:256];
  quals  = nil;
  
  if ([_qualifier isKindOfClass:AndQualifier]) {
    disjunction = NO;
    quals = [[(EOAndQualifier *)_qualifier qualifiers] objectEnumerator];
  }
  else if ([_qualifier isKindOfClass:OrQualifier]) {
    disjunction = YES;
    quals = [[(EOOrQualifier *)_qualifier qualifiers] objectEnumerator];
  }
  else if (([_qualifier isKindOfClass:KeyValueQualifier])      ||
           ([_qualifier isKindOfClass:NotQualifier])) {
    disjunction = NO;
    quals = [[NSArray arrayWithObject:_qualifier] objectEnumerator];
  }
  else {
    THROW_SEARCH_EXCEPTION_RELEASE_SEARCH(@"unexpected qualifier 1");
  }
  if (disjunction == YES)
    [search appendString:@" or"];

  while ((qualifier = [quals nextObject])) {
    NSString *key;
    id       value;
    SEL      selector;
    BOOL     insertNot;

    insertNot = NO;
    
    if ([qualifier isKindOfClass:NotQualifier] == YES) {
      insertNot = YES;
      qualifier = [qualifier qualifier];
    }
    if ([qualifier isKindOfClass:KeyValueQualifier] == NO)
      THROW_SEARCH_EXCEPTION_RELEASE_SEARCH(@"expected EOKeyValueQualifier");
    key      = [[qualifier key] lowercaseString];
    value    = [qualifier value];
    selector = [qualifier selector];
    
    if ([key isEqualToString:@"flags"] == YES) {
      NSEnumerator *enumerator = nil;
      
      if (sel_eq(selector, EOQualifierOperatorEqual) == YES) {
        value = [NSArray arrayWithObject:value];
      }
      else if (sel_eq(selector, EOQualifierOperatorContains) == NO) {
        THROW_SEARCH_EXCEPTION_RELEASE_SEARCH(@"unexpected EOKeyValueQualifier "
                                             @"selector");
      }
      if ([value isKindOfClass:[NSArray class]] == NO)
        THROW_SEARCH_EXCEPTION_RELEASE_SEARCH(@"expected an array in contains-"
                                              @"qualifier");
      enumerator = [value objectEnumerator];
      while ((value = [enumerator nextObject])) {
        value = [value lowercaseString];
        if ([FlagKeyWords containsObject:value] == YES) {

          if (insertNot == YES) {
            [search appendString:@" not "];
          }
          else {
            [search appendString:@" "];
          }
          [search appendString:value];
        }
        else
          THROW_SEARCH_EXCEPTION_RELEASE_SEARCH(@"unexpected keyword for "
                                                @"EOKeyValueQualifier");
      }
    }
    else {
      if (insertNot == YES) 
        [search appendString:@" not"];
      
      if ([key isEqualToString:@"date"] == YES) {
        if ([value isKindOfClass:[NSCalendarDate class]] == NO)
          THROW_SEARCH_EXCEPTION_RELEASE_SEARCH(@"expected a NSDate as value");
        if (sel_eq(selector, EOQualifierOperatorEqual) == YES)
          [search appendString:@" senton "];
        else if (sel_eq(selector, EOQualifierOperatorGreaterThan) == YES)
          [search appendString:@" sentsince "];
        else if (sel_eq(selector, EOQualifierOperatorLessThan) == YES)
          [search appendString:@" sentbefore "];
        else
          THROW_SEARCH_EXCEPTION_RELEASE_SEARCH(@"unexpected selector");
        [search appendString:[value descriptionWithCalendarFormat:@"%d-%b-%Y"]];
      }
      else if ([key isEqualToString:@"uid"] == YES) {
        if (sel_eq(selector, EOQualifierOperatorEqual) == NO)
          THROW_SEARCH_EXCEPTION_RELEASE_SEARCH(@"unexpected qualifier 2");
        [search appendString:@" uid "];
        [search appendString:[value stringValue]];
      }
      else if ([key isEqualToString:@"size"] == YES) {
        if (sel_eq(selector, EOQualifierOperatorGreaterThan) == YES)
          [search appendString:@" larger "];
        else if (sel_eq(selector, EOQualifierOperatorLessThan) == YES)
          [search appendString:@" smaller "];
        else
          THROW_SEARCH_EXCEPTION_RELEASE_SEARCH(@"unexpected qualifier 3");
        [search appendString:[value stringValue]];
      }
      else if ([OtherKeyWords containsObject:key] == YES) {
        if (sel_eq(selector, EOQualifierOperatorEqual) == NO)
          THROW_SEARCH_EXCEPTION_RELEASE_SEARCH(@"unexpected qualifier 4");
        [search appendString:@" "];
        [search appendString:key];
        [search appendString:@" \""];
        [search appendString:[value stringValue]];
        [search appendString:@"\""];
      }
      else {
        if (sel_eq(selector, EOQualifierOperatorEqual) == NO)
          THROW_SEARCH_EXCEPTION_RELEASE_SEARCH(@"unexpected qualifier 5");
        [search appendString:@" header "];
        [search appendString:key];
        [search appendString:@" \""];
        [search appendString:[value stringValue]];
        [search appendString:@"\""];
      }
    }
  }
  return [search autorelease];
}

/* siehe draft-ietf-imapext-thread-12.txt
   returns an array of uids in sort order */

- (NSDictionary *)threadBySubject:(BOOL)_bySubject
  charset:(NSString *)_charSet
{
  NSString *threadStr;
  NSString *threadAlg;

  if (_bySubject)
    threadAlg = @"REFERENCES";
  else
    threadAlg = @"ORDEREDSUBJECT";
    
  
  if (![_charSet length])
    _charSet = @"UTF-8";

  threadStr = [NSString stringWithFormat:@"UID THREAD %@ %@ ALL",
                      threadAlg, _charSet];

  return [self normalizeThreadResponse:[self processCommand:threadStr]];
}

/* siehe draft-ietf-imapext-sort */
/* returns an array of uids in sort order */

- (NSDictionary *)sort:(NSArray *)_sortOrderings
  qualifier:(EOQualifier *)_qual
{
  NSMutableString *sortStr;
  NSEnumerator    *enumerator;
  EOSortOrdering  *so;
  BOOL            isFirst = YES;

  sortStr = [NSMutableString stringWithCapacity:128];

  if (![_sortOrderings count]) {
    return [NSDictionary dictionaryWithObjectsAndKeys:@"result",
                         [NSNumber numberWithBool:NO], nil];
  }
  enumerator = [_sortOrderings objectEnumerator];

  [sortStr appendString:@"UID SORT ("];
  
  while ((so = [enumerator nextObject])) {
    static NSArray *AllowedSortKeys = nil;
    SEL            sel;
    NSString       *key;

    if (AllowedSortKeys == nil) {
      AllowedSortKeys = [[NSArray alloc] initWithObjects:
                                         @"ARRIVAL",
                                         @"CC",
                                         @"DATE",
                                         @"FROM",
                                         @"SIZE",
                                         @"SUBJECT",
                                         @"TO", nil];
    }
    key = [[so key] uppercaseString];

    if (![key length])
      continue;

    if (![AllowedSortKeys containsObject:key]) {
      NSLog(@"ERROR[%s] key %@ is not allowed here!", __PRETTY_FUNCTION__,
            key);
      continue;
    }
    if (isFirst)
      isFirst = NO;
    else
      [sortStr appendString:@" "];
    sel = [so selector];

    if (sel_eq(sel, EOCompareDescending) ||
        sel_eq(sel, EOCompareCaseInsensitiveDescending)) {
      [sortStr appendString:@"REVERSE "];
    }
    [sortStr appendString:[so key]];
  }
  if (isFirst) { /* found no valid key use date sorting */
    [sortStr appendString:@"DATE"];
#if 0    
    return [NSDictionary dictionaryWithObjectsAndKeys:
                         [NSNumber numberWithBool:NO], @"result", nil];
#endif    
  }
  [sortStr appendString:@") "];
  {
    static NSString *Encoding = nil;

    if (Encoding == nil) {
      Encoding = [[NSUserDefaults standardUserDefaults]
                                  stringForKey:@"ImapSortEncoding"];
      [Encoding retain];
      if (!Encoding) {
        Encoding = @"UTF-8";
      }
    }
    [sortStr appendString:Encoding];
  }
  [sortStr appendString:[self _searchExprForQual:_qual]];
  return [self normalizeSortResponse:[self processCommand:sortStr]];
}

- (NSDictionary *)searchWithQualifier:(EOQualifier *)_qualifier {
  return [self normalizeSearchResponse:
               [self processCommand:[@"search" stringByAppendingString:
                [self _searchExprForQual:_qualifier]]]];
}

/*
** Filter for all responses
**       result  : NSNumber (response result)
**       exists  : NSNumber (number of exists mails in selectet folder
**       recent  : NSNumber (number of recent mails in selectet folder
**       expunge : NSArray  (message sequence number of expunged mails
                             in selectet folder)
*/

- (NSMutableDictionary *)normalizeResponse:(NGHashMap *)_map {
  NSMutableDictionary *result;
  id                  obj;
  NSDictionary        *respRes;

  if (!_map)
    return (id)[NSMutableDictionary dictionary];

  respRes = [[_map objectEnumeratorForKey:@"ResponseResult"] nextObject];
  result  = [NSMutableDictionary dictionaryWithCapacity:32];
  [result setObject:_map forKey:@"RawResponse"];


  if ((obj = [_map objectForKey:@"bye"])) {
    [result setObject:NoNumber forKey:@"result"];
    [result setObject:obj forKey:@"reason"];
    [self closeConnection];
    return result;
  }

  if ([[respRes objectForKey:@"result"] isEqual:@"ok"] == YES) {
    [result setObject:YesNumber forKey:@"result"];
  }
  else {
    id tmp = nil;
    [result setObject:NoNumber forKey:@"result"];
    if ((tmp = [respRes objectForKey:@"description"]) != nil) {
      [result setObject:tmp forKey:@"reason"];
    }
    return result;
  }
  if ((obj = [[_map objectEnumeratorForKey:@"exists"] nextObject]) != nil) { // 
    [result setObject:obj forKey:@"exists"];
  }
  if ((obj = [[_map objectEnumeratorForKey:@"recent"] nextObject]) != nil) {
    [result setObject:obj forKey:@"recent"];
  }
  if ((obj = [_map objectsForKey:@"expunge"]) != nil) {
    [result setObject:obj forKey:@"expunge"];
  }
  return result;
}


/*
** filter for sort response
**       search  : NSArray (msn)
*/

- (NSDictionary *)normalizeSortResponse:(NGHashMap *)_map {
  id                  obj;
  NSMutableDictionary *result;

  result = [self normalizeResponse:_map];
  
  if ((obj = [[_map objectEnumeratorForKey:@"sort"] nextObject]))
    [result setObject:obj forKey:@"sort"];
  
  return result;
}

/*
** filter for capability respone
**       capability  : NSArray 
*/

- (NSDictionary *)normalizeCapabilityRespone:(NGHashMap *)_map {
  id                  obj;
  NSMutableDictionary *result;

  result = [self normalizeResponse:_map];
  
  if ((obj = [[_map objectEnumeratorForKey:@"capability"] nextObject]))
    [result setObject:obj forKey:@"capability"];
  
  return result;
}

/*
** filter for thread response
**       thread  : NSArray (msn) 
*/

- (NSDictionary *)normalizeThreadResponse:(NGHashMap *)_map {
  id                  obj;
  NSMutableDictionary *result;

  result = [self normalizeResponse:_map];
  
  if ((obj = [[_map objectEnumeratorForKey:@"thread"] nextObject]))
    [result setObject:obj forKey:@"thread"];
  
  return result;
}


/*
** filter for search response
**       search  : NSArray (msn)
*/

- (NSDictionary *)normalizeSearchResponse:(NGHashMap *)_map {
  id                  obj;
  NSMutableDictionary *result;

  result = [self normalizeResponse:_map];
  
  if ((obj = [[_map objectEnumeratorForKey:@"search"] nextObject]))
    [result setObject:obj forKey:@"search"];
  
  return result;
}

/*
** filter for select response
**       flags  : NSArray
**       unseen : NSNumber
**       access  : NSString ([READ-WRITE], ... )
*/

- (NSDictionary *)normalizeSelectResponse:(NGHashMap *)_map {
  id                  obj;
  NSEnumerator        *enumerator;
  NSMutableDictionary *result;

  result = [self normalizeResponse:_map];
  
  if ((obj = [[_map objectEnumeratorForKey:@"flags"] nextObject]))
    [result setObject:_imapFlags2Flags(self, obj) forKey:@"flags"];
  
  enumerator = [_map objectEnumeratorForKey:@"ok"];
  while ((obj = [enumerator nextObject])) {
    id o;
    
    if ((o = [obj  objectForKey:@"unseen"]))
      [result setObject:o forKey:@"unseen"];
  }
  
  enumerator = [_map objectEnumeratorForKey:@"no"];
  while ((obj = [enumerator nextObject])) {
    id o;
    if ([obj isKindOfClass:[NSDictionary class]]) {
      if ((o = [obj  objectForKey:@"ALERT"]))
        [result setObject:o forKey:@"alert"];
    }
    else {
      [result setObject:obj forKey:@"alert"];
    }
  }
  
  if ((obj = [[_map objectForKey:@"ResponseResult"] objectForKey:@"flag"]))
    [result setObject:obj forKey:@"access"];
  
  return result;
}

/*
** filter for status response
**       messages  : NSNumber
**       recent    : NSNumber
**       unseen    : NSNumber
*/

- (NSDictionary *)normalizeStatusResponse:(NGHashMap *)_map {
  id                  obj;
  NSMutableDictionary *result;
  id                  o;
  
  result = [self normalizeResponse:_map];
  
  obj = [[[_map objectEnumeratorForKey:@"status"] nextObject]
                objectForKey:@"flags"];

  if ((o = [obj  objectForKey:@"messages"]) != nil) {
    [result setObject:o forKey:@"messages"];
  }
  if ((o = [obj  objectForKey:@"recent"]) != nil) {
    if ([result objectForKey:@"recent"] == nil)
      [result setObject:o forKey:@"recent"];
  }
  if ((o = [obj  objectForKey:@"unseen"]) != nil) {
    [result setObject:o forKey:@"unseen"];
  }
  return result;
}

/*
** filter for fetch response
**       fetch : NSArray (fetch responses)
**             'header'  - RFC822.HEADER
**             'text'    - RFC822.TEXT
**             'size'    - SIZE
**             'flags'   - FLAGS
**             'uid'     - UID
**             'msn'     - message sequence number
**             'message' - RFC822
**             'body   ' - (dictionary with bodystructure)
*/

- (NSDictionary *)normalizeFetchResponse:(NGHashMap *)_map {
  NSMutableDictionary *result;
  id                  obj;
  NSEnumerator        *enumerator;
  NSMutableArray      *fetch;
  
  result = [self normalizeResponse:_map];
  fetch  = [NSMutableArray arrayWithCapacity:512];
  
  enumerator = [_map objectEnumeratorForKey:@"fetch"];
  
  while ((obj = [enumerator nextObject])) {
    NSDictionary *entry;
    NSEnumerator *keyEnum;
    id           key;
    id           keys[9];
    id           values[9];
    unsigned     count;
    
    keyEnum = [obj keyEnumerator];    
    count   = 0;
    
    while ((key = [keyEnum nextObject]) && (count < 9)) {
      if ([key isEqualToString:@"rfc822"]) {
        keys[count]   = @"message";
        values[count] = [obj objectForKey:key];
        count++;
      }
      else if ([key isEqualToString:@"rfc822.header"]) {
        keys[count]   = @"header";
        values[count] = [obj objectForKey:key];
        count++;
      }
      else if ([key isEqualToString:@"rfc822.text"]) {
        keys[count]   = @"text";
        values[count] = [obj objectForKey:key];
        count++;
      }
      else if ([key isEqualToString:@"rfc822.size"]) {
        keys[count]   = @"size";
        values[count] = [obj objectForKey:key];
        count++;
      }
      else if ([key isEqualToString:@"flags"]) {
        keys[count]   = @"flags";
        values[count] = _imapFlags2Flags(self, [obj objectForKey:key]);
        count++;
      }
      else if ([key isEqualToString:@"uid"]) {
        keys[count]   = @"uid";
        values[count] = [obj objectForKey:key];
        count++;
      }
      else if ([key isEqualToString:@"msn"]) {
        keys[count]   = @"msn";
        values[count] = [obj objectForKey:key];
        count++;
      }
      else if ([(NSString *)key hasPrefix:@"body"]) {
        keys[count]   = @"body";
        values[count] = [obj objectForKey:key];
        count++;
      }
    }
    
    /* create dictionary */
    
    entry =
      [[NSDictionary alloc] initWithObjects:values forKeys:keys count:count];
    
    if (entry) {
      [fetch addObject:entry];
      [entry release]; entry = nil;
    }
  }
  [result setObject:fetch forKey:@"fetch"];
  return [[result copy] autorelease];
}

/*
** filter for quota responses
*/

- (NSDictionary *)normalizeQuotaResponse:(NGHashMap *)_map {
  NSMutableDictionary *result, *quotaRoot, *quota, *tmp;
  id                  obj;
  NSEnumerator        *enumerator;

  result     = [self normalizeResponse:_map];
  quotaRoot  = [_map objectForKey:@"quotaRoot"];
  quota      = [_map objectForKey:@"quota"];
  enumerator = [quotaRoot keyEnumerator];
  tmp        = [NSMutableDictionary dictionaryWithCapacity:[quota count]];
  
  while ((obj = [enumerator nextObject])) {
    NSString *qRoot;
    NSDictionary *qDesc;

    qRoot = [quotaRoot objectForKey:obj];

    if (![qRoot length]) {
      NSLog(@"ERROR[%s]: missing quotaroot for %@", __PRETTY_FUNCTION__, obj);
      continue;
    }
    qDesc = [quota objectForKey:qRoot];

    if (![qDesc count]) {
      NSLog(@"ERROR[%s]: missing quota description for folder %@ root %@",
            __PRETTY_FUNCTION__, obj, qRoot);
      continue;
    }
    [tmp setObject:qDesc forKey:_imapFolder2Folder(self, obj)];
  }
  [result setObject:tmp forKey:@"quotas"];
  return [[result copy] autorelease];
}


/*
** filter for open connection
*/

- (NSDictionary *)normalizeOpenConnectionResponse:(NGHashMap *)_map {
  NSMutableDictionary *result;
  id                  obj;

  result = [self normalizeResponse:_map];
  
  if ((obj = [[_map objectEnumeratorForKey:@"ok"] nextObject]) != nil) {

    if ([obj isKindOfClass:[NSString class]]) {
      int index;

      static int VLen;
      static NSString *VersionPrefix = nil;

      /*
       Cyrus IMAP4 v1.6.22 server ready
      */
      if (VersionPrefix == nil) {
        VersionPrefix = [[NSUserDefaults standardUserDefaults]
                                         stringForKey:
                                         @"ImapGreetingsVersionPrefix"];
        if (!VersionPrefix) {
          VersionPrefix = @"Cyrus IMAP4 v";
        }
        [VersionPrefix retain];
        VLen = [VersionPrefix cStringLength];
      }
      
      [result setObject:obj forKey:@"server"];
      
      if ((index = [(NSString *)obj indexOfString:VersionPrefix])
          != NSNotFound) {
        /* cyrus server */
        [result setObject:@"cyrus" forKey:@"serverKind"];
        {
          NSArray *vers = [[obj substringFromIndex:index+VLen]
                                componentsSeparatedByString:@"."];

          if ([vers count] > 2) {
            [result setObject:[NSNumber numberWithInt:[[vers objectAtIndex:0]
                                                             intValue]]
                    forKey:@"version"];
            [result setObject:[NSNumber numberWithInt:[[vers objectAtIndex:1]
                                                             intValue]]
                    forKey:@"subversion"];
            [result setObject:[NSNumber numberWithInt:[[vers objectAtIndex:2]
                                                             intValue]]
                    forKey:@"tag"];
          }
        }
      }
    }
    [result setObject:YesNumber forKey:@"result"];
  }
  else
    [result setObject:NoNumber forKey:@"result"];
  return result;
}

/*
** filter for list
**       list : NSDictionary (folder name as key and flags as value)
*/

- (NSDictionary *)normalizeListResponse:(NGHashMap *)_map {
  NSMutableDictionary *result;
  id                  obj;
  NSAutoreleasePool   *pool;
  NSDictionary        *rr;
  
  pool   = [[NSAutoreleasePool alloc] init];
  result = [self normalizeResponse:_map];
  
  if ((obj = [_map objectsForKey:@"list"]) != nil) {
    NSEnumerator        *enumerator = nil;
    id                  o           = nil;
    NSMutableDictionary *folder     = nil;    
    
    enumerator = [obj objectEnumerator];
    folder     = [[NSMutableDictionary alloc] init];
    
    while ((o = [enumerator nextObject])) {
      [folder setObject:_imapFlags2Flags(self, [o objectForKey:@"flags"])
              forKey:_imapFolder2Folder(self, [o objectForKey:@"folderName"])];
    }
    
    {
      NSDictionary *f;
      
      f = [folder copy];
      [result setObject:f forKey:@"list"];
      [f release];      f      = nil;
      [folder release]; folder = nil;
    }
  }
  rr = [result copy];
  [pool release];
  
  return [rr autorelease];
}

- (NSString *)description {
  return [NSString stringWithFormat:@"<Imap4Client[0x%08X]: socket=%@>",
                     (unsigned)self, [self socket]];
}

/* Private Methods */

- (NGHashMap *)processCommand:(NSString *)_command withTag:(BOOL)_tag
  withNoticication:(BOOL)_notification logText:(NSString *)_txt
{
  NGHashMap    *map;
  BOOL         tryReconnect;
  int          reconnectCnt;
  NSException  *exception;

  struct timeval tv;
  double         ti = 0.0;

  if (ProfileImapEnabled) {
    gettimeofday(&tv, NULL);
    ti =  (double)tv.tv_sec + ((double)tv.tv_usec / 1000000.0);
    fprintf(stderr, "{");
  }
  tryReconnect  = NO;
  reconnectCnt = 0;
  map           = nil;
  exception     = nil;

  do {
    NSString *loginStr = @"login";
    
    tryReconnect  = NO;
    [self->context resetLastException];
    NS_DURING {
      [self sendCommand:_command withTag:_tag logText:_txt];
      map = [self->parser parseResponseForTagId:self->tagId];
      tryReconnect = NO;
    }
    NS_HANDLER {
      if ([localException isKindOfClass:
                          [NGImap4ParserException class]] == YES) {
        fprintf(stderr, "ERROR[%s]: Got Parser-Exception %s\n",
                __PRETTY_FUNCTION__, [[localException description] cString]);
      }
      else {
        fprintf(stderr, "ERROR[%s]: !!!! Got Parser-Exception %s\n",
                __PRETTY_FUNCTION__, [[localException description] cString]);
        
        if (reconnectCnt == 0) {
          if (![_command hasPrefix:loginStr]) {
            reconnectCnt++;
            tryReconnect = YES;
            exception    = localException;
          }
        }
      }
      [self closeConnection];
      [self->context setLastException:localException];
    }
    NS_ENDHANDLER;

    if (tryReconnect) {
      [self reconnect];
    }
    else if ([map objectForKey:@"bye"] && ![_command hasPrefix:@"logout"]) {
      if (reconnectCnt == 0) {
        reconnectCnt++;
        tryReconnect = YES;
        [self reconnect];
      }
    }
  } while (tryReconnect == YES);

  if ([self->context lastException]) {
    if (exception) {
      [self->context setLastException:exception];
    }
    return nil;
  }
  if (_notification == YES) [self sendResponseNotification:map];

  if (ProfileImapEnabled) {
    gettimeofday(&tv, NULL);
    ti = (double)tv.tv_sec + ((double)tv.tv_usec / 1000000.0) - ti;
    fprintf(stderr, "}[%s] <Send Command [%s]> : time needed: %4.4fs\n",
           __PRETTY_FUNCTION__, [_command cString], ti < 0.0 ? -1.0 : ti);    
  }
  return map;
}

- (NGHashMap *)processCommand:(NSString *)_command withTag:(BOOL)_tag
  withNoticication:(BOOL)_notification
{
  return [self processCommand:_command withTag:_tag
               withNoticication:_notification
               logText:_command];
}

- (NGHashMap *)processCommand:(NSString *)_command withTag:(BOOL)_tag {
  return [self processCommand:_command withTag:_tag withNoticication:YES
               logText:_command];
}

- (NGHashMap *)processCommand:(NSString *)_command {
  return [self processCommand:_command withTag:YES withNoticication:YES
               logText:_command];
}

- (NGHashMap *)processCommand:(NSString *)_command logText:(NSString *)_txt {
  return [self processCommand:_command withTag:YES withNoticication:YES
               logText:_txt];
}

- (void)sendCommand:(NSString *)_command withTag:(BOOL)_tag
  logText:(NSString *)_txt
{
  NSString      *command;
  NGCTextStream *txtStream;

  txtStream = [self textStream];

  if (_tag == YES) {
    self->tagId++;

    command = [NSString stringWithFormat:@"%d %@", self->tagId, _command];
    if (self->debug) {
      _txt = [NSString stringWithFormat:@"%d %@", self->tagId, _txt];
    }
  }
  else
    command = _command;

  if (self->debug) {
    if ([_txt length] > 5000) {
      fprintf(stderr, "C[%p]: %s...\n", self, [[_txt substringToIndex:5000]
                                                  cString]);
    }
    else {
      fprintf(stderr, "C[%p]: %s\n", self, [_txt cString]);
    }
  }
  if (![txtStream writeString:command])
    [self->context setLastException:[txtStream lastException]];
  
  if (![txtStream writeString:@"\r\n"])
    [self->context setLastException:[txtStream lastException]];
  
  if (![txtStream flush])
    [self->context setLastException:[txtStream lastException]];
}
  
- (void)sendCommand:(NSString *)_command withTag:(BOOL)_tag {
  [self sendCommand:_command withTag:_tag logText:_command];
}

- (void)sendCommand:(NSString *)_command {
  [self sendCommand:_command withTag:YES logText:_command];
}

static inline NSArray *_imapFlags2Flags(NGImap4Client *self, NSArray *_flags) {
  NSEnumerator *enumerator;
  NSArray      *result;
  id           obj, *objs;
  unsigned     cnt;
  
#if NeXT_Foundation_LIBRARY
  objs = NSZoneMalloc([self zone], sizeof(id) * [_flags count]);
#else
  objs = NSZoneMallocAtomic([self zone], sizeof(id) * [_flags count]);
#endif

  enumerator = [_flags objectEnumerator];
  cnt = 0;
  while ((obj = [enumerator nextObject])) {
    if ([obj length] > 0) {
      if ([[obj substringToIndex:1] isEqualToString:@"\\"]) {
        objs[cnt] = [obj substringFromIndex:1];
        cnt++;
      }
    }
  }
  result = [NSArray arrayWithObjects:objs count:cnt];
  NGFree(objs);
  return result;
}

static inline NSArray *_flags2ImapFlags(NGImap4Client *self, NSArray *_flags) {
  NSEnumerator *enumerator;
  NSArray      *result;
  id           obj;
  id           *objs;
  unsigned     cnt;

#if NeXT_Foundation_LIBRARY
  objs = NSZoneMalloc([self zone], sizeof(id) * [_flags count]);
#else
  objs = NSZoneMallocAtomic([self zone], sizeof(id) * [_flags count]);
#endif

  cnt = 0;
  enumerator = [_flags objectEnumerator];
  while ((obj = [enumerator nextObject])) {
    objs[cnt] = [@"\\" stringByAppendingString:obj];
    cnt++;
  }
  result = [NSArray arrayWithObjects:objs count:cnt];
  NGFree(objs);
  return result;
}

static inline NSString *_folder2ImapFolder(NGImap4Client *self,
                                           NSString *_folder)
{
  NSArray *array;
  
  if (self->delimiter == nil) {
    NSDictionary *res;

    res = [self list:@"" pattern:@""];

    if (!_checkResult(self->context, res, __PRETTY_FUNCTION__))
      return nil;
  }

  array = [_folder pathComponents];

  if ([array count] > 0) {
    NSString *o;

    o = [array objectAtIndex:0];
    if (([o isEqualToString:@"/"] == YES) || ([o length] == 0)) {
        array = [array subarrayWithRange:NSMakeRange(1, [array count] - 1)];
    }
    
    o = [array lastObject];
    if (([o length] == 0) || ([o isEqualToString:@"/"] == YES))
      array = [array subarrayWithRange:NSMakeRange(0, [array count] - 1)];
  }
  return [[array componentsJoinedByString:self->delimiter]
                 stringByEncodingImap4FolderName];
}

static inline NSString *_imapFolder2Folder(NGImap4Client *self,
                                           NSString *_folder)
{
  NSArray *array;
  
  array = [NSArray arrayWithObject:@""];

  if (self->delimiter == nil) {
    NSDictionary *res;

    res = [self list:@"" pattern:@""];

    if (!_checkResult(self->context, res, __PRETTY_FUNCTION__))
      return nil;
  }
  
  array = [array arrayByAddingObjectsFromArray:
                   [_folder componentsSeparatedByString:self->delimiter]];
  
  return [[NSString pathWithComponents:array] stringByDecodingImap4FolderName];
}


- (void)setContext:(NGImap4Context *)_ctx {
  self->context = _ctx;
}
- (NGImap4Context *)context {
  return self->context;
}

@end /* NGImap4Client */

@implementation NGImap4Client(ConnectionRegistration)

- (void)removeFromConnectionRegister {
  int cnt;
  
  for (cnt = 0; cnt < MaxImapClients; cnt++) {
    if (ImapClients[cnt] == self) {
      ImapClients[cnt] = nil;
    }
  }
}

- (void)registerConnection {
  int cnt;

  cnt =  CountClient % MaxImapClients;

  if (ImapClients[cnt]) {
    [(NGImap4Context *)ImapClients[cnt] closeConnection];
  }
  ImapClients[cnt] = self;
  CountClient++;
}

- (NGCTextStream *)textStream {
  if (self->text == nil) {
    if ([self->context lastException] == nil)
      [self reconnect];
  }
  return self->text;
}

@end /* NGImap4Client(ConnectionRegistration); */
