// $Id: NGImap4Context.h,v 1.7 2003/01/29 21:15:58 jan Exp $

#ifndef __Networking_NGImap4_NGImap4Context_H__
#define __Networking_NGImap4_NGImap4Context_H__

#import <Foundation/NSObject.h>
#import <NGImap4/NGImap4Support.h>

@class NSDictionary, NGHashMap, NSMutableArray, NSURL;
@class NGImap4Client, NGImap4Folder, NGImap4ServerRoot;

@interface NGImap4Context : NSObject <NGImap4ResponseReceiver>
{
  NSDictionary      *connectionDictionary;
  NSMutableArray    *folderForRefresh;
  NGImap4Client     *client;
  NGImap4Folder     *selectedFolder; /* not retained */
  NGImap4Folder     *trashFolder;    /* not retained */
  NGImap4Folder     *sentFolder;     /* not retained */
  NGImap4Folder     *draftsFolder;   /* not retained */
  NGImap4Folder     *inboxFolder;    /* not retained */  
  NGImap4ServerRoot *serverRoot;     /* not retained */

  NSArray           *capability;

  NSString *sentFolderName;
  NSString *trashFolderName;
  NSString *draftsFolderName;

  NSString *serverName;
  NSString *serverKind;
  NSNumber *serverVersion;
  NSNumber *serverSubVersion;
  NSNumber *serverTag;
  
  BOOL syncMode;
  
  NSException *lastException;

  NSURL *url;

  int canSort;
  int canQuota;
}


+ (id)imap4ContextWithURL:(NSURL *)_url;
+ (id)imap4ContextWithConnectionDictionary:(NSDictionary *)_connection;
- (id)initWithURL:(NSURL *)_url;
- (id)initWithConnectionDictionary:(NSDictionary *)_connection;

- (NGImap4Client *)client;

- (NSURL *)url;

- (BOOL)isSelectedFolder:(NGImap4Folder *)_folder;
- (BOOL)registerAsSelectedFolder:(NGImap4Folder *)_folder;
- (BOOL)removeSelectedFolder:(NGImap4Folder *)_folder;

- (BOOL)openConnection;
- (BOOL)closeConnection;

// NGImap4ResponseReceiver protocol

- (void)responseNotificationFrom:(NGImap4Client *)_client
  response:(NSDictionary *)_dict;

- (id)trashFolder;
- (id)sentFolder;
- (id)draftsFolder;
- (id)inboxFolder;
- (id)serverRoot;
- (void)setSentFolder:(NGImap4Folder *)_folder;
- (void)setTrashFolder:(NGImap4Folder *)_folder;
- (void)setDraftsFolder:(NGImap4Folder *)_folder;

- (void)resetSpecialFolders;

- (BOOL)hasNewMessages;
/* returns all new messages from registered folder and inbox */
- (NSArray *)newMessages;

- (BOOL)registerForRefresh:(NGImap4Folder *)_folder;
- (BOOL)removeFromRefresh:(NGImap4Folder *)_folder;
- (BOOL)removeAllFromRefresh;
- (BOOL)refreshFolder;

- (NGImap4Folder *)folderWithName:(NSString *)_name;
- (NGImap4Folder *)folderWithName:(NSString *)_name caseInsensitive:(BOOL)_b;

- (BOOL)createFolderWithPath:(NSString *)_name;

- (NSString *)host;
- (NSString *)login;

- (id)serverName;
- (id)serverKind;
- (id)serverVersion;
- (id)serverSubVersion;
- (id)serverTag;

- (BOOL)isInSyncMode;
- (void)enterSyncMode;
- (void)leaveSyncMode;
- (void)resetSync;

- (void)setLastException:(NSException *)_exception;
- (NSException *)lastException;
- (void)resetLastException;

@end

@interface NGImap4Context(Capability)

- (BOOL)canSort;
- (BOOL)canQuota;

@end /* NGImap4Context(Capability) */
#endif /* __Networking_NGImap4_NGImap4Context_H__ */
