// $Id: NGImap4Folder.h,v 1.9 2003/01/09 14:25:15 jan Exp $
#ifndef __Networking_NGImap4_NGImap4Folder_H__
#define __Networking_NGImap4_NGImap4Folder_H__

#import <Foundation/Foundation.h>
#import <NGMime/NGPart.h>
#import <NGImap4/NGImap4Support.h>

#define USE_MESSAGE_CACHE 0

@class NSArray, NSString, NSMutableArray, NSNumber;
@class NGHashMap, NGImap4Context, EOQualifier, NGImap4Message;

@interface NGImap4Folder : NSObject <NGImap4Folder>
{
@private  
  NSArray           *flags;
  NSString          *name;
  NSURL             *url;
  NGImap4Context    *context;
  NSArray           *subFolders;

  NSArray           *msn2UidCache;
  
  id<NGImap4Folder> parentFolder; // not retained

  NSNumber *isReadOnly;
  NSArray  *messageFlags;
  
  int exists;
  int recent;
  int unseen;
  
  BOOL responseInProcess;
  BOOL noselect;
  BOOL noinferiors;

  BOOL selectSyncState;

  int maxQuota;
  int usedSpace;
  int overQuota;
  
#if USE_MESSAGE_CACHE
  int cacheIdx;
  
  NSMutableArray *messages;  
  NSMutableArray *qualifierCache;
  NSMutableArray *messagesCache;
#endif  
}

+ (id)imap4FolderWithContext:(NGImap4Context *)_context
  name:(NSString *)_name
  flags:(NSArray *)_flags
  parentFolder:(id<NGImap4Folder>)_folder;

- (id)initImap4FolderWithContext:(NGImap4Context *)_context
  name:(NSString *)_name
  flags:(NSArray *)_flags
  parentFolder:(id<NGImap4Folder>)_folder;

- (BOOL)isEqual:(id)_obj;
- (BOOL)isEqualToImap4Folder:(NGImap4Folder *)_folder;

/* accessors */

- (NGImap4Context *)context;
- (NSException *)lastException;
- (void)resetLastException;
- (id<NGImap4Folder>)parentFolder;

- (NSString *)name;
- (NSString *)absoluteName;
- (NSArray *)flags;

- (NSData *)blobForUid:(unsigned)_mUid
  part:(NSString *)_part;

- (NSArray *)fetchSortedMessages:(NSArray *)_so;

- (void)bulkFetchHeadersFor:(NSArray *)_array inRange:(NSRange)_aRange;

- (NSArray *)fetchSortedMessages:(NSRange)_aRange
  sortOrderings:(NSArray *)_so;
- (NSArray *)messageFlags;
- (NSArray *)messages;
- (NSArray *)messagesForQualifier:(EOQualifier *)_qualifier;
- (NSArray *)messagesForQualifier:(EOQualifier *)_qualifier maxCount:(int)_cnt;
- (NSArray *)subFolders;
- (NGImap4Folder *)subFolderWithName:(NSString *)_name
                     caseInsensitive:(BOOL)_caseIns;
- (NGImap4Message *)messageForUid:(unsigned)_mUid
  sortOrderings:(NSArray *)_so
  onlyUnread:(BOOL)_unread
  nextMessage:(BOOL)_next;

- (BOOL)isReadOnly;
- (BOOL)noselect;
- (BOOL)noinferiors;
- (int)exists;
- (int)recent;
- (int)unseen;

/* this folder or its subfolders */

- (BOOL)hasNewMessagesSearchRecursiv:(BOOL)_recursiv;
- (BOOL)hasUnseenMessagesSearchRecursiv:(BOOL)_recursiv;

/* Notifications */

- (void)processResponse:(NSDictionary *)_dict;

/* Actions */
- (BOOL)status;
- (BOOL)select;
/* if imm == YES syncState will be ignored */
- (BOOL)selectImmediately:(BOOL)_imm;
- (void)expunge;

- (BOOL)addFlag:(NSString *)_flag toMessages:(NSArray *)_messages;
- (BOOL)removeFlag:(NSString *)_flag fromMessages:(NSArray *)_messages;
- (BOOL)renameTo:(NSString *)_name;

/*
  returns quota in kBytes
*/

- (int)usedSpace;
- (int)maxQuota;
- (BOOL)isOverQuota; /* evaluate ALERT sequences during select */

- (BOOL)deleteMessages:(NSArray *)_messages;
- (BOOL)deleteAllMessages;
- (BOOL)moveMessages:(NSArray *)_messages toFolder:(NGImap4Folder *)_folder;
- (BOOL)copyMessages:(NSArray *)_messages toFolder:(NGImap4Folder *)_folder;
- (BOOL)appendMessage:(NSData *)_msg;

- (BOOL)deleteSubFolder:(NGImap4Folder *)_folder;
- (BOOL)createSubFolderWithName:(NSString *)_name;
- (BOOL)copySubFolder:(NGImap4Folder *)_f to:(NGImap4Folder *)_folder;
- (BOOL)moveSubFolder:(NGImap4Folder *)_f to:(NGImap4Folder *)_folder;

- (BOOL)isInTrash;

- (void)resetFolder;
- (void)resetSubFolders;
- (void)resetStatus;

- (void)resetSync;

- (NSURL *)url;
@end

#endif /* __Networking_NGImap4_NGImap4Folder_H__ */
