// $Id: NGImap4Folder.m,v 1.36 2003/01/31 13:58:07 jan Exp $

#import "common.h"
#import "NGImap4Folder.h"
#import "NGImap4Context.h"
#import "NGImap4Client.h"
#import "NGImap4Message.h"
#import "NGImap4Functions.h"

@interface NGImap4Message(Private)

- (void)_setHeaders:(NGHashMap *)_headers
  size:(unsigned)_size
  flags:(NSArray *)_flags;

- (BOOL)isComplete;

@end /* NGImap4Message(Private) */


@interface NGImap4Context(Private)
- (void)setLastException:(NSException *)_exception;
- (void)setSelectedFolder:(NGImap4Folder *)_folder;
@end /*NGImap4Context(Private) */

@interface NGImap4Folder(Private)

- (void)quota;
- (NSArray *)initializeMessagesFrom:(unsigned)_from to:(unsigned)_to;
- (NSArray *)initializeMessages;
- (void)initializeSubFolders;
- (void)addSubFolder:(NGImap4Folder *)_folder;
- (BOOL)flag:(NSString*)_doof toMessages:(NSArray*)_msg add:(NSNumber*)_n;
- (BOOL)flagToAllMessages:(NSString *)_flag add:(NSNumber *)_add;
- (NSArray *)fetchMessagesFrom:(unsigned)_from to:(unsigned)_to;
#if USE_MESSAGE_CACHE
- (void)resetQualifierCache;
#endif
- (void)setRecent:(NSNumber *)_rec exists:(NSNumber *)_exists;
@end /* NGImap4Folder(Private) */

@implementation NGImap4Folder

static inline BOOL    _testMessages(NGImap4Folder*,NSArray*,NSString*);
static inline NSArray *_getMsnRanges(NGImap4Folder*,NSArray*);
static inline NSArray *_calculateSequences(NGImap4Folder  *self,
                                           NSMutableArray *_numbers,
                                           int            _cnt);
  
static NSNumber *YesNumber;
static NSNumber *NoNumber;
static NSArray  *StatusFlags;
static NSArray  *UnseenFlag;

+ (void)initialize {
  if (YesNumber == nil) YesNumber = [[NSNumber numberWithBool:YES] retain];
  if (NoNumber  == nil) NoNumber  = [[NSNumber numberWithBool:NO]  retain];
  if (StatusFlags == nil)
    StatusFlags = [[NSArray alloc]
                            initWithObjects:@"messages", @"recent", @"unseen",
                              nil];
  if (UnseenFlag == nil)
    UnseenFlag = [[NSArray alloc] initWithObjects:@"unseen", nil];
}

+ (id)imap4FolderWithContext:(NGImap4Context *)_context
  name:(NSString *)_name
  flags:(NSArray *)_flags
  parentFolder:(id<NGImap4Folder>)_folder
{
  return [[[self alloc]
                 initImap4FolderWithContext:_context
                 name:_name flags:_flags parentFolder:_folder]
                 autorelease];
} 

- (id)init {
  [self release];
  [self doesNotRecognizeSelector:_cmd];
  return nil;
}

- (id)initImap4FolderWithContext:_context
  name:(NSString *)_name
  flags:(NSArray *)_flags
  parentFolder:(id<NGImap4Folder>)_folder
{
  if ((self = [super init])) {
    NSZone *z = [self zone];
    self->context           = [_context retain];
    self->flags             = [_flags   retain];
    self->noselect          = [self->flags containsObject:@"noselect"];
    self->noinferiors       = [self->flags containsObject:@"noinferiors"];    
    self->name              = [_name copyWithZone:z];
    self->parentFolder      = _folder;
    self->responseInProcess = NO;
    self->exists            = -1;
    self->recent            = -1;
    self->unseen            = -1;
    self->usedSpace         = -1;
    self->maxQuota          = -1;
    self->overQuota         = -1;
    self->selectSyncState   = NO;

    if ([[self->name lowercaseString] isEqualToString:@"/inbox"] &&
        self->noselect == YES) {
      NSDictionary *res;

      [self resetLastException];
      
      res = [[self->context client] subscribe:[self absoluteName]];

      if ([self lastException] != nil) {
        [self release];
        return nil;
      }
      
      if ([[res objectForKey:@"result"] boolValue]) {
        id tmp;
        
        self->noselect = NO;
        tmp = self->flags;

        self->flags = [self->flags mutableCopy];
        [(NSMutableArray *)self->flags removeObject:@"noselect"];
        [tmp release]; tmp = nil;
      }
    }
#if USE_MESSAGE_CACHE
    self->cacheIdx          = 0;    
    self->qualifierCache =
      [[NSMutableArray allocWithZone:z] initWithCapacity:MAX_QUALIFIER_CACHE];
    self->messagesCache  =
      [[NSMutableArray allocWithZone:z] initWithCapacity:MAX_QUALIFIER_CACHE];
#endif    
  }
#if NGIMAP_FOLDER_DEBUG
  return [self->context registerFolder:self];
#else  
  return self;
#endif  
}

- (void)dealloc {
  [self->context removeSelectedFolder:self];
  [self->msn2UidCache release];
  [self->context      release];
  [self->flags        release];
  [self->name         release];
  [self->subFolders   release]; 
  [self->messageFlags release];
  [self->url          release];
#if USE_MESSAGE_CACHE
  [self->messages       release];
  [self->qualifierCache release];
  [self->messagesCache  release];
#endif
  self->isReadOnly   = nil;  
  self->parentFolder = nil;  
  [super dealloc];
}

- (BOOL)isEqual:(id)_obj {
  if (self == _obj)
    return YES;
  if ([_obj isKindOfClass:[NGImap4Folder class]])
    return [self isEqualToImap4Folder:_obj];
  return NO;
}

- (BOOL)isEqualToImap4Folder:(NGImap4Folder *)_folder {
  if (self == _folder)
    return YES;
  if (([[_folder absoluteName] isEqualToString:self->name] == YES) &&
      [_folder context] == self->context) {
    return YES;
  }
  return NO;
}

/* accessors */

- (NSException *)lastException {
  return [self->context lastException];
}
- (void)resetLastException {
  [self->context resetLastException];
}

- (NGImap4Context *)context {
  return self->context;
}

- (NSString *)name {
  return [[self absoluteName] lastPathComponent];
}

- (NSString *)absoluteName {
  return self->name;
}

- (NSArray *)flags {
  return self->flags;
}

- (NSArray *)messages {
  return [self initializeMessages];
}

- (BOOL)_checkResult:(NSDictionary *)_dict cmd:(const char *)_command {
  return _checkResult(self->context, _dict, _command);
}

- (NSArray *)messagesForQualifier:(EOQualifier *)_qualifier
  maxCount:(int)_cnt
{
  NSMutableArray    *mes  = nil;
  NSMutableArray    *msn  = nil;
  id                dict  = nil;
  NSAutoreleasePool *pool = nil;

  if (self->noselect == YES)
    return nil;
  

  if (![self->context registerAsSelectedFolder:self])
    return nil;

  pool = [[NSAutoreleasePool alloc] init];

#if USE_MESSAGE_CACHE  
  if (self->cacheIdx > 0) {
    NSEnumerator *qualifierEnum = nil;
    EOQualifier  *qual          = nil;
    int          cnt            = 0;

    qualifierEnum = [self->qualifierCache objectEnumerator];

    while ((qual = [qualifierEnum nextObject])) {
      if ([qual isEqual:_qualifier] == YES) {
        NSArray *m;
        
        m = [[self->messagesCache objectAtIndex:cnt] retain];
        [pool release];
        return [m autorelease];
      }
      cnt++;
    }
  }
#endif
  [self resetLastException];
  
  dict = [[self->context client] searchWithQualifier:_qualifier];

  if (![self _checkResult:dict cmd:__PRETTY_FUNCTION__])
    return nil;

  msn  = [[dict objectForKey:@"search"] mutableCopy];
  
  if ((msn == nil) || ([msn count] == 0)) {
    mes = [NSArray array];
  }
  else {
    NSEnumerator *seq   = nil;
    id           obj    = nil;
    
    mes = [NSMutableArray arrayWithCapacity:512];
    seq = [_calculateSequences(self, msn, _cnt) objectEnumerator];
    while ((obj = [seq nextObject])) {
      NSArray *a;

      a = [self fetchMessagesFrom:
                [[obj objectForKey:@"start"] unsignedIntValue]
                to: [[obj objectForKey:@"end"] unsignedIntValue]];

      if ([self lastException])
        break;

      if (a)
        [mes addObjectsFromArray:a];
    }
    mes = [[mes copy] autorelease];
  }
  [msn release];

#if USE_MESSAGE_CACHE  
  if (self->cacheIdx == 5)
    self->cacheIdx = 0;

  if ([self->qualifierCache count] == self->cacheIdx)
    [self->qualifierCache addObject:_qualifier];
  else
    [self->qualifierCache replaceObjectAtIndex:self->cacheIdx
         withObject:_qualifier];
  if ([self->messagesCache count] == self->cacheIdx)
    [self->messagesCache addObject:mes];
  else {
    [self->messagesCache replaceObjectAtIndex:self->cacheIdx
                         withObject:mes];
  }
  self->cacheIdx++;
#endif  
  
  mes = [mes retain];
  [pool release];

  if ([self lastException]) {
    [mes release];
    return nil;
  }
  return [mes autorelease];
}

- (NSArray *)messagesForQualifier:(EOQualifier *)_qualifier {
  return [self messagesForQualifier:_qualifier maxCount:-1];
}

- (NSArray *)messageFlags {
  if (self->messageFlags == nil)
    [self->context registerAsSelectedFolder:self];
  return self->messageFlags;
}

- (NSArray *)subFolders {
  if (self->subFolders == nil)
    [self initializeSubFolders];
  return self->subFolders;
}

- (NGImap4Folder *)subFolderWithName:(NSString *)_name
  caseInsensitive:(BOOL)_caseIns {
  return _subFolderWithName(self, _name, _caseIns);
}

- (id<NGImap4Folder>)parentFolder {
  return self->parentFolder;
}

- (BOOL)isReadOnly {
  if (self->isReadOnly == nil)
    [self->context registerAsSelectedFolder:self];
  return (self->isReadOnly == YesNumber) ? YES : NO;
}

- (BOOL)noselect {
  return self->noselect;
}

- (BOOL)noinferiors {
  return self->noinferiors;
}

- (int)exists {
  if (self->exists == -1) {
    [self status];
  }
  return self->exists;
}

- (int)recent {
  if (self->recent == -1) {
    [self status];
  }
  return self->recent;
}

- (int)unseen {
  if (self->unseen == -1) {
    [self status];
  }
  return self->unseen;
}

- (BOOL)isOverQuota {
  if (self->overQuota == -1) {
    [self->context registerAsSelectedFolder:self];
  }
  return (self->overQuota == 1)? YES : NO;
}

- (int)usedSpace {
  if (self->usedSpace == -1) {
    [self quota];
  }
  return self->usedSpace;
}

- (int)maxQuota {
  if (self->maxQuota == -1) {
    [self quota];
  }
  return self->maxQuota;
}

- (BOOL)hasNewMessagesSearchRecursiv:(BOOL)_recursiv {
  if (([self recent] > 0) && ([self unseen] > 0))
    return YES;

  if (_recursiv == YES) {
    return _hasNewMessagesInSubFolder(self);
  }
  return NO;
}

- (BOOL)hasUnseenMessagesSearchRecursiv:(BOOL)_recursiv {
  if ([self unseen] > 0)
    return YES;

  if (_recursiv == YES) {
    return _hasUnseenMessagesInSubFolder(self);
  }
  return NO;
}

// private methods

- (NSArray *)initializeMessages {
  return [self initializeMessagesFrom:0 to:[self exists]];
}

- (NSArray *)initializeMessagesFrom:(unsigned)_from to:(unsigned)_to {
#if USE_MESSAGE_CACHE  
  if (self->messages == nil) {
    self->messages = [[NSMutableArray alloc] initWithCapacity:_to];
  }
  [self->messages addObjectsFromArray:[self fetchMessagesFrom:_from to:_to]];
  return self->messages;
#else
  return [self fetchMessagesFrom:_from to:_to];
#endif  
}

- (NSArray *)_buildMesssagesFromFetch:(NSDictionary *)_fetch
  usingMessages:(NSDictionary *)_messages
{
  NSEnumerator        *mEnum;
  id                  m;
  NGMimeMessageParser *parser;
  NSMutableArray      *mes;
  NSAutoreleasePool   *pool;

  pool  = [[NSAutoreleasePool alloc] init];
  mEnum = [[_fetch objectForKey:@"fetch"] objectEnumerator];
  mes   = nil;
  
  if (_messages == nil)
    mes = [[NSMutableArray alloc] initWithCapacity:512];
  
  parser = [[NGMimeMessageParser alloc] init];
  
  while ((m = [mEnum nextObject])) {
    NGDataStream *stream = nil;
    id headers, uid, f, size;

    headers = [m objectForKey:@"header"];
    uid     = [m objectForKey:@"uid"];
    f       = [m objectForKey:@"flags"];
    size    = [m objectForKey:@"size"];

    if (!headers || !uid || !f || !size) {
      NSLog(@"WARNING[%s]: got no header, uid, flags, size for %@",
            __PRETTY_FUNCTION__, m);
      continue;
    }
    if (([f containsObject:@"recent"]) && ([f containsObject:@"seen"])) {
        f = [f mutableCopy];
        [f removeObject:@"recent"];
        [f autorelease];
    }
    stream = [[NGDataStream alloc] initWithData:headers];
    [parser prepareForParsingFromStream:stream];
    headers = [parser parseHeader];

    if (_messages) {
      NGImap4Message *msg;

      if (!(msg = [_messages objectForKey:uid])) {
        NSLog(@"WARNING[%s]: missing message for uid %@ from fetch %@ in "
              @"dict %@", __PRETTY_FUNCTION__, uid, _fetch, _messages);
        continue;
      }
      [msg _setHeaders:headers size:[size intValue] flags:f];
    }
    else {
      [mes addObject:[NGImap4Message messageWithUid:[uid unsignedIntValue]
                                     headers:headers
                                     size:[size intValue]
                                     flags:f
                                     folder:self
                                     context:self->context]];
    }
    [stream release]; stream = nil;
  }
  [parser release]; parser = nil;
  m = [mes copy];
  [mes release]; mes = nil;
  [pool release];
  
  return [m autorelease];;
}

- (NSArray *)_buildMesssagesFromFetch:(NSDictionary *)_fetch {
  return [self _buildMesssagesFromFetch:_fetch usingMessages:nil];
}

- (NSArray *)_messageIds:(NSArray *)_so onlyUnseen:(BOOL)_unseen {
  NSAutoreleasePool  *pool;
  NSDictionary       *dict;
  NSArray            *uids;
  static EOQualifier *UnseenQual = nil;

  uids = nil;
  
  /* hack for sorting for unseen/seen */
  if (UnseenQual == nil) {
    UnseenQual = [[EOKeyValueQualifier alloc]
                                       initWithKey:@"flags"
                                       operatorSelector:
                                       EOQualifierOperatorEqual
                                       value:@"unseen"];
  }

  pool = [[NSAutoreleasePool alloc] init];
  
  if ([_so count] == 1) {
    EOSortOrdering *so;

    so = [_so lastObject];

    if ([[so key] isEqualToString:@"unseen"]) {
      static NSArray        *DateSo     = nil;
      static EOQualifier    *SeenQual   = nil;
      
      NSMutableArray *muids;
      EOQualifier    *qual1, *qual2; 

      if (DateSo == nil) {
        DateSo = [[NSArray alloc] initWithObjects:
                          [EOSortOrdering sortOrderingWithKey:@"date"
                                           selector:[so selector]], nil];
      }
      if (SeenQual == nil) {
        SeenQual = [[EOKeyValueQualifier alloc]
                                         initWithKey:@"flags"
                                         operatorSelector:
                                         EOQualifierOperatorEqual
                                         value:@"seen"];
      }
      muids = [[NSMutableArray alloc] initWithCapacity:255];

      if (sel_eq([so selector], EOCompareAscending) ||
          sel_eq([so selector], EOCompareCaseInsensitiveAscending)) {
        qual1 = UnseenQual;
        if (_unseen)
          qual2 = nil;
        else
          qual2 = SeenQual;
      }
      else {
        if (_unseen)
          qual1 = nil;
        else
          qual1 = SeenQual;
        
        qual2 = UnseenQual;
      }

      if (qual1) {
        dict = [[self->context client] sort:DateSo qualifier:qual1];

        if ([[dict objectForKey:@"result"] boolValue] == NO) {
          NSLog(@"ERROR[%s](1): sort failed (sortOrderings %@, qual1 %@)",
                __PRETTY_FUNCTION__, DateSo, qual1);
          return nil;
        }
        [muids addObjectsFromArray:[dict objectForKey:@"sort"]];
      }
      if (qual2) {
        dict = [[self->context client] sort:DateSo qualifier:qual2];

        if ([[dict objectForKey:@"result"] boolValue] == NO) {
          NSLog(@"ERROR[%s](2): sort failed (sortOrderings %@, qual2 %@ ",
                __PRETTY_FUNCTION__, DateSo, qual2);
          return nil;
        }
        [muids addObjectsFromArray:[dict objectForKey:@"sort"]];
      }
      uids = [muids copy];
      [muids release]; muids = nil;
    }
  }
  if (uids == nil) {
    EOQualifier *qual;

    if (![_so count]) {
      static NSArray *ArrivalSO = nil;

      if (ArrivalSO == nil) {
        ArrivalSO =
          [[NSArray alloc]
                    initWithObjects:
                    [EOSortOrdering sortOrderingWithKey:@"arrival"
                                    selector:EOCompareAscending], nil];
      }
      _so = ArrivalSO;
    }
    if (_unseen) {
      qual = UnseenQual;
    }
    else
      qual = nil;

    [self resetLastException];
    
    dict = [[self->context client] sort:_so qualifier:qual];

    if (![self _checkResult:dict cmd:__PRETTY_FUNCTION__])
      return nil;

    uids = [[dict objectForKey:@"sort"] retain];
  }
  [pool release]; pool = nil;
  
  return [uids autorelease];
}

- (NSData *)blobForUid:(unsigned)_mUid
  part:(NSString *)_part
{
  NSDictionary *result;
  NSString     *bodyKey;

  if (![self->context registerAsSelectedFolder:self])
    return nil;

  bodyKey = [NSString stringWithFormat:@"body[%@]", _part?_part:@""];

  result = [[self->context client]
                           fetchUids:
                           [NSArray arrayWithObject:
                                    [NSNumber numberWithUnsignedInt:_mUid]]
                           parts:[NSArray arrayWithObject:bodyKey]];

  if (![self _checkResult:result cmd:__PRETTY_FUNCTION__]) {
    return nil;
  }
  result = [[[result objectForKey:@"fetch"]
                     lastObject] objectForKey:@"body"];
  
  return [result objectForKey:@"data"];
}

- (NGImap4Message *)messageForUid:(unsigned)_mUid
  sortOrderings:(NSArray *)_so
  onlyUnread:(BOOL)_unread
  nextMessage:(BOOL)_next
{
  NSArray      *uids, *allSortUids;
  NSEnumerator *enumerator;
  NSNumber     *uid, *eUid, *lastUid;

  uid         = [NSNumber numberWithUnsignedInt:_mUid];
  allSortUids = [self _messageIds:_so onlyUnseen:NO];
  uids        = _unread ? [self _messageIds:_so onlyUnseen:_unread] : nil;
  enumerator  = [allSortUids objectEnumerator];
  lastUid     = nil;
  
  while ((eUid = [enumerator nextObject])) {
    if ([uid isEqual:eUid])
      break;

    if (_unread) {
      if ([uids containsObject:eUid])
        lastUid = eUid;
    }
    else
      lastUid = eUid;
  }
  if (eUid == nil) {
    NSLog(@"WARNING[%s]: Couldn`t found next/prev message (missing orig. "
          @"message %d for sortordering %@",
          __PRETTY_FUNCTION__, _mUid, _so);
    return nil;
  }
  if (_next) {
    if (_unread) {
      while ((uid = [enumerator nextObject])) {
        if ([uids containsObject:uid])
          break;
      }
    }
    else
      uid = [enumerator nextObject];
  }
  else {
    uid = lastUid;
  }
  if (uid) {
    return [NGImap4Message messageWithUid:[uid unsignedIntValue]
                           folder:self context:self->context];
  }
  return nil;
}

/*
  build NGImap4Messages with sorted uids
*/

- (NSArray *)fetchSortedMessages:(NSArray *)_so {
  NSArray           *uids, *array;
  NSMutableArray    *marray;
  NSAutoreleasePool *pool;
  NSEnumerator      *enumerator;
  NSNumber          *uid;
  
  if (self->noselect == YES)
    return nil;

  if (![self->context registerAsSelectedFolder:self])
    return nil;

  pool = [[NSAutoreleasePool alloc] init];

  if (![_so count]) {
    return [self messages];
  }
  
  if (!(uids = [self _messageIds:_so onlyUnseen:NO]))
    return [self messages];

  enumerator = [uids objectEnumerator];
  marray     = [[NSMutableArray alloc] initWithCapacity:[uids count]];
  
  while ((uid = [enumerator nextObject])) {
    [marray addObject:[NGImap4Message messageWithUid:[uid intValue]
                                        folder:self context:self->context]];
  }
  array = [marray shallowCopy];

  [marray release]; marray = nil;
  [pool release];   pool   = nil;
  
  return [array autorelease];
}

/*
  fetch headers for _array in range (for use with fetchSortedMessages)
*/

- (void)bulkFetchHeadersFor:(NSArray *)_array inRange:(NSRange)_aRange {
  NSArray             *messages, *uids;
  NSAutoreleasePool   *pool;
  NSEnumerator        *enumerator;
  NGImap4Message      *message;
  NSMutableDictionary *messageMapping;
  NSDictionary        *dict;
  
  if (self->noselect == YES)
    return;

  if (_aRange.length == 0)
    return;

  if (![self->context registerAsSelectedFolder:self])
    return;

  pool = [[NSAutoreleasePool alloc] init];

  if (_aRange.location >= [_array count]) {
    return;
  }
  if (_aRange.location + _aRange.length > [_array count]) {
    _aRange.length = [_array count] - _aRange.location;
  }
  messages       = [_array subarrayWithRange:_aRange];
  enumerator     = [messages objectEnumerator];
  messageMapping = [NSMutableDictionary dictionaryWithCapacity:
                                        [messages count]];

  while ((message = [enumerator nextObject])) {
    if (![message isComplete])
      [messageMapping setObject:message
                      forKey:[NSNumber numberWithUnsignedInt:[message uid]]];
  }
  if ([messageMapping count] > 0) {
    uids = [messageMapping allKeys];

    dict = [[self->context client] fetchUids:uids
                                   parts:[NSArray arrayWithObjects:@"uid",
                                                  @"rfc822.header",
                                                  @"rfc822.size", @"flags",
                                                  nil]];
    if ([self _checkResult:dict cmd:__PRETTY_FUNCTION__])
      [self _buildMesssagesFromFetch:dict usingMessages:messageMapping];
  }
  [pool release];
  return;    
}

/*
  fetch only sorted messages in range
*/

- (NSArray *)fetchSortedMessages:(NSRange)_aRange
  sortOrderings:(NSArray *)_so
{
  NSDictionary      *dict;
  NSArray           *uids, *m;
  NSAutoreleasePool *pool;
  
  if (self->noselect == YES)
    return nil;

  if (_aRange.length == 0)
    return [NSArray array];

  if (![self->context registerAsSelectedFolder:self])
    return nil;

  pool = [[NSAutoreleasePool alloc] init];

  if (!(uids = [self _messageIds:_so onlyUnseen:NO]))
    return nil;
  
  if (_aRange.location + _aRange.length > [uids count]) {
    _aRange.length = [uids count] - _aRange.location;
  }
  
  uids = [uids subarrayWithRange:_aRange];
  dict = [[self->context client] fetchUids:uids
                                 parts:[NSArray arrayWithObjects:@"uid",
                                                @"rfc822.header",
                                                @"rfc822.size", @"flags",
                                                nil]];

  if (![self _checkResult:dict cmd:__PRETTY_FUNCTION__])
    return nil;
  
  m = [[self _buildMesssagesFromFetch:dict] retain];
  [pool release];
  return [m autorelease];
}


- (NSArray *)fetchMessagesFrom:(unsigned)_from to:(unsigned)_to {
  NSAutoreleasePool   *pool;
  NSDictionary        *dict;
  NSArray             *m;

  if (self->noselect == YES)
    return nil;
  
  if (_from == 0)
    _from = 1;
  
  if (![self->context registerAsSelectedFolder:self])
    return nil;
 
  if (_to == 0)
    return [NSArray array];

  pool = [[NSAutoreleasePool alloc] init];

  [self resetLastException];
  
  dict = [[self->context client] fetchFrom:_from to:_to
                                 parts:[NSArray arrayWithObjects:@"uid",
                                                @"rfc822.header",
                                                @"rfc822.size", @"flags",
                                                nil]];

  if (![self _checkResult:dict cmd:__PRETTY_FUNCTION__])
    return nil;
  
  m = [[self _buildMesssagesFromFetch:dict] retain];
  
  [pool release];
  return [m autorelease];
}

- (void)initializeSubFolders {
  NSEnumerator *folders;
  NSDictionary *res;
  id           folder, *objs;
  unsigned     cnt;
  
  static int ShowOnlySubscribedInSubFolders = -1;

  if (ShowOnlySubscribedInSubFolders == -1) {
    ShowOnlySubscribedInSubFolders =
      [[NSUserDefaults standardUserDefaults]
                       boolForKey:@"ShowOnlySubscribedInSubFolders"]?1:0;
  }
  
  if (self->noinferiors == YES)
    return;
  
  if (self->subFolders != nil) {
    [self resetSubFolders];
  }
  [self resetLastException];

  //  if (![self->context registerAsSelectedFolder:self])
  //    return;

  if (ShowOnlySubscribedInSubFolders) {
    res = [[self->context client]
                          lsub:@""
                          pattern:[[self absoluteName]
                                         stringByAppendingString:@"/%"]];
  }
  else {
    res = [[self->context client]
                          list:@""
                          pattern:[[self absoluteName]
                                         stringByAppendingString:@"/%"]];
  }
  if (![self _checkResult:res cmd:__PRETTY_FUNCTION__]) {
    return;
  }
  
  res = [res objectForKey:@"list"];

#if NeXT_Foundation_LIBRARY
  objs = NSZoneMalloc([self zone], sizeof(id) * [res count]);
#else
  objs = NSZoneMallocAtomic([self zone], sizeof(id) * [res count]);
#endif
  {
    NSArray *names;

    names   = [res allKeys];    
    names   = [names sortedArrayUsingSelector:
                     @selector(caseInsensitiveCompare:)];
    folders = [names objectEnumerator];
  }
  cnt = 0;
  while ((folder = [folders nextObject])) {
    objs[cnt] = [NGImap4Folder imap4FolderWithContext:self->context
                               name:folder
                               flags:[res objectForKey:folder]
                               parentFolder:self];
    if (objs[cnt] == nil)
      break;
    cnt++;
  }
  if (folder == nil) {
    self->subFolders = [[NSArray allocWithZone:[self zone]]
                                 initWithObjects:objs count:cnt];
  }
  NGFree(objs); objs = NULL;
}

- (BOOL)select {
  return [self selectImmediately:NO];
}

- (BOOL)selectImmediately:(BOOL)_imm {
  NSDictionary *dict;

  if (self->noselect == YES) {
    NSLog(@"WARNING[%s]: try to select folder with noselect flag <%@>",
          __PRETTY_FUNCTION__, self);
    return NO;
  }
  if (_imm == NO) {
    if ([[self context] isInSyncMode] == YES &&
        (self->selectSyncState == YES)) {
      return YES;
    }
  }
  [self->context setSelectedFolder:self];

  [self resetLastException];
  
  dict = [[self->context client] select:[self absoluteName]];

  if (![self _checkResult:dict cmd:__PRETTY_FUNCTION__]) {
    [self->context setSelectedFolder:nil];
    return NO;
  }
  
  [self->messageFlags release]; self->messageFlags = nil;
  
  self->messageFlags = [[dict objectForKey:@"flags"] retain];

  self->isReadOnly = 
    [[dict objectForKey:@"access"] isEqualToString:@"READ-WRITE"]
    ? NoNumber : YesNumber;
  
  if ([[dict objectForKey:@"alert"] isEqualToString:@"Mailbox is over quota"])
    self->overQuota = 1;
  else
    self->overQuota = 0;

  [self setRecent:[dict objectForKey:@"recent"]
        exists:[dict objectForKey:@"exists"]];
  
  self->maxQuota  = -1;
  self->usedSpace = -1;
  
  self->selectSyncState = YES;

  return YES;
}

- (BOOL)status {
  NSDictionary *dict;

  if (self->noselect == YES)
    return NO;

  [self->context resetLastException];

  dict = [[self->context client] status:[self absoluteName] flags:StatusFlags];

  if (![self _checkResult:dict cmd:__PRETTY_FUNCTION__]) {
    self->unseen    = -1;
    self->recent    = -1;
    self->exists    = -1;
    self->maxQuota  = -1;
    self->usedSpace = -1;
    self->overQuota = -1;
    return NO;
  }
  [self setRecent:[dict objectForKey:@"recent"]
        exists:[dict objectForKey:@"messages"]];
  self->unseen = [[dict objectForKey:@"unseen"] intValue];

  return YES;
}

// actions

- (void)resetFolder {
#if USE_MESSAGE_CACHE  
  [self resetQualifierCache];
#endif  
  [self resetSubFolders];

  [self->msn2UidCache release]; self->msn2UidCache = nil;       
  [self->flags        release]; self->flags        = nil;
  [self->messageFlags release]; self->messageFlags = nil;
#if USE_MESSAGE_CACHE  
  [self->messages     release]; self->messages     = nil;  
#endif
  self->exists    = -1;
  self->recent    = -1;
  self->unseen    = -1;
  self->maxQuota  = -1;
  self->usedSpace = -1;
  self->overQuota = -1;

  [self expunge];
}

- (void)resetStatus {
  self->unseen = -1;
  self->exists = -1;
  self->recent = -1;
}

- (void)resetSubFolders {
  id ctx = [self context];
  
  if ((self->parentFolder == nil) || (self == [ctx inboxFolder]))
    [ctx resetSpecialFolders];
  
  [self->subFolders release]; self->subFolders = nil;
}

- (BOOL)renameTo:(NSString *)_name {
  NSString     *n;
  NSDictionary *dict;

  if ([self isReadOnly] == YES)
    return NO;
  
  if ((_name == nil) || ([_name length] == 0))
    return NO;

  n = [[self->name stringByDeletingLastPathComponent]
                   stringByAppendingPathComponent:_name];

  [self resetLastException];

  dict = [[self->context client] rename:[self absoluteName] to:n];

  if (![self _checkResult:dict cmd:__PRETTY_FUNCTION__])
    return NO;
                        
  [self->name release]; self->name = nil;
  self->name = [n copy];
 
  [self resetSubFolders];
 
  dict = [[self->context client] subscribe:self->name];

  if (![self _checkResult:dict cmd:__PRETTY_FUNCTION__])
    return NO;

  return YES;
}

// folder

- (BOOL)deleteSubFolder:(NGImap4Folder *)_folder {
  if ([self isReadOnly] == YES)
    return NO;
  return _deleteSubFolder(self, _folder);;
}

- (BOOL)copySubFolder:(NGImap4Folder *)_f to:(NGImap4Folder *)_folder {
  return _copySubFolder(self, _f, _folder);
}

- (BOOL)moveSubFolder:(NGImap4Folder *)_f to:(NGImap4Folder *)_folder {
  if ([self isReadOnly] == YES)
    return NO;
  return _moveSubFolder(self, _f, _folder);
}

- (BOOL)createSubFolderWithName:(NSString *)_name {
  if ([self isReadOnly] == YES)
    return NO;
  return _createSubFolderWithName(self, _name, YES);
}

- (void)expunge {
  if (self->noselect == YES)
    return;

  if ([self isReadOnly] == NO) {
    //    if (self->selectSyncState == YES) {
    //      [self selectImmediately:YES];
    //    }
    if ([self->context registerAsSelectedFolder:self]) {
      NSDictionary *dict;
      
      dict = [[self->context client] expunge];

      if ([self _checkResult:dict cmd:__PRETTY_FUNCTION__]) {
        [self setRecent:[dict objectForKey:@"recent"]
              exists:[dict objectForKey:@"exists"]];
      }
    }
  }
}

- (BOOL)addFlag:(NSString *)_flag toMessages:(NSArray *)_messages {
  return [self flag:_flag toMessages:_messages add:YesNumber];
}

- (BOOL)removeFlag:(NSString *)_flag fromMessages:(NSArray *)_messages {
  return [self flag:_flag toMessages:_messages add:NoNumber];
}

- (BOOL)flag:(NSString *)_flag toMessages:(NSArray *)_messages
        add:(NSNumber *)_add
{
  NSEnumerator *enumerator;
  id           obj;
  BOOL         add;

  add = [_add boolValue];   

  if (self->noselect == YES)
    return NO;

  if (_flag == nil) {
    NSLog(@"WARNING[%s]: try to set an empty flag", __PRETTY_FUNCTION__);
    return NO;
  }
  if ([self isReadOnly] == YES)
    return NO;
  
  if (![self->context registerAsSelectedFolder:self])
    return NO;

  if (!_testMessages(self, _messages, @"store"))
    return NO;

  [self resetLastException];
  
  if (!(enumerator = [_getMsnRanges(self, _messages) objectEnumerator])) {
    [self resetStatus];
    [self->context removeSelectedFolder:self];
    return NO;
  }
  while ((obj = [enumerator nextObject])) {
    int objEnd = [[obj objectForKey:@"end"] unsignedIntValue];
    if (objEnd > 0) {
      NSDictionary *res;
      
      res = [[self->context client]
                            storeFrom:[[obj objectForKey:@"start"]
                                            unsignedIntValue]
                            to:objEnd
                            add:_add
                            flags:[NSArray arrayWithObject:_flag]];

      if (![self _checkResult:res cmd:__PRETTY_FUNCTION__])
        break;
    }
  }
  if (obj) {
    return NO;
  }
  enumerator = [_messages objectEnumerator];
  while ((obj = [enumerator nextObject])) {
    (add == YES)
      ? [obj addFlag:_flag]
      : [obj removeFlag:_flag];
  }
  [self resetStatus];
  return YES;
}

- (BOOL)flagToAllMessages:(NSString *)_flag add:(NSNumber *)_add {
#if USE_MESSAGE_CACHE
  BOOL            add        = [_add boolValue];
  NSEnumerator   *enumerator = nil;
  NGImap4Message *m          = nil;
#endif

  if (self->noselect == YES)
    return NO;
  
  if (_flag == nil)
    return NO;

  if ([self isReadOnly] == YES)
    return NO;

  if (![self->context registerAsSelectedFolder:self])
    return NO;

  
  if ([self exists] > 0) {
    NSDictionary *res;
    
    [self resetLastException];
    
    res = [[self->context client] storeFrom:0 to:[self exists]
                                  add:_add
                                  flags:[NSArray arrayWithObject:_flag]];

    if (![self _checkResult:res cmd:__PRETTY_FUNCTION__])
      return NO;
    
#if USE_MESSAGE_CACHE
    enumerator = [self->messages objectEnumerator];
    while ((m = [enumerator nextObject])) {
      (add == YES)
        ? [m addFlag:_flag]
        : [m removeFlag:_flag];
    }
#endif
  }
  return YES;
}

- (BOOL)deleteAllMessages {
  if ([self isReadOnly] == YES)
    return NO;

  if ([self flagToAllMessages:@"Deleted" add:YesNumber]) {
    [self resetFolder];
    return YES;
  }
  return NO;
}

- (BOOL)deleteMessages:(NSArray *)_messages {
  if ([self isReadOnly] == YES)
    return NO;

  if ([self addFlag:@"Deleted" toMessages:_messages]) {
    [self resetFolder];
    return YES;
  }
  return NO;
}

- (BOOL)moveMessages:(NSArray *)_messages toFolder:(NGImap4Folder *)_folder {
  if ([self isReadOnly] == YES)
    return NO;

  if (_folder != nil) {
    if ([self copyMessages:_messages toFolder:_folder]) {
      return [self deleteMessages:_messages];
    }
  }
  return NO;
}

- (BOOL)copyMessages:(NSArray *)_messages toFolder:(NGImap4Folder *)_folder {
  NSEnumerator *enumerator;
  id           obj;
  NSString     *folderName;

  if (self->noselect == YES)
    return NO;
  
  folderName = [_folder absoluteName];  

  if (_folder == nil) {
    NSLog(@"WARNING[%s]: try to copy to nil folder", __PRETTY_FUNCTION__);
    return NO;
  }
  [self resetLastException];
  
  if ([_folder isReadOnly]) {
    NGImap4ResponseException *exc;

    NSLog(@"WARNING[%s]: try to copy to readonly folder %@",
          __PRETTY_FUNCTION__, _folder);

    exc = [[NGImap4ResponseException alloc] initWithFormat:
                                            @"copy to read only folder"];
    [self->context setLastException:exc];
    [exc release]; exc = nil;
    return NO;
  }
  if (![self->context registerAsSelectedFolder:self])
    return NO;
  
  if (_testMessages(self, _messages, @"copy") == NO)
    return NO;

  if (!(enumerator = [_getMsnRanges(self, _messages) objectEnumerator])) {
    [self resetStatus];
    [self->context removeSelectedFolder:self];
    return NO;
  }
  
  [self resetLastException];
  if (![self->context registerAsSelectedFolder:self]) {
    return NO;
  }
  while ((obj = [enumerator nextObject])) {
    int objEnd = [[obj objectForKey:@"end"] unsignedIntValue];

    if (objEnd > 0) {
      NSDictionary *res;

      
      res = [[self->context client]
                            copyFrom:[[obj objectForKey:@"start"]
                                           unsignedIntValue]
                            to:objEnd
                            toFolder:folderName];

      if (![self _checkResult:res cmd:__PRETTY_FUNCTION__])
        break;
    }
  }
  [_folder resetFolder];
  return (obj == nil) ? YES : NO;
}

- (BOOL)appendMessage:(NSData *)_msg {
  
  if ([self isReadOnly] == YES)
    return NO;

  if (_msg != nil) {
    NSDictionary *dict;

    dict = [[self->context client]
                           append:_msg toFolder:[self absoluteName]
                           withFlags:[NSArray arrayWithObject:@"seen"]];

    if ([self _checkResult:dict cmd:__PRETTY_FUNCTION__]) {
      self->exists    = -1;
      self->recent    = -1;
      self->maxQuota  = -1;
      self->usedSpace = -1;
      self->overQuota = -1;
      return YES;
    }
  }
  return NO;
}

#if USE_MESSAGE_CACHE
- (void)resetQualifierCache {
  self->cacheIdx = 0;
  [self->qualifierCache removeAllObjects];
}
#endif
// notifications

- (void)setRecent:(NSNumber *)_rec exists:(NSNumber *)_exists {
#if USE_MESSAGE_CACHE  
  BOOL resetQualifier = NO;
  if (_rec != nil) {
    int tmp = [_rec intValue];
      if (self->recent != tmp) {
        self->recent   = tmp;
        resetQualifier = YES;
      }
  }
  if (_exists != nil) {
    int tmp = [_exists intValue];
    if (self->exists != tmp) {
      self->exists = tmp;
      resetQualifier = YES;
    }
  }
  if (resetQualifier == YES) 
    [self resetQualifierCache];
#else
  {
    if (_exists != nil) {
      int      e;

      e = [_exists intValue];

      if (e == 0) {
        self->exists = 0;
        self->recent = 0;
        self->unseen = 0;
      }
      else {
        int      r;
    
        r  = [_rec intValue];

        if ((e != self->exists) || (r != self->recent)) {
          self->exists = e;
          self->recent = r;
          self->unseen = -1;
        }
      }
    }
  }
#endif  
}

- (void)processResponse:(NSDictionary *)_dict {
#if USE_MESSAGE_CACHE  
  id exp = [_dict objectForKey:@"expunge"];
#endif

  [self setRecent:[_dict objectForKey:@"recent"]
        exists:[_dict objectForKey:@"exists"]];

#if USE_MESSAGE_CACHE  
  if ((exp != nil) && ([exp count] > 0) && ([self->messages count] > 0)) {
    
    NSEnumerator *enumerator = [exp objectEnumerator];
    id           obj         = nil;
    
    while ((obj = [enumerator nextObject])) {
      int n = [obj intValue] - 1;

      [self->messages removeObjectAtIndex:n];
    }
    [self resetQualifierCache];
  }
#endif  
}

- (BOOL)isInTrash {
  NGImap4Folder *f, *trash;

  trash = [self->context trashFolder];

  if (trash == nil) {
    NSLog(@"WARNING[%s]: No trash folder was set", __PRETTY_FUNCTION__);
    return NO;
  }

  f = self;
  while (f != nil) {
    if ([f isEqual:trash] == YES) {
      return YES;
    }
    else
      f = [f parentFolder];
  }
  return NO;
}

- (NSString *)description {
  NSMutableString *ms;
  NSString        *tmp;

  ms = [NSMutableString stringWithCapacity:64];

  [ms appendFormat:@"<0x%08X[%@]:", self, NSStringFromClass([self class])];
  
  if ((tmp = [self name]))
    [ms appendFormat:@" name=%@", tmp];
  if ((tmp = [self absoluteName]))
    [ms appendFormat:@" absolute=%@", tmp];
  if ((tmp = [[self flags] componentsJoinedByString:@","]))
    [ms appendFormat:@" flags=%@", tmp];
  
  [ms appendString:@">"];

  return ms;
}

- (void)resetSync {
  NSEnumerator *enumerator;
  id           folder;

  self->selectSyncState = NO;

  if (self->subFolders != nil) {
    enumerator = [[self subFolders] objectEnumerator];
    while ((folder = [enumerator nextObject])) {
      [folder resetSync];
    }
  }
}

- (NSURL *)url {
  if (!self->url) {
    NSURL *base;

    base = [self->context url];

    self->url = [[NSURL alloc]
                        initWithScheme:[base scheme]
                        host:[base host]
                        path:[self absoluteName]];
  }
  return self->url;
}

- (void)quota {
  NSString     *n;
  NSDictionary *quota;
  

  if (![self->context canQuota]) {
    NSLog(@"WARNING[%s] call quota but capability contains no quota string",
          __PRETTY_FUNCTION__);
    return;
  }
  n     = [self absoluteName];
  [self resetLastException];

  quota = [[self->context client] getQuotaRoot:n];

  if (![self _checkResult:quota cmd:__PRETTY_FUNCTION__])
    return;
  
  quota = [[quota objectForKey:@"quotas"] objectForKey:n];

  self->maxQuota  = [[quota objectForKey:@"maxQuota"] intValue];
  self->usedSpace = [[quota objectForKey:@"usedSpace"] intValue];
}


static inline BOOL _testMessages(NGImap4Folder *self, NSArray *_messages,
                                 NSString *_operation)
{
  NSEnumerator *enumerator;
  id           obj;

  enumerator = [_messages objectEnumerator];

  while ((obj = [enumerator nextObject])) {
    if ([obj folder] != self) {
      NSLog(@"ERROR: try to %@ mails in folder who didn`t own this mail \n"
            @"Folder %@\nMail %@ allMessages %@",
            _operation, self, obj, _messages);
      return NO;
    }
  }
  return YES;
}

static inline NSArray *_getMsnRanges(NGImap4Folder *self, NSArray *_messages) {
  static NSArray      *UidKey     = nil;

  NSArray             *result;
  NSMutableDictionary *map;
  NSMutableArray      *msn;
  NSEnumerator        *enumerator;  
  id                  obj;
  NSAutoreleasePool   *pool;
  
  if ([self exists] == 0)
    return [NSArray array];

  pool = [[NSAutoreleasePool alloc] init];
  
  if (UidKey == nil) {
    id objs = nil;

    objs   = @"uid";
    UidKey = [[NSArray alloc] initWithObjects:&objs count:1];
  }

  [self resetLastException];
  
  if (![self->context registerAsSelectedFolder:self])
    return nil;

  if (!self->msn2UidCache) {
    NSDictionary *res;

    res = [[self->context client] fetchFrom:1 to:[self exists] parts:UidKey];

    if (![self _checkResult:res cmd:__PRETTY_FUNCTION__])
      return nil;
    
    self->msn2UidCache = [[res objectForKey:@"fetch"] retain];
  }
  map = [[NSMutableDictionary alloc] initWithCapacity:
                                     [self->msn2UidCache count]];
  enumerator = [self->msn2UidCache objectEnumerator];
  
  while ((obj = [enumerator nextObject])) {
    [map setObject:[obj objectForKey:@"msn"] forKey:[obj objectForKey:@"uid"]];
  }
  msn = [[NSMutableArray alloc] initWithCapacity:[_messages count]];
  enumerator = [_messages objectEnumerator];
  while ((obj = [enumerator nextObject])) {
    id m;
    
    m = [map objectForKey:[NSNumber numberWithUnsignedInt:[obj uid]]];
    
    if (m == nil) {
      NSLog(@"WARNING[%s]: Couldn`t map a message sequence number to "
            @"message %@ numbers %@ messages %@",
            __PRETTY_FUNCTION__, obj, map, _messages);
      [msn  release];
      [map  release];
      [pool release];
      return nil;
    }
    [msn addObject:m];
  }
  [map release]; map = nil;
  
  result = _calculateSequences(self, msn, -1);
  
  result = [result retain];
  [msn release]; msn = nil;
  [pool release];
  return [result autorelease];
}

static inline  NSArray *_calculateSequences(NGImap4Folder *self,
                                            NSMutableArray *_numbers,
                                            int            _cnt)
{
  NSAutoreleasePool   *pool;
  NSEnumerator        *enumerator;
  NSMutableDictionary *range;
  NSMutableArray      *ranges;
  id                  obj, buffer;
  int                 cntMsgs;

  pool = [[NSAutoreleasePool alloc] init];

  if (_cnt == -1)
    _cnt = [_numbers count];
  
  [_numbers sortUsingSelector:@selector(compare:)];
  
  ranges     = [NSMutableArray arrayWithCapacity:[_numbers count]];
  enumerator = [_numbers objectEnumerator];
  buffer     = [NSNumber numberWithInt:0];
  range      = nil;
  cntMsgs    = 0;
  while (((obj = [enumerator nextObject])) && (cntMsgs < _cnt)) {
    cntMsgs++;
    if (range == nil) {
      range = [NSMutableDictionary dictionaryWithCapacity:2];
      [range setObject:buffer forKey:@"start"];
    }
    
    if ([obj intValue] != [buffer intValue] + 1) {
      NSDictionary *ir;
      
      [range setObject:buffer forKey:@"end"];
      ir = [range copy];
      [ranges addObject:ir];
      [ir release];
      range = nil;
    }
    buffer = obj;
  }
  if (range != nil) {
    [range setObject:buffer forKey:@"end"];
    [ranges addObject:range];
  }
  else {
    [ranges addObject:[NSDictionary dictionaryWithObjectsAndKeys:
                                      buffer, @"start",
                                      buffer, @"end", nil]];
  }
  range = [ranges objectAtIndex:0];
  if ([[range objectForKey:@"end"] intValue] == 0)
    [ranges removeObjectAtIndex:0];
  
  obj = [ranges copy];
  [pool release];
  return [obj autorelease];
}
 
@end /* NGImap4Folder */

@implementation NSData(xxxx)
- (NSString *)description {
  return [NSString stringWithFormat:@"NSData len: %d",
                   [self length]];
}
@end
