// $Id: NGImap4Functions.m,v 1.8 2003/02/28 17:34:49 jan Exp $

#include <NGImap4/NGImap4Folder.h>
#include <NGImap4/NGImap4Support.h>
#include <NGImap4/NGImap4Context.h>
#include <NGImap4/NGImap4Client.h>
#include "common.h"

@interface NGImap4Context(Private)
- (void)setLastException:(NSException *)_exception;
@end /*NGImap4Context(Private) */

BOOL _checkResult(NGImap4Context *_ctx, NSDictionary *_dict,
                  const char *_command)
{
  if (![[_dict objectForKey:@"result"] boolValue]) {
    if (![_ctx lastException]) {
      NGImap4ResponseException *exc;

      NSLog(@"ERROR[%s]: got error during %s: %@",
            __PRETTY_FUNCTION__, _command, _dict);

      exc = [[NGImap4ResponseException alloc]
                                       initWithFormat:
                                       [_dict objectForKey:@"reason"]];
      [_ctx setLastException:exc];
      [_ctx removeAllFromRefresh];
      [exc release]; exc = nil;
    }
    return NO;
  }
  return YES;
}

BOOL _isSubFolder(id<NGImap4Folder> self,
                                id<NGImap4Folder>_folder) {
  NSEnumerator  *enumerator;
  NGImap4Folder *folder;

  enumerator = [[self subFolders] objectEnumerator];
  while ((folder = [enumerator nextObject])) {
    if ([_folder isEqual:folder])
      break;

    if ([[self context] lastException])
      return NO;
  }
  if (folder != nil)
    return YES;
  return NO;
}

NGImap4Folder *_subFolderWithName(id<NGImap4Folder> self, NSString *_name,
                                  BOOL _caseIns) 
{
  NSEnumerator  *enumerator;
  NGImap4Folder *f;

  if (_caseIns)
    _name = [_name lowercaseString];

  enumerator = [[self subFolders] objectEnumerator];
  while ((f = [enumerator nextObject])) {
    NSString *n;

    n = [f name];
    if (_caseIns) 
      n = [n lowercaseString];

    if ([n isEqualToString:_name])
      return f;

    if ([[self context] lastException])
      return NO;
  }
  return nil;
}

BOOL _hasNewMessagesInSubFolder(id<NGImap4Folder> self) {
  NSEnumerator *enumerator;
  id           obj;

  enumerator = [[self subFolders] objectEnumerator];
  while ((obj = [enumerator nextObject])) {
    if ([obj hasNewMessagesSearchRecursiv:YES])
      return YES;

    if ([[self context] lastException])
      return NO;
  }
  return NO;
}

BOOL _hasUnseenMessagesInSubFolder(id<NGImap4Folder> self) {
  NSEnumerator *enumerator;
  id           obj;
  
  enumerator = [[self subFolders] objectEnumerator];
  while ((obj = [enumerator nextObject])) {
    if ([obj hasUnseenMessagesSearchRecursiv:YES])
      return YES;

    if ([[self context] lastException])
      return NO;
  }
  return NO;
}

BOOL _deleteSubFolder(id<NGImap4Folder> self, NGImap4Folder *_folder) {
  /* TODO: jr, why is this a function and not a method ?
     AW: they will be used in 2 different classes ...
         'grep _deleteSubFolder *.m' */
  
  NSEnumerator  *enumerator;
  NGImap4Folder *folder;
  id            ctx;

  ctx = [self context];

  [ctx resetLastException];
  
  if (!_isSubFolder(self, _folder)) {
    NSLog(@"ERROR: Couldn`t delete %@ because it`s no subfolder of %@",
          _folder, self);
    return NO;
  }
  enumerator = [[_folder subFolders] objectEnumerator];
  
  while ((folder = [enumerator nextObject])) {
    [_folder deleteSubFolder:folder];

    if ([ctx lastException])
      return NO;
  }
  if (![_folder isReadOnly] && ![_folder noselect])  {
    NSDictionary *result;

    if ([self isKindOfClass:[NGImap4Folder class]]) {
      if (![ctx registerAsSelectedFolder:self])
        return NO;
    }
    [ctx resetLastException];
    
    result = [[ctx client] delete:[_folder absoluteName]];

    if (!_checkResult(ctx, result, __PRETTY_FUNCTION__))
      return NO;
    
    [self resetSubFolders];
    [ctx removeFromRefresh:_folder];
    return YES;
  }
  return NO;
}

BOOL _copySubFolder(id<NGImap4Folder> self,
                    id<NGImap4Folder> _f, id<NGImap4Folder> _toFolder)
{
  /* TODO: jr, why is this a function and not a method ? 
     AW: they will be used in 2 different classes ...
         'grep _deleteSubFolder *.m' */
  
  NSEnumerator  *enumerator;
  NGImap4Folder *folder, *subFolder;  
  NSString      *folderName;

  if (_isSubFolder(self, _f) == NO) {
    NSLog(@"ERROR: Couldn`t copy %@ because it`s no subfolder of %@",
          _f, self);
    return NO;
  }
  if ([[_toFolder absoluteName] hasPrefix:[_f absoluteName]] == YES) {
    NSLog(@"ERROR: Couldn`t copy %@ is subFolder from %@",
          _toFolder, _f);
    return NO;
  }
  
  if ([self isEqual:_toFolder] == YES) {
    return YES;
  }
  folderName = [_f name];

  self = [self retain];
  
  if (![_toFolder createSubFolderWithName:folderName]) {
    return NO;
  }
  [self autorelease];
  enumerator = [[_toFolder subFolders] objectEnumerator];
  while ((folder = [enumerator nextObject])) {
    if ([[folder name] isEqualToString:folderName])
      break;
    
    if ([[self context] lastException])
      return NO;
  }
  if (!folder)
    return NO;

  if ([_f noselect] == NO) {
    
    if (![[self context] registerAsSelectedFolder:_f])
      return NO;
    
    if ([_f exists] > 0) {
      NSDictionary *res;
      
      res = [[[self context] client] copyFrom:0 to:[_f exists]
                                     toFolder:[folder absoluteName]];
      if (!_checkResult([self context], res, __PRETTY_FUNCTION__))
        return NO;
    }
  }
  [folder resetFolder];
  enumerator = [[_f subFolders] objectEnumerator];
  while ((subFolder = [enumerator nextObject])) {
    if (![_f copySubFolder:subFolder to:folder])
      break;

    if ([[self context] lastException])
      return NO;
  }
  return (subFolder == nil) ? YES : NO;
}

BOOL _moveSubFolder(id<NGImap4Folder> self, NGImap4Folder *_f,
                    id<NGImap4Folder>_folder) 
{
  if (_isSubFolder(self, _f) == NO)
    return NO;

  if ([[_folder absoluteName] hasSuffix:[_f absoluteName]] == YES)
    /* _folder is the same as or a subfolder from _f */
    return NO;

  if ([_folder isEqual:self])
    return YES;

  if ([self copySubFolder:_f to:_folder] == YES)
    return [self deleteSubFolder:_f];

  return NO;
}

BOOL _createSubFolderWithName(id<NGImap4Folder> self, NSString *_name,
                              BOOL _append)
{
  NGImap4Context *ctx;
  NSRange        r;
  NSDictionary   *res;

  if (_name == nil)
    return NO;

  ctx = [self context];
  r   = [_name rangeOfString:[[ctx client] delimiter]];

  if (r.length > 0) {
    [ctx setLastException:
         [[NGImap4Exception alloc]
                            initWithFormat:@"It`s not allowed to use '%@' in "
                            @"a foldername <%@>", [[ctx client] delimiter],
                            _name]];
    return NO;
  }
  [ctx resetLastException];
  if (_append)
    _name = [[self absoluteName] stringByAppendingPathComponent:_name];

  res = [[ctx client] create:_name];
  if (!_checkResult(ctx, res, __PRETTY_FUNCTION__))
    return NO;

  res = [[ctx client] subscribe:_name];

  if (!_checkResult(ctx, res, __PRETTY_FUNCTION__))
    return NO;

  [self resetSubFolders];
  return YES;;
}
