// $Id: NGImap4Message.m,v 1.20 2003/04/01 13:50:03 gnustep Exp $

#import "NGImap4Message.h"
#import "NGImap4Folder.h"
#import "NGImap4Context.h"
#import "NGImap4Functions.h"
#import "NGImap4Client.h"
#include <NGExtensions/NGFileManager.h>
#import "common.h"

#import "NGImap4Message+BodyStructure.h"

@interface NGImap4Message(Private)
- (void)initializeMessage;
- (void)fetchMessage;
- (void)parseMessage;
- (void)generateBodyStructure;

@end /* NGImap4Message(Private) */



@implementation NGImap4Message

static NSNumber *YesNumber = nil;
static NSNumber *NoNumber = nil;

+ (void)initialize {
  [super initialize];
  if (YesNumber == nil) YesNumber = [[NSNumber numberWithBool:YES] retain];
  if (NoNumber == nil)  NoNumber  = [[NSNumber numberWithBool:NO] retain];  
}

+ (id)messageWithUid:(unsigned)_uid folder:(NGImap4Folder *)_folder
  context:(NGImap4Context *)_ctx
{
  return [[[self alloc] initWithUid:_uid folder:_folder context:_ctx]
                 autorelease];
}

+ (id)messageWithUid:(unsigned)_uid headers:(NGHashMap *)_headers
  size:(unsigned)_size flags:(NSArray *)_flags folder:(NGImap4Folder *)_folder
  context:(NGImap4Context *)_ctx
{
  return [[[self alloc] initWithUid:_uid headers:_headers
                        size:_size flags:_flags
                        folder:_folder context:_ctx] autorelease];
}

- (id)initWithUid:(unsigned)_uid folder:(NGImap4Folder *)_folder
  context:(NGImap4Context *)_ctx
{
  return [self initWithUid:_uid headers:nil size:-1 flags:nil
               folder:_folder context:_ctx];
}

- (id)initWithUid:(unsigned)_uid headers:(NGHashMap *)_headers
  size:(unsigned)_size flags:(NSArray *)_flags folder:(NGImap4Folder *)_folder
  context:(NGImap4Context *)_ctx
{
  if ((self = [super init])) {
    self->uid     = _uid;
    self->size    = _size;
    self->folder  = _folder;
    self->headers = [_headers retain];
    self->flags   = [_flags retain];
    self->context = [_ctx retain];
  }
  return self;
}

/* internal method */

- (void)_setHeaders:(NGHashMap *)_headers
  size:(unsigned)_size
  flags:(NSArray *)_flags
{
  [self->headers release]; self->headers = nil;
  [self->flags release];   self->flags = nil;

  self->size    = _size;
  self->headers = [_headers retain];
  self->flags   = [_flags retain];
}

- (BOOL)isComplete {
  return (self->headers != nil) && (self->flags != nil);
}

- (void)dealloc {
  RELEASE(self->headers);
  RELEASE(self->flags);
  RELEASE(self->context);
  RELEASE(self->rawData);
  RELEASE(self->message);
  RELEASE(self->bodyStructure);
  RELEASE(self->url);
  RELEASE(self->bodyStructureContent);
  self->folder = nil;          
  [super dealloc];
}

// accessors

- (NSException *)lastException {
  return [self->context lastException];
}
- (void)resetLastException {
  [self->context resetLastException];
}

- (unsigned)uid {
  return self->uid;
}

- (NGHashMap *)headers {
  if (self->headers == nil)
    [self initializeMessage];
  return self->headers;
}

- (int)size {
  if (self->size == -1)
    [self initializeMessage];
  return self->size;
}

- (NSArray *)flags {
  if (self->flags == nil)
    [self initializeMessage];
  return self->flags;
}

- (NSData *)rawData {
  if (self->rawData == nil)
    [self fetchMessage];
  return self->rawData;
}

- (NSData *)contentsOfPart:(NSString *)_part {
  NSData *result;

  if (_part == nil)
    _part = @"";
  
  if (self->bodyStructureContent == nil)
    self->bodyStructureContent = [[NSMutableDictionary alloc]
                                                       initWithCapacity:8];

  if (!(result = [self->bodyStructureContent objectForKey:_part])) {
    if ((result = [self->folder blobForUid:self->uid part:_part]))
      [self->bodyStructureContent setObject:result forKey:_part];
  }
  return result;
}

- (id<NGMimePart>)bodyStructure {
  if (self->bodyStructure == nil) {
    [self generateBodyStructure];
  }
  return self->bodyStructure;
}

- (id<NGMimePart>)message {
  if (self->message == nil) 
    [self parseMessage];
  return self->message;
}

- (NGImap4Folder *)folder {
  return self->folder;
}

- (NGImap4Context *)context {
  return self->context;
}

// private methods

- (void)initializeMessage {
  NSDictionary        *dict;
  NSDictionary        *fetch;
  NSAutoreleasePool   *pool;

  pool = [[NSAutoreleasePool alloc] init];

  if (![self->context registerAsSelectedFolder:self->folder])
    return;
  
  [self resetLastException];

  dict = [[self->context client]
                         fetchUid:self->uid
                         parts:[NSArray arrayWithObjects:@"rfc822.header",
                                        @"rfc822.size", @"flags", nil]];

  if (!_checkResult(self->context, dict, __PRETTY_FUNCTION__))
    return;
  
  fetch = [[[dict objectForKey:@"fetch"] objectEnumerator] nextObject];

  if (fetch == nil) {
    NSLog(@"WARNING[%s] : couldn`t fetch message with id %d",
          __PRETTY_FUNCTION__, self->uid);
    return;
  }
  {
    id                  h, f, s;
    NGMimeMessageParser *parser;
    NGDataStream        *stream;  

    h   = [fetch objectForKey:@"header"];
    f   = [fetch objectForKey:@"flags"];
    s   = [fetch objectForKey:@"size"];

    if ((h == nil) || (f == nil) || (s == nil)) {
      NSLog(@"WARNING[%s]: got no header, flags, size for %@",
            __PRETTY_FUNCTION__, fetch);
      return;
    }
    parser = [[NGMimeMessageParser alloc] init];  
    stream = [[NGDataStream alloc] initWithData:h];
    
    [parser prepareForParsingFromStream:stream];

    [self->headers release]; self->headers = nil;
    
    self->headers = [[parser parseHeader] retain];
    self->size    = [s intValue];

    if (([f containsObject:@"recent"]) && ([f containsObject:@"seen"])) {
      f = [f mutableCopy];
      [f removeObject:@"recent"];
      AUTORELEASE(f);
    }
    [self->flags release]; self->flags = nil;
    
    self->flags = [f copy];
    
    [stream release]; stream = nil;
    [parser release]; parser = nil;
  }
  [pool release];
}

- (void)fetchMessage {
  NSDictionary        *dict     = nil;
  NSDictionary        *fetch    = nil;

  if (![self->context registerAsSelectedFolder:self->folder])
    return;

  [self resetLastException];
    
  dict = [[self->context client] fetchUid:self->uid
                                 parts:[NSArray arrayWithObject:@"rfc822"]];

  if (!_checkResult(self->context, dict, __PRETTY_FUNCTION__))
      return;
  
  fetch = [[[dict objectForKey:@"fetch"] objectEnumerator] nextObject];

  if (fetch == nil) {
    NSLog(@"WARNING[%s]: couldn`t fetch message with id %d",
          __PRETTY_FUNCTION__, self->uid);
    return;
  }
  [self->rawData release]; self->rawData = nil;
  self->rawData = [[fetch objectForKey:@"message"] retain];
}

- (void)generateBodyStructure {
  NSDictionary *dict;
  NSDictionary *bStruct;
  
  [self->bodyStructure release]; self->bodyStructure = nil;

  [self resetLastException];

  if (![self->context registerAsSelectedFolder:self->folder])
    return;

  dict = [[self->context client] fetchUid:self->uid
                                 parts:[NSArray arrayWithObject:@"body"]];
  if (!(_checkResult(self->context, dict, __PRETTY_FUNCTION__)))
    return;

  bStruct = [[[dict objectForKey:@"fetch"] lastObject] objectForKey:@"body"];

  if (bStruct == nil) {
    NSLog(@"WARNING[%s]: couldn`t fetch body of message with id %d",
          __PRETTY_FUNCTION__, self->uid);
    return;
  }
  
  self->bodyStructure = [[NGMimeMessage alloc]
                                        initWithHeader:[self headers]];
  [self->bodyStructure setBody:
       _buildMimeMessageBody(self, [self url], bStruct,
                             self->bodyStructure)];
}

- (void)parseMessage {
  NGMimeMessageParser *parser;

  [self->message release]; self->message = nil;
  
  parser = [[NGMimeMessageParser allocWithZone:[self zone]] init];

  self->message = [[parser parsePartFromData:[self rawData]] retain];
  [parser release]; parser = nil;    
}

- (BOOL)isRead {
  return [self->flags containsObject:@"seen"];
}

- (void)markRead {
  [self addFlag:@"seen"];
  [self removeFlag:@"recent"];
}

- (void)markUnread {
  [self removeFlag:@"seen"];
}

- (BOOL)isFlagged {
  return [self->flags containsObject:@"flagged"];
}

- (void)markFlagged {
  [self addFlag:@"flagged"];
}

- (void)markUnFlagged {
  [self removeFlag:@"flagged"];
}

- (BOOL)isAnswered {
  return [self->flags containsObject:@"answered"];
}

- (void)markAnswered {
  [self addFlag:@"answered"];
}

- (void)markNotAnswered {
  [self removeFlag:@"answered"];
}


- (void)addFlag:(NSString *)_flag {
  NSArray *tmp;

  tmp = [self flags];

  if (![self->flags containsObject:_flag]) {
    self->flags = [[self->flags arrayByAddingObject:_flag] retain];
    RELEASE(tmp); tmp = nil;
  }
  if ([[self->folder messageFlags] containsObject:_flag]) {
    NSDictionary *res;
    
    if (![self->context registerAsSelectedFolder:self->folder])
      return;
    
    res = [[self->context client] storeUid:self->uid add:YesNumber
                                   flags:[NSArray arrayWithObject:_flag]];

    if (!_checkResult(self->context, res, __PRETTY_FUNCTION__))
      return;
    
    [self->folder resetStatus];
  }
}

- (void)removeFlag:(NSString *)_flag {
  NSArray *tmp;

  tmp = [self flags];

  if ([self->flags containsObject:_flag]) {
    self->flags = [self->flags mutableCopyWithZone:[self zone]];
    [tmp release];
    tmp = self->flags;
    [(NSMutableArray *)self->flags removeObject:_flag];
    self->flags = [self->flags copy];
    [tmp release]; tmp = nil;
  }
  if ([[self->folder messageFlags] containsObject:_flag]) {
    NSDictionary *res;
    
    if (![self->context registerAsSelectedFolder:self->folder])
      return;
    
    res = [[self->context client] storeUid:self->uid add:NoNumber
                                  flags:[NSArray arrayWithObject:_flag]];

    if (!_checkResult(self->context, res, __PRETTY_FUNCTION__))
      return;
    
    [self->folder resetStatus];
  }
}


/* equality */

- (BOOL)isEqual:(id)_obj {
  if (_obj == self)
    return YES;
  if ([_obj isKindOfClass:[NGImap4Message class]] == YES)
    return [self isEqualToNGImap4Message:_obj];
  return NO;
}

- (BOOL)isEqualToNGImap4Message:(NGImap4Message *)_messages {
  if (([[_messages context] isEqual:self->context] == YES)
      && ([[_messages folder] isEqual:self->folder])
      && ([_messages uid] == self->uid)) {
    return YES;
  }
  return NO;
}

- (unsigned)hash {
  return self->uid;
}

- (EOGlobalID *)globalID {
  EOGlobalID *gid;
  id keys[4];
  
  keys[0] = [NSNumber numberWithUnsignedInt:[self uid]];
  keys[1] = [[self folder] absoluteName];
  keys[2] = [[self context] host];
  keys[3] = [[self context] login];
  
  gid = [EOKeyGlobalID globalIDWithEntityName:@"NGImap4Message"
                       keys:keys keyCount:4
                       zone:NULL];
  
  return gid;
}

/* key-value coding */

- (id)valueForKey:(NSString *)_key {
  unsigned len;
  id v = nil;

  static NGMimeHeaderNames *Fields = NULL;

  if (!Fields)
    Fields = (NGMimeHeaderNames *)[NGMimePartParser headerFieldNames];


  if ((len = [_key length]) == 0)
    return nil;
  
  if ([_key characterAtIndex:0] == 'N') {
    
    if (len < 9)
      ;
    else if ([_key isEqualToString:@"NSFileIdentifier"])
      v = [[self headers] objectForKey:Fields->messageID];
    else if ([_key isEqualToString:NSFileSize])
      v = [NSNumber numberWithInt:[self size]];
    else if ([_key isEqualToString:NSFileModificationDate])
      v = [[self headers] objectForKey:Fields->date];
    else if ([_key isEqualToString:NSFileType])
      v = NSFileTypeRegular;
    else if ([_key isEqualToString:NSFileOwnerAccountName])
      v = [self->context login];
    else if ([_key isEqualToString:@"NGFileSubject"])
      v = [[self headers] objectForKey:Fields->subject];
    else if ([_key isEqualToString:@"NSFileSubject"])
      v = [[self headers] objectForKey:Fields->subject];
    else if ([_key isEqualToString:@"NGFilePath"]) {
      v = [NSString stringWithFormat:@"%@/%d",
		      [self->folder absoluteName],
		      [self uid]];
    }
    if (v) return v;
  }
  
  if ((v = [[self headers] objectForKey:_key]))
    return v;
  
  return [super valueForKey:_key];
}

/* description */

- (NSString *)description {
  NSMutableString *ms;
  id tmp;
  
  ms = [NSMutableString stringWithCapacity:128];
  [ms appendFormat:@"<0x%08X[%@]:", self, NSStringFromClass([self class])];

  [ms appendFormat:@" uid=%d",  self->uid];
  [ms appendFormat:@" size=%d", self->size];

  if (self->headers) {
    tmp = [[self headers] objectForKey:@"subject"];
    if ([tmp length] > 0)
      [ms appendFormat:@" subject='%@'", tmp];
    else
      [ms appendString:@" no-subject"];
  }
  else {
    [ms appendString:@" [header not fetched]"];
  }
  tmp = [self->folder absoluteName];
  if (tmp)
    [ms appendFormat:@" folder='%@'", tmp];


  if (self->flags) {
  tmp = [self flags];

  if ([tmp count] > 0)
    [ms appendFormat:@"flags: (%@)", [tmp componentsJoinedByString:@", "]];
  else
    [ms appendString:@"no-flags"];
  }
  else {
    [ms appendString:@" [flags not fetched]"];
  }
  
  [ms appendString:@">"];
  return ms;
}

- (NSURL *)url {
  if (!self->url) {
    NSURL    *base;
    NSString *path;

    base = [self->folder url];
    path = [[base path] stringByAppendingPathComponent:
                        [[NSNumber numberWithUnsignedInt:[self uid]]
                                   stringValue]];

    self->url = [[NSURL alloc] initWithScheme:[base scheme]
                               host:[base host]
                               path:path];
  }
  return self->url;
}

@end /* NGImap4Message */
