// $Id: NGImap4ResponseParser.m,v 1.26 2003/04/01 13:50:03 gnustep Exp $

#import "NGImap4ResponseParser.h"
#import "NGImap4Support.h"
#import "common.h"

@implementation NGImap4ResponseParser

static NSDictionary *_parseBody(NGImap4ResponseParser *self);
static NSDictionary *_parseBodyContent(NGImap4ResponseParser *self);
static NSString *_parseBodyString(NGImap4ResponseParser *self,
                                  BOOL _convertString);
static NSString *_parseBodyDecodeString(NGImap4ResponseParser *self,
                                        BOOL _convertString,
                                        BOOL _decode);
static NSDictionary *_parseBodyParameterList(NGImap4ResponseParser *self);
static NSArray *_parseAddressStructure(NGImap4ResponseParser *self);
static NSArray *_parseParenthesizedAddressList(NGImap4ResponseParser *self);
static NSDictionary *_parseSingleBody(NGImap4ResponseParser *self);
static NSDictionary *_parseMultipartBody(NGImap4ResponseParser *self);
static BOOL _isUntaggedResponse(NGImap4ResponseParser *self);
static int _parseTaggedResponse(NGImap4ResponseParser *self,
                                NGMutableHashMap *result_);
static void _parseUntaggedResponse(NGImap4ResponseParser *self,
                                   NGMutableHashMap *result_);
static BOOL _parseByeUntaggedResponse(NGImap4ResponseParser *self,
                                      NGMutableHashMap *result_);
static BOOL _parseNumberUntaggedResponse(NGImap4ResponseParser *self,
                                         NGMutableHashMap *result_);
static NSArray *_parseFlagArray(NGImap4ResponseParser *self);
static BOOL _parseFlagsUntaggedResponse(NGImap4ResponseParser *self,
                                        NGMutableHashMap *result_);
static BOOL _parseOkUntaggedResponse(NGImap4ResponseParser *self,
                                     NGMutableHashMap *result_);
static BOOL _parseBadUntaggedResponse(NGImap4ResponseParser *self,
                                      NGMutableHashMap *result_);
static BOOL _parseNoUntaggedResponse(NGImap4ResponseParser *self,
                                     NGMutableHashMap *result_);
static BOOL _parseThreadResponse(NGImap4ResponseParser *self,
                                 NGMutableHashMap *result_);
static BOOL _parseStatusResponse(NGImap4ResponseParser *self,
                                 NGMutableHashMap *result_);
static BOOL _parseListOrLSubResponse(NGImap4ResponseParser *self,
                                     NGMutableHashMap *result_);
static BOOL _parseCapabilityResponse(NGImap4ResponseParser *self,
                                     NGMutableHashMap *result_);
static BOOL _parseSearchResponse(NGImap4ResponseParser *self,
                                 NGMutableHashMap *result_);
static NSArray *_parseThread(NGImap4ResponseParser *self);
static BOOL _parseSortResponse(NGImap4ResponseParser *self,
                               NGMutableHashMap *result_);
static inline int _la(NGImap4ResponseParser *self, unsigned _la);
static NSNumber *_parseUnsigned(NGImap4ResponseParser *self);
static NSString *_parseUntil(NGImap4ResponseParser *self, char _c);
static NSString *_parseUntil2(NGImap4ResponseParser *self, char _c1, char _c2);
static inline void _match(NGImap4ResponseParser *self, char _match);
static inline void _consume(NGImap4ResponseParser *self, unsigned _cnt);
static BOOL _isContinuationResponse(NGImap4ResponseParser *self);
static void _parseContinuationResponse(NGImap4ResponseParser *self,
                                       NGMutableHashMap *result_);
static NSData *_parseData(NGImap4ResponseParser *self);

static void _parseSieveRespone(NGImap4ResponseParser *self,
                               NGMutableHashMap *result_);
static BOOL _parseGreetingsSieveResponse(NGImap4ResponseParser *self,
                                         NGMutableHashMap *result_);
static BOOL _parseDataSieveResponse(NGImap4ResponseParser *self,
                                    NGMutableHashMap *result_);
static BOOL _parseOkSieveResponse(NGImap4ResponseParser *self,
                                  NGMutableHashMap *result_);
static BOOL _parseNoSieveResponse(NGImap4ResponseParser *self,
                                  NGMutableHashMap *result_);
static NSString *_parseContentSieveResponse(NGImap4ResponseParser *self);
static NSString *_parseStringSieveResponse(NGImap4ResponseParser *self);

static BOOL _parseQuotaResponse(NGImap4ResponseParser *self,
                                NGMutableHashMap *result_);
static BOOL _parseQuotaRootResponse(NGImap4ResponseParser *self,
                                    NGMutableHashMap *result_);

static int          USE_CYRUS_2_0 = -1;
static unsigned int LaSize        = 4097;
static int          UseMemoryMappedData = -1;

+ (void)initialize {
  if (USE_CYRUS_2_0 == -1) {
    USE_CYRUS_2_0 = [[NSUserDefaults standardUserDefaults]
                                     boolForKey:@"UseCyrus2_0"]?1:0;
  }
  if (UseMemoryMappedData == -1) {
    UseMemoryMappedData =
      [[NSUserDefaults standardUserDefaults]
                       boolForKey:@"NoMemoryMappedDataForImapBlobs"]?0:1;
  }
}

+ (id)parserWithStream:(id<NGActiveSocket>)_stream {
  return [[[NGImap4ResponseParser alloc] initWithStream:_stream] autorelease];
}

- (id)initWithStream:(id<NGActiveSocket>)_stream {
  // designated initializer
  if (_stream == nil) {
    NSLog(@"%s: got no stream ...", __PRETTY_FUNCTION__);
    [self release];
    return nil;
  }
  if ((self = [super init])) {
    id s;
    
    s = [[NGBufferedStream alloc] initWithSource:_stream];
    self->buffer = [NGByteBuffer alloc];
    self->buffer = [self->buffer initWithSource:s la:LaSize];
    [s release];
    
    if ([self->buffer respondsToSelector:@selector(methodForSelector:)])
      self->la = (int(*)(id, SEL, unsigned))
        [self->buffer methodForSelector:@selector(la:)];
    
    self->debug = [[NSUserDefaults standardUserDefaults]
                                   boolForKey:@"ImapDebugEnabled"];
  }
  return self;
}

- (id)init {
  [self release];
  [NSException raise:@"InvalidUseOfMethodException"
	       format:
		 @"calling -init on the NGImap4ResponseParser is not allowed"];
  return nil;
}

- (void)dealloc {
  [self->buffer release];
  if (self->debug == YES)
    [self->serverResponseDebug release];
  [super dealloc];
}


/*
** Parse Sieve Responses
*/

- (NGHashMap *)parseSieveResponse {
  NGMutableHashMap *result;

  if (self->debug == YES) {
    if (self->serverResponseDebug != nil)
      [self->serverResponseDebug release];
    self->serverResponseDebug = [[NSMutableString alloc] initWithCapacity:512];
  }
  result = [NGMutableHashMap hashMapWithCapacity:64];

  if (_la(self,0) == -1) {
    [[self->buffer lastException] raise];
  }
  
  _parseSieveRespone(self, result);
  return result;
}

/*"
** parse a response from the sever, _tag != -1 parse until tagged response
"*/

- (NGHashMap *)parseResponseForTagId:(int)_tag {
  BOOL             endOfCommand;
  NGMutableHashMap *result;

  if (self->debug == YES) {
    if (self->serverResponseDebug != nil)
      [self->serverResponseDebug release];
    self->serverResponseDebug = [[NSMutableString alloc] initWithCapacity:512];
  }
  
  result = [NGMutableHashMap hashMapWithCapacity:64];

  if (_la(self,0) == -1) {
    NSLog(@"%s: got lastException %@", __PRETTY_FUNCTION__,
          [self->buffer lastException]);
    
    [[self->buffer lastException] raise];
  }
  endOfCommand = NO;
  while (endOfCommand == NO) {
    if (_isUntaggedResponse(self) == YES) {
      _parseUntaggedResponse(self, result);
      if ([result objectForKey:@"bye"]) {
        endOfCommand = YES;
      }
      else {
        if (_tag == -1) {
          if ([result objectForKey:@"ok"] != nil)
            endOfCommand = YES;
        }
      }
    }
    else if (_isContinuationResponse(self)  == YES) {
      _parseContinuationResponse(self, result);
      endOfCommand = YES;
    }
    else
      endOfCommand = (_parseTaggedResponse(self, result) == _tag);
  }
  return result;
}

static void _parseSieveRespone(NGImap4ResponseParser *self,
                               NGMutableHashMap *result_)
{
  if (_parseGreetingsSieveResponse(self, result_) == YES)
    return;
  if (_parseDataSieveResponse(self, result_) == YES)    // la: 1
    return;
  if (_parseOkSieveResponse(self, result_) == YES)     // la: 2
    return; 
  if (_parseNoSieveResponse(self, result_) == YES)     // la: 2
    return;
}

static BOOL _isUntaggedResponse(NGImap4ResponseParser *self) {
  return (_la(self, 0) == '*');
}

static BOOL _isContinuationResponse(NGImap4ResponseParser *self) {
  return (_la(self, 0) == '+');
}

static NSData *_parseData(NGImap4ResponseParser *self) {
  NSData *result;
  int    size;

  static int Imap4MMDataBoundary = -1;

  if (Imap4MMDataBoundary == -1) {
    Imap4MMDataBoundary = [[NSUserDefaults standardUserDefaults]
                                           integerForKey:
                                           @"Imap4MMDataBoundary"];

    if (Imap4MMDataBoundary < 10) {
      Imap4MMDataBoundary = 2*LaSize;
    }
  }
  result = nil;  
  if (_la(self, 0) == '{') { /* got header */
    NSNumber *n;
    
    _consume(self, 1);
    if ((n = _parseUnsigned(self)) == nil) {
      [[[NGImap4ParserException alloc]
                                initWithFormat:@"expect a number "
                                @"between {}"] raise];
    }
    _match(self, '}');
    _match(self, '\n');

    size = [n intValue];
    
    if (UseMemoryMappedData && (size > Imap4MMDataBoundary)) {
      static NSProcessInfo *Pi = nil;
      NGFileStream *stream;
      
      char buf[LaSize+2];
      char tmpBuf[LaSize+2];
      int  wasRead = 0;
      NSString     *path;
      char  lastChar;

      if (Pi == nil)
        Pi = [[NSProcessInfo processInfo] retain];

      path   = [Pi temporaryFileName];
      stream = [[NGFileStream alloc] initWithPath:path];

      if (![stream openInMode:NGFileWriteOnly]) {
        [[[NGImap4ParserException alloc]
                                  initWithFormat:@"Couldn`t open "
                                  @"temporary file %@",
                                  path] raise];
      }
      lastChar = -1;
      
      while (wasRead < size) {
        int    readCnt, bufCnt, tmpSize, cnt, tmpBufCnt;

        bufCnt = 0;
        
        if (lastChar != -1) {
          buf[bufCnt++] = lastChar;
          lastChar = -1;
        }
        
        [self->buffer la:(size - wasRead <  LaSize) ? (size - wasRead)
             : LaSize];
            
        readCnt = [self->buffer readBytes:buf+bufCnt count:size - wasRead];
        
        wasRead+=readCnt;
        bufCnt +=readCnt;

        tmpSize   = bufCnt - 1;
        cnt       = 0;
        tmpBufCnt = 0;
        
        while (cnt < tmpSize) {
          if ((buf[cnt] == '\r') && (buf[cnt+1] == '\n')) {
            cnt++;
          }
          tmpBuf[tmpBufCnt++] = buf[cnt++];
        }
        if (cnt < bufCnt) {
          lastChar = buf[cnt];
        }
        [stream writeBytes:tmpBuf count:tmpBufCnt];
      }
      if (lastChar != -1) {
        [stream writeBytes:&lastChar count:1];
      }
      [stream close];
      [stream release]; stream = nil;
      result = [NSData dataWithContentsOfMappedFile:path];
      [[NSFileManager defaultManager] removeFileAtPath:path handler:nil];
    }
    else {
      char *buf    = NULL;
      int  wasRead = 0;
          
#if NeXT_Foundation_LIBRARY
      buf = NSZoneMalloc([self zone], sizeof(char) * size);
#else
      buf = NSZoneMallocAtomic([self zone], sizeof(char) * size);
#endif
      while (wasRead < size) {
        [self->buffer la:(size - wasRead <  LaSize) ? (size - wasRead)
             : LaSize];
            
        wasRead += [self->buffer readBytes:(buf + wasRead)
                        count:size - wasRead];
      }
      { /* normalize response  \r\n -> \n */
        char *tmpBuf;
        int  cnt, tmpBufCnt, tmpSize;

#if NeXT_Foundation_LIBRARY
        tmpBuf = NSZoneMalloc([self zone], sizeof(char) * size);
#else
        tmpBuf = NSZoneMallocAtomic([self zone], sizeof(char) * size);
#endif
        cnt       = 0;
        tmpBufCnt = 0;
        tmpSize   = size - 1;
        while (cnt < tmpSize) {
          if ((buf[cnt] == '\r') && (buf[cnt+1] == '\n')) {
            cnt++;
          }
          tmpBuf[tmpBufCnt++] = buf[cnt++];
        }
        if (cnt < size)
          tmpBuf[tmpBufCnt++] = buf[cnt++];

        result = [NSData dataWithBytesNoCopy:tmpBuf length:tmpBufCnt];
      }
      NSZoneFree([self zone], buf); buf = NULL;
    }
  }
  return result;
}

static int _parseTaggedResponse(NGImap4ResponseParser *self,
                                NGMutableHashMap *result_) {
  NSNumber *tag  = nil;
  NSString *res  = nil;
  NSString *desc = nil;
  NSString *flag = nil;

  tag  = _parseUnsigned(self);
  if (tag == nil) {
    if (self->debug == YES) {
      [[[NGImap4ParserException alloc]
                                     initWithFormat:@"expect a number at begin "
                                     @"of tagged response <%@>",
                                     self->serverResponseDebug] raise];
    }
    else {
      [[[NGImap4ParserException alloc]
                                initWithFormat:@"expect a number at begin "
                                @"of tagged response"] raise];
    }
  }
  _match(self, ' ');
  res  = [_parseUntil(self, ' ') lowercaseString];
  if (_la(self, 0) == '[') { /* Found flag like [READ-ONLY] */
    _consume(self, 1);
    flag = _parseUntil(self, ']');
  }
  desc = _parseUntil(self, '\n');
  /*
    ATTENTION: if no flag was set, flag == nil, in this case all key-value pairs
    after flag will be ignored
  */
  [result_ addObject:[NSDictionary dictionaryWithObjectsAndKeys:
                                   tag,  @"tagId",
                                   res,  @"result",
                                   desc, @"description",
                                   flag, @"flag", nil]
           forKey:@"ResponseResult"];
  return [tag intValue];
}

static void _parseUntaggedResponse(NGImap4ResponseParser *self,
                                   NGMutableHashMap *result_) {
  _match(self, '*');
  _match(self, ' ');

  if (_parseOkUntaggedResponse(self, result_) == YES)     // la: 2
    return; 
  if (_parseNoUntaggedResponse(self, result_) == YES)     // la: 2
    return;
  if (_parseBadUntaggedResponse(self, result_) == YES)    // la: 3
    return;
  if (_parseByeUntaggedResponse(self, result_) == YES)    // la: 3
    return;
  if (_parseListOrLSubResponse(self, result_) == YES)     // la: 4
    return;
  if (_parseSortResponse(self, result_) == YES)         // la: 4
    return;
  if (_parseNumberUntaggedResponse(self, result_) == YES) // la: 5
    return;
  if (_parseFlagsUntaggedResponse(self, result_) == YES)  // la: 5
    return;
  if (_parseSearchResponse(self, result_) == YES)         // la: 5
    return;
  if (_parseQuotaResponse(self, result_) == YES)          // la: 6
    return;
  if (_parseStatusResponse(self, result_) == YES)         // la: 6
    return;
  if (_parseThreadResponse(self, result_) == YES)         // la: 6
    return;
  if (_parseCapabilityResponse(self, result_) == YES)         // la: 10
    return;
  if (_parseQuotaRootResponse(self, result_) == YES)         // la: 10
    return;
}

static void _parseContinuationResponse(NGImap4ResponseParser *self,
                                       NGMutableHashMap *result_) {
  _match(self, '+');
  _match(self, ' ');

  [result_ addObject:[NSNumber numberWithBool:YES]
           forKey:@"ContinuationResponse"];
  [result_ addObject:_parseUntil(self, '\n') forKey:@"description"];
}

static BOOL _parseListOrLSubResponse(NGImap4ResponseParser *self,
                                     NGMutableHashMap *result_) {
  if (((_la(self, 0) == 'L')
       && (_la(self, 1) == 'I')
       && (_la(self, 2) == 'S')
       && (_la(self, 3) == 'T')
       && (_la(self, 4) == ' ')) ||
      ((_la(self, 0) == 'L')
       && (_la(self, 1) == 'S')
       && (_la(self, 2) == 'U')
       && (_la(self, 3) == 'B')
       && (_la(self, 4) == ' '))) {
    NSArray  *flags = nil;
    NSString *delim = nil;
    NSString *name  = nil;
    
    _consume(self, 5);

    flags = _parseFlagArray(self);

    _match(self, ' ');

    if (_la(self, 0) == '"') {
      _match(self, '"');
      delim = _parseUntil(self, '"');
      _match(self, ' ');
    }
    else {
      _parseUntil(self, ' ');
      delim = nil;
    }
    if (_la(self, 0) == '"') {
      _consume(self, 1);
      name = _parseUntil(self, '"');
      _parseUntil(self, '\n');
    }
    else {
      name = _parseUntil(self, '\n');
    }
    [result_ addObject:[NSDictionary dictionaryWithObjectsAndKeys:
                                     name, @"folderName",
                                     flags, @"flags",
                                     delim, @"delimiter", nil]
             forKey:@"list"];
    return YES;
  }
  return NO;
}

static BOOL _parseCapabilityResponse(NGImap4ResponseParser *self,
                                     NGMutableHashMap *result_) {
  if ((_la(self, 0)     == 'C')
      && (_la(self, 1)  == 'A')
      && (_la(self, 2)  == 'P')
      && (_la(self, 3)  == 'A')
      && (_la(self, 4)  == 'B')
      && (_la(self, 5)  == 'I')
      && (_la(self, 6)  == 'L')
      && (_la(self, 7)  == 'I')
      && (_la(self, 8)  == 'T')
      && (_la(self, 9)  == 'Y')
      && (_la(self, 10) == ' ')) {
    NSString *caps;

    caps = _parseUntil(self, '\n');
    {
      NSEnumerator   *enumerator;
      id             obj;
      NSMutableArray *array;
      NSArray        *tmp;

      array = [[NSMutableArray alloc] initWithCapacity:16];

      enumerator = [[caps componentsSeparatedByString:@" "] objectEnumerator];
      while ((obj = [enumerator nextObject])) {
        [array addObject:[obj lowercaseString]];
      }
      tmp = [array shallowCopy];
      [result_ addObject:tmp forKey:@"capability"];

      [array release]; array = nil;
      [tmp   release]; tmp   = nil;
    }
    return YES;
  }
  return NO;
}

static BOOL _parseSearchResponse(NGImap4ResponseParser *self,
                                 NGMutableHashMap *result_) {
  if ((_la(self, 0) == 'S')
      && (_la(self, 1) == 'E')
      && (_la(self, 2) == 'A')
      && (_la(self, 3) == 'R')
      && (_la(self, 4) == 'C')
      && (_la(self, 5) == 'H')) {

    NSMutableArray *msn = nil;

    _consume(self, 6);

    msn = [NSMutableArray arrayWithCapacity:128];

    while (_la(self, 0) == ' ') {
      _consume(self, 1);
      [msn addObject:_parseUnsigned(self)];
    }
    _parseUntil(self, '\n');
    [result_ addObject:msn forKey:@"search"];
    return YES;
  }
  return NO;
}

static BOOL _parseSortResponse(NGImap4ResponseParser *self,
                                 NGMutableHashMap *result_) {
  if ((_la(self, 0) == 'S')
      && (_la(self, 1) == 'O')
      && (_la(self, 2) == 'R')
      && (_la(self, 3) == 'T')) {

    NSMutableArray *msn = nil;

    _consume(self, 4);

    msn = [NSMutableArray arrayWithCapacity:128];

    while (_la(self, 0) == ' ') {
      _consume(self, 1);
      [msn addObject:_parseUnsigned(self)];
    }
    _parseUntil(self, '\n');
    [result_ addObject:msn forKey:@"sort"];
    return YES;
  }
  return NO;
}

static BOOL _parseQuotaResponse(NGImap4ResponseParser *self,
                                NGMutableHashMap *result_) {
  if ((_la(self, 0) == 'Q')
      && (_la(self, 1) == 'U')
      && (_la(self, 2) == 'O')
      && (_la(self, 3) == 'T')
      && (_la(self, 4) == 'A')
      && (_la(self, 5) == ' ')) {

    NSString            *qRoot;
    NSMutableDictionary *parse;
    NSMutableDictionary *quota;

    _consume(self, 6);

    quota = [result_ objectForKey:@"quota"];

    if (!quota) {
      quota = [NSMutableDictionary dictionaryWithCapacity:2];
      [result_ setObject:quota forKey:@"quota"];
    }
    
    parse = [NSMutableDictionary dictionaryWithCapacity:3];
    qRoot = _parseUntil2(self, ' ', '\n');

    if (_la(self, 0) == ' ') {
      _consume(self, 1);

      if (_la(self, 0) == '(') {
        NSString *key;

        _consume(self,1);

        key = _parseUntil(self, ' ');
        key = [key lowercaseString];
        if ([key isEqualToString:@"storage"]) {
          NSString *used, *max;

          used = _parseUntil(self, ' ');
          max  = _parseUntil(self, ')');

          [parse setObject:used forKey:@"usedSpace"];
          [parse setObject:max  forKey:@"maxQuota"];
          
        }
        else {
          NSString *v;

          v = _parseUntil(self, ')');

          [parse setObject:v forKey:@"resource"];
        }
      }
      [quota setObject:parse forKey:qRoot];
    }
    _parseUntil(self, '\n');
    
    return YES;
  }
  return NO;
}

static BOOL _parseQuotaRootResponse(NGImap4ResponseParser *self,
                                    NGMutableHashMap *result_)
{
  if ((_la(self, 0) == 'Q')
      && (_la(self, 1) == 'U')
      && (_la(self, 2) == 'O')
      && (_la(self, 3) == 'T')
      && (_la(self, 4) == 'A')
      && (_la(self, 5) == 'R')
      && (_la(self, 6) == 'O')
      && (_la(self, 7) == 'O')
      && (_la(self, 8) == 'T')
      && (_la(self, 9) == ' ')) {
    NSString *folderName, *folderRoot;
    NSMutableDictionary *dict;

    _consume(self, 10);

    dict = [result_ objectForKey:@"quotaRoot"];

    if (!dict) {
      dict = [NSMutableDictionary dictionaryWithCapacity:2];
      [result_ setObject:dict forKey:@"quotaRoot"];
    }
    if (_la(self, 0) == '"') {
      _consume(self , 1);
      folderName = _parseUntil(self, '"');
    }
    else {
      folderName = _parseUntil2(self, '\n', ' ');
    }
    if (_la(self, 0) == ' ') {
      _consume(self, 1);
      folderRoot = _parseUntil(self, '\n');
    }
    else {
      _consume(self, 1);
      folderRoot = nil;
    }
    if ([folderName length] && [folderRoot length]) {
      [dict setObject:folderRoot forKey:folderName];
    }
    return YES;
  }
  return NO;
}

static NSArray *_parseThread(NGImap4ResponseParser *self) {
  NSMutableArray *array;
  NSNumber       *msg;
    
  array = [NSMutableArray arrayWithCapacity:64];

  if (_la(self, 0) == '(') {
    _consume(self, 1);
  }
  while (1) {
    if (_la(self, 0) == '(') {
      id a;
      a = _parseThread(self);
      [array addObject:a];
    }
    else if ((msg = _parseUnsigned(self))) {
      [array addObject:msg];
    }
    else {
      return nil;
    }
    if (_la(self, 0) == ')')
      break;
    else if (_la(self, 0) == ' ')
      _consume(self, 1);
  }
  _match(self, ')');
  return array;
}


static BOOL _parseThreadResponse(NGImap4ResponseParser *self,
                                 NGMutableHashMap *result_) {
  if ((_la(self, 0) == 'T')
      && (_la(self, 1) == 'H')
      && (_la(self, 2) == 'R')
      && (_la(self, 3) == 'E')
      && (_la(self, 4) == 'A')
      && (_la(self, 5) == 'D')) {

    NSMutableArray *msn;

    _consume(self, 6);

    if (_la(self, 0) == ' ') {
      _consume(self, 1);
    }
    else {
      [result_ addObject:[NSArray array] forKey:@"thread"];
      return YES;
    }
    msn = [NSMutableArray arrayWithCapacity:64];
    while ((_la(self, 0) == '(')) {
      NSArray *array;
      if ((array = _parseThread(self)))
        [msn addObject:array];
    }
    _parseUntil(self, '\n');
    [result_ addObject:msn forKey:@"thread"];
    return YES;
  }
  return NO;
}

static BOOL _parseStatusResponse(NGImap4ResponseParser *self,
                                 NGMutableHashMap *result_) {
  if ((_la(self, 0) == 'S')
      && (_la(self, 1) == 'T')
      && (_la(self, 2) == 'A')
      && (_la(self, 3) == 'T')
      && (_la(self, 4) == 'U')
      && (_la(self, 5) == 'S')       
      && (_la(self, 6) == ' ')) {
    NSString            *name  = nil;
    NSMutableDictionary *flags = nil;
    
    _consume(self, 7);

    if (_la(self, 0) == '"') {
      _consume(self, 1);
      name = _parseUntil(self, '"');
      _match(self, ' ');
    }
    else {
      name = _parseUntil(self, ' ');
    }
    _match(self, '(');
    flags = [NSMutableDictionary dictionaryWithCapacity:8];
    
    while (_la(self, 0) != ')') {
      NSString *key   = _parseUntil(self, ' ');
      id       value = _parseUntil2(self, ' ', ')');

      if (_la(self, 0) == ' ')
        _consume(self, 1);
      
      [flags setObject:[NSNumber numberWithInt:[value intValue]]
             forKey:[key lowercaseString]];
    }
    _match(self, ')');
    _parseUntil(self, '\n');
    
    [result_ addObject:[NSDictionary dictionaryWithObjectsAndKeys:
                                     name, @"folderName",
                                     flags, @"flags", nil]
             forKey:@"status"];
    return YES;
  }
  return NO;
}

static BOOL _parseByeUntaggedResponse(NGImap4ResponseParser *self,
                                      NGMutableHashMap *result_) {
  if ((_la(self, 0) == 'B')
      && (_la(self, 1) == 'Y')
      && (_la(self, 2) == 'E')
      && (_la(self, 3) == ' ')) {
    NSString *reason = nil;

    _consume(self, 4);
    reason = _parseUntil(self, '\n');
    [result_ addObject:reason forKey:@"bye"];
    return YES;
  }
  return NO;
}

static BOOL _parseNumberUntaggedResponse(NGImap4ResponseParser *self,
                                         NGMutableHashMap *result_) {
  NSNumber *number = nil;
  NSString *key    = nil;

  if ((number = _parseUnsigned(self)) == nil)
    return NO;

  _match(self, ' ');
  
  if ((_la(self, 0) == 'F')
      && (_la(self, 1) == 'E')
      && (_la(self, 2) == 'T')
      && (_la(self, 3) == 'C')
      && (_la(self, 4) == 'H')
      && (_la(self, 5) == ' ')) { /* got a fetch response (fetch request) */
    NSMutableDictionary *fetch = nil;

    fetch = [NSMutableDictionary dictionaryWithCapacity:10];
    
    _consume(self, 6);
    _match(self, '(');
    while (_la(self, 0) != ')') {
      NSString *key = nil;

      key = [_parseUntil(self, ' ') lowercaseString];
      if ([key hasPrefix:@"body["]) {
        [fetch setObject:_parseBodyContent(self) forKey:key];
      }
      else if ([key isEqualToString:@"body"]) {
        [fetch setObject:_parseBody(self) forKey:key];
      }
      else if ([key isEqualToString:@"flags"]) {
        [fetch setObject:_parseFlagArray(self) forKey:key];
      }
      else if ([key isEqualToString:@"uid"]) {
        [fetch setObject:_parseUnsigned(self) forKey:key];
      }
      else if ([key isEqualToString:@"rfc822.size"]) {
        [fetch setObject:_parseUnsigned(self) forKey:key];
      }
      else if ([key hasPrefix:@"rfc822"]) {
        NSData *data;

        if (_la(self, 0) == '"') {
          NSString *str;
          _consume(self,1);

          str = _parseUntil(self, '"');
          data = [str dataUsingEncoding:[NSString defaultCStringEncoding]];
        }
        else 
          data = _parseData(self);

        if (data)
          [fetch setObject:data forKey:key];
        
      }
      else {
        [[[NGImap4ParserException alloc]
                                  initWithFormat:
                                    @"unsupported fetch %@", key]
                                  raise];
      }
      if (_la(self, 0) == ' ')
        _consume(self, 1);
    }
    if (fetch) {
      [fetch setObject:number forKey:@"msn"];
      [result_ addObject:fetch forKey:@"fetch"];
      _consume(self, 1); /* consume ')' */
      _match(self, '\n');
    }
    else { /* no correct fetch line */
      _parseUntil(self, '\n');
    }
  }
  else { /* got a number request from select like  exists or recent */
    key = _parseUntil(self, '\n');
    [result_ addObject:number forKey:[key lowercaseString]];
  }
  return YES;
}

static BOOL _parseGreetingsSieveResponse(NGImap4ResponseParser *self,
                                         NGMutableHashMap *result_) {
  BOOL isOK;
  while (!(isOK = _parseOkSieveResponse(self, result_))) {
    NSString *key, *value;

    if (!(key = _parseStringSieveResponse(self))) {
      break;
    }
    _match(self, ' ');
    if (!(value = _parseStringSieveResponse(self))) {
      break;
    }
    _parseUntil(self, '\n');
    [result_ addObject:value forKey:[key lowercaseString]];

    if (USE_CYRUS_2_0 == 0)
      return YES;
  }
  return isOK;
}

static BOOL _parseDataSieveResponse(NGImap4ResponseParser *self,
                                  NGMutableHashMap *result_) {
  NSData *data;

  if ((data = _parseData(self))) {
    NSString *str;

    str = [[NSString alloc] initWithData:data
                            encoding:[NSString defaultCStringEncoding]];
    [result_ setObject:str forKey:@"data"];
    [str release]; str = nil;
    return YES;
  }
  return NO;
}

static BOOL _parseOkSieveResponse(NGImap4ResponseParser *self,
                                  NGMutableHashMap *result_) {
  if ((_la(self, 0) == 'O')
      && (_la(self, 1) == 'K')) {

    _consume(self, 2);

    if (_la(self, 0) == ' ') {
      NSString *reason;

      if ((reason = _parseContentSieveResponse(self)))
        [result_ addObject:reason forKey:@"reason"];
    }
    _parseUntil(self, '\n');

    [result_ addObject:[NSNumber numberWithBool:YES] forKey:@"ok"];
    return YES;
  }
  return NO;
}

static BOOL _parseNoSieveResponse(NGImap4ResponseParser *self,
                                  NGMutableHashMap *result_) {
  if ((_la(self, 0) == 'N')
      && (_la(self, 1) == 'O')
      && (_la(self, 2) == ' ')) {
    NSString *data;
    
    _consume(self, 3);

    data = _parseContentSieveResponse(self);
    
    [result_ addObject:[NSNumber numberWithBool:NO] forKey:@"ok"];
    if (data) 
      [result_ addObject:data forKey:@"reason"];
    return YES;
  }
  return NO;
}

static NSString *_parseContentSieveResponse(NGImap4ResponseParser *self) {
  NSString *str;
  if ((str = _parseStringSieveResponse(self)))
    return str;
  else {
    NSData *data;

    if ((data = _parseData(self))) {
      return [[[NSString alloc] initWithData:data encoding:
                                [NSString defaultCStringEncoding]]
                         autorelease];
    }
  }
  return nil;  
}

static NSString *_parseStringSieveResponse(NGImap4ResponseParser *self) {
  if (_la(self, 0) == '"') {
    _consume(self, 1);
    return _parseUntil(self, '"');
  }
  return nil;
}

static NSString *_parseBodyDecodeString(NGImap4ResponseParser *self,
                                        BOOL _convertString,
                                        BOOL _decode)
{
  static NSStringEncoding encoding      = -1;
  static Class            NSStringClass = NULL;

  NSString *str;

  if (NSStringClass == NULL) {
    NSStringClass = [NSString class];
  }

  if (encoding == -1) {
    encoding = [NGMimePartParser defaultHeaderFieldEncoding];
  }

  if (_la(self, 0) == '"') {
    _consume(self, 1);
    str = _parseUntil(self, '"');
  }
  else if (_la(self, 0) == '{') {
    NSData *data;

    data = _parseData(self);

    if (_decode) {
      data = [data decodeQuotedPrintableValueOfMIMEHeaderField:nil];
    }
    return [[[NSString alloc] initWithData:data encoding:encoding]
                       autorelease];
  }
  else {
    str = _parseUntil2(self, ' ', ')');
  }
  if (_convertString) {
    if ([[str lowercaseString] isEqualToString:@"nil"])
      str = @"";
  }
  if (_decode) {
    id  d;

    d = [str dataUsingEncoding:[NSString defaultCStringEncoding]];
    d = [d decodeQuotedPrintableValueOfMIMEHeaderField:nil];

    if ([d isKindOfClass:NSStringClass])
      str = d;
    else
      str = [[[NSString alloc] initWithData:d encoding:encoding]
                        autorelease];
  }
  return str;
}


static NSString *_parseBodyString(NGImap4ResponseParser *self,
                                  BOOL _convertString)
{
  return _parseBodyDecodeString(self, _convertString, NO);
}

static NSDictionary *_parseBodyParameterList(NGImap4ResponseParser *self)
{
  NSMutableDictionary *list;

  if (_la(self, 0) == '(') {
    _consume(self, 1);

    list = [NSMutableDictionary dictionaryWithCapacity:4];
  
    while (_la(self,0) != ')') {
      NSString *key, *value;

      if (_la(self, 0) == ' ')
        _consume(self, 1);
      
      key = _parseBodyString(self, YES);
      _match(self, ' ');
      value = _parseBodyDecodeString(self, YES, YES);

      [list setObject:value forKey:[key lowercaseString]];
    }
    _match(self, ')');
  }
  else {
    NSString *str;
    str = _parseBodyString(self, YES);

    if ([str length] > 0) {
      NSLog(@"%s: got unexpected string %@", __PRETTY_FUNCTION__, str);
    }
    list = (id)[NSDictionary dictionary];
  }
  return list;
}

static NSArray *_parseAddressStructure(NGImap4ResponseParser *self) {
  NSString     *personalName, *sourceRoute, *mailboxName, *hostName;
  _match(self, '(');
  personalName = _parseBodyString(self, YES);
  _match(self, ' ');
  sourceRoute = _parseBodyString(self, YES);
  _match(self, ' ');
  mailboxName = _parseBodyString(self, YES);
  _match(self, ' ');
  hostName = _parseBodyString(self, YES);
  _match(self, ')');
  return [NSDictionary dictionaryWithObjectsAndKeys:
                       personalName, @"personalName",
                       sourceRoute,  @"sourceRoute",
                       mailboxName,  @"mailboxName",
                       hostName,     @"hostName", nil];
}

static NSArray *_parseParenthesizedAddressList(NGImap4ResponseParser *self) {
  NSMutableArray *result;
  result = [NSMutableArray arrayWithCapacity:8];

  if (_la(self, 0) == '(') {
    _consume(self, 1);
    while (_la(self, 0) != ')') {
      [result addObject:_parseAddressStructure(self)];
    }
    _consume(self, 1);
  }
  else {
    NSString *str;
    str = _parseBodyString(self, YES);

    if ([str length] > 0) {
      NSLog(@"%s: got unexpected string %@", __PRETTY_FUNCTION__, str);
    }
    result = (id)[NSArray array];
  }
  return result;
}

static NSDictionary *_parseSingleBody(NGImap4ResponseParser *self) {
  NSString            *type, *subtype, *bodyId, *description,
                      *encoding, *bodysize;
  NSDictionary        *parameterList;
  NSMutableDictionary *dict;

  type = _parseBodyString(self, YES);
  _match(self, ' ');
  subtype = _parseBodyString(self, YES);
  _match(self, ' ');
  parameterList = _parseBodyParameterList(self);
  _match(self, ' ');
  bodyId = _parseBodyString(self, YES);
  _match(self, ' ');
  description = _parseBodyString(self, YES);
  _match(self, ' ');
  encoding = _parseBodyString(self, YES);
  _match(self, ' ');
  bodysize = _parseBodyString(self, YES);

  dict = [NSMutableDictionary dictionaryWithObjectsAndKeys:
                              type, @"type",
                              subtype, @"subtype",
                              parameterList, @"parameterList",
                              bodyId,        @"bodyId",
                              description,   @"description",
                              encoding,      @"encoding",
                              bodysize,      @"size", nil];
  
  if ([type isEqualToString:@"TEXT"]) {
    _match(self, ' ');
    [dict setObject:_parseBodyString(self, YES) forKey:@"lines"];
  }
  if ([type isEqualToString:@"MESSAGE"]) {
    if (_la(self, 0) != ')') {
      _match(self, ' ');
      _match(self, '(');
      [dict setObject:_parseBodyString(self, YES) forKey:@"date"];
      _match(self, ' ');
      [dict setObject:_parseBodyString(self, YES) forKey:@"subject"];
      _match(self, ' ');
      [dict setObject:_parseParenthesizedAddressList(self) forKey:@"from"];
      _match(self, ' ');
      [dict setObject:_parseParenthesizedAddressList(self) forKey:@"sender"];
      _match(self, ' ');
      [dict setObject:_parseParenthesizedAddressList(self)
            forKey:@"reply-to"];
      _match(self, ' ');
      [dict setObject:_parseParenthesizedAddressList(self) forKey:@"to"];
      _match(self, ' ');
      [dict setObject:_parseParenthesizedAddressList(self) forKey:@"cc"];
      _match(self, ' ');
      [dict setObject:_parseParenthesizedAddressList(self) forKey:@"bcc"];
      _match(self, ' ');
      [dict setObject:_parseBodyString(self, YES) forKey:@"in-reply-to"];
      _match(self, ' ');
      [dict setObject:_parseBodyString(self, YES) forKey:@"messageId"];
      _match(self, ')');
      _match(self, ' ');
      [dict setObject:_parseBody(self) forKey:@"body"];
      _match(self, ' ');
      [dict setObject:_parseBodyString(self, YES) forKey:@"bodyLines"];
    }
  }
  return dict;
}

static NSDictionary *_parseMultipartBody(NGImap4ResponseParser *self) {
  NSMutableArray *parts;
  NSString       *kind;

  parts = [NSMutableArray arrayWithCapacity:4];

  while (_la(self, 0) == '(') {
    [parts addObject:_parseBody(self)];
  }
  _match(self, ' ');
  kind = _parseBodyString(self, YES);
  return [NSDictionary dictionaryWithObjectsAndKeys:
                       parts,        @"parts",
                       @"multipart", @"type",
                       kind        , @"subtype", nil];
}

static NSDictionary *_parseBody(NGImap4ResponseParser *self) {
  NSDictionary *result;

  _match(self, '(');

  if (_la(self, 0) == '(') {
    result = _parseMultipartBody(self);
  }
  else {
    result = _parseSingleBody(self);
  }
  if (_la(self,0) != ')') {
    NSString *str;

    str = _parseUntil(self, ')');
    NSLog(@"%s: got noparsed content %@", __PRETTY_FUNCTION__,
          str);
  }
  else 
    _consume(self, 1);

  return result;
}

static NSDictionary *_parseBodyContent(NGImap4ResponseParser *self) {
  NSData *data;

  if (_la(self, 0) == '"') {
    NSString *str;
    _consume(self,1);

    str = _parseUntil(self, '"');
    data = [str dataUsingEncoding:[NSString defaultCStringEncoding]];
  }
  else 
    data = _parseData(self);
  
  return [NSDictionary dictionaryWithObject:data forKey:@"data"];
}


static NSArray *_parseFlagArray(NGImap4ResponseParser *self) {
  NSString *flags;
  
  _match(self, '(');
  
  flags = _parseUntil(self, ')');
  if ([flags length] == 0) {
    static NSArray *emptyArray = nil;
    if (emptyArray == nil) emptyArray = [[NSArray alloc] init];
    return emptyArray;
  }
  else
    return [[flags lowercaseString] componentsSeparatedByString:@" "];
}

static BOOL _parseFlagsUntaggedResponse(NGImap4ResponseParser *self,
                                        NGMutableHashMap *result_) {
  if ((_la(self, 0) == 'F')
      && (_la(self, 1) == 'L')
      && (_la(self, 2) == 'A')
      && (_la(self, 3) == 'G')
      && (_la(self, 4) == 'S')      
      && (_la(self, 5) == ' ')) {
    _consume(self, 6);
    [result_ addObject:_parseFlagArray(self) forKey:@"flags"];
    _match(self, '\n');
    return YES;
  }
  return NO;
}

static BOOL _parseBadUntaggedResponse(NGImap4ResponseParser *self,
                                     NGMutableHashMap *result_) {
  if ((_la(self, 0) == 'B')
      && (_la(self, 1) == 'A')
      && (_la(self, 2) == 'D')      
      && (_la(self, 3) == ' ')) {
    _consume(self, 4);
    [result_ addObject:_parseUntil(self, '\n')
             forKey:@"bad"];
    return YES;
  }
  return NO;
}

static BOOL _parseNoOrOkArguments(NGImap4ResponseParser *self,
                                  NGMutableHashMap *result_, NSString *_key) {
  NSString *obj;

  obj = nil;
  
  if (_la(self, 0) == '[') {
    NSString *key;
      
    _consume(self, 1);
    key = _parseUntil2(self, ']', ' ');

    /* possible kinds of untagged OK responses are either
     * OK [ALERT] System shutdown in 10 minutes
     or               
     * OK [UNSEEN 14]
     or
     * OK [PERMANENTFLAGS (\Answered \Flagged \Draft \Deleted \Seen \*)]
     */
    if (_la(self, 0) == ']') {

      _consume(self, 1);
      if (_la(self, 0) == ' ') {
        id value;
        
        _consume(self, 1);
        value = _parseUntil(self, '\n');
        if ([value length])
          obj = [NSDictionary dictionaryWithObject:value forKey:key];
        else
          obj = key;
      }
      else {
        obj = key;
        _parseUntil(self, '\n');
      }
    }
    else { /* _la(self, 0) should be ' ' */
      id value;

      value = nil;
      
      _consume(self, 1);
      if (_la(self, 0) == '(') {
        value = _parseFlagArray(self);
        _consume(self, 1); /* consume ']' */
      }
      else {
        value = _parseUntil(self, ']');
      }
      obj = [NSDictionary dictionaryWithObject:value forKey:key];
      _parseUntil(self, '\n');      
    }
  }
  else
    obj = _parseUntil(self, '\n');

  [result_ addObject:obj forKey:_key];
  return YES;
}


static BOOL _parseNoUntaggedResponse(NGImap4ResponseParser *self,
                                     NGMutableHashMap *result_) {
  if ((_la(self, 0) == 'N')
      && (_la(self, 1) == 'O')
      && (_la(self, 2) == ' ')) {
    _consume(self, 3);
    return _parseNoOrOkArguments(self, result_, @"no");
  }
  return NO;
}

static BOOL _parseOkUntaggedResponse(NGImap4ResponseParser *self,
                                     NGMutableHashMap *result_) {
  if ((_la(self, 0) == 'O')
      && (_la(self, 1) == 'K')
      && (_la(self, 2) == ' ')) {
    _consume(self, 3);
    return _parseNoOrOkArguments(self, result_, @"ok");
  }
  return NO;
}


static NSNumber *_parseUnsigned(NGImap4ResponseParser *self) {
  unsigned n;
  char     c;
  BOOL     isNumber;

  isNumber = NO;  
  n        = 0;  
  c        = _la(self, 0);
  
  while ((c >= '0') && (c <= '9')) {
    _consume(self, 1);
    isNumber = YES;
    n        = 10 * n + (c - 48);
    c        = _la(self, 0);
  }
  if (isNumber == YES)
    return [NSNumber numberWithUnsignedInt:n];
  else
    return nil;
}

/*"
** _parseUntil(self, char) consume the stop char
** normalize \r\n constructions
"*/

static NSString *_parseUntil(NGImap4ResponseParser *self, char _c) {
  char            buf[1024], c;
  NSMutableString *str;
  unsigned        cnt;

  cnt = 0;
  str = nil;  
  while ((c = _la(self, 0)) != _c) {
    buf[cnt] = c;
    _consume(self, 1);
    cnt++;
    if (cnt == 1024) {
      if (str == nil)
        str = (NSMutableString *)
          [NSMutableString stringWithCString:buf length:1024];
      else
        [str appendString:[NSString stringWithCString:buf length:1024]];
      cnt = 0;
    }
  }
  _consume(self,1); /* consume known stop char */
  if (_c == '\n' && cnt > 0) {
    if (buf[cnt-1] == '\r')
      cnt--;
  }
  return (str == nil)
    ? [NSString stringWithCString:buf length:cnt]
    : [str stringByAppendingString:
           [NSString stringWithCString:buf length:cnt]];
}

/*"
** _parseUntil2(self, char, char) doesn`t consume the stop-chars
"*/

static NSString *_parseUntil2(NGImap4ResponseParser *self, char _c1, char _c2)
{
  char            buf[1024], c;
  NSMutableString *str;
  unsigned        cnt;

  cnt = 0;
  c   = _la(self, 0);
  str = nil;
  
  while ((c != _c1) && (c != _c2)) {
    buf[cnt] = c;
    _consume(self, 1);
    cnt++;
    if (cnt == 1024) {
      if (str == nil)
        str = (NSMutableString *)
                 [NSMutableString stringWithCString:buf length:1024];
      else
        [str appendString:[NSString stringWithCString:buf length:1024]];
      
      cnt = 0;
    }
    c = _la(self, 0);    
  }
  return (str == nil)
    ? [NSString stringWithCString:buf length:cnt]
    : [str stringByAppendingString:
           [NSString stringWithCString:buf length:cnt]];
}



static inline void _match(NGImap4ResponseParser *self, char _match) {
  if (_la(self,0) == _match) 
    _consume(self, 1);
  else {
    if (self->debug == YES) {
      [[[NGImap4ParserException alloc]
                                     initWithFormat:@"unexpected char <%c> "
                                     @"expected <%c> <%@>",
                                     _la(self, 0), _match,
                                     self->serverResponseDebug] raise];
    }
    else {
      [[[NGImap4ParserException alloc]
                                     initWithFormat:@"unexpected char <%c> "
                                     @"expected <%c>",
                                     _la(self, 0), _match] raise];
    }
  }
}

static inline int __la(NGImap4ResponseParser *self, unsigned _laCnt) {
  return (self->la == NULL)
               ? [self->buffer la:_laCnt]
               : self->la(self->buffer, @selector(la:), _laCnt);
}

static inline int _la(NGImap4ResponseParser *self, unsigned _laCnt) {
  char c = __la(self, _laCnt);

  if (c == '\r') {
    return _la(self, _laCnt + 1);
  }
  return c;
}


static inline void _consume(NGImap4ResponseParser *self, unsigned _cnt) {

  /* Normalize end of line */

  if (_cnt == 0)
    return;
  
  _cnt +=  (__la(self, _cnt - 1) == '\r') ? 1 : 0;
  
  if (self->debug == YES) {
    int cnt = 0;

    while (cnt < _cnt) {
      char c = _la(self, cnt);
      if (c != '\r') {
        [self->serverResponseDebug appendString:
             [NSString stringWithCString:&c length:1]];
        if (c == '\n') {
          if ([self->serverResponseDebug cStringLength] > 2) {
            fprintf(stderr, "S[%p]: %s", self,
                    [self->serverResponseDebug cString]);
          }
          [self->serverResponseDebug release];
          self->serverResponseDebug = [[NSMutableString alloc]
                                                        initWithCapacity:512];
        }
      }
      cnt++;      
    }
  }
  [self->buffer consume:_cnt];
}

@end /* NGImap4ResponseParser */
