// $Id: NGImap4ServerRoot.m,v 1.11 2003/02/26 16:53:00 jan Exp $

#import "common.h"
#import "NGImap4ServerRoot.h"
#import "NGImap4Context.h"
#import "NGImap4Client.h"
#import "NGImap4Message.h"
#import "NGImap4Functions.h"
#import "NGImap4Folder.h"

@interface NGImap4ServerRoot(Private)
- (void)initializeSubFolders;
@end

@implementation NGImap4ServerRoot

+ (id)serverRootWithContext:(NGImap4Context *)_context {
  return [[[self alloc] initServerRootWithContext:_context] autorelease];
}

- (id)init {
  [self release];
  [self doesNotRecognizeSelector:_cmd];
  return nil;
}

- (id)initServerRootWithContext:(NGImap4Context *)_context {
  if ((self = [super init])) {
    self->context    = [_context retain];
    self->name       = [[_context host] copy];
    self->subFolders = nil;
  }
  return self;
}

- (void)dealloc {
  [self->context resetSpecialFolders];
  [self->context release];
  [self->name    release];
  [self->subFolders release];
  [super dealloc];
}

- (NSException *)lastException {
  return [self->context lastException];
}
- (void)resetLastException {
  [self->context resetLastException];
}

- (BOOL)isEqual:(id)_obj {
  if (self == _obj)
    return YES;
  if ([_obj isKindOfClass:[NGImap4ServerRoot class]])
    return [self isEqualToServerRoot:_obj];
  return NO;
}

- (BOOL)isEqualToServerRoot:(NGImap4ServerRoot *)_root {
  if (_root == self)
    return YES;
  if ([_root context] == self->context) {
    return YES;
  }
  return NO;
}

/* accessors */

- (NGImap4Context *)context {
  return self->context;
}

- (NSString *)name {
  return self->name;
}

- (NSString *)absoluteName {
  return self->name;
}

- (NSArray *)messages {
  return nil;
}

- (NSArray *)messagesForQualifier:(EOQualifier *)_qualifier {
  return nil;
}

- (NSArray *)messagesForQualifier:(EOQualifier *)_qualifier maxCount:(int)_cnt {
  return nil;
}

- (NSArray *)fetchSortedMessages:(NSArray *)_so {
  return nil;
}

- (void)bulkFetchHeadersFor:(NSArray *)_array inRange:(NSRange)_aRange {
}


- (NSArray *)messageFlags { 
  return nil;
}

- (NSArray *)subFolders {
  if (self->subFolders == nil)
    [self initializeSubFolders];
  return self->subFolders;
}

- (NGImap4Folder *)subFolderWithName:(NSString *)_name
  caseInsensitive:(BOOL)_caseIns
{
  return _subFolderWithName(self, _name, _caseIns);
}

- (id<NGImap4Folder>)parentFolder {
  return nil;
}

- (BOOL)isReadOnly {
  return YES;
}

- (BOOL)noinferiors {
  return NO;
}

- (BOOL)noselect {
  return YES;
}

- (int)exists {
  return 0;
}

- (int)recent {
  return 0;
}

- (int)unseen {
  return 0;
}

- (int)usedSpace {
  return 0;
}

- (int)maxQuota {
  return 0;
}

- (BOOL)isOverQuota {
  return NO;
}

- (BOOL)hasNewMessagesSearchRecursiv:(BOOL)_recursiv {
  if (_recursiv == YES)
    return _hasNewMessagesInSubFolder(self);
  return NO;
}

- (BOOL)hasUnseenMessagesSearchRecursiv:(BOOL)_recursiv {
  if (_recursiv == YES)  
    return _hasUnseenMessagesInSubFolder(self);
  return NO;
}

// private methods

/*"
** Should only happens if folder is rootfolder
"*/

- (void)initializeSubFolders {
  NSEnumerator *folders;
  NSDictionary *res;
  id           *objs, folder;
  unsigned     cnt;
  BOOL         gotInbox;
  
  static int ShowOnlySubscribedInRoot = -1;

  if (ShowOnlySubscribedInRoot == -1) {
    ShowOnlySubscribedInRoot =
      [[NSUserDefaults standardUserDefaults]
                       boolForKey:@"ShowOnlySubscribedInRoot"]?1:0;
  }

  if (self->subFolders != nil) {
    [self resetSubFolders];
  }

  if (ShowOnlySubscribedInRoot) {
    res = [[self->context client] lsub:@"" pattern:@"%"];
  }
  else {
    res = [[self->context client] list:@"" pattern:@"%"];
  }

  if (!_checkResult(self->context, res, __PRETTY_FUNCTION__))
    return;
  
  res = [res objectForKey:@"list"];

#if NeXT_Foundation_LIBRARY
  objs = NSZoneMalloc([self zone], sizeof(id) * ([res count] + 1));
#else
  objs = NSZoneMallocAtomic([self zone], sizeof(id) * ([res count] + 1));
#endif
  {
    NSArray *names;
    
    names   = [res allKeys];
    names   = [names sortedArrayUsingSelector:
                       @selector(caseInsensitiveCompare:)];
    folders = [names objectEnumerator];
  }
  cnt      = 0;
  gotInbox = NO;
  
  while ((folder = [folders nextObject])) {
    objs[cnt++] = [NGImap4Folder imap4FolderWithContext:self->context
                               name:folder
                               flags:[res objectForKey:folder]
                               parentFolder:self];
    if ([[folder lowercaseString] isEqualToString:@"/inbox"])
      gotInbox = YES;
  }
  if (!gotInbox && ShowOnlySubscribedInRoot) { /* try unsubscribed */
    res = [[[self->context client] list:@"" pattern:@"%"] objectForKey:@"list"];

    {
      NSArray *names;
    
      names   = [res allKeys];
      names   = [names sortedArrayUsingSelector:
                       @selector(caseInsensitiveCompare:)];
      folders = [names objectEnumerator];
    }
    while ((folder = [folders nextObject])) {
      if ([[folder lowercaseString] isEqualToString:@"/inbox"]) {
        objs[cnt++] = [NGImap4Folder imap4FolderWithContext:self->context
                                     name:folder
                                     flags:[res objectForKey:folder]
                                     parentFolder:self];
        break;
      }
    }
  }
  self->subFolders = [[NSArray allocWithZone:[self zone]]
                              initWithObjects:objs count:cnt];
  NGFree(objs); objs = NULL;
}

- (void)select {
}

- (void)status {
}

/* actions */

- (void)resetFolder {
}

- (void)resetStatus {
}

- (void)resetSubFolders {
  [self->context resetSpecialFolders];
  [self->subFolders release]; self->subFolders = nil;
}

- (BOOL)renameTo:(NSString *)_name {
  return NO;
}

/* folder */

- (BOOL)deleteSubFolder:(NGImap4Folder *)_folder {
  return _deleteSubFolder(self, _folder);
}

- (BOOL)copySubFolder:(NGImap4Folder *)_f to:(id<NGImap4Folder>)_folder {
  return _copySubFolder(self, _f, _folder);
}

- (BOOL)moveSubFolder:(NGImap4Folder *)_f to:(id<NGImap4Folder>)_folder {
  return _moveSubFolder(self, _f, _folder);
}

- (BOOL)createSubFolderWithName:(NSString *)_name {
  return _createSubFolderWithName(self, _name, NO);
}

- (void)expunge {
}

- (BOOL)addFlag:(NSString *)_flag toMessages:(NSArray *)_messages {
  return NO;
}

- (BOOL)removeFlag:(NSString *)_flag fromMessages:(NSArray *)_messages {
  return NO;
}

- (BOOL)flag:(NSString *)_flag toMessages:(NSArray *)_messages
        add:(NSNumber *)_add
{
  return NO;
}

- (BOOL)flagToAllMessages:(NSString *)_flag add:(NSNumber *)_add {
  return NO;
}

- (BOOL)deleteAllMessages {
  return NO;
}

- (BOOL)deleteMessages:(NSArray *)_messages {
  return NO;
}

- (BOOL)moveMessages:(NSArray *)_messages toFolder:(NGImap4Folder *)_folder {
  return NO;
}

- (BOOL)copyMessages:(NSArray *)_messages toFolder:(NGImap4Folder *)_folder {
  return NO;
}

- (BOOL)appendMessage:(NSData *)_msg {
  return NO;
}

- (BOOL)isInTrash {
  return NO;
}

- (NSString *)description {
  NSMutableString *ms;
  NSString *tmp;

  ms = [NSMutableString stringWithCapacity:64];

  [ms appendFormat:@"<0x%08X[%@]:", self, NSStringFromClass([self class])];
  
  if ((tmp = [self name]))
    [ms appendFormat:@" name=%@", tmp];
  if ((tmp = [self absoluteName]))
    [ms appendFormat:@" absolute=%@", tmp];
  
  [ms appendString:@">"];

  return ms;
}

- (void)resetSync {
  NSEnumerator *enumerator;
  id           folder;

  enumerator = [[self subFolders] objectEnumerator];
  
  while ((folder = [enumerator nextObject])) {
    [folder resetSync];
  }
}

@end /* NGImap4ServerRoot */
