// $Id: NGMailAddressParser.m,v 1.3 2003/01/07 11:38:28 helge Exp $

#include "NGMailAddressParser.h"
#include "NGMailAddress.h"
#include "NGMailAddressList.h"
#include "common.h"

@interface NGMailAddressParser(PrivateMethods)
- (id)parseQuotedString:(BOOL)_guestMode;
- (id)parseWord:(BOOL)_guestMode;
- (id)parsePhrase:(BOOL)_guestMode;
- (id)parseLocalPart:(BOOL)_guestMode;
- (id)parseDomain:(BOOL)_guestMode;
- (id)parseAddrSpec:(BOOL)_guestMode;
- (id)parseRouteAddr:(BOOL)_guessMode;
- (id)parseGroup:(BOOL)_guessMode;
- (id)parseMailBox:(BOOL)_guessMode;
- (id)parseAddress:(BOOL)_guessMode;
@end

@implementation NGMailAddressParser

+ (int)version {
  return 2;
}

static inline id parseWhiteSpaces(NGMailAddressParser *self, BOOL _guessMode) {
  id   returnValue = nil;
  char text[self->maxLength];
  int  length      = 0;
  
  while ((self->data[self->dataPos] == ' ')  ||
         (self->data[self->dataPos] == '\n')) {
    text[length++] = ' ';    
    self->dataPos++;
  }
  if (length) {
    if (_guessMode) { 
      returnValue = [NSNumber numberWithBool:YES];
    }
    else {
      returnValue = [NSString stringWithCString:text length:length]; 
    }
  }
  return returnValue;
}


static inline id parseAtom(NGMailAddressParser *self, BOOL _guessMode) {
  int  keepPos     = self->dataPos; // keep reference for backtracking
  id   returnValue = nil;
  BOOL isAtom      = YES;
  char text[self->maxLength];  // token text
  int  length      = 0;  // token text length
  BOOL done        = NO;

  do {
    if (self->dataPos == self->maxLength) { // end of text is reached
      isAtom = (length > 0);
      done   = YES;
    }
    else {
      register unsigned char c = self->data[self->dataPos];
      
      switch (c) {
        case '(' :  case ')': case '<': case '>':
        case '@' :  case ',': case ';': case ':':
        case '\\':  case '"': case '.': case '[':
        case ']' :  case ' ': case 127:
          isAtom = (length > 0);
          done   = YES;
          break;

        default:
          if (c < 32) {
            isAtom = (length > 0);
            done   = YES;
          }
          else {
            text[length] = c;  // store char in text
            length++;          // increase text size
            (self->dataPos)++; // go ahead
          }
      }
    }
  }
  while (!done);
  
  if (isAtom) {
    if (_guessMode) {
      NSCAssert(length > 0, @"no atom with length=0");
      returnValue =  [NSNumber numberWithBool:YES];
    }
    else {
      NSCAssert(length > 0, @"no atom with length=0");
      returnValue = [NSString stringWithCString:text length:length];
      NSCAssert([returnValue isKindOfClass:[NSString class]],
                     @"got no string ..");
    }
  }
  else {
    self->dataPos = keepPos;
    returnValue = nil;
  }
  return returnValue;
}

static inline id parseQuotedPair(NGMailAddressParser *self, BOOL _guessMode) {
  id   returnValue  = nil;

  if ((self->maxLength - (self->dataPos)) < 3) {
    returnValue =  nil;
  }
  else {
    if (self->data[self->dataPos] == '\\') {
      self->dataPos = self->dataPos + 2;      
      if (_guessMode)
        returnValue = [NSNumber numberWithBool:YES];
      else
        returnValue = [NSString stringWithCString:
                                   &(self->data[self->dataPos - 1])
                                length:1];
    }
  }
  return returnValue;
}

static inline id parseQText(NGMailAddressParser *self, BOOL _guessMode) {
  int  keepPos     = self->dataPos; // keep reference for backtracking
  id   returnValue = nil;
  BOOL isQText    = YES;
  char text[self->maxLength];  // token text
  int  length      = 0;  // token text length
  BOOL done        = YES;

  do {
    if (self->dataPos == self->maxLength) { // end of text is reached
      isQText = (length > 0);
      done    = YES;
    }
    else {
      register char c = self->data[self->dataPos];
      
      switch ((int)c) {
        case '"' :  
        case '\\':
        case 13  :
          isQText = (length > 0);
          done    = YES;
          break;

        default: {
            text[length] = c;  // store char in text
            length++;          // increase text size
            (self->dataPos)++; // go ahead
          }
      }
    }
  }
  while (!done);

  if (isQText) {
    if (_guessMode) {
      NSCAssert(length > 0, @"no qtext with length=0");
      returnValue = [NSNumber numberWithBool:YES];
    }
    else {
      NSCAssert(length > 0, @"no qtext with length=0");
      returnValue = [NSString stringWithCString:text length:length];
      NSCAssert([returnValue isKindOfClass:[NSString class]],
               @"got no string ..");
    }
  }
  else {
    self->dataPos = keepPos;
    returnValue = nil;
  }
  return returnValue;
}

static inline id parseDText(NGMailAddressParser *self, BOOL _guessMode) {
  int  keepPos     = self->dataPos; // keep reference for backtracking
  id   returnValue = nil;
  BOOL isDText    = YES;
  char text[self->maxLength];  // token text
  int  length      = 0;  // token text length
  BOOL done        = YES;

  do {
    if (self->dataPos == self->maxLength) { // end of text is reached
      isDText = (length > 0);
      done    = YES;
    }
    else {
      register char c = self->data[self->dataPos];
      
      switch ((int)c) {
        case '[':  case ']':
        case '\\': case 13:
          isDText = (length > 0);
          done    = YES;
          break;

        default: {
            text[length] = c;  // store char in text
            length++;          // increase text size
            (self->dataPos)++; // go ahead
          }
      }
    }
  }
  while (!done);

  if (isDText) {
    if (_guessMode) {
      NSCAssert(length > 0, @"no dtext with length=0");
      returnValue =  [NSNumber numberWithBool:YES];
    }
    else {
      NSCAssert(length > 0, @"no dtext with length=0");
      returnValue = [NSString stringWithCString:text length:length];
      NSCAssert([returnValue isKindOfClass:[NSString class]],
               @"got no string ..");
    }
  }
  else {
    self->dataPos = keepPos;
    returnValue = nil;
  }
  return returnValue;
}

static inline id parseDomainLiteral(NGMailAddressParser *self, BOOL _guessMode) {
  int             keepPos          = self->dataPos;
  id              returnValue      = nil;
  BOOL            returnOK         = NO;

  if (_guessMode) {
    if (self->data[self->dataPos] != '[')
      return nil;

    (self->dataPos)++; // skip starting '"'

    // parses: "suafdjksfd \"sdafsadf"
    while (self->data[self->dataPos] != ']') {
      if (self->data[self->dataPos] == '\\') {// skip quoted chars 
        (self->dataPos)++;
      }
      (self->dataPos)++;
      if (self->dataPos >= self->maxLength) {
        return nil;
      }
    }
    (self->dataPos)++; // skip closing '"'
    returnValue = [NSNumber numberWithBool:YES];
  }
  else {
    if (self->data[self->dataPos++] == '[') {
      id  result     = nil;      
      returnValue    = [NSMutableString stringWithCapacity:10];
      do {
        if ((result = parseQuotedPair(self, NO)))
          [returnValue appendString:result];
        else
          if ((result = parseDText(self, NO))) 
            [returnValue appendString:result];        
      } while (result);
      
      if (self->data[self->dataPos++] == ']')
        returnOK = YES;
    }
    if (!returnOK) {
      if (returnValue) {
        returnValue = nil;
      }
      self->dataPos = keepPos;
    }
  }
  return returnValue;
}

// constructors

+ (id)mailAddressParserWithString:(NSString *)_string {
  return [[[self alloc] initWithCString:[_string cString]
                        length:[_string cStringLength]] autorelease];
}
+ (id)mailAddressParserWithData:(NSData *)_data {
  return [[[self alloc] initWithCString:(char *)[_data bytes]
                        length:[_data length]] autorelease];
}
+ (id)mailAddressParserWithCString:(char *)_cString {
  return [[[self alloc] initWithCString:_cString
                        length:strlen(_cString)] autorelease];
}
- (id)initWithCString:(const char*)_cString length:(int unsigned)_length{
  if ((self = [super init])) {
    (const char *)self->data = _cString;
    self->maxLength          = _length;
    self->dataPos            = 0;
    self->errorPos           = -1;
  }
  return self;
}

- (void)dealloc {
  self->data      = NULL;
  self->maxLength = 0;
  self->dataPos   = 0;
  [super dealloc];
}

- (id)parseQuotedString:(BOOL)_guessMode {
  int  keepPos     = self->dataPos;
  id   returnValue = nil;
  BOOL returnOK    = NO;

  if (_guessMode) {
    if (self->data[self->dataPos] != '"')
      return nil;

    (self->dataPos)++; // skip starting '"'

    // parses: "suafdjksfd \"sdafsadf"
    while (self->data[self->dataPos] != '"') {
      if (self->data[self->dataPos] == '\\') {// skip quoted chars 
        (self->dataPos)++;
      }
      (self->dataPos)++;
      if (self->dataPos >= self->maxLength) {
        self->dataPos = keepPos;
        return nil;
      }
    }
    (self->dataPos)++; // skip closing '"'
    returnValue = [NSNumber numberWithBool:YES];
  }
  else {
    if (data[dataPos++] == '"') {
      id result = nil;
      
      returnValue = [NSMutableString stringWithCapacity:10];
      do {
        if ((result = parseQuotedPair(self, NO)))
          [returnValue appendString:result];
        else
          if ((result = parseQText(self, NO))) 
            [returnValue appendString:result];        
      } while (result);
      
      if (data[dataPos++] == '"')
        returnOK = YES;
    }
    if (!returnOK) {
      if (returnValue) {
        returnValue = nil;
      }
      dataPos = keepPos;
    }
  }
  return returnValue;  
}

- (id)parseWord:(BOOL)_guessMode {
  id returnValue = nil;

  if (!((returnValue = [self parseQuotedString:_guessMode])))
    returnValue = parseAtom(self, _guessMode);
  
  return returnValue;
}

- (id)parsePhrase:(BOOL)_guessMode {
  BOOL isPhrase    = NO;
  id   returnValue = nil;
  id   result      = nil;      

  if (_guessMode) {
    do {
      if ((result = parseWhiteSpaces(self, _guessMode))) {
        isPhrase = YES;
      }
      else
        if ((result = [self parseWord:_guessMode])) {
          isPhrase = YES;
          [returnValue appendString:result];
          if ((result = parseWhiteSpaces(self, _guessMode))) {}
        }
    } while (result);

    if (!isPhrase) {
      returnValue = nil;
    }
    else 
      returnValue = [NSNumber numberWithBool:YES];
  }
  else {
    NSString *tmp;
    
    returnValue = [NSMutableString stringWithCapacity:10];
    tmp         = nil;

    do {
      if ((result = parseWhiteSpaces(self, _guessMode))) {
        tmp = result;
        ;
        //        isPhrase = YES;
        //        [returnValue appendString:result];
      }
      else if ((result = [self parseWord:_guessMode])) {
        isPhrase = YES;

        if (tmp)
          [returnValue appendString:tmp];

        tmp = nil;
          
        [returnValue appendString:result];
        if (self->dataPos < self->maxLength) {
          if (self->data[self->dataPos] == '.') {
            [returnValue appendString:@"."];
            self->dataPos++;
          }
        }
      }
    } while (result);

    if (!isPhrase || (![returnValue length])) {
      returnValue = nil;
    }
  }
  return returnValue;
}

- (id)parseLocalPart:(BOOL) _guessMode {
  id returnValue = nil;
  
  if (_guessMode) {
    if ([self parseWord:YES]) {
      id result = nil;

      do {
        result = nil;
        if (self->data[self->dataPos] == '.') {
          self->dataPos++;
          result = [self parseWord:YES];
        }
      } while (result);
      returnValue = [NSNumber numberWithBool:YES];
    }
    else {
      returnValue = nil;
    }
  }
  else {
    returnValue = [self parseWord:NO];
    
    if (returnValue) {
      NSString *result = nil;

      returnValue = [returnValue mutableCopyWithZone:[self zone]];
      
      do {
        if (self->data[self->dataPos] == '.') {
          self->dataPos++;
          result = [self parseWord:NO];
          
          if (result) {
            NSAssert([result isKindOfClass:[NSString class]],
                     @"parseWord should return string");
            
            [returnValue appendString:@"."];
            [returnValue appendString:result];
          }
        }
        else
          result = nil;
      } while (result);
      
      returnValue = [returnValue autorelease];
    }
  }
  return returnValue;
}

- (id)parseDomain:(BOOL)_guessMode {
  id returnValue = nil;
  id result      = nil;

  if (_guessMode) {
    returnValue = parseAtom(self, YES);
    if (!result)
      returnValue = parseDomainLiteral(self, YES);
    if (returnValue) {
      do {
        result = nil;
        if (self->data[self->dataPos] == '.') {
          self->dataPos++;
          result = parseAtom(self,YES);
          if (!result)
            result = parseDomainLiteral(self, YES);
        }
      } while (result);
    }
  }
  else {
    result = parseAtom(self, NO);
    if (!result)
      result = parseDomainLiteral(self, NO);

    if (result) {
      returnValue = [result mutableCopyWithZone:[self zone]];
      
      do {
        if (self->data[self->dataPos] == '.') {
          self->dataPos++;
          
          result = parseAtom(self,NO);
          if (!result)
            result = parseDomainLiteral(self, NO);
          
          if (result) {
            [returnValue appendString:@"."];
            [returnValue appendString:result];
          }
        }
        else
          result = nil;
      } while (result);

      returnValue = [returnValue autorelease];
    }
    else
      returnValue = nil;
  }
  return returnValue;
}

- (id)parseAddrSpec:(BOOL)_guessMode {
  id   returnValue  = nil;
  id   result       = nil;
  int  keepPos      = self->dataPos;
  BOOL returnStatus = NO;
  
  if (_guessMode) {
    if ([self parseLocalPart:YES])
      if (self->data[self->dataPos] == '@') {
        dataPos++;
        if ([self parseDomain:YES]) {
          returnValue = [NSNumber numberWithBool:YES];
        }
      }
  }
  else {
    returnValue = [self parseLocalPart:NO];

    if (returnValue) {
      returnValue = [returnValue mutableCopyWithZone:[self zone]];

      if (self->data[self->dataPos] == '@') {
        self->dataPos++;
        
        if ((result = [self parseDomain:NO])) {
          [returnValue appendString:@"@"];
          [returnValue appendString:result];
          returnStatus = YES;
        }
      }
      AUTORELEASE(returnValue);
    }
    if (!returnStatus) {
      returnValue = nil;
      dataPos = keepPos;
    }
  }
  return returnValue;
}

- (id)parseRoute:(BOOL)_guessMode {
  id  returnValue = nil;
  id  result      = nil;
  int keepPos     = self->dataPos;
  
  if (_guessMode) {
    BOOL status = YES;
    
    if (self->data[self->dataPos] == '@') {
      status = NO;
      if ((result = [self parseDomain:YES])) {
        status = YES;
      }
    }
    if (status) {
      parseWhiteSpaces(self,YES);
      if (self->data[self->dataPos] == ':') {
        status = YES;
      }
      else {
        status = NO;
      }
    }
    if (status) {
      returnValue = [NSNumber numberWithBool:YES];
    }
    else {
      returnValue = nil;
      self->dataPos = keepPos;
    }
  }
  else {
    BOOL status = YES;
    returnValue = [NSMutableString stringWithCapacity:10];
    if (self->data[self->dataPos] == '@') {
      status = NO;
      self->dataPos++;
      if ((result = [self parseDomain:NO])) {
        status = YES;
        [returnValue appendString:result];
      }
    }
    if (status) {
      parseWhiteSpaces(self,NO);
      if (self->data[self->dataPos] == ':') {
        status = YES;
        self->dataPos++;        
      }
      else {
        status = NO;
      }
    }
    if (!status) {
      returnValue = nil;
      self->dataPos = keepPos;
    }
  }
  return returnValue;
}

- (id)parseRouteAddr:(BOOL)_guessMode {
  int  keepPos      = self->dataPos;
  id   returnValue  = nil;
  id   result       = nil;
  BOOL returnStatus = NO;
  
  if (_guessMode) {
    if (self->data[self->dataPos] == '<') {
      dataPos++;
      result = [self parseRoute:YES];
      parseWhiteSpaces(self, YES);      
      if ((result = [self parseAddrSpec:YES])) {
        parseWhiteSpaces(self, YES);
        if (self->data[self->dataPos] == '>') {
          self->dataPos++;
          returnStatus = YES;
        }
      }
      else if ((result = [self parseWord:YES])) {
        parseWhiteSpaces(self, YES);
        if (self->data[self->dataPos] == '>') {
          self->dataPos++;
          returnStatus = YES;
        }
      }
    }
    if (returnStatus) {
      returnValue = [NSNumber numberWithBool:YES];
    }
    else {
      returnValue = nil;
      dataPos     = keepPos;      
    }
  }
  else {
    returnValue = [NSMutableDictionary dictionaryWithCapacity:2];
    if (self->data[self->dataPos] == '<') {
      dataPos++;
      if ((result = [self parseRoute:NO])) {
        [returnValue setObject:result forKey:@"route"];
      }
      parseWhiteSpaces(self, NO);
      if ((result = [self parseAddrSpec:NO])) {
        parseWhiteSpaces(self, NO);
        if (self->data[self->dataPos] == '>') {
          self->dataPos++;
          [returnValue setObject:result forKey:@"address"];
          returnStatus = YES;
        }
      }
      else if ((result = [self parseWord:NO])) {
        parseWhiteSpaces(self, NO);
        if (self->data[self->dataPos] == '>') {
          self->dataPos++;
          [returnValue setObject:result forKey:@"address"];
          returnStatus = YES;
        }
      }
    }
    if (!returnStatus) {
      returnValue = nil;
      if (!(self->errorPos == -1))
        self->errorPos = self->dataPos;
      self->dataPos    = keepPos;
    }
  }
  return returnValue;
}

- (id)parseMailBox:(BOOL)_guessMode {
  id   returnValue  = nil;
  id   result       = nil;
  int  keepPos      = self->dataPos;
  BOOL returnStatus = NO;

  if (_guessMode) {
    if ((result = [self parseAddrSpec:YES])) {
      returnStatus = YES;
    }
    else {
      if ((result = [self parsePhrase:YES])) {
        parseWhiteSpaces(self, YES);
        if ((result = [self parseRouteAddr:YES])) {
          returnStatus = YES;
        }
      }
    }
    if (!returnStatus) {
      self->dataPos = keepPos;
      returnValue  = nil;
    }
    else {
      returnValue = [NSNumber numberWithBool:YES];
    }
  }
  else {
    if ((result = [self parseAddrSpec:NO])) {
      returnValue = [NGMailAddress mailAddressWithAddress:result
                                   displayName:nil
                                   route:nil];
      returnStatus = YES;
    }
    else if ((result = [self parseRouteAddr:NO])) {
      returnValue = [NGMailAddress mailAddressWithAddress:
                                   [result objectForKey:@"address"]
                                   displayName:nil
                                   route:nil];
      returnStatus = YES;
    }
    else {
      returnValue = [[[NGMailAddress alloc] init] autorelease];
      if ((result = [self parsePhrase:NO])) {
        [returnValue setDisplayName:result];
        parseWhiteSpaces(self, NO);        
        if ((result = [self parseRouteAddr:NO])) {
          [returnValue setAddress:[result objectForKey:@"address"]];
          [returnValue setRoute:[result objectForKey:@"route"]];
          returnStatus = YES;
        }
      }
    }
    if (!returnStatus) { /* try to read until eof or next ',' */
      self->dataPos = keepPos;
      
      if ((result = [self parseRouteAddr:NO])) {
        returnValue = [[[NGMailAddress alloc] init] autorelease];
        [returnValue setAddress:[result objectForKey:@"address"]];
        returnStatus = YES;
      }
    }
    if (!returnStatus) { /* try to read until eof or next ',' */
      self->dataPos = keepPos;
      
      if ((result = [self parseWord:NO])) {
        returnValue = [[[NGMailAddress alloc] init] autorelease];
        [returnValue setAddress:result];
        returnStatus = YES;
      }
    }
    if (!returnStatus) {
      if (!(self->errorPos == -1))
        self->errorPos = self->dataPos;

      self->dataPos = keepPos;
      returnValue  = nil;
    }
  }
  return returnValue;
}

- (id)parseGroup:(BOOL)_guessMode {
  id   returnValue  = nil;
  id   result       = nil;
  int  keepPos      = self->dataPos;
  BOOL returnStatus = NO;
  
  if (_guessMode) {
    if ((result = [self parsePhrase:YES])) {
      if (self->data[self->dataPos] == ':') { 
        self->dataPos++;
        parseWhiteSpaces(self, YES);                      
        if ((result = [self parseMailBox:YES])) {
          do {
            parseWhiteSpaces(self, YES);              
            result = nil;
            if (self->data[self->dataPos] == ',') {
              self->dataPos++;
              parseWhiteSpaces(self, YES);              
              result = [self parseMailBox:YES];
            }
          } while (result);
          parseWhiteSpaces(self, YES);                        
          if (self->data[self->dataPos] == ';') {
            self->dataPos++;
            returnStatus = YES;
          }
        }
      }
    }
    if (!returnStatus) {
      returnValue = nil;
      self->dataPos = keepPos;
    }
    else {
      returnValue = [NSNumber numberWithBool:YES];
    }
  }
  else {
    returnValue = [[[NGMailAddressList alloc] init] autorelease];
    if ((result = [self parsePhrase:NO])) {
      [returnValue setGroupName:result];
      if (self->data[self->dataPos] == ':') {
        self->dataPos++;
        parseWhiteSpaces(self, NO);
        if ((result = [self parseMailBox:NO])) {
          [returnValue addAddress:result];
          do {
            parseWhiteSpaces(self, NO);
            result = nil;            
            if (self->data[self->dataPos] == ',') {
              self->dataPos++;
              parseWhiteSpaces(self, NO);              
              result = [self parseMailBox:NO];
              if (result) {
                [returnValue addAddress:result];
              }
            }
          } while (result);
          parseWhiteSpaces(self, NO);                        
          if (self->data[self->dataPos] == ';') {
            self->dataPos++;
            returnStatus = YES;
          }
        }
      }
    }
    if (!returnStatus) {
      returnValue = nil;
      self->dataPos = keepPos;
    }
  }
  return returnValue;
}

- (id)parseAddress:(BOOL)_guessMode {
  id  returnValue = nil;
  int keepPos     = self->dataPos;
  
  if (_guessMode) {
    returnValue = [self parseMailBox:YES];
    if (!returnValue)
      returnValue = [self parseGroup:YES];
    if (!returnValue)
      self->dataPos = keepPos;
  }
  else {
    returnValue = [self parseMailBox:NO];
    if (!returnValue)
      returnValue = [self parseGroup:NO];
    if (!returnValue)
      self->dataPos = keepPos;
  }
  return returnValue;
}

- (NSArray *)parseAddressList {
  NGMailAddress  *address = nil;
  NSMutableArray *addrs   = nil;

  addrs = [NSMutableArray arrayWithCapacity:16];
  while (self->dataPos < self->maxLength) {
    address = [self parseAddress:NO];
    if (address)
      [addrs addObject:address];
    else
      break;
    
    if (self->dataPos < self->maxLength) {
      parseWhiteSpaces(self, NO);
      if (self->dataPos < self->maxLength) {
        if (self->data[self->dataPos] == ',') {
          self->dataPos++;
          if (self->dataPos < self->maxLength)
            parseWhiteSpaces(self, NO);
        }
      }
    }
  }
  return [[addrs copy] autorelease];
}

- (id)parse { 
  dataPos  = 0;
  errorPos = -1;
  return [self parseAddress:NO];
}

- (int)errorPosition {
  return self->errorPos;
}

/* description */

- (NSString *)description {
  return [NSString stringWithFormat:@"<%@[0x%08X]>",
                     NSStringFromClass([self class]),
                     (unsigned)self];
}

@end /* NGMailAddressParser */
