// $Id: NGMimeMessageGenerator.m,v 1.5 2003/01/27 18:22:59 jan Exp $

#include "NGMimeMessageGenerator.h"
#include "NGMimeMessage.h"
#include "NGMimeFileData.h"
#include "common.h"

/* Defaults

  Mail_Use_8bit_Encoding_For_Text[BOOL] --
    Use 8bit content-transfer-encoding for
    text messages
*/

static inline NSData *
_base64Encoding(NGMimeBodyGenerator *self,
                NSData *_data_,
                id<NGMimePart>_part,
                NGMutableHashMap *_addHeaders)
{
  NSString   *transEnc = nil;  
  const char *bytes    = NULL;
  unsigned   length    = 0;

  if ([_data_ isKindOfClass:[NGMimeFileData class]])
    return _data_;
  
  bytes  = [_data_ bytes];
  length = [_data_ length];

  while (length > 0) {
    if ((unsigned char)*bytes > 127) {
      break;
    }
    bytes++;
    length--;
  }
  if (length > 0) { // should be encoded
    NGMimeType *type;

    type = [_part contentType];
    
    if ([[type type] isEqualToString:@"text"] == YES) {
      BOOL use8bit;

      use8bit = [[NSUserDefaults standardUserDefaults]
                                 boolForKey:
                                   @"Mail_Use_8bit_Encoding_For_Text"];

      if (use8bit) {
        transEnc = @"8bit";
      }
      else {
        _data_   = [_data_ dataByEncodingQuotedPrintable];
        transEnc = @"quoted-printable";
      }
    }
    else {
      _data_   = [_data_ dataByEncodingBase64];
      transEnc = @"base64";

      if (type == nil) {
        [_addHeaders setObject:[NGMimeType mimeType:@"application"
                                           subType:@"octet-stream"]
                     forKey:@"content-type"];
      }
    }
  }
  else { // no encoding
    transEnc = @"7bit";
  }
  [_addHeaders setObject:transEnc forKey:@"content-transfer-encoding"];
  [_addHeaders setObject:[NSNumber numberWithInt:[_data_ length]]
                      forKey:@"content-length"];
  return _data_;
}

@implementation NGMimeMessageTextBodyGenerator

+ (int)version {
  return 2;
}
+ (void)initialize {
  NSAssert2([super version] == 2,
            @"invalid superclass (%@) version %i !",
            NSStringFromClass([self superclass]), [super version]);
}

- (NSData *)encodeData:(NSData *)_data
  forPart:(id<NGMimePart>)_part
  additionalHeaders:(NGMutableHashMap *)_addHeaders
{
  return _base64Encoding(self, _data, _part, _addHeaders);
}

@end

@implementation NGMimeMessageBodyGenerator

+ (int)version {
  return 2;
}
+ (void)initialize {
  NSAssert2([super version] == 2,
            @"invalid superclass (%@) version %i !",
            NSStringFromClass([self superclass]), [super version]);
}

- (NSData *)encodeData:(NSData *)_data
  forPart:(id<NGMimePart>)_part
  additionalHeaders:(NGMutableHashMap *)_addHeaders
{
  return _base64Encoding(self, _data, _part, _addHeaders);
}

@end

@implementation NGMimeMessageRfc822BodyGenerator

+ (int)version {
  return 2;
}
+ (void)initialize {
  NSAssert2([super version] == 2,
            @"invalid superclass (%@) version %i !",
            NSStringFromClass([self superclass]), [super version]);
}

- (id<NGMimePartGenerator>)generatorForPart:(id<NGMimePart>)_part {
  id gen;

  gen = [[[NGMimeMessageGenerator alloc] init] autorelease];

  [gen setUseMimeData:self->useMimeData];
  
  return gen;
}

@end /* NGMimeMessageRfc822BodyGenerator */

@implementation NGMimeMessageMultipartBodyGenerator

+ (int)version {
  return 2;
}
+ (void)initialize {
  NSAssert2([super version] == 2,
            @"invalid superclass (%@) version %i !",
            NSStringFromClass([self superclass]), [super version]);
}

- (id<NGMimePartGenerator>)multipartBodyGenerator:(NGMimeBodyGenerator *)_gen
  generatorForPart:(id<NGMimePart>)_part 
{
  id gen;

  gen = [[[NGMimeMessageGenerator alloc] init] autorelease];
  [gen setUseMimeData:self->useMimeData];
  return gen;
}

@end /* NGMimeMessageMultipartBodyGenerator */

@implementation NGMimeMessageGenerator

+ (int)version {
  return 2;
}
+ (void)initialize {
  NSAssert2([super version] == 2,
            @"invalid superclass (%@) version %i !",
            NSStringFromClass([self superclass]), [super version]);
}

/* header field specifics */

static int NGEncodeQuotedPrintableMime(const char *_src, unsigned _srcLen,
                                       char *_dest, unsigned _destLen) {
  /* decode also spaces*/
  unsigned cnt      = 0;
  unsigned destCnt  = 0;
  char     hexT[16] = {'0','1','2','3','4','5','6','7','8',
                       '9','A','B','C','D','E','F'};
  
  if (_srcLen > _destLen)
    return -1;
  
  for (cnt = 0; (cnt < _srcLen) && (destCnt < _destLen); cnt++) {
    char c = _src[cnt];

    if (((c > 47) && (c < 58)) ||
        ((c > 64) && (c < 91)) ||
        ((c > 96) && (c < 123))) { // no quoting
      _dest[destCnt++] = c;
    }
    else { // need to be quoted
      if (_destLen - destCnt > 2) {
        _dest[destCnt++] = '=';
        _dest[destCnt++] = hexT[(c >> 4) & 15];
        _dest[destCnt++] = hexT[c & 15];
      }
      else 
        break;
    }
  }
  if (cnt < _srcLen)
    return -1;
  return destCnt;
}

- (NSData *)generateDataForHeaderField:(NSString *)_headerField
  value:(id)_value
{
  NSData *data = nil;

  data = [super generateDataForHeaderField:_headerField
                value:_value];
  {
    const char   *bytes  = NULL;
    unsigned int length  = 0;
    unsigned int desLen  = 0;
    char         *des    = NULL;
    unsigned int cnt     = 0;
    BOOL         doEnc   = NO;
    NSString     *str;

#if APPLE_Foundation_LIBRARY || NeXT_Foundation_LIBRARY
    str = [[NSString alloc] initWithData:data
                            encoding:NSISOLatin1StringEncoding];
#else
    str = [[NSString alloc] initWithData:data
                            encoding:NSISOLatin9StringEncoding];
#endif
    str = [str autorelease];
    
    bytes  = [str cString];
    length = [str cStringLength];

    while (cnt < length) {
      if ((unsigned char)bytes[cnt] > 127) {
        doEnc = YES;
        break;
      }
      cnt++;
    }
    if (doEnc == YES) {
      char        iso[]     = "=?iso-8859-15?q?";
      unsigned    isoLen    = 16;
      char        isoEnd[]  = "?=";
      unsigned    isoEndLen = 2;
      
      desLen = length * 3 + 20;

      des    = NGMallocAtomic(sizeof(char) * desLen);

      //      memcpy(des, bytes, cnt);
      memcpy(des, iso, isoLen);
      if ((desLen = NGEncodeQuotedPrintableMime(bytes, length,
                                                des + isoLen,
                                                desLen - isoLen)) != -1) {
        memcpy(des + isoLen + desLen, isoEnd, isoEndLen);
        
        data = [NSData dataWithBytesNoCopy:des
                         length:isoLen +desLen + isoEndLen];
      }
      else {
        NSLog(@"WARNING: An error occour during quoted-printable decoding");
        NGFree(des);
      }
    }
  }
  return data;
}


/*
  content-transfer-encoding
*/

- (id<NGMimeBodyGenerator>)defaultBodyGenerator {
  id gen;

  gen  = [[NGMimeMessageBodyGenerator allocWithZone:[self zone]] init];
  [gen setUseMimeData:self->useMimeData];
  return gen;
}

- (id<NGMimeBodyGenerator>)generatorForBodyOfPart:(id<NGMimePart>)_part {
  id<NGMimeBodyGenerator> bodyGen      = nil;
  NGMimeType              *contentType = nil;
  NSString                *type        = nil;
  
  if (self->delegateRespondsTo.generatorGeneratorForBodyOfPart)
    bodyGen = [self->delegate mimePartGenerator:self
                              generatorForBodyOfPart:self->part];

  if (bodyGen)
    return bodyGen;

  contentType = [_part contentType];
  if (contentType == nil) {
    contentType = [self defaultContentTypeForPart:_part];
  }
  if (contentType == nil) {
    NSLog(@"WARNING: no content-type");
    return nil;
  }
  type = [contentType type];
  if ([type isEqualToString:NGMimeTypeMultipart]) {
    bodyGen = [[[NGMimeMessageMultipartBodyGenerator alloc] init] autorelease];
  }
  else if ([type isEqualToString:NGMimeTypeText]) {
    bodyGen = [[[NGMimeMessageTextBodyGenerator alloc] init] autorelease];
  }
  else if (([type isEqualToString:NGMimeTypeMessage]) &&
           [[contentType subType] isEqualToString:@"rfc822"]) {
    bodyGen = [[[NGMimeMessageRfc822BodyGenerator alloc] init] autorelease];
  }
  [(id)bodyGen setUseMimeData:self->useMimeData];
  return bodyGen;
}

@end /* NGMimeMessageGenerator */
