// $Id: NGMimeHeaderFieldGenerator.m,v 1.9 2003/01/20 10:46:07 helge Exp $

#import "NGMimeHeaderFieldGenerator.h"
#import "NGMimeHeaderFields.h"
#import <NGMail/NGMailAddressParser.h>
#import <NGMime/NGMimePartParser.h>
#import "common.h"

@interface NSObject(UsedProtocols)
- (NSString *)displayName; // hh: where is that implemented ?
@end

@implementation NGMimeHeaderFieldGenerator

+ (int)version {
  return 2;
}

+ (id)headerFieldGenerator {
  return AUTORELEASE([[self alloc] init]);
}

- (NSData *)generateDataForHeaderFieldNamed:(NSString *)_headerField
  value:(id)_value
{
  [self subclassResponsibility:_cmd];
  return nil;
}

@end /* NGMimeHeaderFieldGenerator */

@implementation NGMimeRFC822DateHeaderFieldGenerator

+ (int)version {
  return 2;
}

- (NSData *)generateDataForHeaderFieldNamed:(NSString *)_headerField
  value:(id)_value
{
  NSCalendarDate *date       = nil;
  NSString       *dateString = nil;
  
  if ([_value isKindOfClass:[NSString class]]) {
    return [_value dataUsingEncoding:NSUTF8StringEncoding];
  }
  date = _value;

  if (date == nil)
    return [NSData data];
  
  if ([date respondsToSelector:@selector(descriptionWithCalendarFormat:)]) {
    dateString = [date descriptionWithCalendarFormat:
                         @" %a, %d %b %Y %H:%M:%S %z"];
  }
  else
    dateString = [date stringValue];
  
  return [dateString dataUsingEncoding:NSUTF8StringEncoding];
}

@end /* NGMimeRFC822DateHeaderFieldGenerator */

@implementation NGMimeAddressHeaderFieldGenerator

+ (int)version {
  return 2;
}

static int NGEncodeQuotedPrintableMime
(const unsigned char *_src, unsigned _srcLen,
 unsigned char *_dest, unsigned _destLen) {
  /* decode also spaces*/
  unsigned cnt      = 0;
  unsigned destCnt  = 0;
  unsigned char
    hexT[16] = {'0','1','2','3','4','5','6','7','8',
                '9','A','B','C','D','E','F'};
  
  if (_srcLen > _destLen)
    return -1;
  
  for (cnt = 0; (cnt < _srcLen) && (destCnt < _destLen); cnt++) {
    register unsigned char c = _src[cnt];
    
    if (((c > 47) && (c < 58)) ||
        ((c > 64) && (c < 91)) ||
        ((c > 96) && (c < 123))) {
      // no quoting
      _dest[destCnt] = c;
      destCnt++;
    }
    else { // need to be quoted
      if (_destLen - destCnt > 2) {
        _dest[destCnt] = '='; destCnt++;
        _dest[destCnt] = hexT[(c >> 4) & 15]; destCnt++;
        _dest[destCnt] = hexT[c & 15]; destCnt++;
      }
      else 
        break;
    }
  }
  if (cnt < _srcLen)
    return -1;
  return destCnt;
}

- (NSData *)generateDataForHeaderFieldNamed:(NSString *)_headerField
  value:(id)_value
{
  NGMailAddressParser *parser;
  NSMutableString     *result;
  NSData              *data;
  id                  obj;
  NSEnumerator        *enumerator;
  
  parser = ([_value isKindOfClass:[NSString class]])
    ? [NGMailAddressParser mailAddressParserWithString:_value]
    : [NGMailAddressParser mailAddressParserWithData:_value];
  
  enumerator = [[parser parseAddressList] objectEnumerator];
  result     = [[NSMutableString alloc] initWithCapacity:128];
  
  while ((obj = [enumerator nextObject])) {
    NSString   *tmp;
    char       *buffer;
    unsigned   bufLen, cnt;
    BOOL       doEnc;

    static int UseLFSeperatedAddressEntries = -1;

    if (UseLFSeperatedAddressEntries == -1) {
      id o;

      o = [[NSUserDefaults standardUserDefaults]
                           objectForKey:@"UseLFSeperatedAddressEntries"];

      if (o) {
        UseLFSeperatedAddressEntries = [o boolValue]?1:0;
      }
      else {
        UseLFSeperatedAddressEntries = 1;
      }
        
    }

    if ([result length]) {
      if (UseLFSeperatedAddressEntries == 1)
        [result appendString:@",\n   "];
      else
        [result appendString:@", "];
    }
    
    tmp    = [obj displayName];
    bufLen = [tmp cStringLength];
    
    buffer = calloc(bufLen + 10, sizeof(char));
    [tmp getCString:buffer];
    
    cnt   = 0;
    doEnc = NO;
    
    while (cnt < bufLen) {
      if ((unsigned char)buffer[cnt] > 127) {
        doEnc = YES;
        break;
      }
      cnt++;
    }
    
    if (doEnc == YES) {
      char        iso[]     = "=?iso-8859-15?q?";
      unsigned    isoLen    = 16;
      char        isoEnd[]  = "?=";
      unsigned    isoEndLen = 2;
      unsigned    desLen;
      char        *des;
      
      free(buffer);
      {
        NSData *data;

#if APPLE_Foundation_LIBRARY || NeXT_Foundation_LIBRARY
        data = [tmp dataUsingEncoding:NSISOLatin1StringEncoding];
#else
        data = [tmp dataUsingEncoding:NSISOLatin9StringEncoding];
#endif

        bufLen  = [data length];
        buffer =  malloc(bufLen+1);
        [data getBytes:buffer];  buffer[bufLen] = '\0';
      }
          
      desLen = bufLen * 3 + 20;
      des    = calloc(desLen + 10, sizeof(char));
      
      memcpy(des, buffer, cnt);
      memcpy(des + cnt, iso, isoLen);
      if ((desLen =
           NGEncodeQuotedPrintableMime(buffer + cnt, bufLen - cnt,
                                       des + cnt + isoLen,
                                       desLen - cnt - isoLen)) != -1) {
        memcpy(des + cnt + isoLen + desLen, isoEnd, isoEndLen);
        tmp = [NSString stringWithCString:des
                        length:(cnt + isoLen + desLen + isoEndLen)];
      }
      else {
        NSLog(@"WARNING: An error occour during quoted-printable decoding");
      }
      if (des) free(des);
    }
    if (buffer) free(buffer); buffer = NULL;

    if ([tmp length]) {
      [result appendString:@"\""];
      [result appendString:tmp];
      [result appendString:@"\""];
      if ((tmp = [(NSHost *)obj address])) {
        [result appendString:@" <"];
        [result appendString:tmp];
        [result appendString:@">"];
      }
    }
    else if ((tmp = [(NSHost *)obj address])) {
      [result appendString:tmp];
    }
  }
  
#if APPLE_Foundation_LIBRARY || NeXT_Foundation_LIBRARY
  data = [result dataUsingEncoding:NSISOLatin1StringEncoding];
#else
  data = [result dataUsingEncoding:NSISOLatin9StringEncoding];
#endif
  RELEASE(result);
  
  return data;
}

@end /* NGMimeAddressHeaderFieldGenerator */

@implementation NGMimeContentTypeHeaderFieldGenerator

+ (int)version {
  return 2;
}

- (NSData *)generateDataForHeaderFieldNamed:(NSString *)_headerField
  value:(id)_value
{
  NGMimeType    *type = nil; // only one content-type field
  NSString      *tmp  = nil;
  NSMutableData *data = nil;
  unsigned char *ctmp = NULL;
  unsigned      len   = 0;
  
  type = _value;

  if (type == nil) {
    NSLog(@"WARNING(%s): empty content type field", __PRETTY_FUNCTION__);
    return [NSData dataWithBytes:"application/octet" length:17];
  }
  if ([_value isKindOfClass:[NSString class]]) {
    return [_value dataUsingEncoding:NSUTF8StringEncoding];
  }
  
  if (![type isKindOfClass:[NGMimeType class]]) {
    NSLog(@"WARNING(%s): invalid MIME type value (%@) !", __PRETTY_FUNCTION__,
          type);
    return [NSData dataWithBytes:"application/octet" length:17];
  }
  
  data = [NSMutableData dataWithCapacity:64];
  
  tmp = [type type];
  NSAssert(tmp, @"type should not be nil");
  len  = [tmp length];
  ctmp = malloc(len + 1);
  [tmp getCString:ctmp]; ctmp[len] = '\0';
  [data appendBytes:ctmp length:len];
  free(ctmp);
  
  [data appendBytes:"//" length:1];
  
  tmp = [type subType];
  if (tmp != nil) {
    len  = [tmp length];
    ctmp = malloc(len + 1);
    [tmp getCString:ctmp]; ctmp[len] = '\0';
    [data appendBytes:ctmp length:len];
    free(ctmp);
  }
  else
    [data appendBytes:"*" length:1];

  {  // parameters
    NSEnumerator *enumerator = [type parameterNames];
    NSString     *name       = nil;
    NSString     *value      = nil;
    
    while ((name = [enumerator nextObject])) {
      value = [type valueOfParameter:name];
      if (![value isKindOfClass:[NSString class]]) {
        NSLog(@"ERROR[%s]: parameter should be a NSString headerField: %@ "
              @"value %@", __PRETTY_FUNCTION__, _headerField, _value);
        continue;
      }
      [data appendBytes:"; " length:2];
      
      len  = [name cStringLength];
      ctmp = malloc(len + 1);
      [name getCString:ctmp]; ctmp[len] = '\0';
      [data appendBytes:ctmp length:len];
      free(ctmp);

      /*
        this confuses GroupWise: "= \"" (a space)
      */
      [data appendBytes:"=\"" length:2];

      /* check for encoding */
      {
        int  cnt;
        BOOL doEnc;
        
        len  = [value cStringLength];
        ctmp = malloc(len + 1);
        [value getCString:ctmp]; ctmp[len] = '\0';
        cnt  = 0;
        doEnc = NO;
        while (cnt < len) {
          if ((unsigned char)ctmp[cnt] > 127) {
            doEnc = YES;
            break;
          }
          cnt++;
        }
        if (doEnc == YES) {
          char        iso[]     = "=?iso-8859-15?q?";
          unsigned    isoLen    = 16;
          char        isoEnd[]  = "?=";
          unsigned    isoEndLen = 2;
          unsigned    desLen;
          char        *des;

          free(ctmp);
          {
            NSData *data;

#if APPLE_Foundation_LIBRARY || NeXT_Foundation_LIBRARY
            data = [value dataUsingEncoding:NSISOLatin1StringEncoding];
#else
            data = [value dataUsingEncoding:NSISOLatin9StringEncoding];
#endif

            len  = [data length];
            ctmp =  malloc(len+1);
            [data getBytes:ctmp];  ctmp[len] = '\0';
          }
          
          desLen = len * 3 + 20;
          des    = calloc(desLen + 10, sizeof(char));
      
          memcpy(des, ctmp, cnt);
          memcpy(des + cnt, iso, isoLen);
          if ((desLen =
               NGEncodeQuotedPrintableMime(ctmp + cnt, len - cnt,
                                           des + cnt + isoLen,
                                           desLen - cnt - isoLen)) != -1) {
            memcpy(des + cnt + isoLen + desLen, isoEnd, isoEndLen);
            [data appendBytes:des length:(cnt + isoLen + desLen + isoEndLen)];
          }
          else {
            NSLog(@"WARNING: An error occour during quoted-printable decoding");
          }
          if (des) free(des);
        }
        else {
          [data appendBytes:ctmp length:len];
        }
          free(ctmp);
      }
      [data appendBytes:"\"" length:1];      
    }
  }
  return data;
}

@end /* NGMimeContentTypeHeaderFieldGenerator */

@implementation NGMimeContentLengthHeaderFieldGenerator

+ (int)version {
  return 2;
}

- (NSData *)generateDataForHeaderFieldNamed:(NSString *)_headerField
  value:(id)_value
{
  if (_value == nil) {
    NSLog(@"WARNING(%s): empty content length field", __PRETTY_FUNCTION__);
    return [NSData data];
  }
  
  return [[_value stringValue] dataUsingEncoding:NSUTF8StringEncoding];
}
  
@end /* NGMimeContentLengthHeaderFieldGenerator */

@implementation NGMimeContentDispositionHeaderFieldGenerator

+ (int)version {
  return 2;
}

- (NSData *)generateDataForHeaderFieldNamed:(NSString *)_headerField
  value:(id)_value
{
  NGMimeContentDispositionHeaderField *field = nil;
  NSString      *tmp  = nil;
  NSMutableData *data = nil;

  field = _value;
  if (field == nil) {
    NSLog(@"WARNING(%s): Content-Disposition field is empty",
          __PRETTY_FUNCTION__);
    return [NSData data];
  }
  if ([_value isKindOfClass:[NSString class]]) {
    return [_value dataUsingEncoding:NSUTF8StringEncoding];
  }
  data = [NSMutableData dataWithCapacity:64];
  tmp = [field type];
  [data appendBytes:[tmp cString] length:[tmp length]];
  tmp = [field filename];
  if (tmp != nil) {
    [data appendBytes:"; " length:2];
    [data appendBytes:"filename=\"" length:10];
    {
      char *ctmp;
      int  cnt, len;
      BOOL doEnc;
        
      len  = [tmp cStringLength];
      ctmp = malloc(len + 1);
      [tmp getCString:ctmp]; ctmp[len] = '\0';
      cnt  = 0;
      doEnc = NO;
      while (cnt < len) {
        if ((unsigned char)ctmp[cnt] > 127) {
          doEnc = YES;
          break;
        }
        cnt++;
      }
      if (doEnc == YES) {
        char        iso[]     = "=?iso-8859-15?q?";
        unsigned    isoLen    = 16;
        char        isoEnd[]  = "?=";
        unsigned    isoEndLen = 2;
        unsigned    desLen;
        char        *des;
      
        free(ctmp);
        {
          NSData *data;

#if APPLE_Foundation_LIBRARY || NeXT_Foundation_LIBRARY
          data = [tmp dataUsingEncoding:NSISOLatin1StringEncoding];
#else
          data = [tmp dataUsingEncoding:NSISOLatin9StringEncoding];
#endif

          len  = [data length];
          ctmp =  malloc(len+1);
          [data getBytes:ctmp];  ctmp[len] = '\0';
        }
          
        desLen = len * 3 + 20;
        des    = calloc(desLen + 10, sizeof(char));
      
        memcpy(des, ctmp, cnt);
        memcpy(des + cnt, iso, isoLen);
        if ((desLen =
             NGEncodeQuotedPrintableMime(ctmp + cnt, len - cnt,
                                         des + cnt + isoLen,
                                         desLen - cnt - isoLen)) != -1) {
          memcpy(des + cnt + isoLen + desLen, isoEnd, isoEndLen);
          [data appendBytes:des length:(cnt + isoLen + desLen + isoEndLen)];
        }
        else {
          NSLog(@"WARNING: An error occour during quoted-printable decoding");
        }
        if (des) free(des);
      }
      else {
        [data appendBytes:ctmp length:len];
      }
    }
      //      [data appendBytes:[tmp cString] length:[tmp length]];
      [data appendBytes:"\"" length:1];
  }
  return data;
}
  
@end /* NGMimeContentDispositionHeaderFieldGenerator */

@implementation NGMimeStringHeaderFieldGenerator

+ (int)version {
  return 2;
}

- (NSData *)generateDataForHeaderFieldNamed:(NSString *)_headerField
  value:(id)_value
{
  if (_value == nil)
    return [NSData data];

  if ([_value isKindOfClass:[NSData class]])
    return _value;

  if (![_value isKindOfClass:[NSString class]])
    _value = [_value stringValue];

#if APPLE_Foundation_LIBRARY || NeXT_Foundation_LIBRARY
  return [_value dataUsingEncoding:NSISOLatin1StringEncoding];
#else
  return [_value dataUsingEncoding:NSISOLatin9StringEncoding];
#endif
}

@end /* NGMimeStringHeaderFieldGenerator */

@implementation NGMimeHeaderFieldGeneratorSet

+ (int)version {
  return 2;
}

static NGMimeHeaderFieldGeneratorSet *rfc822Set = nil;

+ (id)headerFieldGenerator {
  return [[[self alloc] init] autorelease];
}

+ (id)headerFieldGeneratorSet {
  return [[[self alloc] init] autorelease];
}

+ (id)defaultRfc822HeaderFieldGeneratorSet {
  static NGMimeHeaderNames *Fields = NULL;

  if (!Fields)
    Fields = (NGMimeHeaderNames *)[NGMimePartParser headerFieldNames];
                    
  if (rfc822Set == nil) {
    rfc822Set = [[self alloc] init];
    
    [rfc822Set setGenerator:
                 [NGMimeContentDispositionHeaderFieldGenerator
                    headerFieldGenerator]
               forField:Fields->contentDisposition];
    [rfc822Set setGenerator:
                 [NGMimeContentLengthHeaderFieldGenerator headerFieldGenerator]
               forField:Fields->contentLength];
    [rfc822Set setGenerator:
                 [NGMimeContentTypeHeaderFieldGenerator headerFieldGenerator]
               forField:Fields->contentType];
    [rfc822Set setGenerator:
                 [NGMimeRFC822DateHeaderFieldGenerator headerFieldGenerator]
               forField:Fields->date];

    [rfc822Set setGenerator:
                 [NGMimeAddressHeaderFieldGenerator headerFieldGenerator]
               forField:@"resent-from"];
    [rfc822Set setGenerator:
                 [NGMimeAddressHeaderFieldGenerator headerFieldGenerator]
               forField:@"resent-to"];
    [rfc822Set setGenerator:
                 [NGMimeAddressHeaderFieldGenerator headerFieldGenerator]
               forField:Fields->to];
    [rfc822Set setGenerator:
                 [NGMimeAddressHeaderFieldGenerator headerFieldGenerator]
               forField:Fields->cc];
    [rfc822Set setGenerator:
                 [NGMimeAddressHeaderFieldGenerator headerFieldGenerator]
               forField:@"bcc"];
    [rfc822Set setGenerator:
                 [NGMimeAddressHeaderFieldGenerator headerFieldGenerator]
               forField:Fields->from];

    [rfc822Set setDefaultGenerator:
                 [NGMimeStringHeaderFieldGenerator headerFieldGenerator]];
  }
  return rfc822Set;
}

- (id)init {
  return [self initWithDefaultGenerator:nil];
}

- (id)initWithDefaultGenerator:(id<NGMimeHeaderFieldGenerator>)_gen {
  if ((self = [super init])) {
    self->fieldNameToGenerate =
      [[NSMutableDictionary allocWithZone:[self zone]]
                            initWithCapacity:16];
    self->defaultGenerator = [_gen retain];
  }
  return self;
}

- (void)dealloc {
  [self->fieldNameToGenerate release];
  [self->defaultGenerator    release];
  [super dealloc];
}

/* accessors */

- (void)setGenerator:(id<NGMimeHeaderFieldGenerator>)_gen
  forField:(NSString *)_name
{
  [self->fieldNameToGenerate setObject:_gen forKey:_name];
}

- (void)setDefaultGenerator:(id<NGMimeHeaderFieldGenerator>)_gen {
  ASSIGN(self->defaultGenerator, _gen);
}

- (id<NGMimeHeaderFieldGenerator>)_gen {
  return self->defaultGenerator;
}

/* operation */

- (NSData *)generateDataForHeaderFieldNamed:(NSString *)_headerField
  value:(id)_value
{
  NGMimeHeaderFieldGenerator *gen = nil;

  if (!(gen = [self->fieldNameToGenerate objectForKey:_headerField]))
    gen = self->defaultGenerator;

  if (gen == nil) {
    NSLog(@"WARNING(%s): no defaultGenerator is set", __PRETTY_FUNCTION__);
    return [NSData data];
  }
  return [gen generateDataForHeaderFieldNamed:_headerField
              value:_value];
}

@end /* NGMimeHeaderFieldGeneratorSet */

