// $Id: NGMimeHeaderFieldParser.h,v 1.2 2002/06/07 15:11:31 jan41 Exp $

#ifndef __NGMime_NGHeaderFieldParser_H__
#define __NGMime_NGHeaderFieldParser_H__

#import <Foundation/NSObject.h>

@class NSData, NSString, NSMutableDictionary;

@protocol NGMimeHeaderFieldParser < NSObject >

- (id)parseValue:(NSString *)_data ofHeaderField:(NSString *)_field;

@end

@interface NGMimeHeaderFieldParser : NSObject < NGMimeHeaderFieldParser >

- (NSString *)removeCommentsFromValue:(NSString *)_rawValue;
- (id)parseValue:(id)_data ofHeaderField:(NSString *)_field; // abstract

@end

@interface NGMimeContentTypeHeaderFieldParser : NGMimeHeaderFieldParser
@end

@interface NGMimeRFC822DateHeaderFieldParser : NGMimeHeaderFieldParser
@end

@interface NGMimeContentLengthHeaderFieldParser : NGMimeHeaderFieldParser
@end

/*
  Content-Disposition headers have the form:

    disposition := "Content-Disposition" ":"
                   disposition-type
                   *(";" disposition-parm)

    disposition-type := "inline"
                      / "attachment"
                      / extension-token
                      ; values are not case-sensitive

    disposition-parm := filename-parm / parameter
    filename-parm := "filename" "=" value;

  Content-Disposition values may not contain comments !
*/
@interface NGMimeContentDispositionHeaderFieldParser : NGMimeHeaderFieldParser
@end

/*
  This strips spaces at the beginning and the end of the value, then it removes
  all comments
*/
@interface NGMimeStringHeaderFieldParser : NGMimeHeaderFieldParser
{
@protected
  BOOL removeComments; // default=YES
}

- (id)initWithRemoveComments:(BOOL)_flag;
- (id)init;

- (id)parseValue:(NSString *)_data ofHeaderField:(NSString *)_field;

@end

/*
  This stores a mapping between header field parsers and field names.
*/
@interface NGMimeHeaderFieldParserSet : NSObject <NGMimeHeaderFieldParser,NSCopying>
{
@protected
  NSMutableDictionary         *fieldNameToParser;
  id<NGMimeHeaderFieldParser> defaultParser;
}

+ (id)headerFieldParserSet;
+ (id)defaultRfc822HeaderFieldParserSet;
- (id)init;
- (id)initWithDefaultParser:(id<NGMimeHeaderFieldParser>)_parser;
- (id)initWithParseSet:(NGMimeHeaderFieldParserSet *)_set;

// accessors

- (void)setParser:(id<NGMimeHeaderFieldParser>)_parser
  forField:(NSString *)_name;

- (void)setDefaultParser:(id<NGMimeHeaderFieldParser>)_parser;
- (id<NGMimeHeaderFieldParser>)defaultParser;

// operation

- (id)parseValue:(NSString *)_data ofHeaderField:(NSString *)_field;

@end

#endif /* __NGMime_NGHeaderFieldParser_H__ */
