// $Id: NGMimeHeaderFieldParser.m,v 1.8 2003/01/20 10:46:02 helge Exp $

#include "NGMimeHeaderFieldParser.h"
#include "NGMimeHeaderFields.h"
#include "NGMimeUtilities.h"
#include "common.h"
#include <NGMime/NGMimePartParser.h>

@implementation NGMimeHeaderFieldParser

+ (int)version {
  return 2;
}

- (NSString *)removeCommentsFromValue:(NSString *)_rawValue {
  const char   *bytes;
  const char   *tmp;
  unsigned int len, cnt;

  bytes = [_rawValue cString];
  tmp   = bytes;
  len   = [_rawValue cStringLength];
  cnt   = 0;

  if (_rawValue == NULL) return nil;

  while ((cnt < len) && (tmp[cnt] != '(')) cnt++;
  
  if (cnt < len) {
    NSMutableString *result;
    int             commentNesting = 0;
    int             begin          = 0;

    result = [NSMutableString stringWithCapacity:len];    
    
    for (cnt = 0; cnt < len; cnt++) {
      if (commentNesting == 0) {
        if (isRfc822_QUOTE(bytes[cnt])) {
          cnt++;
          while ((cnt < len) && !isRfc822_QUOTE(bytes[cnt]))
            cnt++;
        }
        else if (bytes[cnt] == '(') {
          if ((cnt - begin) > 0)
            [result appendString:[NSString stringWithCString:&(bytes[begin])
                                           length:(cnt - begin)]];
          commentNesting++;
        }
      }
      else {
        if (bytes[cnt] == ')') {
          commentNesting--;
          if (commentNesting == 0)
            begin = (cnt + 1);
        }
        else if (bytes[cnt] == '(')
          commentNesting++;
      }
    }
    if ((cnt - begin) > 0) {
      [result appendString:[NSString stringWithCString:&(bytes[begin])
                                   length:(cnt - begin)]];
    }
    return result;
  }
  else
    return _rawValue;
}

- (NSData *)quotedPrintableDecoding:(NSData *)_value {
  return _value;
}

- (id)parseValue:(NSString *)_data ofHeaderField:(NSString *)_field { // abstract
  [self subclassResponsibility:_cmd];
  return nil;
}

@end

@implementation NGMimeRFC822DateHeaderFieldParser

+ (int)version {
  return 2;
}

static NSString *dateFormats[] = {
  /*
    short-weekday, day short-month year hour:minute:second timezoneoffset
    eg: Mon, 01 Mar 1999 13:13:13 +0000
  */
  @"%a, %d %b %Y %H:%M:%S %z",

  /*
    short-weekday, day short-month year hour:minute:second timezonename
    eg: Mon, 01 Mar 1999 13:13:13 GMT
  */
  @"%a, %d %b %Y %H:%M:%S %Z",

  /*
    short-weekday, day short-month year hour:minute:second timezonename
    eg: Mon, 01 Mar 1999 13:13 GMT
  */
  @"%a, %d %b %Y %H:%M %z",

  /*
    short-weekday, day short-month year hour:minute:second timezonename
    eg: Mon, 01 Mar 1999 13:13 (GMT)
  */
  @"%a, %d %b %Y %H:%M (%z)",

  /*
    short-weekday, day short-month year hour:minute:second
    eg: Mon, 01 Mar 1999 13:13:13
  */
  @"%a, %d %b %Y %H:%M:%S",

  /*
    day short-month year hour:minute:second timezoneoffset
    eg: 01 Oct 1999 18:20:12 +0200
  */
  @"%d %b %Y %H:%M:%S %z",
  
  /*
    day short-month year hour:minute:second timezonename
    eg: 01 Oct 1999 18:20:12 EST
  */
  @"%d %b %Y %H:%M:%S %Z",
  
  /*
    day short-month year hour:minute:second (timezoneoffset)
    eg: 30 Sep 1999 21:00:05  (+0200)
  */
  @"%d %b %Y %H:%M:%S  (%z)",

  /*
    day short-month year hour:minute:second (timezonename)
    eg: 30 Sep 1999 21:00:05  (MEST)
  */
  @"%d %b %Y %H:%M:%S  (%Z)",
  /*
    day short-month year hour:minute:second (timezonename)
    eg: 30 Sep 1999 21:00:05  (MEST)
  */
  @"%d %b %Y %H:%M:%S  (%Z)",

  /*
    short-weekday, day short-month year hour:minute:second timezoneoffset
    eg: Mon, 01 Mar 1999 13:13:13 +0000
  */
  @"%a %b %d %H:%M:%S %Y  (%Z)",

  /*
    eg: '16 Jun 2002 10:28 GMT'
  */
  @"%d %b %Y %H:%M %Z",

  /* terminate list */
  nil
};

- (id)parseValue:(NSString *)_data ofHeaderField:_field {
  NSCalendarDate *date       = nil;
  NSString       *dateString = nil;
  const char     *bytes      = NULL;
  unsigned       length      = 0;

  bytes  = [_data cString];
  length = [_data cStringLength];
  
  
  if (length == 0) {
    NSLog(@"WARNING(%s): empty value for header field %@ ..",
          __PRETTY_FUNCTION__, _field);
    return [NSCalendarDate date];
  }

  while (length > 0) {
    if (*bytes != ' ')
      break;
    bytes++;
    length--;
  }
  
  dateString =
    [[NSString alloc] initWithCString:bytes length:length];

  /* check various date formats */
  {
    int i;

    for (i = 0, date = nil; (date == nil) && (dateFormats[i] != nil); i++) {
      date = [NSCalendarDate dateWithString:dateString
                             calendarFormat:dateFormats[i]];
    }
  }
  
  RELEASE(dateString);
  
  return [date y2kDate];
}

@end /* NGMimeRFC822DateHeaderFieldParser */

@implementation NSCalendarDate(RFC822Dates)

+ (NSCalendarDate *)calendarDateWithRfc822DateString:(NSString *)_str {
  NSCalendarDate *date       = nil;
  NSString       *dateString = nil;
  
  dateString = [_str stringByTrimmingSpaces];
  if ([dateString length] == 0)
    return nil;
  
  /* check various date formats */
  {
    int i;

    for (i = 0, date = nil; (date == nil) && (dateFormats[i] != nil); i++) {
      date = [NSCalendarDate dateWithString:dateString
                             calendarFormat:dateFormats[i]];
    }
  }
  return [date y2kDate];
}

@end /* NSCalendarDate(RFC822Dates) */

@implementation NGMimeContentTypeHeaderFieldParser

+ (int)version {
  return 2;
}

- (id)parseValue:(NSString *)_data ofHeaderField:(NSString *)_field {
  unsigned      len   = 0;
  const char *src  = NULL;

  NSString   *typeString = nil;
  NGMimeType *type       = nil;

  _data = [self removeCommentsFromValue:_data];

  len   = [_data length];

  if (len == 0) {
    NSLog(@"WARNING(%s): empty value for header field %@ ..",
          __PRETTY_FUNCTION__, _field);
    return [NGMimeType mimeType:@"text/plain"];
  }

  src = [_data cString];
  len = [_data cStringLength];

  // strip leading spaces
  while (isRfc822_LWSP(*src) && (len > 0)) {
    src++;
    len--;
  }

  typeString = [[NSString alloc] initWithCString:src length:len];
  NSAssert(typeString, @"type string allocation failed ..");

  if ([typeString isEqualToString:@"text"])
    type = [NGMimeType mimeType:@"text/plain"];
  else {
    type = [NGMimeType mimeType:typeString];
    //NSAssert(type, @"mime type creation failed ..");
  }
  type = [type retain];

  [typeString release]; typeString = nil;

  return [type autorelease];
}

// description

- (NSString *)description {
  return [NSString stringWithFormat:
                     @"<MimeContentTypeHeaderFieldParser: object=0x%08X>",
                     (unsigned)self];
}

@end

@implementation NGMimeContentLengthHeaderFieldParser

+ (int)version {
  return 2;
}

- (id)parseValue:(NSString *)_data ofHeaderField:(NSString *)_field {
  const char *buf, *ptr;
    
  _data = [self removeCommentsFromValue:_data];
  buf   = [_data cString];
  ptr = buf;
  while (!isRfc822_DIGIT(*ptr) && (*ptr != '\0'))
    ptr++;

  if (isRfc822_DIGIT(*ptr))
    return [NSNumber numberWithUnsignedInt:atol(ptr)];
  else {
    NSLog(@"WARNING(%s): invalid content-length field value (value='%s')",
          __PRETTY_FUNCTION__, buf);
    return nil;
  }
}

// description

- (NSString *)description {
  return [NSString stringWithFormat:
                     @"<MimeContentLengthHeaderFieldParser: object=0x%08X>",
                     (unsigned)self];
}

@end

@implementation NGMimeContentDispositionHeaderFieldParser

+ (int)version {
  return 2;
}

- (id)parseValue:(NSString *)_data ofHeaderField:(NSString *)_field {
  unsigned   len;
  const char *src;

  src = [_data cString];
  len = [_data cStringLength];
  
  // strip leading spaces
  while (isRfc822_LWSP(*src) && (len > 0)) {
    src++;
    len--;
  }
  // strip trailing spaces
  while (len > 0) {
    if (isRfc822_LWSP(src[len - 1])) len--;
    else break;
  }

  if (len == 0) {
    NSLog(@"WARNING(%s): empty value for header field %@ ..",
          __PRETTY_FUNCTION__, _field);
    return nil;
  }

  {
    NSString *str = [[NSString alloc] initWithCString:src length:len];
    NGMimeContentDispositionHeaderField *result = nil;

    result = [[NGMimeContentDispositionHeaderField alloc] initWithString:str];
    RELEASE(str);
    
    return AUTORELEASE(result);
  }
}

@end

@implementation NGMimeStringHeaderFieldParser

+ (int)version {
  return 2;
}

- (id)initWithRemoveComments:(BOOL)_flag {
  if ((self = [super init])) {
    removeComments = _flag;
  }
  return self;
}
- (id)init {
  return [self initWithRemoveComments:YES];
}

// operation

- (id)parseValue:(NSString *)_data ofHeaderField:(NSString *)_field {
  NSString   *data;
  unsigned   len;
  const char *src;

  if (_data == nil)
    return nil;
  
  data = removeComments ? [self removeCommentsFromValue:_data] : _data;
  len  = [data length];

  if (len == 0) return @"";

  src = [data cString];
  len = [data cStringLength];
  
  // strip leading spaces
  while (isRfc822_LWSP(*src) && (len > 0)) {
    src++;
    len--;
  }
  return [NSString stringWithCString:src length:len];
}

// description

- (NSString *)description {
  return [NSString stringWithFormat:
                     @"<MimeStringHeaderFieldParser: id=0x%08X removesComments=%s>",
                     (unsigned)self,
                     removeComments ? "YES" : "NO"
                   ];
}

@end

@implementation NGMimeHeaderFieldParserSet

+ (int)version {
  return 2;
}

static NGMimeHeaderFieldParserSet *rfc822set = nil;

+ (id)headerFieldParserSet {
  return AUTORELEASE([[self alloc] init]);
}

+ (id)defaultRfc822HeaderFieldParserSet {
  if (rfc822set == nil) {
    id parser = nil;
    static NGMimeHeaderNames *Fields = NULL;

    if (!Fields)
      Fields = (NGMimeHeaderNames *)[NGMimePartParser headerFieldNames];
  
    
    
    rfc822set = [[self alloc] init];

    [rfc822set setParser:
                 (parser = [[NGMimeContentTypeHeaderFieldParser alloc] init])
               forField:Fields->contentType];
    [parser release]; parser = nil;
    
    [rfc822set setParser:
                 (parser = [[NGMimeContentLengthHeaderFieldParser alloc] init])
               forField:Fields->contentLength];
    [parser release]; parser = nil;
    
    [rfc822set setParser:
                 (parser = [[NGMimeStringHeaderFieldParser alloc]
                                 initWithRemoveComments:NO])
               forField:Fields->received];
    [parser release]; parser = nil;
    
    [rfc822set setParser:
                 (parser = [[NGMimeStringHeaderFieldParser alloc]
                                 initWithRemoveComments:NO])
               forField:Fields->subject];
    [parser release]; parser = nil;
    
    [rfc822set setParser:
                 (parser = [[NGMimeStringHeaderFieldParser alloc]
                                 initWithRemoveComments:NO])
               forField:@"x-face"];
    [parser release]; parser = nil;

    [rfc822set setParser:(parser = [NGMimeContentDispositionHeaderFieldParser new])
               forField:Fields->contentDisposition];
    [parser release]; parser = nil;
    [rfc822set setParser:(parser = [NGMimeRFC822DateHeaderFieldParser new])
               forField:Fields->date];
    [parser release]; parser = nil;
  }
  return rfc822set;
}

- (id)init {
  return [self initWithDefaultParser:
                 AUTORELEASE([[NGMimeStringHeaderFieldParser alloc] init])];
}
- (id)initWithDefaultParser:(id<NGMimeHeaderFieldParser>)_parser {
  if ((self = [super init])) {
    fieldNameToParser = [[NSMutableDictionary alloc] initWithCapacity:32];
    [self setDefaultParser:_parser];
  }
  return self;
}
- (id)initWithParseSet:(NGMimeHeaderFieldParserSet *)_set {
  if ((self = [self initWithDefaultParser:[_set defaultParser]])) {
    [self->fieldNameToParser addEntriesFromDictionary:_set->fieldNameToParser];
  }
  return self;
}

#if !LIB_FOUNDATION_BOEHM_GC
- (void)dealloc {
  RELEASE(self->fieldNameToParser); self->fieldNameToParser = nil;
  RELEASE(self->defaultParser);     self->defaultParser     = nil;
  [super dealloc];
}
#endif

// accessors

- (void)setParser:(id<NGMimeHeaderFieldParser>)_parser
  forField:(NSString *)_name {

  [self->fieldNameToParser setObject:_parser forKey:_name];
}

- (void)setDefaultParser:(id<NGMimeHeaderFieldParser>)_parser {
  ASSIGN(self->defaultParser, _parser);
}
- (id<NGMimeHeaderFieldParser>)defaultParser {
  return self->defaultParser;
}

// operation

- (id)parseValue:(NSString *)_data ofHeaderField:(NSString *)_field {
  id parser = [fieldNameToParser objectForKey:_field];
  if (parser == nil)
    parser = [self defaultParser];

  return [parser parseValue:_data ofHeaderField:_field];
}

// NSCopying

- (id)copyWithZone:(NSZone *)_zone {
  id copy = [[NGMimeHeaderFieldParserSet allocWithZone:_zone]
                initWithDefaultParser:[self defaultParser]];
  NSEnumerator *keys = [fieldNameToParser keyEnumerator];
  NSString *key;

  while ((key = [keys nextObject])) {
    id value = [fieldNameToParser objectForKey:key];

    [copy setParser:value forField:key];
  }
  return copy;
}

// description

- (NSString *)description {
  return [NSString stringWithFormat:
                     @"<HeaderFieldParserSet: id=0x%08X map=%@ default=%@>",
                     (unsigned)self, fieldNameToParser, [self defaultParser]
                   ];
}

@end
