/* 
   NGByteCountStream.m

   Copyright (C) 2002 SKYRIX Software AG, Helge Hess
   All rights reserved.

   Author: Helge Hess (helge.hess@skyrix.com)

   This file is part of the SKYRiX Core Framework

   Permission to use, copy, modify, and distribute this software and its
   documentation for any purpose and without fee is hereby granted, provided
   that the above copyright notice appear in all copies and that both that
   copyright notice and this permission notice appear in supporting
   documentation.

   We disclaim all warranties with regard to this software, including all
   implied warranties of merchantability and fitness, in no event shall
   we be liable for any special, indirect or consequential damages or any
   damages whatsoever resulting from loss of use, data or profits, whether in
   an action of contract, negligence or other tortious action, arising out of
   or in connection with the use or performance of this software.
*/
// $Id: NGByteCountStream.m,v 1.3 2003/01/07 11:36:47 helge Exp $

#include "common.h"
#include "NGByteCountStream.h"

@implementation NGByteCountStream

+ (id)byteCounterForStream:(id<NGStream>)_stream byte:(unsigned char)_byte {
  return [[[self alloc] initWithSource:_stream byte:_byte] autorelease];
}

- (id)initWithSource:(id<NGStream>)_source byte:(unsigned char)_byte {
  if ((self = [super initWithSource:_source])) {
    [self setByteToCount:_byte];
  }
  return self;
}
- (id)initWithSource:(id<NGStream>)_source {
  return [self initWithSource:_source byte:'\n'];
}

// accessors

- (void)setByteToCount:(unsigned char)_byte {
  if (_byte != byteToCount) {
    byteReadCount  = 0;
    byteWriteCount = 0;
    byteToCount    = _byte;
  }
}
- (unsigned char)byteToCount {
  return byteToCount;
}

- (unsigned)readCount {
  return byteReadCount;
}
- (unsigned)writeCount {
  return byteWriteCount;
}

- (unsigned)totalReadCount {
  return totalReadCount;
}
- (unsigned)totalWriteCount {
  return totalWriteCount;
}

// operations

- (void)resetCounters {
  totalReadCount  = 0;
  totalWriteCount = 0;
  byteReadCount   = 0;
  byteWriteCount  = 0;
}

// primitives

- (unsigned)readBytes:(void *)_buf count:(unsigned)_len {
  unsigned result;

  result = (readBytes != NULL)
    ? readBytes(source, _cmd, _buf, _len)
    : [source readBytes:_buf count:_len];

  totalReadCount += result;
  {
    register unsigned char *byteBuffer = _buf;

    for (_len = result - 1; _len >= 0; _len--, byteBuffer++) {
      if (*byteBuffer == byteToCount)
        byteReadCount++;
    }
  }
  return result;
}

- (unsigned)writeBytes:(const void *)_buf count:(unsigned)_len {
  unsigned result;

  result = (writeBytes != NULL)
    ? writeBytes(source, _cmd, _buf, _len)
    : [source writeBytes:_buf count:_len];

  totalWriteCount += result;
  {
    register unsigned char *byteBuffer = (unsigned char *)_buf;

    for (_len = result - 1; _len >= 0; _len--, byteBuffer++) {
      if (*byteBuffer == byteToCount)
        byteWriteCount++;
    }
  }
  return result;
}

@end
