/* 
   NGFilterStream.m

   Copyright (C) 2002 SKYRIX Software AG, Helge Hess
   All rights reserved.

   Author: Helge Hess (helge.hess@skyrix.com)

   This file is part of the SKYRiX Core Framework

   Permission to use, copy, modify, and distribute this software and its
   documentation for any purpose and without fee is hereby granted, provided
   that the above copyright notice appear in all copies and that both that
   copyright notice and this permission notice appear in supporting
   documentation.

   We disclaim all warranties with regard to this software, including all
   implied warranties of merchantability and fitness, in no event shall
   we be liable for any special, indirect or consequential damages or any
   damages whatsoever resulting from loss of use, data or profits, whether in
   an action of contract, negligence or other tortious action, arising out of
   or in connection with the use or performance of this software.
*/
// $Id: NGFilterStream.m,v 1.4 2003/01/07 11:36:47 helge Exp $

#include "common.h"
#include "NGFilterStream.h"

@implementation NGFilterStream

+ (id)filterWithInputSource:(id<NGInputStream>)_source {
  return [[[self alloc] initWithInputSource:_source] autorelease];
}
+ (id)filterWithOutputSource:(id<NGOutputStream>)_source {
  return [[[self alloc] initWithOutputSource:_source] autorelease];
}
+ (id)filterWithSource:(id<NGStream>)_source {
  return [[[self alloc] initWithSource:_source] autorelease];
}

- (id)init {
  return [self initWithSource:nil];
}

- (id)initWithSource:(id<NGStream>)_source {
  if ((self = [super init])) {
    self->source = [_source retain];

    if ([source isKindOfClass:[NSObject class]]) {
      self->readBytes  = (NGIOReadMethodType)
        [(NSObject *)self->source methodForSelector:@selector(readBytes:count:)];
      self->writeBytes = (NGIOWriteMethodType)
        [(NSObject *)self->source methodForSelector:@selector(writeBytes:count:)];
    }
  }
  return self;
}

- (id)initWithInputSource:(id<NGInputStream>)_source {
  if ((self = [super init])) {
    self->source = [_source retain];

    if ([source isKindOfClass:[NSObject class]]) {
      self->readBytes  = (NGIOReadMethodType)
        [(NSObject *)self->source methodForSelector:@selector(readBytes:count:)];
    }
  }
  return self;
}
- (id)initWithOutputSource:(id<NGOutputStream>)_source {
  if ((self = [super init])) {
    self->source = [_source retain];

    if ([source isKindOfClass:[NSObject class]]) {
      self->writeBytes  = (NGIOWriteMethodType)
        [(NSObject *)self->source methodForSelector:@selector(writeBytes:count:)];
    }
  }
  return self;
}

- (void)dealloc {
  [self->source release];
  self->readBytes  = NULL;
  self->writeBytes = NULL;
  [super dealloc];
}

/* accessors */

- (id<NGInputStream>)inputStream {
  return [self source];
}
- (id<NGOutputStream>)outputStream {
  return [self source];
}
- (id<NGStream>)source {
  return self->source;
}

/* primitives */

- (NSException *)lastException {
  return [self->source lastException];
}
- (void)resetLastException {
  [self->source resetLastException];
}
- (void)setLastException:(NSException *)_exception {
  [self->source setLastException:_exception];
}

- (BOOL)isOpen {
  return [self->source isOpen];
}

- (unsigned)readBytes:(void *)_buf count:(unsigned)_len {
  if (self->readBytes)
    return (unsigned)readBytes(self->source, _cmd, _buf, _len);
  else
    return [self->source readBytes:_buf count:_len];
}
- (unsigned)writeBytes:(const void *)_buf count:(unsigned)_len {
  if (self->writeBytes)
    return (unsigned)writeBytes(self->source, _cmd, _buf, _len);
  else
    return [self->source writeBytes:_buf count:_len];
}

- (BOOL)flush {
  return [self->source flush];
}
- (BOOL)close {
  return [((NGStream *)self->source) close];
}

- (NGStreamMode)mode {
  return [self->source mode];
}
- (BOOL)isRootStream {
  return NO;
}

// all other things are forward

- (void)forwardInvocation:(NSInvocation *)_invocation {
  if ([self->source respondsToSelector:[_invocation selector]]) {
    [_invocation setTarget:self->source];
    [_invocation invoke];
  }
  else
    [self doesNotRecognizeSelector:[_invocation selector]];
}

// description

- (NSString *)description {
  return [NSString stringWithFormat:
                     @"<%@[0x%08X] source=%@ mode=%@>",
                     NSStringFromClass([self class]), (unsigned)self,
                     self->source ? (id)self->source : (id)@"nil",
                     [self modeDescription]];
}

@end /* NGFilterStream */
