/* 
   NGInternetSocketDomain.m

   Copyright (C) 2002 SKYRIX Software AG, Helge Hess
   All rights reserved.

   Author: Helge Hess (helge.hess@skyrix.com)

   This file is part of the SKYRiX Core Framework

   Permission to use, copy, modify, and distribute this software and its
   documentation for any purpose and without fee is hereby granted, provided
   that the above copyright notice appear in all copies and that both that
   copyright notice and this permission notice appear in supporting
   documentation.

   We disclaim all warranties with regard to this software, including all
   implied warranties of merchantability and fitness, in no event shall
   we be liable for any special, indirect or consequential damages or any
   damages whatsoever resulting from loss of use, data or profits, whether in
   an action of contract, negligence or other tortious action, arising out of
   or in connection with the use or performance of this software.
*/
// $Id: NGInternetSocketDomain.m,v 1.4 2003/01/14 13:02:23 helge Exp $

#include "NGInternetSocketDomain.h"
#include "NGInternetSocketAddress.h"
#include "common.h"

#ifndef __MINGW32__
#  include <netinet/in.h>
#endif

@implementation NGInternetSocketDomain

static NGInternetSocketDomain *domain = nil;

+ (int)version {
  return 1;
}
+ (void)initialize {
  if (domain == nil) domain = [[NGInternetSocketDomain alloc] init];
}
+ (id)domain {
  return domain;
}

/* NGSocketDomain */

- (id<NGSocketAddress>)addressWithRepresentation:(void *)_data
  size:(unsigned int)_size
{
  NGInternetSocketAddress *address = nil;

  if ([self addressRepresentationSize] != _size) {
    NSLog(@"%@: invalid address size %i ..", NSStringFromSelector(_cmd), _size);
    return nil;
  }
  
  address = [[NGInternetSocketAddress allocWithZone:[self zone]]
                                      initWithDomain:self
                                      internalRepresentation:_data
                                      size:_size];
  return [address autorelease];
}

- (BOOL)prepareAddress:(id<NGSocketAddress>)_address
  forBindWithSocket:(id<NGSocket>)_socket
{
  // nothing to prepare
  return YES;
}
- (BOOL)cleanupAddress:(id<NGSocketAddress>)_address
  afterCloseOfSocket:(id<NGSocket>)_socket
{
  // nothing to cleanup
  return YES;
}

- (int)socketDomain {
  return AF_INET;
}

- (int)addressRepresentationSize {
  return sizeof(struct sockaddr_in);
}

- (int)protocol {
  return 0;
}

/* NSCopying */

- (id)copyWithZone:(NSZone *)_zone {
  /* domain objects are immutable, just retain on copy */
  return [self retain];
}

/* NSCoding */

- (void)encodeWithCoder:(NSCoder *)_encoder {
}
- (id)initWithCoder:(NSCoder *)_decoder {
  [self release]; self = nil;
  return [domain retain];
}

- (id)awakeAfterUsingCoder:(NSCoder *)_decoder {
  if (self != domain) {
    [self release]; self = nil;
    return [domain retain];
  }
  else
    return self;
}

/* description */

- (NSString *)description {
  return [NSString stringWithFormat:@"<InternetDomain[0x%08X]>", (unsigned)self];
}

@end /* NGInternetSocketDomain */
