/* 
   NGLockingStream.m

   Copyright (C) 2002 SKYRIX Software AG, Helge Hess
   All rights reserved.

   Author: Helge Hess (helge.hess@skyrix.com)

   This file is part of the SKYRiX Core Framework

   Permission to use, copy, modify, and distribute this software and its
   documentation for any purpose and without fee is hereby granted, provided
   that the above copyright notice appear in all copies and that both that
   copyright notice and this permission notice appear in supporting
   documentation.

   We disclaim all warranties with regard to this software, including all
   implied warranties of merchantability and fitness, in no event shall
   we be liable for any special, indirect or consequential damages or any
   damages whatsoever resulting from loss of use, data or profits, whether in
   an action of contract, negligence or other tortious action, arising out of
   or in connection with the use or performance of this software.
*/
// $Id: NGLockingStream.m,v 1.4 2003/01/07 11:36:47 helge Exp $

#include "common.h"
#include "NGLockingStream.h"

@implementation NGLockingStream

+ (id)filterWithSource:(id<NGStream>)_source lock:(id<NSObject,NSLocking>)_lock {
  return [[[self alloc] initWithSource:_source lock:_lock] autorelease];
}

- (id)initWithSource:(id<NGStream>)_source lock:(id<NSObject,NSLocking>)_lock {
  if ((self = [super initWithSource:_source])) {
    if (_lock == nil) {
      readLock  = [[NSRecursiveLock allocWithZone:[self zone]] init];
      writeLock = [readLock retain];
    }
    else {
      readLock  = [_lock    retain];
      writeLock = [readLock retain];
    }
  }
  return self;
}

- (id)initWithSource:(id<NGStream>)_source {
  return [self initWithSource:_source lock:nil];
}

- (void)dealloc {
  [self->readLock  release];
  [self->writeLock release];
  [super dealloc];
}

// primitives

- (unsigned)readBytes:(void *)_buf count:(unsigned)_len {
  volatile unsigned result = 0;

  [readLock lock];

  NS_DURING {
    result = (readBytes != NULL)
      ? (unsigned)readBytes(source, _cmd, _buf, _len)
      : [source readBytes:_buf count:_len];
  }
  NS_HANDLER {
    [readLock unlock];
    [localException raise];
  }
  NS_ENDHANDLER;
  [readLock unlock];

  return result;
}

- (unsigned)writeBytes:(const void *)_buf count:(unsigned)_len {
  volatile unsigned result = 0;

  [writeLock lock];

  NS_DURING {
    result = (writeBytes != NULL)
      ? (unsigned)writeBytes(source, _cmd, _buf, _len)
      : [source writeBytes:_buf count:_len];
  }
  NS_HANDLER {
    [writeLock unlock];
    [localException raise];
  }
  NS_ENDHANDLER;
  [writeLock unlock];

  return result;
}

- (BOOL)flush {
  BOOL res = NO;
  
  [writeLock lock];

  NS_DURING {
    res = [super flush];
  }
  NS_HANDLER {
    [writeLock unlock];
    [localException raise];
  }
  NS_ENDHANDLER;
  [writeLock unlock];
  return res;
}

- (BOOL)safeReadBytes:(void *)_buf  count:(unsigned)_len {
  BOOL res;
  
  [readLock lock];

  NS_DURING {
    *(&res) = [super safeReadBytes:_buf count:_len];
  }
  NS_HANDLER {
    [readLock unlock];
    [localException raise];
  }
  NS_ENDHANDLER;
  [readLock unlock];

  return res;
}

- (BOOL)safeWriteBytes:(const void *)_buf count:(unsigned)_len {
  BOOL res = NO;
  
  [writeLock lock];

  NS_DURING {
    res = [super safeWriteBytes:_buf count:_len];
  }
  NS_HANDLER {
    [writeLock unlock];
    [localException raise];
  }
  NS_ENDHANDLER;
  [writeLock unlock];
  return res;
}

@end /* NGLockingStream */
