/* 
   NGMD5Stream.m

   Copyright (C) 2002 SKYRIX Software AG, Helge Hess
   All rights reserved.

   Author: Helge Hess (helge.hess@skyrix.com)

   This file is part of the SKYRiX Core Framework

   Permission to use, copy, modify, and distribute this software and its
   documentation for any purpose and without fee is hereby granted, provided
   that the above copyright notice appear in all copies and that both that
   copyright notice and this permission notice appear in supporting
   documentation.

   We disclaim all warranties with regard to this software, including all
   implied warranties of merchantability and fitness, in no event shall
   we be liable for any special, indirect or consequential damages or any
   damages whatsoever resulting from loss of use, data or profits, whether in
   an action of contract, negligence or other tortious action, arising out of
   or in connection with the use or performance of this software.
*/
// $Id: NGMD5Stream.m,v 1.3 2003/01/07 11:36:47 helge Exp $

#import <NGExtensions/NGMD5Generator.h>
#include "common.h"
#include "NGMD5Stream.h"

@implementation NGMD5Stream

- (id)initWithSource:(id<NGStream>)_source {
  if ((self = [super initWithSource:_source])) {
    self->inDigest  = [[NGMD5Generator allocWithZone:[self zone]] init];
    self->outDigest = [[NGMD5Generator allocWithZone:[self zone]] init];
  }
  return self;
}
- (id)initWithInputSource:(id<NGInputStream>)_source {
  if ((self = [super initWithInputSource:_source])) {
    self->inDigest  = [[NGMD5Generator allocWithZone:[self zone]] init];
  }
  return self;
}
- (id)initWithOutputSource:(id<NGOutputStream>)_source {
  if ((self = [super initWithOutputSource:_source])) {
    self->outDigest = [[NGMD5Generator allocWithZone:[self zone]] init];
  }
  return self;
}

- (void)dealloc {
  [self->inDigest  release];
  [self->outDigest release];
  [super dealloc];
}

- (unsigned)readBytes:(void *)_buf count:(unsigned)_len {
  unsigned len;

  len = (self->readBytes)
    ? (unsigned)readBytes(self->source, _cmd, _buf, _len)
    : [self->source readBytes:_buf count:_len];

  [self->inDigest encodeBytes:_buf count:len];

  return len;
}
- (unsigned)writeBytes:(const void *)_buf count:(unsigned)_len {
  unsigned len;

  len = (self->writeBytes)
    ? (unsigned)writeBytes(self->source, _cmd, _buf, _len)
    : [self->source writeBytes:_buf count:_len];

  [self->outDigest encodeBytes:_buf count:len];

  return len;
}
  
// accessors

- (NSData *)inDigestAsData {
  return [self->inDigest digestAsData];
}
- (NSData *)outDigestAsData {
  return [self->outDigest digestAsData];
}

- (NSString *)inDigestAsString {
  return [self->inDigest digestAsString];
}
- (NSString *)outDigestAsString {
  return [self->outDigest digestAsString];
}

@end
