/* 
   NGSocketExceptions.m

   Copyright (C) 2002 SKYRIX Software AG, Helge Hess
   All rights reserved.

   Author: Helge Hess (helge.hess@skyrix.com)

   This file is part of the SKYRiX Core Framework

   Permission to use, copy, modify, and distribute this software and its
   documentation for any purpose and without fee is hereby granted, provided
   that the above copyright notice appear in all copies and that both that
   copyright notice and this permission notice appear in supporting
   documentation.

   We disclaim all warranties with regard to this software, including all
   implied warranties of merchantability and fitness, in no event shall
   we be liable for any special, indirect or consequential damages or any
   damages whatsoever resulting from loss of use, data or profits, whether in
   an action of contract, negligence or other tortious action, arising out of
   or in connection with the use or performance of this software.
*/
// $Id: NGSocketExceptions.m,v 1.4 2003/01/07 11:36:47 helge Exp $

#import <Foundation/Foundation.h>
#import <NGExtensions/NGExtensions.h>
#import <NGExtensions/NSException+misc.h>
#include "NGSocketExceptions.h"

@implementation NGSocketException

- (id)init {
  return [self initWithReason:@"a socket exception occured" socket:nil];
}
- (id)initWithReason:(NSString *)_reason {
  return [self initWithReason:_reason socket:nil];
}

- (id)initWithReason:(NSString *)_reason socket:(id<NGSocket>)_socket {
  self = [super initWithName:NSStringFromClass([self class])
                reason:_reason
                userInfo:nil];
  if (self) {
    self->socket = [_socket retain];
  }
  return self;
}

- (void)dealloc {
  [self->socket release];
  [super dealloc];
}

- (id<NGSocket>)socket {
  return self->socket;
}

@end /* NGSocketException */

@implementation NGCouldNotResolveHostNameException

- (id)init {
  return [self initWithHostName:@"<noname>" reason:nil];
}

- (id)initWithHostName:(NSString *)_name reason:(NSString *)_reason {
  self = [super initWithReason:
                  [NSString stringWithFormat:
                              @"Could not resolve host %@: %@",
                            _name, _reason ? _reason : @"error"]
                socket:nil];
  if (self) {
    self->hostName = [_name copy];
  }
  return self;
}

- (void)dealloc {
  RELEASE(self->hostName);
  [super dealloc];
}

- (NSString *)hostName {
  return self->hostName;
}

@end

@implementation NGDidNotFindServiceException

- (id)init {
  return [self initWithServiceName:nil];
}
- (id)initWithServiceName:(NSString *)_service {
  self = [super initWithReason:
                  [NSString stringWithFormat:@"did not find service %@", _service]
                socket:nil];
  if (self) {
    self->serviceName = [_service copy];
  }
  return self;
}

- (void)dealloc {
  RELEASE(self->serviceName);
  [super dealloc];
}

- (NSString *)serviceName {
  return self->serviceName;
}

@end

@implementation NGInvalidSocketDomainException

- (id)initWithReason:(NSString *)_reason
  socket:(id<NGSocket>)_socket
  domain:(id<NGSocketDomain>)_domain {

  if ((self = [super initWithReason:_reason socket:nil])) {
    self->domain = [_domain retain];
  }
  return self;
}

- (void)dealloc {
  [self->domain release];
  [super dealloc];
}

- (id<NGSocketDomain>)domain {
  return self->domain;
}

@end /* NGInvalidSocketDomainException */

@implementation NGCouldNotCreateSocketException

- (id)init {
  return [self initWithReason:@"Could not create socket" domain:nil];
}
- (id)initWithReason:(NSString *)_reason domain:(id<NGSocketDomain>)_domain {
  if ((self = [super initWithReason:_reason socket:nil])) {
    self->domain = [_domain retain];
  }
  return self;
}

- (void)dealloc {
  [self->domain release];
  [super dealloc];
}

- (id<NGSocketDomain>)domain {
  return self->domain;
}

@end /* NGCouldNotCreateSocketException */

// ******************** bind ***********************

@implementation NGSocketBindException
@end /* NGSocketBindException */

@implementation NGSocketAlreadyBoundException

- (id)init {
  return [self initWithReason:@"Socket is already bound"];
}

@end /* NGSocketAlreadyBoundException */

@implementation NGCouldNotBindSocketException

- (id)init {
  return [self initWithReason:@"could not bind socket" socket:nil address:nil];
}

- (id)initWithReason:(NSString *)_reason
  socket:(id<NGSocket>)_socket
  address:(id<NGSocketAddress>)_address {

  if ((self = [super initWithReason:_reason socket:_socket])) {
    self->address = [_address retain];
  }
  return self;
}

- (void)dealloc {
  [self->address release];
  [super dealloc];
}

- (id<NGSocketAddress>)address {
  return self->address;
}

@end /* NGCouldNotBindSocketException */

// ******************** connect ********************

@implementation NGSocketConnectException
@end /* NGSocketConnectException */

@implementation NGSocketNotConnectedException

- (id)init {
  return [self initWithReason:@"Socket is not connected"];
}

@end /* NGSocketNotConnectedException */

@implementation NGSocketAlreadyConnectedException

- (id)init {
  return [self initWithReason:@"Socket is already connected"];
}

@end /* NGSocketAlreadyConnectedException */

@implementation NGCouldNotConnectException

- (id)init {
  return [self initWithReason:@"could not connect socket" socket:nil address:nil];
}

- (id)initWithReason:(NSString *)_reason
  socket:(id<NGActiveSocket>)_socket
  address:(id<NGSocketAddress>)_address {

  if ((self = [super initWithReason:_reason socket:_socket])) {
    self->address = [_address retain];
  }
  return self;
}

- (void)dealloc {
  [self->address release];
  [super dealloc];
}

- (id<NGSocketAddress>)address {
  return self->address;
}

@end /* NGCouldNotConnectException */

// ******************** listen ********************

@implementation NGSocketIsAlreadyListeningException

- (id)init {
  return [self initWithReason:@"Socket is already listening"];
}

@end /* NGSocketIsAlreadyListeningException */

@implementation NGCouldNotListenException
@end /* NGCouldNotListenException */

// ******************** accept ********************

@implementation NGCouldNotAcceptException
@end /* NGCouldNotAcceptException */

// ******************** options ********************

@implementation NGSocketOptionException

- (id)init {
  return [self initWithReason:@"Could not get/set socket option" option:-1 level:-1];
}
- (id)initWithReason:(NSString *)_reason option:(int)_option level:(int)_level {
  if ((self = [super initWithReason:_reason])) {
    option = _option;
    level  = _level;
  }
  return self;
}

- (int)option {
  return option;
}
- (int)level {
  return level;
}

@end /* NGSocketOptionException */

@implementation NGCouldNotSetSocketOptionException
@end /* NGCouldNotSetSocketOptionException */

@implementation NGCouldNotGetSocketOptionException
@end /* NGCouldNotGetSocketOptionException */

// ******************** socket closed **************

@implementation NGSocketShutdownException

- (id)init {
  return [self initWithStream:nil reason:@"the socket was shutdown"];
}
- (id)initWithReason:(NSString *)_reason {
  return [self initWithStream:nil reason:_reason];
}
- (id)initWithReason:(NSString *)_reason socket:(id<NGActiveSocket>)_socket {
  return [self initWithStream:_socket reason:_reason];
}

- (id)initWithSocket:(id<NGActiveSocket>)_socket {
  return [self initWithStream:_socket reason:@"the socket was shutdown"];
}

/* accessors */

- (id<NGActiveSocket>)socket {
  return [self->streamPointer nonretainedObjectValue];
}

@end /* NGSocketShutdownException */

@implementation NGSocketShutdownDuringReadException
@end

@implementation NGSocketShutdownDuringWriteException
@end

@implementation NGSocketTimedOutException
@end

@implementation NGSocketConnectionResetException
@end
