/* 
   NGBufferedStream.h

   Copyright (C) 2002 SKYRIX Software AG, Helge Hess
   All rights reserved.

   Author: Helge Hess (helge.hess@skyrix.com)

   This file is part of the SKYRiX Core Framework

   Permission to use, copy, modify, and distribute this software and its
   documentation for any purpose and without fee is hereby granted, provided
   that the above copyright notice appear in all copies and that both that
   copyright notice and this permission notice appear in supporting
   documentation.

   We disclaim all warranties with regard to this software, including all
   implied warranties of merchantability and fitness, in no event shall
   we be liable for any special, indirect or consequential damages or any
   damages whatsoever resulting from loss of use, data or profits, whether in
   an action of contract, negligence or other tortious action, arising out of
   or in connection with the use or performance of this software.
*/
// $Id: NGBufferedStream.h,v 1.2 2002/12/27 09:45:33 helge Exp $

#ifndef __NGStreams_NGBufferedStream_H__
#define __NGStreams_NGBufferedStream_H__

#include <NGStreams/NGFilterStream.h>

@interface NGBufferedStream : NGFilterStream
{
@private
  void     *readBuffer;
  void     *readBufferPos;     // current position (ptr) in buffer
  unsigned readBufferFillSize; // number of 'read' bytes in the buffer
  unsigned readBufferSize;     // maximum capacity in bytes

  void     *writeBuffer;
  unsigned writeBufferFillSize;
  unsigned writeBufferSize;

  struct {
    unsigned int _flushOnNewline:1;
  } flags;
}

+ (id)filterWithSource:(id<NGStream>)_source bufferSize:(unsigned)_size;
- (id)initWithSource:(id<NGStream>)_source bufferSize:(unsigned)_size;
- (id)initWithSource:(id<NGStream>)_source;

// accessors

- (void)setReadBufferSize:(unsigned)_size;
- (unsigned)readBufferSize;

- (void)setWriteBufferSize:(unsigned)_size;
- (unsigned)writeBufferSize;

// blocking ..

- (BOOL)wouldBlockInMode:(NGStreamMode)_mode;

// primitives

- (unsigned)readBytes:(void *)_buf count:(unsigned)_len;
- (unsigned)writeBytes:(const void *)_buf count:(unsigned)_len;
- (BOOL)flush;

@end

@interface NGStream(NGBufferedStreamExtensions)

- (NGBufferedStream *)bufferedStream;

@end

#endif /* __NGStreams_NGBufferedStream_H__ */
