/* 
   NGByteBuffer.h

   Copyright (C) 2002 SKYRIX Software AG, Helge Hess
   All rights reserved.

   Author: Jan Reichmann (jan.reichmann@skyrix.com)

   This file is part of the SKYRiX Core Framework

   Permission to use, copy, modify, and distribute this software and its
   documentation for any purpose and without fee is hereby granted, provided
   that the above copyright notice appear in all copies and that both that
   copyright notice and this permission notice appear in supporting
   documentation.

   We disclaim all warranties with regard to this software, including all
   implied warranties of merchantability and fitness, in no event shall
   we be liable for any special, indirect or consequential damages or any
   damages whatsoever resulting from loss of use, data or profits, whether in
   an action of contract, negligence or other tortious action, arising out of
   or in connection with the use or performance of this software.
*/
// $Id: NGByteBuffer.h,v 1.2 2002/12/27 09:45:33 helge Exp $

#ifndef __NGStreams_NGByteBuffer_H__
#define __NGStreams_NGByteBuffer_H__

#include <NGStreams/NGFilterStream.h>
#include <NGStreams/NGStreamProtocols.h>

struct NGByteBufferLA;

/*
  Although NGByteBuffer is defined to be a stream, it is usually not used as
  such. Instead most parsers implemented using NGByteBuffer will only call
  -la: and -consume.

  The stream interface is provided to read large blocks with a know length. Eg
  if you have a structure that prefixes some data with the data's length, you
  can first parse the length and then call -safeReadBytes:count: to read the
  content.

  Note that -readByte and -la: return -1 on EOF.
*/

@interface NGByteBuffer : NGFilterStream
{
@protected
  struct NGByteBufferLA *la;

  unsigned bufLen;
  BOOL     wasEOF;
  unsigned headIdx;
  unsigned sizeLessOne;

  int (*readByte)(id, SEL);
  int (*laFunction)(id, SEL, unsigned);
}

/*
  Initialize a byte buffer with a lookahead depth of _la bytes.
*/
+ (id)byteBufferWithSource:(id<NGStream>)_stream la:(unsigned)_la;
- (id)initWithSource:(id<NGStream>)_stream la:(unsigned)_la;

// LA
- (int)la:(unsigned)_lookaheadPosition;

- (void)consume;                // consume one byte
- (void)consume:(unsigned)_cnt; // consume _cnt bytes

// NGStream
- (int)readByte;
- (unsigned)readBytes:(void *)_buf count:(unsigned)_len;

@end

#endif /* __NGStreams_NGByteBuffer_H__ */
