/* 
   NGCharBuffer.h

   Copyright (C) 2002 SKYRIX Software AG, Helge Hess
   All rights reserved.

   Author: Jan Reichmann (jan.reichmann@skyrix.com)

   This file is part of the SKYRiX Core Framework

   Permission to use, copy, modify, and distribute this software and its
   documentation for any purpose and without fee is hereby granted, provided
   that the above copyright notice appear in all copies and that both that
   copyright notice and this permission notice appear in supporting
   documentation.

   We disclaim all warranties with regard to this software, including all
   implied warranties of merchantability and fitness, in no event shall
   we be liable for any special, indirect or consequential damages or any
   damages whatsoever resulting from loss of use, data or profits, whether in
   an action of contract, negligence or other tortious action, arising out of
   or in connection with the use or performance of this software.
*/
// $Id: NGCharBuffer.h,v 1.2 2002/12/27 09:45:33 helge Exp $

#ifndef __NGStreams_NGCharBuffer_H__
#define __NGStreams_NGCharBuffer_H__

#include <NGStreams/NGFilterTextStream.h>
#include <NGStreams/NGTextStreamProtocols.h>
#include <NGStreams/NGStreamProtocols.h>

struct NGCharBufferLA;

/*
  Although NGCharBuffer is defined to be a stream, it is usually not used as
  such. Instead most parsers implemented using NGCharBuffer will only call
  -la: and -consume.

  Note that -la: return -1 on EOF and -readCharacter throws an
  NGEndOfStreamException. -readCharacter is basically a -la:0 followed by a
  -consume.
*/

@interface NGCharBuffer : NGFilterTextStream
{
@protected
  struct NGCharBufferLA *la;

  int  bufLen;
  BOOL wasEOF;
  int  headIdx;
  int  sizeLessOne;

  IMP readCharacter;
}

+ (id)charBufferWithSource:(id<NGTextStream>)_source la:(int)_la;
- (id)initWithSource:(id<NGTextStream>)_source la:(int)_la;

// LA
- (int)la:(int)_ls;

- (void)consume;           // consume one character
- (void)consume:(int)_cnt; // consume _cnt characters

// NGTextStream
- (unichar)readCharacter;

@end

#endif /* __NGStreams_NGCharBuffer_H__ */
