/* 
   NGDataStream.h

   Copyright (C) 2002 SKYRIX Software AG, Helge Hess
   All rights reserved.

   Author: Helge Hess (helge.hess@skyrix.com)

   This file is part of the SKYRiX Core Framework

   Permission to use, copy, modify, and distribute this software and its
   documentation for any purpose and without fee is hereby granted, provided
   that the above copyright notice appear in all copies and that both that
   copyright notice and this permission notice appear in supporting
   documentation.

   We disclaim all warranties with regard to this software, including all
   implied warranties of merchantability and fitness, in no event shall
   we be liable for any special, indirect or consequential damages or any
   damages whatsoever resulting from loss of use, data or profits, whether in
   an action of contract, negligence or other tortious action, arising out of
   or in connection with the use or performance of this software.
*/
// $Id: NGDataStream.h,v 1.3 2002/12/27 09:45:34 helge Exp $

#ifndef __NGStreams_NGDataStream_H__
#define __NGStreams_NGDataStream_H__

#include <NGStreams/NGStream.h>
#include <NGStreams/NGStreamProtocols.h>

@class NSData;

@interface NGDataStream : NGStream < NGPositionableStream >
{
@private
  NSException  *lastException;
  NGStreamMode streamMode;
  NSData       *data;
  unsigned     position;

  unsigned int (*dataLength)(id, SEL);
  const void   *(*dataBytes)(id, SEL);  
}

+ (id)dataStream;
+ (id)dataStreamWithCapacity:(int)_capacity;
+ (id)streamWithData:(NSData *)_data;
- (id)initWithData:(NSData *)_data;

- (NSException *)lastException;
- (void)setLastException:(NSException *)_exception;
- (void)resetLastException;

// accessors

- (NSData *)data;

- (unsigned)availableBytes; // returns number of available bytes

// primitives

// throws
//   NGStreamNotOpenException   when the stream is not open
//   NGEndOfStreamException     when the end of the stream is reached
- (unsigned)readBytes:(void *)_buf count:(unsigned)_len;

// throws
//   NGReadOnlyStreamException when the stream is not writeable
//   NGStreamNotOpenException  when the stream is not open
- (unsigned)writeBytes:(const void *)_buf count:(unsigned)_len;

- (BOOL)close;

- (NGStreamMode)mode;
- (BOOL)isRootStream;

// NGPositionableStream

- (BOOL)moveToLocation:(unsigned)_location;
- (BOOL)moveByOffset:(int)_delta;

// blocking ..

- (BOOL)wouldBlockInMode:(NGStreamMode)_mode; // always NO ..

@end

#endif /* __NGStreams_NGDataStream_H__ */
