/* 
   NGDatagramPacket.h

   Copyright (C) 2002 SKYRIX Software AG, Helge Hess
   All rights reserved.

   Author: Helge Hess (helge.hess@skyrix.com)

   This file is part of the SKYRiX Core Framework

   Permission to use, copy, modify, and distribute this software and its
   documentation for any purpose and without fee is hereby granted, provided
   that the above copyright notice appear in all copies and that both that
   copyright notice and this permission notice appear in supporting
   documentation.

   We disclaim all warranties with regard to this software, including all
   implied warranties of merchantability and fitness, in no event shall
   we be liable for any special, indirect or consequential damages or any
   damages whatsoever resulting from loss of use, data or profits, whether in
   an action of contract, negligence or other tortious action, arising out of
   or in connection with the use or performance of this software.
*/
// $Id: NGDatagramPacket.h,v 1.2 2002/12/27 09:45:34 helge Exp $

#ifndef __NGNet_NGDatagramPacket_H__
#define __NGNet_NGDatagramPacket_H__

#import <Foundation/NSObject.h>
#include <NGStreams/NGSocketProtocols.h>

@class NSData;

/*
  This class represents an UDP datagram. It contains the addresses of it's sender
  and it's receiver.
*/

@interface NGDatagramPacket : NSObject < NGDatagramPacket >
{
@protected
  id<NGSocketAddress> sender;
  id<NGSocketAddress> receiver;
  NSData              *packet;
}

// packet factory

+ (id)packetWithData:(NSData *)_data;
+ (id)packetWithBytes:(const void *)_bytes size:(int)_packetSize;

- (id)initWithBytes:(const void *)_bytes size:(int)_size;
- (id)initWithData:(NSData *)_data;

// accessors

- (void)setSender:(id<NGSocketAddress>)_address;
- (id<NGSocketAddress>)sender;
- (void)setReceiver:(id<NGSocketAddress>)_address;
- (id<NGSocketAddress>)receiver;

- (void)setData:(NSData *)_data;
- (NSData *)data;

- (int)packetSize;

// operations

- (void)reverseAddresses;

@end

#endif /* __NGNet_NGDatagramPacket_H__ */
