/*
   NGDescriptorFunctions.h

   Copyright (C) 2002 SKYRIX Software AG, Helge Hess
   All rights reserved.

   Author: Helge Hess (helge.hess@skyrix.com)

   This file is part of the SKYRiX Core Framework

   Permission to use, copy, modify, and distribute this software and its
   documentation for any purpose and without fee is hereby granted, provided
   that the above copyright notice appear in all copies and that both that
   copyright notice and this permission notice appear in supporting
   documentation.

   We disclaim all warranties with regard to this software, including all
   implied warranties of merchantability and fitness, in no event shall
   we be liable for any special, indirect or consequential damages or any
   damages whatsoever resulting from loss of use, data or profits, whether in
   an action of contract, negligence or other tortious action, arising out of
   or in connection with the use or performance of this software.
*/
// $Id: NGDescriptorFunctions.h,v 1.2 2002/12/27 09:45:34 helge Exp $

#ifndef __NGStreams_NGDescriptorFunctions_H__
#define __NGStreams_NGDescriptorFunctions_H__

#import <Foundation/NSObject.h>

#if !defined(WIN32)

/*
  Polls a descriptor. Returns 1 if events occurred, 0 if a timeout occured
  and -1 if an error other than EINTR or EAGAIN occured.
*/
extern int NGPollDescriptor(int _fd, short _events, int _timeout);

/*
  Set/Get descriptor flags
*/
extern int  NGGetDescriptorFlags(int _fd);
extern void NGSetDescriptorFlags(int _fd, int _flags);
extern void NGAddDescriptorFlag (int _fd, int _flag);

/*
  Reading and writing with non-blocking IO support.
  The functions return
    -1  on error, with errno set to either recv's or poll's errno
    0   on the end of file condition
    -2  if the operation timed out

  Enable login topic 'nonblock' to find out about timeouts.
*/
extern int NGDescriptorRecv(int _fd, char *_buf, int _len,
                            int _flags, int _timeout);
extern int NGDescriptorSend(int _fd, const char *_buf, int _len,
                            int _flags, int _timeout);

/*
  Check whether the descriptor is associated to a terminal device.
  Get the name of the associated terminal device.
*/
extern BOOL     NGDescriptorIsAtty(int _fd);
extern NSString *NGDescriptorGetTtyName(int _fd);

#endif /* !WIN32 */

#endif /* __NGStreams_NGDescriptorFunctions_H__ */
