/* 
   NGInternetSocketAddress.h

   Copyright (C) 2002 SKYRIX Software AG, Helge Hess
   All rights reserved.

   Author: Helge Hess (helge.hess@skyrix.com)

   This file is part of the SKYRiX Core Framework

   Permission to use, copy, modify, and distribute this software and its
   documentation for any purpose and without fee is hereby granted, provided
   that the above copyright notice appear in all copies and that both that
   copyright notice and this permission notice appear in supporting
   documentation.

   We disclaim all warranties with regard to this software, including all
   implied warranties of merchantability and fitness, in no event shall
   we be liable for any special, indirect or consequential damages or any
   damages whatsoever resulting from loss of use, data or profits, whether in
   an action of contract, negligence or other tortious action, arising out of
   or in connection with the use or performance of this software.
*/
// $Id: NGInternetSocketAddress.h,v 1.3 2003/01/13 09:37:28 helge Exp $

#ifndef __NGNet_NGInternetSocketAddress_H__
#define __NGNet_NGInternetSocketAddress_H__

#include <NGStreams/NGSocketProtocols.h>
#include <NGStreams/NGActiveSocket.h>

/*
  Represents an Internet socket address (AF_INET).
  
  Socket addresses are immutable. -copy therefore returns a retained self.

  The host arguments are id because they are allowed to be either NSString
  or NSHost objects (although NSString's are preferred). If the host is nil,
  then a wildcard (INADDR_ANY) is used.

  Note that the hostName is resolved when the internalAddressRepresentation
  is requested.
 */

@interface NGInternetSocketAddress : NSObject < NSCopying, NSCoding, NGSocketAddress >
{
@private
  void     *address; /* ptr to struct sockaddr_in */
  NSString *hostName;
  BOOL     isAddressFilled;
  BOOL     isHostFilled;
  BOOL     isWildcardHost;
}

+ (id)addressWithPort:(int)_port onHost:(id)_host;
+ (id)addressWithPort:(int)_port; // localhost
- (id)initWithPort:(int)_port onHost:(id)_host; // designated init
- (id)initWithPort:(int)_port;    // localhost

// these throw NGDidNotFindServiceException if the service is not found
+ (id)addressWithService:(NSString *)_serviceName
  onHost:(id)_host protocol:(NSString *)_protocol;
+ (id)addressWithService:(NSString *)_serviceName protocol:(NSString *)_protocol;
- (id)initWithService:(NSString *)_serviceName onHost:(id)_host
  protocol:(NSString *)_protocol;
- (id)initWithService:(NSString *)_serviceName protocol:(NSString *)_protocol;

+ (id)wildcardAddress;
+ (id)wildcardAddressWithPort:(int)_port;

/* accessors */

- (NSString *)hostName;
- (NSString *)address;
- (int)port;

- (BOOL)isWildcardAddress;

/* testing for equality */

- (BOOL)isEqualToAddress:(NGInternetSocketAddress *)_addr;
- (BOOL)isEqual:(id)_obj;

/* description */

- (NSString *)stringValue; // returns 'hostname:port' as used in URLs
- (NSString *)description;

/* NGSocketAddress */

// throws NGCouldNotResolveHostNameException
- (void *)internalAddressRepresentation;

- (int)addressRepresentationSize;
- (id)domain;

@end

@interface NGActiveSocket(NGInternetActiveSocket)

// this method calls +socketConnectedToAddress: with an NGInternetSocketAddress
+ (id)socketConnectedToPort:(int)_port onHost:(id)_host;

// this method calls -connectToAddress: with an NGInternetSocketAddress
- (BOOL)connectToPort:(int)_port onHost:(id)_host;

@end

#endif /* __NGNet_NGInternetSocketAddress_H__ */
