/* 
   NGLocalSocketAddress.h

   Copyright (C) 2002 SKYRIX Software AG, Helge Hess
   All rights reserved.

   Author: Helge Hess (helge.hess@skyrix.com)

   This file is part of the SKYRiX Core Framework

   Permission to use, copy, modify, and distribute this software and its
   documentation for any purpose and without fee is hereby granted, provided
   that the above copyright notice appear in all copies and that both that
   copyright notice and this permission notice appear in supporting
   documentation.

   We disclaim all warranties with regard to this software, including all
   implied warranties of merchantability and fitness, in no event shall
   we be liable for any special, indirect or consequential damages or any
   damages whatsoever resulting from loss of use, data or profits, whether in
   an action of contract, negligence or other tortious action, arising out of
   or in connection with the use or performance of this software.
*/
// $Id: NGLocalSocketAddress.h,v 1.3 2002/12/27 09:45:35 helge Exp $

#ifndef __NGNet_NGLocalSocketAddress_H__
#define __NGNet_NGLocalSocketAddress_H__

#if !defined(WIN32) || defined(__CYGWIN32__)

/*
  UNIX domain sockets, currently not available on Windows

  The Win32 local sockets are done using so called 'named pipes'.
*/

#include <NGStreams/NGSocketProtocols.h>

/*
  Represents a UNIX domain socket address (AF_LOCAL) or a named pipe (Win32).

  Socket addresses are immutable. -copy therefore returns a retained self.

  Note that when a local socket address is archived it stores the host together
  with the path. This ensures that the address-space will be the same on
  unarchiving, otherwise it will return an error.
*/

@interface NGLocalSocketAddress : NSObject < NSCopying, NGSocketAddress >
{
@private
#if defined(__WIN32__) && !defined(__CYGWIN32__)
  NSString *path;
#else
  void *address; /* ptr to struct sockaddr_un */
#endif
}

+ (id)addressWithPath:(NSString *)_path;
+ (id)address;
- (id)initWithPath:(NSString *)_path; // designated initializer
- (id)init;                           // creates unique path (pid,thread-id,cnt)

/* accessors */

- (NSString *)path;

/* testing for equality */

- (BOOL)isEqualToAddress:(NGLocalSocketAddress *)_addr;
- (BOOL)isEqual:(id)_obj;

/* test for accessibility */

- (BOOL)canSendOnAddress;
- (BOOL)canReceiveOnAddress;

@end

#endif /* !WIN32 */

#endif /* __NGNet_NGLocalSocketAddress_H__ */
