/* 
   NGPassiveSocket.h

   Copyright (C) 2002 SKYRIX Software AG, Helge Hess
   All rights reserved.

   Author: Helge Hess (helge.hess@skyrix.com)

   This file is part of the SKYRiX Core Framework

   Permission to use, copy, modify, and distribute this software and its
   documentation for any purpose and without fee is hereby granted, provided
   that the above copyright notice appear in all copies and that both that
   copyright notice and this permission notice appear in supporting
   documentation.

   We disclaim all warranties with regard to this software, including all
   implied warranties of merchantability and fitness, in no event shall
   we be liable for any special, indirect or consequential damages or any
   damages whatsoever resulting from loss of use, data or profits, whether in
   an action of contract, negligence or other tortious action, arising out of
   or in connection with the use or performance of this software.
*/
// $Id: NGPassiveSocket.h,v 1.2 2002/12/27 09:45:35 helge Exp $

#ifndef __NGNet_NGPassiveSocket_H__
#define __NGNet_NGPassiveSocket_H__

#import <Foundation/NSLock.h>
#include <NGStreams/NGSocket.h>
#include <NGStreams/NGSocketProtocols.h>

/*
  Represents a STREAM server socket based on the standard Unix sockets library.

  A passive socket has exactly one address, the address the socket is bound to.
  If you do not bind the socket, the address is determined after the listen()
  call was executed through the getsockname() call.

  Note that if the socket is bound it's still an active socket in the
  system's view, it becomes an passive one when the listen call is executed.

  NOTE: Currently the passive _must_ be bound. This is because during the
        creation of the socket the domain is needed. The domain is encapsulated
        in the socket-address.
        Therefore the method of letting the kernel determine a socket address,
        as described above, currently does not work.
*/

@interface NGPassiveSocket : NGSocket < NGPassiveSocket >
{
@protected
  id<NSObject,NSLocking> acceptLock; // prevents file-locking
  int backlogSize;
}

+ (id)socketBoundToAddress:(id<NGSocketAddress>)_address;

/* accessors */

- (BOOL)isListening;
- (BOOL)isOpen;

/* operations */

// throws
//   NGSocketIsAlreadyListeningException  when the socket is in the listen state
//   NGCouldNotListenException            when the listen call failed
- (BOOL)listenWithBacklog:(int)_backlogSize;

// accept blocks when multiple threads try to accept (using acceptLock)
// throws
//   NGCouldNotAcceptException  when the socket is not listening
//   NGCouldNotAcceptException  when the accept call failed
- (id<NGActiveSocket>)accept;

@end

#endif /* __NGNet_NGPassiveSocket_H__ */
