/* 
   NGStream.h

   Copyright (C) 2002 SKYRIX Software AG, Helge Hess
   All rights reserved.

   Author: Helge Hess (helge.hess@skyrix.com)

   This file is part of the SKYRiX Core Framework

   Permission to use, copy, modify, and distribute this software and its
   documentation for any purpose and without fee is hereby granted, provided
   that the above copyright notice appear in all copies and that both that
   copyright notice and this permission notice appear in supporting
   documentation.

   We disclaim all warranties with regard to this software, including all
   implied warranties of merchantability and fitness, in no event shall
   we be liable for any special, indirect or consequential damages or any
   damages whatsoever resulting from loss of use, data or profits, whether in
   an action of contract, negligence or other tortious action, arising out of
   or in connection with the use or performance of this software.
*/
// $Id: NGStream.h,v 1.2 2002/12/27 09:46:03 helge Exp $

#ifndef __NGStreams_NGStream_H__
#define __NGStreams_NGStream_H__

#import <Foundation/NSObject.h>
#include <NGStreams/NGStreamsDecls.h>
#include <NGStreams/NGStreamProtocols.h>

@class NSData, NSException;

static inline BOOL NGCanReadInStreamMode(NGStreamMode _mode) {
  return ((_mode == NGStreamMode_readOnly) ||
          (_mode == NGStreamMode_readWrite));
}
static inline BOOL NGCanWriteInStreamMode(NGStreamMode _mode) {
  return ((_mode == NGStreamMode_writeOnly) ||
          (_mode == NGStreamMode_readWrite));
}

@interface NGStream : NSObject < NGStream, NGByteSequenceStream >

// ******************** primitives ********************

// Never returns 0. If an EOF like condition occures, NGEndOfStreamException
// is thrown.
- (unsigned)readBytes:(void *)_buf count:(unsigned)_len;       // abstract
- (unsigned)writeBytes:(const void *)_buf count:(unsigned)_len; // abstract

- (void)setLastException:(NSException *)_exception;

- (BOOL)flush; // empty
- (BOOL)close; // empty

- (NGStreamMode)mode; // abstract
- (BOOL)isRootStream; // abstract

/* methods which read/write exactly _len bytes */

- (BOOL)safeReadBytes:(void *)_buf count:(unsigned)_len;
- (BOOL)safeWriteBytes:(const void *)_buf count:(unsigned)_len;

/* marking */

- (BOOL)mark;           // does nothing
- (BOOL)rewind;         // does nothing
- (BOOL)markSupported;  // returns NO

/* convenience methods */

- (int)readByte; // java semantics (-1 returned on EOF)

// description

- (NSString *)modeDescription;

@end

@interface NGStream(DataMethods)

- (NSData *)readDataOfLength:(unsigned int)_length;
- (NSData *)safeReadDataOfLength:(unsigned int)_length;
- (unsigned int)writeData:(NSData *)_data;
- (BOOL)safeWriteData:(NSData *)_data;

@end

// concrete implementations as functions (to be used in non-subclasses)

NGStreams_EXPORT int NGReadByteFromStream(id<NGInputStream> _stream);

NGStreams_EXPORT BOOL
NGSafeReadBytesFromStream(id<NGInputStream> _in, void *_buf, unsigned _len);

NGStreams_EXPORT BOOL
NGSafeWriteBytesToStream(id<NGOutputStream> _out,const void *_buf,unsigned _len);

#endif /* __NGStreams_NGStream_H__ */
