/* 
   NGStreamCoder.h

   Copyright (C) 2002 SKYRIX Software AG, Helge Hess
   All rights reserved.

   Author: Helge Hess (helge.hess@skyrix.com)

   This file is part of the SKYRiX Core Framework

   Permission to use, copy, modify, and distribute this software and its
   documentation for any purpose and without fee is hereby granted, provided
   that the above copyright notice appear in all copies and that both that
   copyright notice and this permission notice appear in supporting
   documentation.

   We disclaim all warranties with regard to this software, including all
   implied warranties of merchantability and fitness, in no event shall
   we be liable for any special, indirect or consequential damages or any
   damages whatsoever resulting from loss of use, data or profits, whether in
   an action of contract, negligence or other tortious action, arising out of
   or in connection with the use or performance of this software.
*/
// $Id: NGStreamCoder.h,v 1.3 2002/12/27 09:46:03 helge Exp $

#ifndef __NGStreams_NGStreamCoder_H__
#define __NGStreams_NGStreamCoder_H__

#import <Foundation/NSCoder.h>

#if !(MAC_OS_X_VERSION_10_2 <= MAC_OS_X_VERSION_MAX_ALLOWED)
#  define USE_SERIALIZER 1
#  import <Foundation/NSSerialization.h>
#endif

#import <Foundation/NSMapTable.h>
#import <Foundation/NSHashTable.h>
#include <NGStreams/NGStreamProtocols.h>

@interface NGStreamCoder : NSCoder 
#if USE_SERIALIZER
  < NSObjCTypeSerializationCallBack >
#endif
{
@protected
  id<NGSerializer,NGStream> stream;   // destination/source stream
  NGIOSafeReadMethodType    readIMP;  // safe read method
  NGIOSafeWriteMethodType   writeIMP; // safe write method

  // used during encoding
  NSHashTable *outObjects;          // objects written so far
  NSHashTable *outConditionals;     // conditional objects
  NSHashTable *outPointers;         // set of pointers
  NSMapTable  *replacements;        // src-object to replacement
  BOOL        traceMode;            // YES if finding conditionals
  BOOL        didWriteHeader;
  SEL         classForCoder;        // default: classForCoder:
  SEL         replObjectForCoder;   // default: replacementObjectForCoder:
  BOOL        encodingRoot;

  // used during decoding
  unsigned    inArchiverVersion;    // archiver's version that wrote the data
  NSMapTable  *inObjects;           // decoded objects: key -> object
  NSMapTable  *inClasses;           // decoded classes: key -> class info
  NSMapTable  *inPointers;          // decoded pointers: key -> pointer
  NSMapTable  *inClassAlias;        // archive name -> decoded name
  NSMapTable  *inClassVersions;     // archive name -> class info
  NSZone      *objectZone;
  BOOL        decodingRoot;
  BOOL        didReadHeader;
}

+ (id)coderWithStream:(id<NGSerializer,NGStream>)_stream;
- (id)initWithStream:(id<NGSerializer,NGStream>)_stream;

// accessors

- (id<NGStream>)stream;
- (NSString *)coderSignature; // ID of the coder used
- (int)coderVersion;          // Version of the coder used

// encoding

- (void)encodeConditionalObject:(id)_object;
- (void)encodeRootObject:(id)_object;

// decoding

- (unsigned int)systemVersion;
- (id)decodeObject;

// Substituting One Class for Another

+ (NSString *)classNameDecodedForArchiveClassName:(NSString *)nameInArchive;
+ (void)decodeClassName:(NSString *)nameInArch asClassName:(NSString *)trueName;
- (NSString *)classNameDecodedForArchiveClassName:(NSString *)nameInArchive;
- (void)decodeClassName:(NSString *)nameInArch asClassName:(NSString *)trueName;

@end

#endif /* __NGStreams_NGStreamCoder_H__ */
