/* 
   NGStreamProtocols.h

   Copyright (C) 2002 SKYRIX Software AG, Helge Hess
   All rights reserved.

   Author: Helge Hess (helge.hess@skyrix.com)

   This file is part of the SKYRiX Core Framework

   Permission to use, copy, modify, and distribute this software and its
   documentation for any purpose and without fee is hereby granted, provided
   that the above copyright notice appear in all copies and that both that
   copyright notice and this permission notice appear in supporting
   documentation.

   We disclaim all warranties with regard to this software, including all
   implied warranties of merchantability and fitness, in no event shall
   we be liable for any special, indirect or consequential damages or any
   damages whatsoever resulting from loss of use, data or profits, whether in
   an action of contract, negligence or other tortious action, arising out of
   or in connection with the use or performance of this software.
*/
// $Id: NGStreamProtocols.h,v 1.3 2002/12/27 09:46:03 helge Exp $

#ifndef __NGStreams_NGStreamProtocols_H__
#define __NGStreams_NGStreamProtocols_H__

#import <Foundation/NSObject.h>

#if !(MAC_OS_X_VERSION_10_2 <= MAC_OS_X_VERSION_MAX_ALLOWED)
#  define USE_SERIALIZER 1
#  import <Foundation/NSSerialization.h>
#endif

@class NSException;

typedef enum {
  NGStreamMode_undefined = 0,
  NGStreamMode_readOnly  = 1,
  NGStreamMode_writeOnly = 2,
  NGStreamMode_readWrite = 4
} NGStreamMode;

/* if this value is returned by -read, -lastException is set ... */
enum {NGStreamError = 0x7fffffff};

typedef unsigned (*NGIOReadMethodType )(id, SEL, void *, unsigned);
typedef unsigned (*NGIOWriteMethodType)(id, SEL, const void *, unsigned);
typedef BOOL (*NGIOSafeReadMethodType )(id, SEL, void *, unsigned);
typedef BOOL (*NGIOSafeWriteMethodType)(id, SEL, const void *, unsigned);

@protocol NGInputStream < NSObject >

- (unsigned)readBytes:(void *)_buf count:(unsigned)_len;
- (BOOL)safeReadBytes:(void *)_buf count:(unsigned)_len;
- (BOOL)close;

// marks

- (BOOL)mark;
- (BOOL)rewind;
- (BOOL)markSupported;

@end

@protocol NGOutputStream < NSObject >

- (unsigned)writeBytes:(const void *)_buf count:(unsigned)_len;
- (BOOL)safeWriteBytes:(const void *)_buf count:(unsigned)_len;

- (BOOL)flush;
- (BOOL)close;

@end

@protocol NGPositionableStream < NSObject >

- (BOOL)moveToLocation:(unsigned)_location;
- (BOOL)moveByOffset:(int)_delta;

@end

@protocol NGStream < NGInputStream, NGOutputStream >

- (BOOL)close;
- (NGStreamMode)mode;
- (NSException *)lastException;

@end

@protocol NGByteSequenceStream < NGInputStream >

- (int)readByte; // Java semantics (-1 on EOF)
  
@end

typedef int (*NGSequenceReadByteMethod)(id<NGByteSequenceStream> self, SEL _cmd);

// push streams

@class NSData;

@protocol NGPushStream < NSObject >

- (void)pushChar:(char)_c;
- (void)pushCString:(const char *)_cstr;
- (void)pushData:(NSData *)_block;
- (void)pushBytes:(const void *)_buffer count:(unsigned)_len;
- (void)abort;

@end

// serializer

@protocol NGSerializer < NSObject >

- (void)serializeChar:(char)_value;
- (void)serializeShort:(short)_value;
- (void)serializeInt:(int)_value;
- (void)serializeLong:(long)_value;
- (void)serializeFloat:(float)_value;
- (void)serializeDouble:(double)_value;
- (void)serializeLongLong:(long long)_value;

- (char)deserializeChar;
- (short)deserializeShort;
- (int)deserializeInt;
- (long)deserializeLong;
- (float)deserializeFloat;
- (double)deserializeDouble;
- (long long)deserializeLongLong;

- (void)serializeCString:(const char *)_value;
- (char *)deserializeCString;

#if USE_SERIALIZER
- (void)serializeDataAt:(const void*)data ofObjCType:(const char*)type
  context:(id<NSObjCTypeSerializationCallBack>)_callback;
- (void)deserializeDataAt:(const void*)data ofObjCType:(const char*)type
  context:(id<NSObjCTypeSerializationCallBack>)_callback;
#endif

@end

#endif /* __NGStreams_NGStreamProtocols_H__ */
