/* 
   NGTaskStream.m

   Copyright (C) 2002 SKYRIX Software AG, Helge Hess
   All rights reserved.

   Author: Helge Hess (helge.hess@skyrix.com)

   This file is part of the SKYRiX Core Framework

   Permission to use, copy, modify, and distribute this software and its
   documentation for any purpose and without fee is hereby granted, provided
   that the above copyright notice appear in all copies and that both that
   copyright notice and this permission notice appear in supporting
   documentation.

   We disclaim all warranties with regard to this software, including all
   implied warranties of merchantability and fitness, in no event shall
   we be liable for any special, indirect or consequential damages or any
   damages whatsoever resulting from loss of use, data or profits, whether in
   an action of contract, negligence or other tortious action, arising out of
   or in connection with the use or performance of this software.
*/
// $Id: NGTaskStream.m,v 1.3 2002/12/27 09:49:00 helge Exp $

#include "config.h"
#include "common.h"
#include "NGTaskStream.h"

@implementation NGTaskStream

- (id)initWithPath:(NSString *)_executable
  arguments:(NSArray *)_args
  environment:(NSDictionary *)_env
{
  return nil;
}

- (void)dealloc {
  [super dealloc];
}

// state

- (BOOL)isOpen {
  return [self->task isRunning];
}

// primitives

- (unsigned)readBytes:(void *)_buf count:(unsigned)_len {
  return NGStreamError;
}
- (unsigned)writeBytes:(const void *)_buf count:(unsigned)_len {
  return NGStreamError;
}

- (BOOL)close {
  if (![self isOpen]) {
    NSLog(@"tried to close already closed stream %@", self);
    return NO;
  }
  [self->task terminate];
  return YES;
}

- (NGStreamMode)mode {
  return NGStreamMode_readWrite;
}
- (BOOL)isRootStream {
  return YES;
}

// marking

- (BOOL)mark {
  NSLog(@"WARNING: called mark on a stream which doesn't support marking !");
  return NO;
}
- (BOOL)rewind {
  THROW([[NGStreamException alloc] initWithStream:self
                                   reason:@"marking not supported"]);
  return NO;
}
- (BOOL)markSupported {
  return -1;
}

// description

- (NSString *)description {
  return [NSString stringWithFormat:
                     @"<%@[0x%08X] task=%@ mode=%@>",
                     NSStringFromClass([self class]), (unsigned)self,
                     self->task,
                     [self modeDescription]];
}

@end /* NGTaskStream */
