// $Id: ical3.m,v 1.1 2003/02/12 17:28:17 helge Exp $

#import <Foundation/Foundation.h>

@class EOQualifier, NSString, EOSortOrdering;

@interface iCal3Tool : NSObject
{
  EOQualifier *qualifier;
  NSString    *entityName;
  NSArray     *sortOrderings;
}

- (int)runWithArguments:(NSArray *)_args;

@end

#include <NGiCal/iCalDataSource.h>
#include <NGiCal/iCalObject.h>
#include <EOControl/EOQualifier.h>
#include <EOControl/EOSortOrdering.h>
#include "common.h"

@implementation iCal3Tool

- (id)init {
  if ((self = [super init])) {
    NSUserDefaults *ud;
    id tmp;
    
    /* collect options */
    ud = [NSUserDefaults standardUserDefaults];

    self->entityName = [[ud stringForKey:@"entity"] copy];
    
    if ((tmp = [ud objectForKey:@"qualifier"]))
      self->qualifier = [[EOQualifier alloc] initWithPropertyList:tmp];

    if ((tmp = [ud objectForKey:@"sort"])) {
      if ((tmp = [[EOSortOrdering alloc] initWithPropertyList:tmp])) {
	self->sortOrderings = [[NSArray alloc] initWithObjects:tmp,nil];
	[tmp release];
      }
    }
  }
  return self;
}
- (void)dealloc {
  [self->sortOrderings release];
  [self->qualifier     release];
  [self->entityName    release];
  [super dealloc];
}

/* run */

- (void)printObject:(id)_object {
  printf("object: %s\n", [[_object description] cString]);
}

- (int)runWithArguments:(NSArray *)_args {
  NSEnumerator *args;
  NSString     *arg;
  
  /* begin processing */
  
  args = [_args objectEnumerator];
  [args nextObject]; // process name ...
  
  while ((arg = [args nextObject])) {
    NSAutoreleasePool *pool2;
    
    if ([arg hasPrefix:@"-"]) { /* consume defaults */
      [args nextObject];
      continue;
    }
    
    pool2 = [[NSAutoreleasePool alloc] init];
    {    
      iCalDataSource       *ds;
      EOFetchSpecification *fspec;
      NSArray    *objs;
      iCalObject *obj;
      
      /* setup fetch specification */
      
      fspec = [[[EOFetchSpecification alloc] init] autorelease];
      [fspec setEntityName:self->entityName];
      [fspec setQualifier:self->qualifier];
      
      /* setup datasource */
      
      ds = [[[iCalDataSource alloc] initWithPath:arg] autorelease];
      [ds setFetchSpecification:fspec];

      /* perform fetch */
      
      if ((objs = [ds fetchObjects]) == nil) {
	/* fetch failed */
	
	NSLog(@"fetch on ical file failed: %@", arg);
      }
      else {
	/* process results */
	NSEnumerator *e;
	
	e = [objs objectEnumerator];
	while ((obj = [e nextObject])) {
	  [self printObject:obj];
	}
      }
    }
    [pool2 release];
  }
  return 0;
}

@end /* iCal3Tool */

int main(int argc, char **argv, char **env)  {
  NSAutoreleasePool *pool;
  iCal3Tool *tool;
  int rc;

#if LIB_FOUNDATION_LIBRARY  
  [NSProcessInfo initializeWithArguments:argv count:argc environment:env];
#endif
  
  pool = [[NSAutoreleasePool alloc] init];
  
  if ((tool = [[iCal3Tool alloc] init])) {
    NS_DURING
      rc = [tool runWithArguments:[[NSProcessInfo processInfo] arguments]];
    NS_HANDLER
      abort();
    NS_ENDHANDLER;
    
    [tool release];
  }
  else
    rc = 1;
  
  [pool release];
  return rc;
}
